/* GStreamer
 *
 * Copyright (C) 2008-2010 Nokia Corporation <multimedia@maemo.org>
 *
 * gstcameraphotoiface.c: Photo interface implementation for camerasrc
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>
#include <gst/gst.h>

#include "gstcameraphotoiface.h"
#include "gstcamerasrc.h"

GST_DEBUG_CATEGORY_EXTERN (gst_camerasrc_debug);
#define GST_CAT_DEFAULT gst_camerasrc_debug


static gboolean
gst_camerasrc_photo_parse_capture_caps (GstCameraSrc *camerasrc,
    GstCaps *op_mode_caps);
static gboolean
gst_camerasrc_photo_parse_preview_caps (GstCameraSrc *camerasrc,
    GstCaps *op_mode_caps);


/*
 *
 */
GstPhotoCaps
gst_camerasrc_photo_get_capabilities (GstCameraSrc *camerasrc)
{
  GstCameraSrcClass *bclass;
  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

  /* FIXME: driver might not be initialized yet */
  return bclass->get_capabilities (camerasrc);
}



/*
 *
 */
gboolean
gst_camerasrc_photo_prepare_for_capture (GstCameraSrc *camerasrc,
                                    GstPhotoCapturePrepared func,
                                    GstCaps *desired_caps,
                                    gpointer user_data)
{
  GstCameraSrcClass *bclass;
  gboolean use_vf_caps = TRUE;

  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

  g_mutex_lock (camerasrc->state_lock);

  camerasrc->prep_func = func;
  camerasrc->prep_udata = user_data;

  if (!camerasrc->capture_resolution_set) {
    if (desired_caps) {
      GstStructure *cstr;

      GST_DEBUG_OBJECT (camerasrc, "parsing capture caps");

      cstr = gst_caps_get_structure (desired_caps, 0);

      /* FIXME: Don't require FPS in capture caps */
      if (gst_structure_get_int (cstr, "width",
              (gint *) &camerasrc->capture_w) &&
          gst_structure_get_int (cstr, "height",
              (gint *) &camerasrc->capture_h) &&
          gst_structure_get_fraction (cstr, "framerate",
              (gint *) &camerasrc->capture_fps_n,
              (gint *) &camerasrc->capture_fps_d))
      {
        if (gst_structure_has_field (cstr, "format")) {
          gst_structure_get_fourcc (cstr, "format", &camerasrc->capture_fourcc);
        }
        else {
          /* If color format is not set, use the viewfinder format then */
          GST_DEBUG_OBJECT (camerasrc, "using viewfinder color format");
          camerasrc->capture_fourcc = camerasrc->current_fourcc;
        }
        use_vf_caps = FALSE;
      }
    }

    if (use_vf_caps) {
      GST_DEBUG_OBJECT (camerasrc, "given caps inadequate, using VF caps");

      camerasrc->capture_w = camerasrc->current_w;
      camerasrc->capture_h = camerasrc->current_h;
      camerasrc->capture_fps_n = camerasrc->fps_n;
      camerasrc->capture_fps_d = camerasrc->fps_d;
      camerasrc->capture_fourcc = camerasrc->current_fourcc;
    }
  }

  GST_DEBUG_OBJECT (camerasrc, "starting preparations for capture (%dx%d @ %f)",
      camerasrc->capture_w, camerasrc->capture_h,
      (gfloat) camerasrc->capture_fps_n / camerasrc->capture_fps_d);

  camerasrc->photo_capture_phase = GST_CAMERA_CAPTURE_START;

  GST_DEBUG_OBJECT (camerasrc, "calling unlock");
  /* Tell subclass to interrupt frame grabbing */
  if (bclass->unlock) {
    bclass->unlock (camerasrc);
  }
  GST_DEBUG_OBJECT (camerasrc, "unlock returned");

  g_mutex_unlock (camerasrc->state_lock);

  return TRUE;
}


/*
 *
 */
void
gst_camerasrc_photo_ready_for_capture (GstCameraSrc *camerasrc,
                                        GstCaps *newcaps)
{
  GST_DEBUG_OBJECT (camerasrc, "prepare for capture is complete");

  if (camerasrc->prep_func) {
    camerasrc->prep_func (camerasrc->prep_udata, newcaps);
    camerasrc->prep_func = NULL;
    camerasrc->prep_udata = NULL;
  }
  GST_DEBUG_OBJECT (camerasrc, "callback returned");
}


/*
 *
 */
static char *
create_debug_string (const char *base_str, GType type, gint value)
{
  GTypeClass *t = g_type_class_ref (type);
  GEnumValue *val = g_enum_get_value (G_ENUM_CLASS (t), value);
  gchar *ret = g_strconcat (base_str, val->value_nick, NULL);
  g_type_class_unref (t);

  return ret;
}


/*
 *
 */
gboolean
gst_camerasrc_photo_set_ev_compensation (GstCameraSrc *camerasrc,
                                          gfloat ev_comp)
{
  GstCameraSrcClass *bclass;
  gboolean ret = TRUE;
  
  if (ev_comp >= -2.5 && ev_comp <= 2.5)
  {

    bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

    camerasrc->photoconf.ev_compensation = ev_comp;

    if (bclass->is_active (camerasrc)) {
      ret = bclass->write_settings (camerasrc, &camerasrc->photoconf, FALSE);
    }
    GST_DEBUG_OBJECT (camerasrc, "set EV: %.2f, ret = %d", ev_comp, ret);
  } else {
    GST_DEBUG_OBJECT (camerasrc, "requested ev compensation value out of range");
    ret = FALSE;
  }
  return ret;
}


/*
 *
 */
gboolean
gst_camerasrc_photo_get_ev_compensation (GstCameraSrc *camerasrc,
                                          gfloat *ev_comp)
{
  GstCameraSrcClass *bclass;
  gboolean ret = TRUE;

  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

  if (bclass->is_active (camerasrc)) {
  ret = bclass->read_settings (camerasrc, &camerasrc->photoconf);
  }
  *ev_comp = camerasrc->photoconf.ev_compensation;
  GST_DEBUG_OBJECT (camerasrc, "got ev compensation: %.2f, ret = %d",
                    *ev_comp, ret);

  return ret;
}


/*
 *
 */
gboolean
gst_camerasrc_photo_set_iso_speed (GstCameraSrc *camerasrc,
                                    guint iso_speed)
{
  GstCameraSrcClass *bclass;
  gboolean ret = TRUE;

  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

  camerasrc->photoconf.iso_speed = iso_speed;

  if (bclass->is_active (camerasrc)) {
  return bclass->write_settings (camerasrc, &camerasrc->photoconf, FALSE);
}
  GST_DEBUG_OBJECT (camerasrc, "set ISO: %d, ret = %d", iso_speed, ret);

  return ret;
}


/*
 *
 */
gboolean
gst_camerasrc_photo_get_iso_speed (GstCameraSrc *camerasrc,
                                    guint *iso_speed)
{
  GstCameraSrcClass *bclass;
  gboolean ret = TRUE;

  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

  if (bclass->is_active (camerasrc)) {
  ret = bclass->read_settings (camerasrc, &camerasrc->photoconf);
  }
  *iso_speed = camerasrc->photoconf.iso_speed;
  GST_DEBUG_OBJECT (camerasrc, "got iso speed: %d, ret = %d", *iso_speed, ret);

  return ret;
}


/*
 *
 */
gboolean
gst_camerasrc_photo_set_aperture (GstCameraSrc *camerasrc,
                                   guint aperture)
{
  GstCameraSrcClass *bclass;
  gboolean ret = TRUE;

  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

  camerasrc->photoconf.aperture = aperture;

  if (bclass->is_active (camerasrc)) {
    ret = bclass->write_settings (camerasrc, &camerasrc->photoconf, FALSE);
}
  GST_DEBUG_OBJECT (camerasrc, "set aperture: %d, ret = %d", aperture, ret);

  return ret;
}


/*
 *
 */
gboolean
gst_camerasrc_photo_get_aperture (GstCameraSrc *camerasrc,
                                   guint *aperture)
{
  GstCameraSrcClass *bclass;
  gboolean ret = TRUE;

  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

  if (bclass->is_active (camerasrc)) {
  ret = bclass->read_settings (camerasrc, &camerasrc->photoconf);
  }
  *aperture = camerasrc->photoconf.aperture;
  GST_DEBUG_OBJECT (camerasrc, "got aperture: %d, ret = %d", *aperture, ret);

  return ret;
}


/*
 *
 */
gboolean
gst_camerasrc_photo_set_exposure (GstCameraSrc *camerasrc,
                                   guint32 exposure)
{
  GstCameraSrcClass *bclass;
  gboolean ret = TRUE;

  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

  camerasrc->photoconf.exposure = exposure;

  if (bclass->is_active (camerasrc)) {
    ret = bclass->write_settings (camerasrc, &camerasrc->photoconf, FALSE);
  }
  GST_DEBUG_OBJECT (camerasrc, "set exposure: %d, ret = %d", exposure, ret);

  return ret;
}


/*
 *
 */
gboolean
gst_camerasrc_photo_get_exposure (GstCameraSrc *camerasrc,
                                   guint32 *exposure)
{
  GstCameraSrcClass *bclass;
  gboolean ret = TRUE;

  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

  if (bclass->is_active (camerasrc)) {
  ret = bclass->read_settings (camerasrc, &camerasrc->photoconf);
  }
  *exposure = camerasrc->photoconf.exposure;
  GST_DEBUG_OBJECT (camerasrc, "got exposure: %d, ret = %d", *exposure, ret);

  return ret;
}


/*
 *
 */
gboolean
gst_camerasrc_photo_set_wb_mode (GstCameraSrc *camerasrc,
                                  GstWhiteBalanceMode mode)
{
  GstCameraSrcClass *bclass;
  gboolean ret = TRUE;
  gchar *dstr;

  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

  camerasrc->photoconf.wb_mode = mode;

  if (bclass->is_active (camerasrc)) {
    ret = bclass->write_settings (camerasrc, &camerasrc->photoconf, FALSE);
  }
  dstr = create_debug_string ("AWB:", GST_TYPE_WHITE_BALANCE_MODE, mode);
  GST_DEBUG_OBJECT (camerasrc, "set %s, ret = %d", dstr, ret);
  g_free (dstr);

  return ret;
}


/*
 *
 */
gboolean
gst_camerasrc_photo_get_wb_mode (GstCameraSrc *camerasrc,
                                  GstWhiteBalanceMode *mode)
{
  GstCameraSrcClass *bclass;
  gboolean ret = TRUE;

  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

  if (bclass->is_active (camerasrc)) {
  ret = bclass->read_settings (camerasrc, &camerasrc->photoconf);
  }
  *mode = camerasrc->photoconf.wb_mode;
  GST_DEBUG_OBJECT (camerasrc, "got AWB mode:%d, ret = %d", *mode, ret);

  return ret;
}


/*
 *
 */
gboolean
gst_camerasrc_photo_set_tone_mode (GstCameraSrc *camerasrc,
                                    GstColourToneMode mode)
{
  GstCameraSrcClass *bclass;
  gboolean ret = TRUE;
  gchar *dstr;

  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

  camerasrc->photoconf.tone_mode = mode;

  if (bclass->is_active (camerasrc)) {
    ret = bclass->write_settings (camerasrc, &camerasrc->photoconf, FALSE);
  }
  dstr = create_debug_string ("tone:", GST_TYPE_COLOUR_TONE_MODE, mode);
  GST_DEBUG_OBJECT (camerasrc, "set %s, ret = %d", dstr, ret);
  g_free (dstr);

  return ret;
}


/*
 *
 */
gboolean
gst_camerasrc_photo_get_tone_mode (GstCameraSrc *camerasrc,
                                    GstColourToneMode *mode)
{
  GstCameraSrcClass *bclass;
  gboolean ret = TRUE;

  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

  if (bclass->is_active (camerasrc)) {
  ret = bclass->read_settings (camerasrc, &camerasrc->photoconf);
  }
  *mode = camerasrc->photoconf.tone_mode;
  GST_DEBUG_OBJECT (camerasrc, "got tone mode: %d, ret = %d", *mode, ret);

  return ret;
}


/*
 *
 */
gboolean
gst_camerasrc_photo_set_scene_mode (GstCameraSrc *camerasrc,
                                     GstSceneMode mode)
{
  GstCameraSrcClass *bclass;
  gboolean ret = TRUE;
  gchar *dstr;
  GParamSpec **properties;
  gpointer photo_iface;
  guint i, n_properties = 0;

  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

  camerasrc->photoconf.scene_mode = mode;

  if (bclass->is_active (camerasrc)) {
  ret = bclass->write_settings (camerasrc, &camerasrc->photoconf, TRUE);
  if (ret) {
    /* Read the changed parameters to local settings cache */
    bclass->read_settings (camerasrc, &camerasrc->photoconf);
    /* Notify that changing scene mode might have changed also other
       interface properties */
    GST_DEBUG_OBJECT (camerasrc, "notifying interface property changes");
    photo_iface = g_type_default_interface_ref (GST_TYPE_PHOTOGRAPHY);
    properties =
        g_object_interface_list_properties (photo_iface, &n_properties);
    if (properties) {
      for (i = 0; i < n_properties; i++) {
        const gchar *name = g_param_spec_get_name (properties[i]);
        g_object_notify (G_OBJECT (camerasrc), name);
      }
      g_free (properties);
    }
    g_type_default_interface_unref (photo_iface);
  }
  }
  dstr = create_debug_string ("scene:", GST_TYPE_SCENE_MODE, mode);
  GST_DEBUG_OBJECT (camerasrc, "set %s, ret = %d", dstr, ret);
  g_free (dstr);

  return ret;
}


/*
 *
 */
gboolean
gst_camerasrc_photo_get_scene_mode (GstCameraSrc *camerasrc,
                                     GstSceneMode *mode)
{
  GstCameraSrcClass *bclass;
  gboolean ret = TRUE;

  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

  if (bclass->is_active (camerasrc)) {
  ret = bclass->read_settings (camerasrc, &camerasrc->photoconf);
  }
  *mode = camerasrc->photoconf.scene_mode;
  GST_DEBUG_OBJECT (camerasrc, "got scene mode: %d, ret = %d", *mode, ret);

  return ret;
}


/*
 *
 */
gboolean
gst_camerasrc_photo_set_flash_mode (GstCameraSrc *camerasrc,
                                     GstFlashMode mode)
{
  GstCameraSrcClass *bclass;
  gboolean ret = TRUE;
  gchar *dstr;

  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

  camerasrc->photoconf.flash_mode = mode;

  if (bclass->is_active (camerasrc)) {
    ret = bclass->write_settings (camerasrc, &camerasrc->photoconf, FALSE);
  }
  dstr = create_debug_string ("flash:", GST_TYPE_FLASH_MODE, mode);
  GST_DEBUG_OBJECT (camerasrc, "set %s, ret = %d", dstr, ret);
  g_free (dstr);

  return ret;
}


/*
 *
 */
gboolean
gst_camerasrc_photo_get_flash_mode (GstCameraSrc *camerasrc,
                                     GstFlashMode *mode)
{
  GstCameraSrcClass *bclass;
  gboolean ret = TRUE;

  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

  if (bclass->is_active (camerasrc)) {
  ret = bclass->read_settings (camerasrc, &camerasrc->photoconf);
  }
  *mode = camerasrc->photoconf.flash_mode;
  GST_DEBUG_OBJECT (camerasrc, "got flash mode: %d, ret = %d", *mode, ret);

  return ret;
}


/*
 *
 */
gboolean
gst_camerasrc_photo_set_zoom (GstCameraSrc *camerasrc, gfloat zoom)
{
  GstCameraSrcClass *bclass;
  gboolean ret = TRUE;

  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

  camerasrc->photoconf.zoom = zoom;

  GST_DEBUG_OBJECT (camerasrc, "setting zoom to %f", camerasrc->photoconf.zoom);

  if (bclass->set_zoom) {
    ret = bclass->set_zoom (camerasrc, camerasrc->photoconf.zoom);
  }
  else {
    /* FIXME: Check if the zoom is within known limits */
  }
  GST_DEBUG_OBJECT (camerasrc, "setting zoom %s", ret ? "success" : "failed");

  return ret;
}


/*
 *
 */
gboolean
gst_camerasrc_photo_get_zoom (GstCameraSrc *camerasrc, gfloat *zoom)
{
  *zoom = camerasrc->photoconf.zoom;

  GST_DEBUG_OBJECT (camerasrc, "current zoom = %f", *zoom);
  return TRUE;
}


/*
*
*/
gboolean
gst_camerasrc_photo_set_flicker_mode (GstCameraSrc *camerasrc,
                                       GstFlickerReductionMode mode)
{
  GstCameraSrcClass *bclass;
  gboolean ret = TRUE;
  gchar *dstr;

  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

  camerasrc->photoconf.flicker_mode = mode;

  if (bclass->is_active (camerasrc)) {
    ret = bclass->write_settings (camerasrc, &camerasrc->photoconf, FALSE);
}
  dstr = create_debug_string ("flicker mode:",
      GST_TYPE_FLICKER_REDUCTION_MODE, mode);

  GST_DEBUG_OBJECT (camerasrc, "set %s, ret = %d", dstr, ret);
  g_free (dstr);

  return ret;
}


/*
 *
 */
gboolean
gst_camerasrc_photo_get_flicker_mode (GstCameraSrc *camerasrc,
                                       GstFlickerReductionMode *mode)
{
  GstCameraSrcClass *bclass;
  gboolean ret = TRUE;

  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

  if (bclass->is_active (camerasrc)) {
  ret = bclass->read_settings (camerasrc, &camerasrc->photoconf);
  }
  *mode = camerasrc->photoconf.flicker_mode;
  GST_DEBUG_OBJECT (camerasrc, "got flicker mode: %d, ret = %d", *mode, ret);

  return ret;
}


/*
 *
 */
gboolean
gst_camerasrc_photo_set_focus_mode (GstCameraSrc *camerasrc,
                                     GstFocusMode mode)
{
  GstCameraSrcClass *bclass;
  gboolean ret = TRUE;
  gchar *dstr;

  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

  camerasrc->photoconf.focus_mode = mode;

  if (bclass->is_active (camerasrc)) {
    ret = bclass->write_settings (camerasrc, &camerasrc->photoconf, FALSE);
  }
  dstr = create_debug_string ("focus mode:", GST_TYPE_FOCUS_MODE, mode);
  GST_DEBUG_OBJECT (camerasrc, "set %s, ret = %d", dstr, ret);
  g_free (dstr);

  return ret;
}


/*
 *
 */
gboolean
gst_camerasrc_photo_get_focus_mode (GstCameraSrc *camerasrc,
                                     GstFocusMode *mode)
{
  GstCameraSrcClass *bclass;
  gboolean ret = TRUE;

  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

  if (bclass->is_active (camerasrc)) {
    ret = bclass->read_settings (camerasrc, &camerasrc->photoconf);
  }
  *mode = camerasrc->photoconf.focus_mode;
  GST_DEBUG_OBJECT (camerasrc, "got focus mode: %d, ret = %d", *mode, ret);

  return ret;
}


/*
 *
 */
void
gst_camerasrc_photo_set_autofocus (GstCameraSrc *camerasrc,
                                    gboolean on)
{
  GST_DEBUG_OBJECT (camerasrc, "setting autofocus %s", on ? "ON" : "OFF");

  g_mutex_lock (camerasrc->af_lock);

  if (on) {
    camerasrc->requested_af_mode = AF_ON_REQUESTED;
  }
  else {
    camerasrc->requested_af_mode = AF_OFF_REQUESTED;
  }

  g_mutex_unlock (camerasrc->af_lock);

  GST_DEBUG_OBJECT (camerasrc, "setting autofocus done");
}


/*
 *
 */
gboolean
gst_camerasrc_photo_set_config (GstCameraSrc *camerasrc,
                                 GstPhotoSettings * config)
{
  GstCameraSrcClass *bclass;
  gboolean ret = TRUE;

  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

  if (bclass->is_active (camerasrc)) {
    gboolean scene_override =
        config->scene_mode != GST_PHOTOGRAPHY_SCENE_MODE_MANUAL ? TRUE : FALSE;

    ret = bclass->write_settings (camerasrc, config, scene_override);

    if (ret && scene_override) {
      ret = bclass->read_settings (camerasrc, config);
    }
  }
  GST_DEBUG_OBJECT (camerasrc, "set config, ret = %d", ret);

  if (ret) {
    memcpy (&camerasrc->photoconf, config, sizeof (GstPhotoSettings));
  }

  return ret;

}


/*
 *
 */
gboolean
gst_camerasrc_photo_get_config (GstCameraSrc *camerasrc,
                                 GstPhotoSettings * config)
{
  memcpy (config, &camerasrc->photoconf, sizeof (GstPhotoSettings));

  GST_DEBUG_OBJECT (camerasrc, "got config");
  return TRUE;
}


/*
 *
 */
gboolean
gst_camerasrc_photo_set_format (GstCameraSrc *camerasrc,
                                GstOperationMode op_mode,
                                GstCaps * op_mode_caps)
{
  gboolean ret = TRUE;

  GST_DEBUG_OBJECT (camerasrc, "Caps received: %" GST_PTR_FORMAT, op_mode_caps);

  if (op_mode == GST_PHOTOGRAPHY_OPERATION_MODE_IMAGE_CAPTURE) {
    ret = gst_camerasrc_photo_parse_capture_caps (camerasrc, op_mode_caps);
  }
  else if (op_mode == GST_PHOTOGRAPHY_OPERATION_MODE_PREVIEW) {
    ret = gst_camerasrc_photo_parse_preview_caps (camerasrc, op_mode_caps);
  }
  else {
    GST_WARNING ("Trying to set unsupported operation mode");
    ret = FALSE;
  }

  return ret;
}


/*
 *
 */
GstCaps *
gst_camerasrc_photo_get_format (GstCameraSrc *camerasrc,
                                GstOperationMode op_mode)
{
  GstCaps *ret = NULL;

  /* FIXME: Check if the format is set or not */

  if (op_mode == GST_PHOTOGRAPHY_OPERATION_MODE_IMAGE_CAPTURE &&
      camerasrc->capture_resolution_set)
  {
    ret = gst_caps_new_simple ("video/x-raw-yuv",
        "format", GST_TYPE_FOURCC, camerasrc->capture_fourcc,
        "width", G_TYPE_INT, camerasrc->capture_w,
        "height", G_TYPE_INT, camerasrc->capture_h,
        "framerate", GST_TYPE_FRACTION, camerasrc->capture_fps_n,
            camerasrc->capture_fps_d,
        NULL);

    GST_DEBUG_OBJECT (camerasrc, "get format (capture): %" GST_PTR_FORMAT, ret);
  }
  else if (op_mode == GST_PHOTOGRAPHY_OPERATION_MODE_PREVIEW &&
            camerasrc->preview_resolution_set)
  {
    ret = gst_caps_new_simple ("video/x-raw-yuv",
        "format", GST_TYPE_FOURCC, camerasrc->preview_fourcc,
        "width", G_TYPE_INT, camerasrc->preview_w,
        "height", G_TYPE_INT, camerasrc->preview_h,
        NULL);

    GST_DEBUG_OBJECT (camerasrc, "get format (preview): %" GST_PTR_FORMAT, ret);
  }

  return ret;
}


/*
 *
 */
gboolean
gst_camerasrc_photo_set_noise_reduction (GstCameraSrc *camerasrc,
    GstPhotographyNoiseReduction noise_reduction)
{
  GstCameraSrcClass *bclass;
  gboolean ret = TRUE;
  gchar *dstr;

  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

  camerasrc->photoconf.noise_reduction = noise_reduction;

  if (bclass->is_active (camerasrc)) {
    ret = bclass->write_settings (camerasrc, &camerasrc->photoconf, FALSE);
  }
  dstr = create_debug_string ("noise reduction mode:",
      GST_TYPE_PHOTOGRAPHY_NOISE_REDUCTION, noise_reduction);
  GST_DEBUG_OBJECT (camerasrc, "set %s, ret = %d", dstr, ret);
  g_free (dstr);

  return ret;
}


/*
 *
 */
gboolean
gst_camerasrc_photo_get_noise_reduction (GstCameraSrc *camerasrc,
    GstPhotographyNoiseReduction *noise_reduction)
{
  /* This isn't implemented in lower levels yet */
  *noise_reduction = camerasrc->photoconf.noise_reduction;
  GST_DEBUG_OBJECT (camerasrc, "got noise reduction: %d, ret = %d",
      *noise_reduction);

  return TRUE;
}


/*
 *
 */
gboolean
gst_camerasrc_photo_get_property (GstCameraSrc *camerasrc,
                                   guint prop_id, GValue * value)
{
  GstCameraSrcClass *bclass;
  gboolean ret = FALSE;

  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

  switch (prop_id) {
    case PROP_WB_MODE:
    {
      GstWhiteBalanceMode wb_mode;
      GST_DEBUG_OBJECT (camerasrc, "==== GETTING PROP_WB_MODE ====");
      if (gst_camerasrc_photo_get_wb_mode (camerasrc, &wb_mode)) {
        g_value_set_enum (value, wb_mode);
      }
      ret = TRUE;
      break;
    }
    case PROP_COLOUR_TONE:
    {
      GstColourToneMode tone;
      GST_DEBUG_OBJECT (camerasrc, "==== GETTING PROP_COLOUR_TONE ====");
      if (gst_camerasrc_photo_get_tone_mode (camerasrc, &tone)) {
        g_value_set_enum (value, tone);
      }
      ret = TRUE;
      break;
    }
    case PROP_SCENE_MODE:
    {
      GstSceneMode scene;
      GST_DEBUG_OBJECT (camerasrc, "==== GETTING PROP_SCENE_MODE ====");
      if (gst_camerasrc_photo_get_scene_mode (camerasrc, &scene)) {
        g_value_set_enum (value, scene);
      }
      ret = TRUE;
      break;
    }
    case PROP_FLASH_MODE:
    {
      GstFlashMode flash;
      GST_DEBUG_OBJECT (camerasrc, "==== GETTING PROP_FLASH_MODE ====");
      if (gst_camerasrc_photo_get_flash_mode (camerasrc, &flash)) {
        g_value_set_enum (value, flash);
      }
      ret = TRUE;
      break;
    }
    case PROP_CAPABILITIES:
    {
      gulong capabilities;
      GST_DEBUG_OBJECT (camerasrc, "==== GETTING PROP_CAPABILITIES ====");
      capabilities = (gulong) gst_camerasrc_photo_get_capabilities (camerasrc);
      g_value_set_ulong (value, capabilities);
      ret = TRUE;
      break;
    }
    case PROP_EV_COMP:
    {
      gfloat ev_comp;
      GST_DEBUG_OBJECT (camerasrc, "==== GETTING PROP_EV_COMP ====");
      if (gst_camerasrc_photo_get_ev_compensation (camerasrc, &ev_comp)) {
        g_value_set_float (value, ev_comp);
      }
      ret = TRUE;
      break;
    }
    case PROP_ISO_SPEED:
    {
      guint iso_speed;
      GST_DEBUG_OBJECT (camerasrc, "==== GETTING PROP_ISO_SPEED ====");
      if (gst_camerasrc_photo_get_iso_speed (camerasrc, &iso_speed)) {
        g_value_set_uint (value, iso_speed);
      }
      ret = TRUE;
      break;
    }
    case PROP_APERTURE:
    {
      guint aperture;
      GST_DEBUG_OBJECT (camerasrc, "==== GETTING PROP_APERTURE ====");
      if (gst_camerasrc_photo_get_aperture (camerasrc, &aperture)) {
        g_value_set_uint (value, aperture);
      }
      ret = TRUE;
      break;
    }
    case PROP_EXPOSURE:
    {
      guint32 exposure;
      GST_DEBUG_OBJECT (camerasrc, "==== GETTING PROP_EXPOSURE ====");
      if (gst_camerasrc_photo_get_exposure (camerasrc, &exposure)) {
        g_value_set_uint (value, exposure);
      }
      ret = TRUE;
      break;
    }
    case PROP_ZOOM:
    {
      gfloat zoom;
      GST_DEBUG_OBJECT (camerasrc, "==== GETTING PROP_ZOOM ====");
      if (gst_camerasrc_photo_get_zoom (camerasrc, &zoom)) {
        g_value_set_float (value, zoom);
      }
      ret = TRUE;
      break;
    }
    case PROP_FLICKER_MODE:
    {
      GstFlickerReductionMode flicker_mode;
      GST_DEBUG_OBJECT (camerasrc, "==== GETTING PROP_FLICKER_MODE ====");
      if (gst_camerasrc_photo_get_flicker_mode (camerasrc, &flicker_mode)) {
        g_value_set_enum (value, flicker_mode);
      }
      ret = TRUE;
      break;
    }
    case PROP_FOCUS_MODE:
    {
      GstFocusMode focus_mode;
      GST_DEBUG_OBJECT (camerasrc, "==== GETTING PROP_FOCUS_MODE ====");
      if (gst_camerasrc_photo_get_focus_mode (camerasrc, &focus_mode)) {
        g_value_set_enum (value, focus_mode);
      }
      ret = TRUE;
      break;
    }
    case PROP_IMAGE_CAPTURE_CAPS:
    {
      GstCaps *caps;
      GST_DEBUG_OBJECT (camerasrc, "==== GETTING PROP_IMAGE_CAPTURE_CAPS ====");
      caps = bclass->get_supported_caps (camerasrc,
          GST_PHOTOGRAPHY_OPERATION_MODE_IMAGE_CAPTURE);
      if (!caps) {
        caps = gst_caps_new_empty ();
      }
      gst_value_set_caps (value, caps);
      gst_caps_unref (caps);
      ret = TRUE;
      break;
    }
    case PROP_IMAGE_PREVIEW_CAPS:
    {
      GstCaps *caps;
      GST_DEBUG_OBJECT (camerasrc, "==== GETTING PROP_IMAGE_PREVIEW_CAPS ====");
      caps = bclass->get_supported_caps (camerasrc,
          GST_PHOTOGRAPHY_OPERATION_MODE_PREVIEW);
      if (!caps) {
        caps = gst_caps_new_empty ();
      }
      gst_value_set_caps (value, caps);
      gst_caps_unref (caps);
      ret = TRUE;
      break;
    }
    case PROP_AUTOFOCUS:
    {
      GST_DEBUG_OBJECT (camerasrc, "==== GETTING PROP_AUTOFOCUS ====");
      g_mutex_lock (camerasrc->af_lock);
      if  (camerasrc->photo_capture_phase == GST_CAMERA_AUTOFOCUS ||
           camerasrc->requested_af_mode == AF_ON_REQUESTED)
      {
        g_value_set_boolean (value, TRUE);
      } else {
        g_value_set_boolean (value, FALSE);
      }
      g_mutex_unlock (camerasrc->af_lock);

      ret = TRUE;
      break;
    }
    case PROP_NOISE_REDUCTION:
    {
      GstPhotographyNoiseReduction nr;
      GST_DEBUG_OBJECT (camerasrc, "==== GETTING PROP_NOISE_REDUCTION ====");
      if (gst_camerasrc_photo_get_noise_reduction (camerasrc, &nr)) {
        g_value_set_enum (value, nr);
      }
      ret = TRUE;
    }
    default:
      break;
  }

  return ret;
}


/*
 *
 */
gboolean
gst_camerasrc_photo_set_property (GstCameraSrc *camerasrc,
                                   guint prop_id, const GValue * value)
{
  gboolean ret = FALSE;

  switch (prop_id) {
    case PROP_WB_MODE:
      GST_DEBUG_OBJECT (camerasrc, "==== SETTING PROP_WB_MODE ====");
      gst_camerasrc_photo_set_wb_mode (camerasrc, g_value_get_enum (value));
      ret = TRUE;
      break;
    case PROP_COLOUR_TONE:
      GST_DEBUG_OBJECT (camerasrc, "==== SETTING PROP_COLOUR_TONE ====");
      gst_camerasrc_photo_set_tone_mode (camerasrc, g_value_get_enum (value));
      ret = TRUE;
      break;
    case PROP_SCENE_MODE:
      GST_DEBUG_OBJECT (camerasrc, "==== SETTING PROP_SCENE_MODE ====");
      gst_camerasrc_photo_set_scene_mode (camerasrc, g_value_get_enum (value));
      ret = TRUE;
      break;
    case PROP_FLASH_MODE:
      GST_DEBUG_OBJECT (camerasrc, "==== SETTING PROP_FLASH_MODE ====");
      gst_camerasrc_photo_set_flash_mode (camerasrc, g_value_get_enum (value));
      ret = TRUE;
      break;
    case PROP_EV_COMP:
      GST_DEBUG_OBJECT (camerasrc, "==== SETTING PROP_EV_COMP ====");
      gst_camerasrc_photo_set_ev_compensation (camerasrc, g_value_get_float (value));
      ret = TRUE;
      break;
    case PROP_ISO_SPEED:
      GST_DEBUG_OBJECT (camerasrc, "==== SETTING PROP_ISO_SPEED ====");
      gst_camerasrc_photo_set_iso_speed (camerasrc, g_value_get_uint (value));
      ret = TRUE;
      break;
    case PROP_APERTURE:
      GST_DEBUG_OBJECT (camerasrc, "==== SETTING PROP_APERTURE ====");
      gst_camerasrc_photo_set_aperture (camerasrc, g_value_get_uint (value));
      ret = TRUE;
      break;
    case PROP_EXPOSURE:
      GST_DEBUG_OBJECT (camerasrc, "==== SETTING PROP_EXPOSURE ====");
      gst_camerasrc_photo_set_exposure (camerasrc, g_value_get_uint (value));
      ret = TRUE;
      break;
    case PROP_ZOOM:
      GST_DEBUG_OBJECT (camerasrc, "==== SETTING PROP_ZOOM ====");
      gst_camerasrc_photo_set_zoom (camerasrc, g_value_get_float (value));
      ret = TRUE;
      break;
    case PROP_FLICKER_MODE:
      GST_DEBUG_OBJECT (camerasrc, "==== SETTING PROP_FLICKER_MODE ====");
      gst_camerasrc_photo_set_flicker_mode (camerasrc, g_value_get_enum (value));
      ret = TRUE;
      break;
    case PROP_FOCUS_MODE:
      GST_DEBUG_OBJECT (camerasrc, "==== SETTING PROP_FOCUS_MODE ====");
      gst_camerasrc_photo_set_focus_mode (camerasrc, g_value_get_enum (value));
      ret = TRUE;
      break;
    case PROP_AUTOFOCUS:
      GST_DEBUG_OBJECT (camerasrc, "==== SETTING PROP_AUTOFOCUS ====");
      gst_camerasrc_photo_set_autofocus (camerasrc, g_value_get_boolean (value));
      ret = TRUE;
      break;
    case PROP_NOISE_REDUCTION:
      GST_DEBUG_OBJECT (camerasrc, "==== SETTING PROP_NOISE_REDUCTION ====");
      gst_camerasrc_photo_set_noise_reduction (camerasrc, g_value_get_enum (value));
      ret = TRUE;
      break;
    default:
      break;
  }

  return ret;
}


/*** Internal API ***/

static gboolean
gst_camerasrc_photo_parse_capture_caps (GstCameraSrc *camerasrc,
    GstCaps *op_mode_caps)
{
  GstCameraSrcClass *bclass;

  gboolean ret = TRUE;
  GstStructure *cstr;
  guint32 fcc_format;
  gint tmp_fps_n = 0;
  gint tmp_fps_d = 0;
  gint tmp_w = 0;
  gint tmp_h = 0;

  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

  GST_DEBUG_OBJECT (camerasrc, "Parsing image capture caps");

  if (op_mode_caps == NULL) {
    camerasrc->capture_resolution_set = FALSE;
    GST_DEBUG_OBJECT (camerasrc, "NULL caps received for image capture");
    goto done;
  }

  cstr = gst_caps_get_structure (op_mode_caps, 0);

  /* FIXME: Use VF format if fourcc is not given */
  /* FIXME: Don't require FPS here */
  if (gst_structure_get_fourcc (cstr, "format", &fcc_format) &&
      gst_structure_get_int (cstr, "width",  &tmp_w) &&
      gst_structure_get_int (cstr, "height", &tmp_h) &&
      gst_structure_get_fraction (cstr, "framerate", &tmp_fps_n, &tmp_fps_d))
  {
    /* check if the requested fourcc format is supported   */
    GstCaps * s_caps = NULL;

   /* get camsrc sourcepad caps, because the pad has
    * caps-template including supported formats */
    s_caps = gst_caps_copy (gst_pad_get_pad_template_caps (GST_BASE_SRC_PAD
          (camerasrc)));

    /* if intersection is empty -> provided caps not supported */
    if (!gst_caps_can_intersect (op_mode_caps, s_caps)) {
      GST_WARNING_OBJECT (camerasrc,
          "Unsupported fourcc format provided by caller");
      ret = FALSE;
    }

    /* Check that the requested resolution is not the same as what is
     * currently configured to be in use. If so, no need to set it again.
     */
    if ((camerasrc->capture_w != tmp_w ||
         camerasrc->capture_h != tmp_h ||
         camerasrc->capture_fourcc != fcc_format ||
         camerasrc->capture_fps_n != tmp_fps_n ||
         camerasrc->capture_fps_d != tmp_fps_d) &&
         ret == TRUE)
    {
      GST_DEBUG_OBJECT (camerasrc,
            "set width: %d , height: %d , fps_n: %d , fps_d :%d , format: %"
            GST_FOURCC_FORMAT, tmp_w, tmp_h, tmp_fps_n, tmp_fps_d,
            GST_FOURCC_ARGS (fcc_format));

      ret = bclass->set_capture (camerasrc,
          GST_PHOTOGRAPHY_OPERATION_MODE_IMAGE_CAPTURE, FALSE, &fcc_format,
          (guint *) &tmp_w, (guint *) &tmp_h, NULL, NULL);

      if (ret) {
        camerasrc->capture_w = tmp_w;
        camerasrc->capture_h = tmp_h;
        camerasrc->capture_fps_n = tmp_fps_n;
        camerasrc->capture_fps_d = tmp_fps_d;
        camerasrc->capture_fourcc = fcc_format;
        camerasrc->capture_resolution_set = TRUE;

        /* Write the assigned values back to the caps structure if possible */
        if (GST_CAPS_REFCOUNT_VALUE (op_mode_caps) == 1) {
          gst_caps_set_simple (op_mode_caps,
              "width", G_TYPE_INT, camerasrc->capture_w,
              "height", G_TYPE_INT, camerasrc->capture_h,
              "format", GST_TYPE_FOURCC, camerasrc->capture_fourcc,
              NULL);
        }

        /* It may not be possible to create the preview with previously
         * given resolution. Therefore we cancel the preview creation */
        gst_camerasrc_photo_parse_preview_caps (camerasrc, NULL);

        /* Notify that supported preview caps may have changed */
        g_object_notify (G_OBJECT (camerasrc),
            GST_PHOTOGRAPHY_PROP_IMAGE_PREVIEW_SUPPORTED_CAPS);
      }
    } else {
      GST_DEBUG_OBJECT (camerasrc, "format not set");
      ret = FALSE;
    }
    if (s_caps)
      gst_caps_unref (s_caps);

  } else {
    GST_DEBUG_OBJECT (camerasrc, "Unable to parse given caps");
    ret = FALSE;
  }

done:

  return ret;
}


static gboolean
gst_camerasrc_photo_parse_preview_caps (GstCameraSrc *camerasrc,
    GstCaps *op_mode_caps)
{
  GstCameraSrcClass *bclass;
  gboolean ret = FALSE;
  guint32 fourcc = 0;
  gint tmp_w = 0;
  gint tmp_h = 0;

  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

  if (op_mode_caps == NULL) {
    /* Setting NULL caps means canceling the preview image creation process.
     * In this case resolution 0x0 will be given to subclass */
    GST_DEBUG_OBJECT (camerasrc, "NULL caps received for preview image");
    ret = TRUE;
  }
  else {
    GstStructure *cstr;

    GST_DEBUG_OBJECT (camerasrc, "parsing preview caps");

    cstr = gst_caps_get_structure (op_mode_caps, 0);

    /* FIXME: Use VF format if fourcc is not given */
    if (gst_structure_get_fourcc (cstr, "format", &fourcc) &&
        gst_structure_get_int (cstr, "width",  &tmp_w) &&
        gst_structure_get_int (cstr, "height", &tmp_h))
    {
#if 0
      /* check if the requested fourcc format is supported */
      GstCaps * s_caps = NULL;
      GstCaps * r_caps = NULL;

     /* get camsrc sourcepad caps, because the pad has
      * caps-template including supported formats */
      s_caps = gst_caps_copy (gst_pad_get_pad_template_caps (GST_BASE_SRC_PAD
            (camerasrc)));
      /* take an intersection between sourcepad caps and provided caps */
      r_caps = gst_caps_intersect (op_mode_caps, s_caps);

      /* if EMPTY or NULL => provided caps not supported as not set in sourcepad */
      if (r_caps == NULL || gst_caps_is_empty (r_caps)) {
        GST_WARNING_OBJECT (camerasrc,
            "Unsupported fourcc format provided by caller");
        ret = FALSE;
      }
      if (s_caps)
        gst_caps_unref (s_caps);
      if (r_caps)
        gst_caps_unref (r_caps);
#endif
      ret = TRUE;
    }
    else {
      GST_DEBUG_OBJECT (camerasrc, "Unable to parse given caps");
      ret = FALSE;
    }
  }

  /* Check that the requested resolution is not the same as what is
   * currently configured to be in use. If so, no need to set it again.
   */
  if (ret && (camerasrc->preview_w != tmp_w || camerasrc->preview_h != tmp_h ||
      camerasrc->preview_fourcc != fourcc))
  {
    GST_DEBUG_OBJECT (camerasrc,
          "set preview width: %d, height: %d, format: %"
          GST_FOURCC_FORMAT, tmp_w, tmp_h, GST_FOURCC_ARGS (fourcc));

    ret = bclass->set_capture (camerasrc,
        GST_PHOTOGRAPHY_OPERATION_MODE_PREVIEW, FALSE,
        &fourcc, (guint *) &tmp_w, (guint *) &tmp_h, NULL, NULL);

    if (fourcc == 0) {
      /* Special case: preview image creation is canceled (NULL caps)*/
      camerasrc->preview_resolution_set = FALSE;
      ret = TRUE;
    }
    else {
      if (ret && op_mode_caps && GST_CAPS_REFCOUNT_VALUE (op_mode_caps) == 1) {
        /* Write the assigned values back to the caps structure */
        gst_caps_set_simple (op_mode_caps,
            "width", G_TYPE_INT, tmp_w,
            "height", G_TYPE_INT, tmp_h,
            "format", GST_TYPE_FOURCC, fourcc,
            NULL);
      }

      GST_DEBUG_OBJECT (camerasrc,
          "selected preview width: %d, height: %d, format: %"
          GST_FOURCC_FORMAT, tmp_w, tmp_h, GST_FOURCC_ARGS (fourcc));

      camerasrc->preview_resolution_set = ret;
    }
  } else {
    GST_DEBUG_OBJECT (camerasrc, "format not set");
    ret = FALSE;
  }

  if (ret) {
    camerasrc->preview_w = tmp_w;
    camerasrc->preview_h = tmp_h;
    camerasrc->preview_fourcc = fourcc;
  }

  return ret;
}





