/* GStreamer
 *
 * Copyright (C) 2001-2002 Ronald Bultje <rbultje@ronald.bitfreak.net>
 *               2006 Edgard Lima <edgard.lima@indt.org.br>
 *               2008-2010 Nokia Corporation <multimedia@maemo.org>
 *
 * gstcamerasrc.c: Abstract camera base class
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:gstcamerasrc
 * @short_description: Base class for camera sources
 *
 * <refsect2>
 * camerasrc is a generic base class for video source elements.
 * </refsect2>
 * <refsect2>
 * <title>Capturing modes</title>
 * <para>
 * GstCameraSrc provides three different operating modes: VIEWFINDER, IMAGE and
 * VIDEO capturing. This mode is selected by using "capture-mode" property.
 * Viewfinder mode is used for capturing preview frames before performing the
 * high-quality image capture, or before starting video recording. Image and
 * video modes should be switched just before starting the corresponding
 * action. GstCameraSrc itself doesn't make any difference between these modes,
 * and this setting is mainly targeted for subclasses (see below).
 * </para>
 * </refsect2>
 * <refsect2>
 * <title>Image capture</title>
 * <para>
 * In order to capture and image of higher quality than negotiated in element's
 * source pad, #GstPhotography's prepare_for_capture() must be called. It
 * Takes a #GstCaps object as a parameter. These caps will define the desired
 * high-quality capture resolution. However, this resolution is just a
 * suggestion; GstCameraSrc uses the given #GstPhotoCapturePrepared callback
 * to inform the selected capturing resolution back to the application. During
 * this callback application also has responsibility to adjust image pipeline
 * capsfilters according to this resolution. After HQ image is captured and
 * pushed, GstCameraSrc switches automatically back to old viewfinder resolution.
 * </para>
 * </refsect2>
 * <refsect2>
 * <title>Video capture</title>
 * <para>
 * By default GstCameraSrc works just like v4l2src, so it can be used for normal
 * video stream capturing as well. There is no separate GstPhotography API
 * function for starting video capturing (in contrary to image capturing), but
 * the notification is just given by setting the #GstCameraSrc:capture-mode
 * to video.
 * </para>
 * </refsect2>
 * <refsect2>
 * <title>Messages</title>
 * <para>
 * During capturing process GstCameraSrc sends a bunch of GstMessages to bus:
 * <itemizedlist>
 * <listitem>
 * <para>
 * GST_PHOTOGRAPHY_AUTOFOCUS_DONE is sent when application has started
 * autofocus operation and it finishes. This message contains following fields:
 * "status" is an integer which contains GstFocusStatus enumeration value.
 * It tells whether the focusing succeeded or failed.
 * "focus-window-rows" and "focus-window-columns" are integers and they tell
 * the focus matrix size.
 * "focus-windows" is a GArray containing a list of integers. They define
 * window numbers where the focus is. Window numbering starts from one, which
 * is top-left window, increasing from left-to-right, top-to-bottom order. This
 * field is only valid when focusing status is SUCCESS.
 * "focus-window-coverage" field is an integer and it tells the percentage how
 * big area from wiewfinder window was used for autofocusing.
 * </para>
 * </listitem>
 * <listitem>
 * <para>
 * GST_PHOTOGRAPHY_SHAKE_RISK is sent during autofocusing process. It keeps
 * application updated about the possible risk for shaken image. Sending of
 * this message is done in subclasses, so it is implementation-specific and
 * may not be always sent.
 * </para>
 * </listitem>
 * <listitem>
 * <para>
 * "photo-capture-start" is sent just before the actual high-quality capture
 * operation is about to happen.
 * </para>
 * </listitem>
 * <listitem>
 * <para>
 * "photo-capture-end" is sent immediately after high-quality image has been
 * captured. This can be used for example for playing the shutter sound
 * in sync with the capture operation.
 * </para>
 * </listitem>
 * <listitem>
 * <para>
 * "caf-update" messages are sent when camera subsystem supports continuous
 * autofocusing (CAF). The message contains information about CAF status. It
 * has one G_TYPE_INT field called "status" and its content is one of the
 * values defined in GstPhotography API's GstFocusStatus enumeration.
 * </para>
 * </listitem>
 * </itemizedlist>
 * </para>
 * </refsect2>
 * <refsect2>
 * <title>Supported formats</title>
 * <para>
 * Currently GstCameraSrc only supports UYVY and YUY2 color formats.
 * </para>
 * </refsect2>
 * <refsect2>
 * <title>Subclasses</title>
 * <para>
 * Hardware-specific imaging functionalities are implemented in subclasses.
 * GstCameraSrc provides vmethods for this purpose. Subclasses tasks are to
 * are handle e.g. autofocusing, flash, image capturing,
 * post-processing and all other non-V4L2 standard operations. Subclass
 * can also implement the API only partially; by default GstCameraSrc uses dummy
 * functions for all vmethods that are not implemented by the subclass.
 * </para>
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>
#include <sys/time.h>
#include <unistd.h>

#include <gst/tag/tag.h>

#include "gstcameracolorbalance.h"
#include "gstcameraphotoiface.h"


GST_DEBUG_CATEGORY (gst_camerasrc_debug);
#define GST_CAT_DEFAULT gst_camerasrc_debug


#define DEFAULT_PROP_ALWAYS_COPY        FALSE

#define MIN_ZOOM 0.0
#define MAX_ZOOM 1000.0
#define ZOOM_1X 100.0

GST_IMPLEMENT_CAMERA_SRC_COLOR_BALANCE_METHODS (GstCameraSrc, gst_camsrc);
GST_IMPLEMENT_CAMERA_SRC_PHOTO_METHODS (GstCameraSrc, gst_camsrc);

static gboolean
gst_camerasrc_iface_supported (GstImplementsInterface * iface, GType iface_type)
{
  GstCameraSrc *camsrc;
  GstCameraSrcClass *bclass;

  camsrc = GST_CAMERA_SRC (iface);
  bclass = GST_CAMERA_SRC_GET_CLASS (camsrc);

  g_assert (iface_type == GST_TYPE_PHOTOGRAPHY ||
      iface_type == GST_TYPE_COLOR_BALANCE);

  /* We implement GstPhotography in NULL state as well */
  if (iface_type == GST_TYPE_PHOTOGRAPHY)
    return TRUE;

  else if (!bclass->is_open (camsrc))
    return FALSE;

  return TRUE;
}

static void
gst_camerasrc_interface_init (GstImplementsInterfaceClass * klass)
{
  /*
   * default virtual functions 
   */
  klass->supported = gst_camerasrc_iface_supported;
}

void
gst_camerasrc_init_interfaces (GType type)
{
  static const GInterfaceInfo camsrc_iface_info = {
    (GInterfaceInitFunc) gst_camerasrc_interface_init,
    NULL,
    NULL,
  };
  static const GInterfaceInfo camsrc_photo_info = {
      (GInterfaceInitFunc) gst_camsrc_photo_interface_init,
       NULL,
       NULL,
  };
  static const GInterfaceInfo camsrc_colorbalance_info = {
    (GInterfaceInitFunc) gst_camsrc_color_balance_interface_init,
    NULL,
    NULL,
  };

  g_type_add_interface_static (type,
      GST_TYPE_IMPLEMENTS_INTERFACE, &camsrc_iface_info);
  g_type_add_interface_static (type,
     GST_TYPE_PHOTOGRAPHY, &camsrc_photo_info);
  g_type_add_interface_static (type,
      GST_TYPE_COLOR_BALANCE, &camsrc_colorbalance_info);
}


GType
gst_camerasrc_capture_mode_get_type (void)
{
  static GType gst_camerasrc_capture_mode_type = 0;
  static GEnumValue gst_camerasrc_capture_modes[] = {
    {GST_CAMERA_SRC_CAPTURE_MODE_VIEWFINDER, "Viewfinder mode", "viewfinder"},
    {GST_CAMERA_SRC_CAPTURE_MODE_STILL, "Still image capture mode", "still"},
    {GST_CAMERA_SRC_CAPTURE_MODE_VIDEO, "Video capturing mode", "video"},
    {0, NULL, NULL},
  };

  if (G_UNLIKELY (!gst_camerasrc_capture_mode_type)) {
    gst_camerasrc_capture_mode_type =
        g_enum_register_static ("GstCameraCaptureMode",
                                gst_camerasrc_capture_modes);
  }
  return gst_camerasrc_capture_mode_type;
}

GType
gst_camerasrc_viewfinder_mode_get_type (void)
{
  static GType gst_camerasrc_viewfinder_mode_type = 0;
  static GEnumValue gst_camerasrc_viewfinder_modes[] = {
    {GST_CAMERA_SRC_VIEWFINDER_MODE_STILL, "Still image capture mode", "still"},
    {GST_CAMERA_SRC_VIEWFINDER_MODE_VIDEO, "Video capturing mode", "video"},
    {0, NULL, NULL},
  };

  if (G_UNLIKELY (!gst_camerasrc_viewfinder_mode_type)) {
    gst_camerasrc_viewfinder_mode_type =
    g_enum_register_static ("GstCameraViewfinderMode",
        gst_camerasrc_viewfinder_modes);
  }
  return gst_camerasrc_viewfinder_mode_type;
}

GST_BOILERPLATE_FULL (GstCameraSrc, gst_camerasrc, GstPushSrc,
                      GST_TYPE_PUSH_SRC, gst_camerasrc_init_interfaces);

static void gst_camerasrc_dispose (GObject * object);

/* element methods */
static GstStateChangeReturn
gst_camerasrc_change_state (GstElement * element, GstStateChange transition);

/* basesrc methods */
static gboolean gst_camerasrc_start (GstBaseSrc * src);

static gboolean gst_camerasrc_unlock (GstBaseSrc * src);

static gboolean gst_camerasrc_unlock_stop (GstBaseSrc * src);

static gboolean gst_camerasrc_stop (GstBaseSrc * src);

static gboolean gst_camerasrc_set_caps (GstBaseSrc * src, GstCaps * caps);

static GstCaps *gst_camerasrc_get_caps (GstBaseSrc * src);

static gboolean gst_camerasrc_query (GstBaseSrc * bsrc, GstQuery * query);

static GstFlowReturn gst_camerasrc_create (GstPushSrc * src, GstBuffer ** out);

static void gst_camerasrc_fixate (GstBaseSrc * basesrc, GstCaps * caps);

static gboolean gst_camerasrc_negotiate (GstBaseSrc * basesrc);

static void gst_camerasrc_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);

static void gst_camerasrc_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);

static void gst_camerasrc_override_photo_properties (
    GObjectClass *gobject_class);

static void gst_camerasrc_default_functions_init (
    GstCameraSrcClass *camera_class);

static gboolean gst_camerasrc_event (GstBaseSrc * src, GstEvent * event);

static gboolean gst_camerasrc_send_event (GstElement * element,
    GstEvent * event);

static void gst_camerasrc_update_max_zoom (GstCameraSrc *camerasrc);

/*
 */
static void
gst_camerasrc_base_init (gpointer g_class)
{
//   GstElementClass *gstelement_class = GST_ELEMENT_CLASS (g_class);
  gst_tag_register_musicbrainz_tags ();

  GST_DEBUG_CATEGORY_INIT (gst_camerasrc_debug, "camerasrc", 0,
    "camerasrc element");
}

/*
 */
static void
gst_camerasrc_class_init (GstCameraSrcClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *element_class;
  GstBaseSrcClass *basesrc_class;
  GstPushSrcClass *pushsrc_class;

  gobject_class = G_OBJECT_CLASS (klass);
  element_class = GST_ELEMENT_CLASS (klass);
  basesrc_class = GST_BASE_SRC_CLASS (klass);
  pushsrc_class = GST_PUSH_SRC_CLASS (klass);

  gobject_class->dispose = gst_camerasrc_dispose;
  gobject_class->set_property = gst_camerasrc_set_property;
  gobject_class->get_property = gst_camerasrc_get_property;

  element_class->change_state = gst_camerasrc_change_state;
  element_class->send_event = gst_camerasrc_send_event;

  g_object_class_install_property (gobject_class, PROP_ALWAYS_COPY,
      g_param_spec_boolean ("always-copy", "Always Copy",
          "If the buffer will or not be used directly from mmap",
          DEFAULT_PROP_ALWAYS_COPY,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

  /**
   * GstCameraSrc:capture-mode:
   *
   * Defines capturing mode to be used. This information is needed by extension
   * modules, which control hardware-specific functionality.
   */
  g_object_class_install_property (gobject_class, PROP_CAPTURE_MODE,
      g_param_spec_enum ("capture-mode", "Capturing mode",
          "Defines what kind of capturing mode to be used",
          GST_TYPE_CAMERA_SRC_CAPTURE_MODE,
          GST_CAMERA_SRC_CAPTURE_MODE_VIEWFINDER,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

  /**
   * GstCameraSrc:viewfinder-mode:
   *
   * Defines which mode viewfinder should run in. This information is needed
   * by extension modules, which control hardware-specific functionality.
   */
  g_object_class_install_property (gobject_class, PROP_VIEWFINDER_MODE,
      g_param_spec_enum ("viewfinder-mode", "Viewfinder mode",
          "Defines in what mode viewfinder should be run",
          GST_TYPE_CAMERA_SRC_VIEWFINDER_MODE,
          GST_CAMERA_SRC_VIEWFINDER_MODE_STILL,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

  gst_camerasrc_override_photo_properties (gobject_class);

  basesrc_class->get_caps = GST_DEBUG_FUNCPTR (gst_camerasrc_get_caps);
  basesrc_class->set_caps = GST_DEBUG_FUNCPTR (gst_camerasrc_set_caps);
  basesrc_class->start = GST_DEBUG_FUNCPTR (gst_camerasrc_start);
  basesrc_class->unlock = GST_DEBUG_FUNCPTR (gst_camerasrc_unlock);
  basesrc_class->unlock_stop = GST_DEBUG_FUNCPTR (gst_camerasrc_unlock_stop);
  basesrc_class->stop = GST_DEBUG_FUNCPTR (gst_camerasrc_stop);
  basesrc_class->query = GST_DEBUG_FUNCPTR (gst_camerasrc_query);
  basesrc_class->fixate = GST_DEBUG_FUNCPTR (gst_camerasrc_fixate);
  basesrc_class->negotiate = GST_DEBUG_FUNCPTR (gst_camerasrc_negotiate);
  basesrc_class->event = GST_DEBUG_FUNCPTR (gst_camerasrc_event);

  pushsrc_class->create = GST_DEBUG_FUNCPTR (gst_camerasrc_create);

  /* Initialize vmethods with default implementations */
  gst_camerasrc_default_functions_init (klass);
}

/*
 */
static void
gst_camerasrc_init (GstCameraSrc * camerasrc, GstCameraSrcClass * klass)
{
  /* number of buffers requested */
  camerasrc->always_copy = DEFAULT_PROP_ALWAYS_COPY;

  camerasrc->state_lock = g_mutex_new ();
  camerasrc->af_lock = g_mutex_new ();

  gst_base_src_set_format (GST_BASE_SRC (camerasrc), GST_FORMAT_TIME);
  gst_base_src_set_live (GST_BASE_SRC (camerasrc), TRUE);

  /* Photo interface */
  camerasrc->photoconf.zoom = 1.0;
  camerasrc->photoconf.ev_compensation = 0.0;
  camerasrc->photoconf.exposure = 0;   /* 0 = auto */
  camerasrc->photoconf.aperture = 0;   /* 0 = auto */
  camerasrc->photoconf.iso_speed = 0;  /* 0 = auto */
  camerasrc->photoconf.wb_mode = GST_PHOTOGRAPHY_WB_MODE_AUTO;
  camerasrc->photoconf.scene_mode = GST_PHOTOGRAPHY_SCENE_MODE_MANUAL;
  camerasrc->photoconf.flash_mode = GST_PHOTOGRAPHY_FLASH_MODE_AUTO;
  camerasrc->photoconf.tone_mode = GST_PHOTOGRAPHY_COLOUR_TONE_MODE_NORMAL;
  camerasrc->photoconf.flicker_mode = GST_PHOTOGRAPHY_FLICKER_REDUCTION_AUTO;
  camerasrc->photoconf.focus_mode = GST_PHOTOGRAPHY_FOCUS_MODE_AUTO;

  camerasrc->photo_capture_phase = GST_CAMERA_VIEWFINDER;
  camerasrc->capture_mode = GST_CAMERA_SRC_CAPTURE_MODE_VIEWFINDER;
  camerasrc->viewfinder_mode = GST_CAMERA_SRC_VIEWFINDER_MODE_STILL;
  camerasrc->capture_resolution_set = FALSE;
  camerasrc->preview_resolution_set = FALSE;

  camerasrc->capture_w = 0;
  camerasrc->capture_h = 0;
  camerasrc->capture_fps_n = 0;
  camerasrc->capture_fps_d = 0;
  camerasrc->capture_fourcc = 0;

  camerasrc->preview_w = 0;
  camerasrc->preview_h = 0;
  camerasrc->preview_fourcc = 0;

  camerasrc->requested_af_mode = AF_NONE_REQUESTED;
}

/*
 */
static void
gst_camerasrc_dispose (GObject * object)
{
  GstCameraSrc *camerasrc = GST_CAMERA_SRC_CAST (object);

  if (camerasrc->vf_caps) {
    gst_caps_unref (camerasrc->vf_caps);
    camerasrc->vf_caps = NULL;
  }
  if (camerasrc->state_lock) {
    g_mutex_free (camerasrc->state_lock);
    camerasrc->state_lock = NULL;
  }
  if (camerasrc->af_lock) {
    g_mutex_free (camerasrc->af_lock);
    camerasrc->af_lock = NULL;
  }
  gst_camerasrc_clear_color_channels (camerasrc);

  G_OBJECT_CLASS (parent_class)->dispose (object);
}

static void
gst_camerasrc_override_photo_properties (GObjectClass *gobject_class)
{
  g_object_class_override_property (gobject_class, PROP_WB_MODE,
      GST_PHOTOGRAPHY_PROP_WB_MODE);

  g_object_class_override_property (gobject_class, PROP_COLOUR_TONE,
      GST_PHOTOGRAPHY_PROP_COLOUR_TONE);

  g_object_class_override_property (gobject_class, PROP_SCENE_MODE,
      GST_PHOTOGRAPHY_PROP_SCENE_MODE);

  g_object_class_override_property (gobject_class, PROP_FLASH_MODE,
      GST_PHOTOGRAPHY_PROP_FLASH_MODE);

  g_object_class_override_property (gobject_class, PROP_CAPABILITIES,
      GST_PHOTOGRAPHY_PROP_CAPABILITIES);

  g_object_class_override_property (gobject_class, PROP_EV_COMP,
      GST_PHOTOGRAPHY_PROP_EV_COMP);

  g_object_class_override_property (gobject_class, PROP_ISO_SPEED,
      GST_PHOTOGRAPHY_PROP_ISO_SPEED);

  g_object_class_override_property (gobject_class, PROP_APERTURE,
      GST_PHOTOGRAPHY_PROP_APERTURE);

  g_object_class_override_property (gobject_class, PROP_EXPOSURE,
      GST_PHOTOGRAPHY_PROP_EXPOSURE);

  g_object_class_override_property (gobject_class, PROP_ZOOM,
      GST_PHOTOGRAPHY_PROP_ZOOM);

  g_object_class_override_property (gobject_class, PROP_FLICKER_MODE,
      GST_PHOTOGRAPHY_PROP_FLICKER_MODE);

  g_object_class_override_property (gobject_class, PROP_FOCUS_MODE,
      GST_PHOTOGRAPHY_PROP_FOCUS_MODE);

  g_object_class_override_property (gobject_class, PROP_IMAGE_CAPTURE_CAPS,
      GST_PHOTOGRAPHY_PROP_IMAGE_CAPTURE_SUPPORTED_CAPS);

  g_object_class_override_property (gobject_class, PROP_IMAGE_PREVIEW_CAPS,
      GST_PHOTOGRAPHY_PROP_IMAGE_PREVIEW_SUPPORTED_CAPS);

  g_object_class_override_property (gobject_class, PROP_AUTOFOCUS,
      GST_PHOTOGRAPHY_PROP_AUTOFOCUS);

    g_object_class_override_property (gobject_class, PROP_NOISE_REDUCTION,
      GST_PHOTOGRAPHY_PROP_NOISE_REDUCTION);

}


/*
 */
static void
gst_camerasrc_set_property (GObject * object,
    guint prop_id, const GValue * value, GParamSpec * pspec)
{
  GstCameraSrc *camerasrc;
  GstCameraSrcClass *bclass;

  camerasrc = GST_CAMERA_SRC_CAST (object);
  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

  if (gst_camerasrc_photo_set_property (camerasrc, prop_id, value))
    return;

  switch (prop_id) {
    case PROP_ALWAYS_COPY:
      camerasrc->always_copy = g_value_get_boolean (value);
      break;
    case PROP_CAPTURE_MODE:
      camerasrc->capture_mode = g_value_get_enum (value);
      GST_INFO_OBJECT (camerasrc, "Setting capture mode: %d",
          camerasrc->capture_mode);

      if (camerasrc->capture_mode == GST_CAMERA_SRC_CAPTURE_MODE_VIDEO &&
        bclass->is_active (camerasrc))
      {
        /* Stop autofocus in video mode */
        bclass->set_autofocus (camerasrc, FALSE);
        /* AutoExposure must be run always in video mode */
        bclass->set_autoexposure (camerasrc, TRUE);
      }
      if (bclass->set_capture_mode) {
        bclass->set_capture_mode (camerasrc, camerasrc->capture_mode);
      }
      break;

    case PROP_VIEWFINDER_MODE:
      camerasrc->viewfinder_mode = g_value_get_enum (value);
      GST_INFO_OBJECT (camerasrc, "Setting viewfinder mode: %d",
          camerasrc->viewfinder_mode);

      if (bclass->set_vfinder_mode) {
        bclass->set_vfinder_mode (camerasrc, camerasrc->viewfinder_mode);
      }
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

/*
 */
static void
gst_camerasrc_get_property (GObject * object,
    guint prop_id, GValue * value, GParamSpec * pspec)
{
  GstCameraSrc *camerasrc = GST_CAMERA_SRC_CAST (object);

  if (gst_camerasrc_photo_get_property (camerasrc, prop_id, value))
    return;

  switch (prop_id) {
    case PROP_ALWAYS_COPY:
      g_value_set_boolean (value, camerasrc->always_copy);
      break;
    case PROP_CAPTURE_MODE:
      g_value_set_enum (value, camerasrc->capture_mode);
      break;
    case PROP_VIEWFINDER_MODE:
      g_value_set_enum (value, camerasrc->viewfinder_mode);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

/* this function is a bit of a last resort */
static void
gst_camerasrc_fixate (GstBaseSrc * basesrc, GstCaps * caps)
{
  GstStructure *structure;

  gint i;

  GST_DEBUG_OBJECT (basesrc, "fixating caps %" GST_PTR_FORMAT, caps);

  for (i = 0; i < gst_caps_get_size (caps); ++i) {
    const GValue *v;

    structure = gst_caps_get_structure (caps, i);

    /* FIXME such sizes? we usually fixate to something in the 320x200
     * range... */
    /* We are fixating to greatest possble size (limited to GST_CAMERA_SRC_MAX_SIZE)
       and the maximum framerate resolution for that size */
    gst_structure_fixate_field_nearest_int (structure, "width",
        GST_CAMERA_SRC_MAX_SIZE);
    gst_structure_fixate_field_nearest_int (structure, "height",
        GST_CAMERA_SRC_MAX_SIZE);
    gst_structure_fixate_field_nearest_fraction (structure, "framerate",
        G_MAXINT, 1);

    v = gst_structure_get_value (structure, "format");
    if (v && G_VALUE_TYPE (v) != GST_TYPE_FOURCC) {
      guint32 fourcc;

      g_return_if_fail (G_VALUE_TYPE (v) == GST_TYPE_LIST);

      fourcc = gst_value_get_fourcc (gst_value_list_get_value (v, 0));
      gst_structure_set (structure, "format", GST_TYPE_FOURCC, fourcc, NULL);
    }
  }

  GST_DEBUG_OBJECT (basesrc, "fixated caps %" GST_PTR_FORMAT, caps);
}

/*
 */
static gboolean
gst_camerasrc_negotiate (GstBaseSrc * basesrc)
{
  GstCaps *thiscaps;

  GstCaps *caps = NULL;

  GstCaps *peercaps = NULL;

  gboolean result = FALSE;

  /* first see what is possible on our source pad */
  thiscaps = gst_pad_get_caps (GST_BASE_SRC_PAD (basesrc));
  GST_DEBUG_OBJECT (basesrc, "caps of src: %" GST_PTR_FORMAT, thiscaps);
  /* nothing or anything is allowed, we're done */
  if (thiscaps == NULL || gst_caps_is_any (thiscaps))
    goto no_nego_needed;

  /* get the peer caps */
  peercaps = gst_pad_peer_get_caps (GST_BASE_SRC_PAD (basesrc));
  GST_DEBUG_OBJECT (basesrc, "caps of peer: %" GST_PTR_FORMAT, peercaps);
  if (peercaps && !gst_caps_is_any (peercaps)) {
    GstCaps *icaps = NULL;
    int i;

    /* Prefer the first caps we are compatible with that the peer proposed */
    for (i = 0; i < gst_caps_get_size (peercaps); i++) {
      /* get intersection */
      GstCaps *ipcaps = gst_caps_copy_nth (peercaps, i);

      GST_DEBUG_OBJECT (basesrc, "peer: %" GST_PTR_FORMAT, ipcaps);

      icaps = gst_caps_intersect (thiscaps, ipcaps);
      gst_caps_unref (ipcaps);

      if (!gst_caps_is_empty (icaps))
        break;

      gst_caps_unref (icaps);
      icaps = NULL;
    }

    GST_DEBUG_OBJECT (basesrc, "intersect: %" GST_PTR_FORMAT, icaps);
    if (icaps) {
      /* If there are multiple intersections pick the one with the smallest
       * resolution strictly bigger then the first peer caps */
      if (gst_caps_get_size (icaps) > 1) {
        GstStructure *s = gst_caps_get_structure (peercaps, 0);

        int best = 0;

        int twidth, theight;

        int width = G_MAXINT, height = G_MAXINT;

        if (gst_structure_get_int (s, "width", &twidth)
            && gst_structure_get_int (s, "height", &theight)) {

          /* Walk the structure backwards to get the first entry of the
           * smallest resolution bigger (or equal to) the preferred resolution)
           */
          for (i = gst_caps_get_size (icaps) - 1; i >= 0; i--) {
            GstStructure *is = gst_caps_get_structure (icaps, i);

            int w, h;

            if (gst_structure_get_int (is, "width", &w)
                && gst_structure_get_int (is, "height", &h)) {
              if (w >= twidth && w <= width && h >= theight && h <= height) {
                width = w;
                height = h;
                best = i;
              }
            }
          }
        }

        caps = gst_caps_copy_nth (icaps, best);
        gst_caps_unref (icaps);
      } else {
        caps = icaps;
      }
    }
    gst_caps_unref (thiscaps);
    gst_caps_unref (peercaps);
  } else {
    /* no peer or peer have ANY caps, work with our own caps then */
    caps = thiscaps;
  }
  if (caps) {
    caps = gst_caps_make_writable (caps);
    gst_caps_truncate (caps);

    /* now fixate */
    if (!gst_caps_is_empty (caps)) {
      gst_pad_fixate_caps (GST_BASE_SRC_PAD (basesrc), caps);
      GST_DEBUG_OBJECT (basesrc, "fixated to: %" GST_PTR_FORMAT, caps);

      if (gst_caps_is_any (caps)) {
        /* hmm, still anything, so element can do anything and
         * nego is not needed */
        result = TRUE;
      } else if (gst_caps_is_fixed (caps)) {
        /* yay, fixed caps, use those then */
        result = gst_pad_set_caps (GST_BASE_SRC_PAD (basesrc), caps);
        GST_DEBUG_OBJECT (basesrc, "Set caps returned: %d", result);
      }
    }
    gst_caps_unref (caps);
  }
  return result;

no_nego_needed:
  {
    GST_DEBUG_OBJECT (basesrc, "no negotiation needed");
    if (thiscaps)
      gst_caps_unref (thiscaps);
    return TRUE;
  }
}


/*
 */
static GstCaps *
gst_camerasrc_get_caps (GstBaseSrc * src)
{
  GstCameraSrc *camerasrc;
  GstCameraSrcClass *bclass;

  camerasrc = GST_CAMERA_SRC_CAST (src);
  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

  if (!bclass->is_open (camerasrc) || !bclass->get_caps) {
    return gst_caps_copy (gst_pad_get_pad_template_caps (GST_BASE_SRC_PAD
            (camerasrc)));
  }

  return bclass->get_caps (camerasrc);
}


/**
 * gst_camerasrc_get_caps_info:
 * @caps: given input caps
 * @four_cc: location for the fourcc
 * @w/@h: location for width and height
 * @fps_n/@fps_d: location for framerate
 *
 * Collect data for the given caps.
 */
static gboolean
gst_camerasrc_get_caps_info (GstCameraSrc * camerasrc, GstCaps * caps,
    guint32 *fourcc, guint * w, guint * h, guint * fps_n, guint * fps_d)
{
  GstCameraSrcClass *bclass;
  GstStructure *structure;
  const GValue *framerate;

  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

  structure = gst_caps_get_structure (caps, 0);

  if (!gst_structure_get_int (structure, "width", (gint *) w))
    return FALSE;

  if (!gst_structure_get_int (structure, "height", (gint *) h))
    return FALSE;

  framerate = gst_structure_get_value (structure, "framerate");
  if (!framerate)
    return FALSE;

  *fps_n = gst_value_get_fraction_numerator (framerate);
  *fps_d = gst_value_get_fraction_denominator (framerate);

  gst_structure_get_fourcc (structure, "format", fourcc);

  return TRUE;
}


/**
 * gst_camerasrc_get_caps_from_info:
 * @camsrc: #GstCameraSrc object
 * @fourcc: fourcc code
 * @width: width to be set
 * @height: height to be set
 * @fps_n: FPS numerator to be set or 0
 * @fps_d: FPS denominator to be set or 0
 *
 * Converts given parameters into GstCaps structure.
 *
 * Returns: GstCaps representing the given values.
 */
GstCaps *
gst_camerasrc_get_caps_from_info (GstCameraSrc *camsrc, guint32 fourcc,
    guint width, guint height, guint fps_n, guint fps_d)
{
  GstCaps *newcaps;
  GstStructure *s;

  s = gst_structure_new ("video/x-raw-yuv",
                         "format", GST_TYPE_FOURCC, fourcc,
                         "width", G_TYPE_INT, width,
                         "height", G_TYPE_INT, height,
                         NULL);

  if (fps_n != 0 && fps_d != 0) {
    gst_structure_set (s, "framerate", GST_TYPE_FRACTION, fps_n, fps_d, NULL);
  }

  newcaps = gst_caps_new_empty ();
  gst_caps_append_structure (newcaps, s);

  return newcaps;
}


/*
 */
static gboolean
gst_camerasrc_configure_device (GstCameraSrc *camerasrc, guint *w, guint *h,
    guint32 *fourcc, guint *fps_n, guint *fps_d, GstCaps *buffer_caps)
{
  GstCameraSrcClass *bclass;
  GstOperationMode opmode;

  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

  /* Stop the capturing */
  if (!bclass->stop (camerasrc))
    return FALSE;

  GST_DEBUG_OBJECT (camerasrc, "trying to set_capture %dx%d at %d/%d fps",
                    *w, *h, *fps_n, *fps_d);

  if (bclass->set_vfinder_mode) {
    bclass->set_vfinder_mode (camerasrc, camerasrc->viewfinder_mode);
  }
  if (bclass->set_capture_mode) {
    bclass->set_capture_mode (camerasrc, camerasrc->capture_mode);
  }

  opmode = camerasrc->photo_capture_phase == GST_CAMERA_CAPTURE ?
            GST_PHOTOGRAPHY_OPERATION_MODE_IMAGE_CAPTURE :
            GST_PHOTOGRAPHY_OPERATION_MODE_VIEWFINDER;

  if (!bclass->set_capture (camerasrc, opmode, FALSE, fourcc, w, h,
      fps_n, fps_d))
  {
    GST_ELEMENT_ERROR (camerasrc, RESOURCE, SETTINGS,
                       ("Failed configuring device for capturing"), (NULL));
    return FALSE;
  }

  gst_camerasrc_update_max_zoom (camerasrc);

  /* Only start the driver when not in HQ capture mode, since in HQ mode */
  /* we have already called start_capture() above */
  if (camerasrc->photo_capture_phase != GST_CAMERA_CAPTURE) {

    if (!bclass->write_settings (camerasrc, &camerasrc->photoconf, FALSE)) {
      GST_ELEMENT_ERROR (camerasrc, RESOURCE, SETTINGS,
                         ("Failed to configure driver module"), (NULL));

      return FALSE;
    }
  }

  /* Use automatic flicker detection every time */
  camerasrc->photoconf.flicker_mode = GST_PHOTOGRAPHY_FLICKER_REDUCTION_AUTO;

  if (!bclass->start (camerasrc, buffer_caps))
    return FALSE;

  return TRUE;
}


/*
 */
static gboolean
gst_camerasrc_init_from_caps (GstCameraSrc *camerasrc, GstCaps * caps)
{
  guint w = 0, h = 0;
  guint32 fourcc;
  guint fps_n, fps_d;
  gboolean ret;

  /* we want our own type of fourcc codes */
  if (!gst_camerasrc_get_caps_info (camerasrc, caps, &fourcc, &w, &h, &fps_n,
       &fps_d))
  {
    GST_DEBUG_OBJECT (camerasrc,
        "can't get capture format from caps %" GST_PTR_FORMAT, caps);
    return FALSE;
  }

  ret = gst_camerasrc_configure_device (camerasrc, &w, &h, &fourcc,
      &fps_n, &fps_d, caps);

  if (ret) {
    camerasrc->current_w = w;
    camerasrc->current_h = h;
    camerasrc->fps_n = fps_n;
    camerasrc->fps_d = fps_d;
    camerasrc->current_fourcc = fourcc;

    camerasrc->duration = gst_util_uint64_scale_int (GST_SECOND, fps_d, fps_n);
    GST_DEBUG_OBJECT (camerasrc, "Buffer duration: %" GST_TIME_FORMAT,
        GST_TIME_ARGS (camerasrc->duration));
  }

  return ret;
}

/*
 */
static gboolean
gst_camerasrc_set_caps (GstBaseSrc * src, GstCaps * caps)
{
  GstCameraSrc *camerasrc;
  GstCameraSrcClass *bclass;

  camerasrc = GST_CAMERA_SRC_CAST (src);
  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

  GST_DEBUG_OBJECT (camerasrc, "set_caps called: %" GST_PTR_FORMAT, caps);

  /* if we're not open, punt -- we'll get setcaps'd later via negotiate */
  if (!bclass->is_open (camerasrc))
    return FALSE;

  return gst_camerasrc_init_from_caps (camerasrc, caps);
}

/*
 */
static gboolean
gst_camerasrc_query (GstBaseSrc * bsrc, GstQuery * query)
{
  GstCameraSrc *src;
  GstCameraSrcClass *bclass;
  gboolean res = FALSE;
  guint num_buffers;

  src = GST_CAMERA_SRC_CAST (bsrc);
  bclass = GST_CAMERA_SRC_GET_CLASS (src);

  switch (GST_QUERY_TYPE (query)) {
    case GST_QUERY_LATENCY:{
      GstClockTime max_latency;

      /* device must be open */
      if (!bclass->is_open (src)) {
        GST_WARNING_OBJECT (src, "Can't give latency since device isn't open !");
        goto done;
      }

      /* we must have a framerate */
      if (src->fps_n <= 0 || src->fps_d <= 0) {
        GST_WARNING_OBJECT (src,
            "Can't give latency since framerate isn't fixated !");
        goto done;
      }

      num_buffers = bclass->get_num_buffers (src);
      /* max latency is total duration of the frame buffer */
      max_latency = num_buffers * src->duration;

      GST_DEBUG_OBJECT (bsrc,
          "report latency min %" GST_TIME_FORMAT " max %" GST_TIME_FORMAT,
          GST_TIME_ARGS (src->duration), GST_TIME_ARGS (max_latency));

      /* we are always live, the min latency is 1 frame and the max latency is
       * the complete buffer of frames. */
      gst_query_set_latency (query, TRUE, src->duration, max_latency);

      res = TRUE;
      break;
    }
    default:
      res = GST_BASE_SRC_CLASS (parent_class)->query (bsrc, query);
      break;
  }

done:

  return res;
}

static gboolean
gst_camerasrc_start (GstBaseSrc * src)
{
//   GstCameraSrc *camerasrc = GST_CAMERA_SRC_CAST (src);

  return TRUE;
}

/*
 */
static gboolean
gst_camerasrc_stop (GstBaseSrc * src)
{
  GstCameraSrc *camerasrc;
  GstCameraSrcClass *bclass;

  camerasrc = GST_CAMERA_SRC_CAST (src);
  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

  GST_DEBUG_OBJECT (camerasrc, "Stopping");

  bclass->stop (camerasrc);

  camerasrc->photo_capture_phase = GST_CAMERA_VIEWFINDER;

  return TRUE;
}

static gboolean
gst_camerasrc_unlock (GstBaseSrc * src)
{
  GstCameraSrcClass *pclass;
  GstCameraSrc *camerasrc;
  gboolean ret = TRUE;

  camerasrc = GST_CAMERA_SRC_CAST (src);
  pclass = GST_CAMERA_SRC_GET_CLASS (src);

  GST_DEBUG_OBJECT (camerasrc, "Unlock");

  if (pclass->unlock)
    ret = pclass->unlock (camerasrc);

  return ret;
}

static gboolean
gst_camerasrc_unlock_stop (GstBaseSrc * src)
{
  GstCameraSrcClass *pclass;
  GstCameraSrc *camerasrc;
  gboolean ret = TRUE;

  camerasrc = GST_CAMERA_SRC_CAST (src);
  pclass = GST_CAMERA_SRC_GET_CLASS (src);

  GST_DEBUG_OBJECT (camerasrc, "Unlock stop");

  if (pclass->unlock_stop)
    ret = pclass->unlock_stop (camerasrc);

  return ret;
}


/*
 * gst_camerasrc_send_image_tags:
 * @camerasrc: #GstCameraSrc object.
 *
 */
static gboolean
gst_camerasrc_send_image_tags (GstCameraSrc *camerasrc)
{
  GstCameraSrcClass *bclass;
  GstEvent *tagevent;
  GstTagList *tlist;
  guint wbalance;

  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

  tlist = gst_tag_list_new ();
  wbalance =
    (camerasrc->photoconf.wb_mode == GST_PHOTOGRAPHY_WB_MODE_AUTO) ? 0 : 1;

  /* Add all possible tags here */
  gst_tag_list_add (tlist, GST_TAG_MERGE_REPLACE,
      GST_TAG_CAPTURING_DIGITAL_ZOOM_RATIO,     /* G_TYPE_DOUBLE */
          camerasrc->photoconf.zoom,
      GST_TAG_CAPTURING_ISO_SPEED,              /* G_TYPE_INT */
          camerasrc->photoconf.iso_speed,
      GST_TAG_CAPTURING_WHITE_BALANCE,          /* G_TYPE_STRING */
          wbalance == 0 ? "auto" : "manual",
      GST_TAG_CAPTURING_EXPOSURE_PROGRAM,       /* G_TYPE_STRING */
          "undefined",
      GST_TAG_CAPTURING_EXPOSURE_COMPENSATION,  /* G_TYPE_DOUBLE */
          camerasrc->photoconf.ev_compensation,
      NULL);

  if (bclass->fill_image_tags)
    bclass->fill_image_tags (camerasrc, tlist);

  tagevent = gst_event_new_tag (gst_tag_list_copy (tlist));
  gst_pad_push_event (GST_BASE_SRC_PAD (camerasrc), tagevent);
  GST_DEBUG_OBJECT (camerasrc, "image tags sent: %" GST_PTR_FORMAT, tlist);
  gst_tag_list_free (tlist);

  return TRUE;
}


/*
 * gst_camerasrc_set_capture_caps:
 * @camerasrc: #GstCameraSrc object.
 *
 * Set the capture caps on element's src pad.
 *
 * Returns: TRUE on success.
 */
static gboolean
gst_camerasrc_set_capture_caps (GstCameraSrc *camerasrc)
{
  GstCameraSrcClass *bclass;
  GstCaps *newcaps = NULL;
  gboolean ret = TRUE;

  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

  GST_DEBUG_OBJECT (camerasrc, "Calling set_capture()");

  /* Try what resolution the subclass can capture */
  ret = bclass->set_capture (camerasrc,
      GST_PHOTOGRAPHY_OPERATION_MODE_IMAGE_CAPTURE, TRUE,
      &camerasrc->capture_fourcc, &camerasrc->capture_w, &camerasrc->capture_h,
      &camerasrc->capture_fps_n, &camerasrc->capture_fps_d);

  if (!ret) {
    goto done;
  }

  /* FIXME: FPS definition should be removed from capture caps */
  newcaps = gst_camerasrc_get_caps_from_info (camerasrc,
      camerasrc->capture_fourcc, camerasrc->capture_w, camerasrc->capture_h,
      camerasrc->capture_fps_n, camerasrc->capture_fps_d);

  GST_DEBUG_OBJECT (camerasrc, "Set capture caps: %" GST_PTR_FORMAT, newcaps);

  /* Notify application that we are ready now. This must be called before */
  /* the set_caps() call below, since application needs to know the new caps */
  /* so that it can adjust the caps filter to accept new format before */
  /* srcpad caps are actually changed */
  gst_camerasrc_photo_ready_for_capture (camerasrc, newcaps);

  camerasrc->vf_caps =
        gst_pad_get_negotiated_caps (GST_BASE_SRC_PAD (camerasrc));

  /* This causes caps nego and switches resolution to hi-res mode */
  /* FIXME: Do we even need to set this? Application has already set the
   * capsfilter */
  if (!gst_caps_is_equal (camerasrc->vf_caps, newcaps)) {
    GST_DEBUG_OBJECT (camerasrc, "Setting image capture caps");
    ret = gst_pad_set_caps (GST_BASE_SRC_PAD (camerasrc), newcaps);
    GST_DEBUG_OBJECT (camerasrc, "Setting image capture caps FINISHED");
  }
  else {
    GST_DEBUG_OBJECT (camerasrc, "Forcing the re-initialization");
    ret = gst_camerasrc_configure_device (camerasrc, &camerasrc->current_w,
        &camerasrc->current_h, &camerasrc->current_fourcc, &camerasrc->fps_n,
        &camerasrc->fps_d, newcaps);
    gst_caps_unref (camerasrc->vf_caps);
    camerasrc->vf_caps = NULL;
  }

  gst_caps_unref (newcaps);

done:

  return ret;
}


/*
 *
 */
static void
gst_camerasrc_apply_timestamp (GstCameraSrc *camerasrc, GstBuffer *buf)
{
  GstClock *clock;
  GstClockTime timestamp;

  GST_OBJECT_LOCK (camerasrc);
  if ((clock = GST_ELEMENT_CLOCK (camerasrc))) {
    /* we have a clock, get base time and ref clock */
    timestamp = GST_ELEMENT (camerasrc)->base_time;
    gst_object_ref (clock);
  } else {
    /* no clock, can't set timestamps */
    timestamp = GST_CLOCK_TIME_NONE;
  }
  GST_OBJECT_UNLOCK (camerasrc);

  if (clock) {
    /* the time now is the time of the clock minus the base time */
    timestamp = gst_clock_get_time (clock) - timestamp;
    gst_object_unref (clock);

    if (timestamp > camerasrc->duration)
      timestamp -= camerasrc->duration;
    else
      timestamp = 0;
  }

  /* we could also use the timestamp from the buffer itself */
  GST_BUFFER_TIMESTAMP (buf) = timestamp;
  GST_BUFFER_DURATION (buf) = camerasrc->duration;
}


/*
 */
static GstFlowReturn
gst_camerasrc_create (GstPushSrc * src, GstBuffer ** buf)
{
  GstCameraSrcClass *bclass;
  GstCameraSrc *camerasrc;
  GstFlowReturn ret = GST_FLOW_OK;
  GstBuffer *temp;

  camerasrc = GST_CAMERA_SRC_CAST (src);
  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

start_over:

  g_mutex_lock (camerasrc->state_lock);

  if (camerasrc->photo_capture_phase == GST_CAMERA_CAPTURE_START) {
    /* Tell subclass to stop flushing buffers */
    if (bclass->unlock_stop) {
      GST_DEBUG_OBJECT (camerasrc, "Stop flushing, capture is starting");
      bclass->unlock_stop (camerasrc);
    }

    camerasrc->photo_capture_phase = GST_CAMERA_CAPTURE;

    GST_DEBUG_OBJECT (camerasrc, "Flushing old buffers before starting HQ capture");
    gst_pad_push_event (GST_BASE_SRC_PAD (camerasrc), gst_event_new_flush_start ());
    gst_pad_push_event (GST_BASE_SRC_PAD (camerasrc), gst_event_new_flush_stop ());

    if (!gst_camerasrc_set_capture_caps (camerasrc)) {
      goto hq_capture_failed;
    }
  }

  else if (camerasrc->photo_capture_phase == GST_CAMERA_CAPTURE_DONE) {

    gboolean ret;

    camerasrc->photo_capture_phase = GST_CAMERA_VIEWFINDER;
    camerasrc->requested_af_mode = AF_OFF_REQUESTED;

    GST_DEBUG_OBJECT (camerasrc, "capture done. switching to viewfinder");

    /* Set the normal viewfinder resolution back */
    if (camerasrc->vf_caps) {
      GST_DEBUG_OBJECT (camerasrc, "set VF caps");
      ret = gst_pad_set_caps (GST_BASE_SRC_PAD (camerasrc), camerasrc->vf_caps);
      gst_caps_unref (camerasrc->vf_caps);
      camerasrc->vf_caps = NULL;
    }
    else {
      GstCaps *tmp;

      tmp = gst_pad_get_negotiated_caps (GST_BASE_SRC_PAD (camerasrc));

      /* Reconfigure the device to run viewfinder again */
      ret = gst_camerasrc_configure_device (camerasrc, &camerasrc->current_w,
        &camerasrc->current_h, &camerasrc->current_fourcc, &camerasrc->fps_n,
        &camerasrc->fps_d, tmp);

        gst_caps_unref (tmp);
    }

    if (!ret) {
      GST_WARNING_OBJECT (camerasrc, "Reinitializing viewfinder failed");
      g_mutex_unlock (camerasrc->state_lock);
      return GST_FLOW_ERROR;
    }

    GST_DEBUG_OBJECT (camerasrc, "viewfinder running");
  }

  g_mutex_lock (camerasrc->af_lock);

  /* Handle AF requests only in VIEWFINDER and AUTOFOCUS states */
  if ((camerasrc->photo_capture_phase == GST_CAMERA_VIEWFINDER ||
    camerasrc->photo_capture_phase == GST_CAMERA_AUTOFOCUS ||
    camerasrc->photo_capture_phase == GST_CAMERA_AUTOFOCUS_DONE) &&
    camerasrc->requested_af_mode != AF_NONE_REQUESTED)
  {
    if (camerasrc->requested_af_mode == AF_ON_REQUESTED) {
      gboolean ret;

      /* In still capture mode AE will be locked during AF operation */
      if (camerasrc->viewfinder_mode == GST_CAMERA_SRC_VIEWFINDER_MODE_STILL) {
        bclass->set_autoexposure (camerasrc, FALSE);
      }
      ret = bclass->set_autofocus (camerasrc, TRUE);

      if (ret) {
        camerasrc->photo_capture_phase = GST_CAMERA_AUTOFOCUS;
      } else {
        /* Starting AF failed, so start AE again */
        bclass->set_autoexposure (camerasrc, TRUE);
      }
    }
    else {
      bclass->set_autofocus (camerasrc, FALSE);
      bclass->set_autoexposure (camerasrc, TRUE);
      camerasrc->photo_capture_phase = GST_CAMERA_VIEWFINDER;
    }

    camerasrc->requested_af_mode = AF_NONE_REQUESTED;
  }

  g_mutex_unlock (camerasrc->af_lock);
  g_mutex_unlock (camerasrc->state_lock);

  ret = bclass->grab_frame (camerasrc, &temp, camerasrc->photo_capture_phase);

  g_mutex_lock (camerasrc->state_lock);

  if (ret != GST_FLOW_OK) {
    /* _prepare_for_capture() may have interrupted frame grabbing. */
    if (ret == GST_FLOW_WRONG_STATE &&
        camerasrc->photo_capture_phase == GST_CAMERA_CAPTURE_START)
    {
      g_mutex_unlock (camerasrc->state_lock);
      ret = GST_FLOW_OK;
      goto start_over;
    } else {
      g_mutex_unlock (camerasrc->state_lock);
      goto leave;
    }
  }

  *buf = temp;

  /* Post-capture phase */

  if (camerasrc->photo_capture_phase == GST_CAMERA_CAPTURE) {
    GstCaps *src_caps;

    src_caps = gst_pad_get_negotiated_caps (GST_BASE_SRC_PAD (camerasrc));
    gst_buffer_set_caps (*buf, src_caps);
    gst_caps_unref (src_caps);

    /* Restore the original number of buffers after capture is done */
    /* FIXME: Commented out */
//     camerasrc->num_buffers = tmp_num_buffers;

    gst_camerasrc_send_image_tags (camerasrc);

    camerasrc->photo_capture_phase = GST_CAMERA_CAPTURE_DONE;
  }

  if (GST_BUFFER_TIMESTAMP (*buf) == GST_CLOCK_TIME_NONE) {
    gst_camerasrc_apply_timestamp (camerasrc, *buf);
  }

done:

  g_mutex_unlock (camerasrc->state_lock);

leave:
  return ret;

  /* ERRORS */
hq_capture_failed:
  GST_ELEMENT_ERROR (camerasrc, RESOURCE, READ,
                     ("Error during HQ capture"), (NULL));
  ret = GST_FLOW_ERROR;
  goto done;
}


static GstStateChangeReturn
gst_camerasrc_change_state (GstElement * element, GstStateChange transition)
{
  GstCameraSrcClass *bclass;
  GstCameraSrc *camerasrc;
  GstStateChangeReturn ret = GST_STATE_CHANGE_SUCCESS;

  camerasrc = GST_CAMERA_SRC_CAST (element);
  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);

  GST_LOG_OBJECT (camerasrc, "State change: %s -> %s",
      gst_element_state_get_name (GST_STATE_TRANSITION_CURRENT (transition)),
      gst_element_state_get_name (GST_STATE_TRANSITION_NEXT (transition)));

  switch (transition) {
    case GST_STATE_CHANGE_NULL_TO_READY:
      /* Open the device */
      if (!bclass->open (camerasrc))
        return GST_STATE_CHANGE_FAILURE;
      break;
    default:
      break;
  }

  ret = GST_ELEMENT_CLASS (parent_class)->change_state (element, transition);

  switch (transition) {
    case GST_STATE_CHANGE_READY_TO_NULL:
      if (bclass->is_open (camerasrc)) {
        bclass->close (camerasrc);
      }
      camerasrc->capture_w = camerasrc->preview_w = camerasrc->current_w = 0;
      camerasrc->capture_h = camerasrc->preview_h = camerasrc->current_h = 0;
      camerasrc->capture_fourcc = camerasrc->preview_fourcc = 0;
      camerasrc->current_fourcc = 0;
      camerasrc->fps_d = camerasrc->fps_n = 0;
      camerasrc->capture_resolution_set = FALSE;
      camerasrc->preview_resolution_set = FALSE;

      /* Notify that preview caps have changed (to NULL) */
      g_object_notify (G_OBJECT (camerasrc),
          GST_PHOTOGRAPHY_PROP_IMAGE_PREVIEW_SUPPORTED_CAPS);

      break;
    default:
      break;
  }

  return ret;
}


/**
 * gst_camerasrc_add_color_channel:
 * @camsrc: #GstCameraSrc object.
 * @channel: #GstColorBalanceChannel object.
 *
 * Add a new color channel to list.
 */
void
gst_camerasrc_add_color_channel (GstCameraSrc *camsrc,
    GstColorBalanceChannel *channel)
{
  camsrc->colors = g_list_append (camsrc->colors, (gpointer) channel);
}


/**
 * gst_camerasrc_clear_color_channels:
 * @camsrc: #GstCameraSrc object.
 *
 * Delete all color channels.
 */
void gst_camerasrc_clear_color_channels (GstCameraSrc *camsrc)
{
  g_list_foreach (camsrc->colors, (GFunc) g_object_unref, NULL);
  g_list_free (camsrc->colors);
  camsrc->colors = NULL;
}


/**
 * gst_camerasrc_send_capture_start_message:
 * @camsrc: GstCameraSrc object
 *
 * Sends a GstMessage notification to GstBus that capture operation is
 * about to start.
 */
void
gst_camerasrc_send_capture_start_message (GstCameraSrc *camsrc)
{
  GstStructure *s;
  GstMessage *msg;

  GST_DEBUG_OBJECT (camsrc, "Sending capture-start message");

  /* Send custom GstMessage "photo-capture-start" */
  s = gst_structure_new (GST_CAMERA_SRC_CAPTURE_START, NULL);
  msg = gst_message_new_element (GST_OBJECT (camsrc), s);

  if (gst_element_post_message (GST_ELEMENT (camsrc), msg) == FALSE) {
    GST_WARNING ("This element has no bus, therefore no message sent!");
  }

  GST_LOG_OBJECT (camsrc, "Capture-start message sent");
}


/**
 * gst_camerasrc_send_capture_stop_message:
 * @camsrc: GstCameraSrc object
 *
 * Sends a GstMessage notification to GstBus that capture operation has
 * just finished.
 */
void
gst_camerasrc_send_capture_stop_message (GstCameraSrc *camsrc)
{
  GstStructure *s;
  GstMessage *msg;

  GST_DEBUG_OBJECT (camsrc, "Sending capture-stop message");

  /* Send custom GstMessage "photo-capture-end" */
  s = gst_structure_new (GST_CAMERA_SRC_CAPTURE_END, NULL);
  msg = gst_message_new_element (GST_OBJECT (camsrc), s);

  if (gst_element_post_message (GST_ELEMENT (camsrc), msg) == FALSE) {
    GST_WARNING ("This element has no bus, therefore no message sent!");
  }

  GST_LOG_OBJECT (camsrc, "Capture-stop message sent");
}


/**
 * gst_camerasrc_send_preview:
 * @camsrc: #GstCameraSrc object
 * @buf: GstBuffer that contains the preview image
 *
 * Sends HQ image preview image (snapshot) as a GstMessage.
 */
void
gst_camerasrc_send_preview (GstCameraSrc *camsrc, GstBuffer *buf)
{
  GstCameraSrcClass *bclass;
  GstStructure *msg_s = NULL;
  GstMessage *msg = NULL;

  g_return_if_fail (buf != NULL);

  bclass = GST_CAMERA_SRC_GET_CLASS (camsrc);

  GST_DEBUG_OBJECT (camsrc, "Sending preview image");

  if (GST_BUFFER_CAPS (buf) == NULL) {
    GstCaps *prvcaps = NULL;

    prvcaps = gst_caps_new_simple ("video/x-raw-yuv",
        "format", GST_TYPE_FOURCC, camsrc->preview_fourcc,
        "width", G_TYPE_INT, camsrc->preview_w,
        "height", G_TYPE_INT, camsrc->preview_h,
        "framerate", GST_TYPE_FRACTION, 1, 1,
        NULL);

    gst_buffer_set_caps (buf, prvcaps);
    gst_caps_unref (prvcaps);
  }

  /* Send custom preview image GstMessage */
  msg_s = gst_structure_new (GST_CAMERA_SRC_PREVIEW_IMAGE, "buffer",
                             GST_TYPE_BUFFER, buf, NULL);

  msg = gst_message_new_element (GST_OBJECT (camsrc), msg_s);

  if (gst_element_post_message (GST_ELEMENT (camsrc), msg) == FALSE) {
    GST_WARNING ("This element has no bus, therefore no message sent!");
  }
}


/**
 * gst_camerasrc_af_update:
 * @camsrc: #GstCameraSrc object.
 * @fs: #GstCameraFocusStatus structure. Owned by caller.
 *
 * Tell GstCameraSrc that previously started autofocus operation has finished.
 */
void
gst_camerasrc_af_update (GstCameraSrc *camsrc, GstCameraFocusStatus *fs)
{
  GstCameraSrcClass *bclass;

  bclass = GST_CAMERA_SRC_GET_CLASS (camsrc);

  g_mutex_lock (camsrc->state_lock);
  g_mutex_lock (camsrc->af_lock);

  if (fs->status == GST_PHOTOGRAPHY_FOCUS_STATUS_SUCCESS ||
      fs->status == GST_PHOTOGRAPHY_FOCUS_STATUS_FAIL)
  {
    GstStructure *s;
    GstMessage *msg;

    GST_DEBUG_OBJECT (camsrc, "autofocusing ended");

    /* Send custom GstMessage "autofocus-done" */
    s = gst_structure_new (GST_PHOTOGRAPHY_AUTOFOCUS_DONE,
                           "status", G_TYPE_INT, fs->status,
                           NULL);

    /* If autofocus succeeded, send the bitmask that defines focused
     * windows too */
    if (fs->status == GST_PHOTOGRAPHY_FOCUS_STATUS_SUCCESS) {
      GArray *windows;
      guint i;

      windows = g_array_new (FALSE, FALSE, sizeof (gint));
      for (i=0; i < fs->num_windows; i++) {
        GST_DEBUG_OBJECT (camsrc, "focus window: %d", fs->windows[i]);
        g_array_append_val (windows, fs->windows[i]);
      }

      gst_structure_set (s,
                         "focus-windows", G_TYPE_ARRAY, windows,
                         "focus-window-rows", G_TYPE_INT, fs->focus_rows,
                         "focus-window-columns", G_TYPE_INT, fs->focus_columns,
                         "focus-window-coverage", G_TYPE_INT, fs->coverage,
                         NULL);

      GST_DEBUG_OBJECT (camsrc, "focus rows: %d", fs->focus_rows);
      GST_DEBUG_OBJECT (camsrc, "focus columns: %d", fs->focus_columns);
      GST_DEBUG_OBJECT (camsrc, "focus coverage: %d", fs->coverage);
    }

    msg = gst_message_new_element (GST_OBJECT (camsrc), s);

    if (gst_element_post_message (GST_ELEMENT (camsrc), msg) == FALSE) {
      GST_WARNING ("This element has no bus, therefore no message sent!");
    }

    /* In still capture mode we don't turn off AF algorithm yet, since it  */
    /* would enable CAF. Instead, it is turned off when application        */
    /* explicitly calls set_autofocus (FALSE), which in turn raises        */
    /* af_requested = OFF flag and AF is finally stopped.                  */

    /* In video capture mode AF will be stopped immediately to enable AE   */
    if (camsrc->viewfinder_mode == GST_CAMERA_SRC_VIEWFINDER_MODE_VIDEO) {
      bclass->set_autofocus (camsrc, FALSE);
    }

    /* We don't turn on autoexposure here either. This way AE stays        */
    /* "locked" until application explicitly calls set_autofocus (FALSE).  */

    camsrc->photo_capture_phase = GST_CAMERA_AUTOFOCUS_DONE;
  }

  g_mutex_unlock (camsrc->af_lock);
  g_mutex_unlock (camsrc->state_lock);
}


/**
 * gst_camerasrc_caf_update:
 * @camsrc: GstCameraSrc object.
 * @fs: #GstCameraFocusStatus structure. Owned by caller.
 *
 * Tell GstCameraSrc that continuous autofocus algorithm has changed its state.
 * Sends a GstMessage notification to GstBus indicating a change in
 * continuous autofocus status.
 */
void
gst_camerasrc_caf_update (GstCameraSrc *camsrc, GstCameraFocusStatus *fs)
{
  GstStructure *s;
  GstMessage *msg;

  GST_DEBUG_OBJECT (camsrc, "Sending CAF status: %d", fs->status);

  /* Send custom GstMessage "caf-update" */
  s = gst_structure_new (GST_CAMERA_SRC_CAF_STATUS, NULL);
  gst_structure_set (s, "status", G_TYPE_INT, fs->status, NULL);
  msg = gst_message_new_element (GST_OBJECT (camsrc), s);

  if (gst_element_post_message (GST_ELEMENT (camsrc), msg) == FALSE) {
    GST_WARNING ("This element has no bus, therefore no message sent!");
  }

  GST_LOG_OBJECT (camsrc, "CAF update message sent");
}


/**
 * gst_camerasrc_shake_update:
 * @camsrc: GstCameraSrc object.
 * @risk: GstPhotoShakeRisk value.
 *
 * Tell GstCameraSrc that shake risk has changed. Sends a GstMessage
 * notification to GstBus indicating a change in shake risk status.
 */
void
gst_camerasrc_shake_update (GstCameraSrc *camsrc, GstPhotoShakeRisk risk)
{
  GstStructure *s;
  GstMessage *msg;

  GST_DEBUG_OBJECT (camsrc, "Sending shake risk update: %d", risk);

  /* Send custom GstMessage telling the changed shake risk level */
  s = gst_structure_new (GST_PHOTOGRAPHY_SHAKE_RISK, NULL);
  gst_structure_set (s, "shake_risk", G_TYPE_INT, risk, NULL);
  msg = gst_message_new_element (GST_OBJECT (camsrc), s);

  if (gst_element_post_message (GST_ELEMENT (camsrc), msg) == FALSE) {
    GST_WARNING ("This element has no bus, therefore no message sent!");
  }

  GST_LOG_OBJECT (camsrc, "Shake indicator message sent");
}


/**
 * gst_camerasrc_shake_update:
 * @camsrc: GstCameraSrc object.
 * @status: GstCameraFlashStatus value.
 *
 * Tell GstCameraSrc that flash status has changed. Sends a GstMessage
 * notification to GstBus indicating a change in flash status.
 */
void
gst_camerasrc_flash_status_update (GstCameraSrc *camsrc,
    GstCameraFlashStatus status)
{
  GstStructure *s;
  GstMessage *msg;

  GST_DEBUG_OBJECT (camsrc, "Sending flash status update: %d", status);

  /* Send custom GstMessage telling the changed shake risk level */
  s = gst_structure_new (GST_CAMERA_SRC_FLASH_STATUS, NULL);
  gst_structure_set (s, "status", G_TYPE_INT, status, NULL);
  msg = gst_message_new_element (GST_OBJECT (camsrc), s);

  if (gst_element_post_message (GST_ELEMENT (camsrc), msg) == FALSE) {
    GST_WARNING ("This element has no bus, therefore no message sent!");
  }

  GST_LOG_OBJECT (camsrc, "Flash status update message sent");
}


/* Tag helper functions */

/**
 * gst_camerasrc_exposure_mode_from_exif_value:
 * @value: exposure mode in EXIF format.
 *
 * Convert exposure mode to string.
 *
 * Returns: String representation of the scene capture type, or NULL if invalid
 * value was given. Possible values: "auto-exposure", "manual-exposure" and
 * "auto-bracket".
 */
const gchar *
gst_camerasrc_exposure_mode_from_exif_value (gint value)
{
  switch (value) {
    case 0:
      return "auto-exposure";
    case 1:
      return "manual-exposure";
    case 2:
      return "auto-bracket";
    default:
      GST_WARNING ("Invalid exif exposure mode: %d", value);
      return NULL;
  }
}

/**
 * gst_camerasrc_scene_capture_type_from_exif_value:
 * @value: scene capture type in EXIF format.
 *
 * Convert scene capture type to string.
 *
 * Returns: String representation of the scene capture type, or NULL if invalid
 * value was given. Possible values: "standard", "landscape", "portrait" and
 * "night-scene".
 */
const gchar *
gst_camerasrc_scene_capture_type_from_exif_value (gint value)
{
  switch (value) {
    case 0:
      return "standard";
    case 1:
      return "landscape";
    case 2:
      return "portrait";
    case 3:
      return "night-scene";
    default:
      GST_WARNING ("Invalid exif scene capture type: %d", value);
      return NULL;
  }
}


/**
 * gst_camerasrc_gain_adjustment_from_exif_value:
 * @value: gain adjustment type in EXIF format.
 *
 * Convert gain adjustment type to string.
 *
 * Returns: String representation of the gain adjustment type, or NULL if
 * invalid value was given. Possible values: "none", "low-gain-up",
 * "high-gain-up", "low-gain-down" and "high-gain-down".
 */
const gchar *
gst_camerasrc_gain_adjustment_from_exif_value (gint value)
{
  switch (value) {
    case 0:
      return "none";
    case 1:
      return "low-gain-up";
    case 2:
      return "high-gain-up";
    case 3:
      return "low-gain-down";
    case 4:
      return "high-gain-down";
    default:
      GST_WARNING ("Invalid exif gain control type: %d", value);
      return NULL;
  }
}


/**
 * gst_camerasrc_contrast_from_exif_value:
 * @value: contrast type in EXIF format.
 *
 * Convert contrast type to string.
 *
 * Returns: String representation of the contrast type, or NULL if invalid
 * value was given. Possible values: "normal", "soft" and "hard".
 */
const gchar *
gst_camerasrc_contrast_from_exif_value (gint value)
{
  switch (value) {
    case 0:
      return "normal";
    case 1:
      return "soft";
    case 2:
      return "hard";
    default:
      GST_WARNING ("Invalid contrast type: %d", value);
      return NULL;
  }
}


/**
 * gst_camerasrc_saturation_from_exif_value:
 * @value: saturation type in EXIF format.
 *
 * Convert saturation type to string.
 *
 * Returns: String representation of the saturation type, or NULL if invalid
 * value was given. Possible values: "normal", "low-saturation" and
 * "high-saturation".
 */
const gchar *
gst_camerasrc_saturation_from_exif_value (gint value)
{
  switch (value) {
    case 0:
      return "normal";
    case 1:
      return "low-saturation";
    case 2:
      return "high-saturation";
    default:
      GST_WARNING ("Invalid saturation type: %d", value);
      return NULL;
  }
}


/**
 * gst_camerasrc_flash_mode_from_exif_value:
 * @value: flash EXIF field
 *
 * Convert "flash" EXIF field into text string.
 *
 * Returns: String representation of the flash mode, or NULL if invalid
 * value was given. Possible values: "always", "never" and "auto".
 */
const gchar *
gst_camerasrc_flash_mode_from_exif_value (gint value)
{
  // frame_info_t

  // bits 3 and 4 indicate the flash mode,
  //
  // Values for bits 3 and 4 indicating the camera's flash mode.
  // 00b              =   unknown
  // 01b              =   Compulsory flash firing
  // 10b              =   Compulsory flash suppression
  // 11b              =   Auto mode
  //

  /* Bit numbering in EXIF spec starts from 0 */
  value >>= 3;

  switch (value & 0x3) {
    case 1:
      return "always";
    case 2:
      return "never";
    case 3:
      return "auto";
    default:
      GST_WARNING ("Invalid flash mode type: %d", value);
      return NULL;
  }
}


/**
 * gst_camerasrc_sharpness_from_exif_value:
 * @value: sharpness type in EXIF format.
 *
 * Convert sharpness type to string.
 *
 * Returns: String representation of the sharpness type, or NULL if invalid
 * value was given. Possible values: "normal", "soft" and "hard".
 */
const gchar *
gst_camerasrc_sharpness_from_exif_value (gint value)
{
  switch (value) {
    case 0:
      return "normal";
    case 1:
      return "soft";
    case 2:
      return "hard";
    default:
      GST_WARNING ("Invalid sharpness type: %d", value);
      return NULL;
  }
}


/**
 * gst_camerasrc_metering_mode_from_exif_value:
 * @value: metering mode type in EXIF format.
 *
 * Convert metering mode type to string.
 *
 * Returns: String representation of the metering mode type, or NULL if invalid
 * value was given. Possible values: "unknown", "average",
 * "center-weighted-average", "spot", "multi-spot", "pattern", "partial" and
 * "other".
 */
const gchar *
gst_camerasrc_metering_mode_from_exif_value (gint value)
{
  switch (value) {
    case 0:
      return "unknown";
    case 1:
      return "average";
    case 2:
      return "center-weighted-average";
    case 3:
      return "spot";
    case 4:
      return "multi-spot";
    case 5:
      return "pattern";
    case 6:
      return "partial";
    case 255:
      return "other";
    default:
      GST_WARNING ("Invalid metering mode type: %d", value);
      return NULL;
  }
}


/**
 * gst_camerasrc_file_source_from_exif_value:
 * @value: FileSource type in EXIF format.
 *
 * Convert FileSource type to string.
 *
 * Returns: String representation of the FileSource type, or NULL if invalid
 * value was given. Possible values: "dsc", "transparent-scanner",
 * "reflex-scanner" and "other".
 */
const gchar *
gst_camerasrc_file_source_from_exif_value (gint value)
{
  switch (value) {
    case 0:
      return "other";
    case 1:
      return "transparent-scanner";
    case 2:
      return "reflex-scanner";
    case 3:
      return "dsc";
    default:
      GST_WARNING ("Invalid file source type: %d", value);
      return NULL;
  }
}


/* Default implementations for vmethods */

static GstPhotoCaps
gst_camerasrc_default_capabilities (GstCameraSrc *camsrc)
{
  return GST_PHOTOGRAPHY_CAPS_NONE;
}

static gboolean
gst_camerasrc_default_ret_true_with_settings (GstCameraSrc *camsrc,
                                        GstPhotoSettings *photoconf)
{
  return TRUE;
}

static gboolean
gst_camerasrc_default_write_settings (GstCameraSrc *camsrc,
                                GstPhotoSettings *photoconf,
                                gboolean scene_mode_override)
{
  return TRUE;
}

static gboolean
gst_camerasrc_default_set_onoff (GstCameraSrc *camsrc, gboolean on_off)
{
  return TRUE;
}

static GstCaps *
gst_camerasrc_default_get_caps (GstCameraSrc *camsrc, GstOperationMode mode)
{
  GST_DEBUG_OBJECT (camsrc, "Returning NULL caps for mode %d", mode);

  return NULL;
}

static void
gst_camerasrc_default_functions_init (GstCameraSrcClass *camera_class)
{
  camera_class->get_capabilities    = gst_camerasrc_default_capabilities;
  camera_class->set_autofocus       = gst_camerasrc_default_set_onoff;
  camera_class->set_autoexposure    = gst_camerasrc_default_set_onoff;
  camera_class->read_settings       = gst_camerasrc_default_ret_true_with_settings;
  camera_class->write_settings      = gst_camerasrc_default_write_settings;
  camera_class->get_supported_caps  = gst_camerasrc_default_get_caps;

  GST_DEBUG ("Default functions set");
}


static gboolean
gst_camerasrc_handle_event (GstCameraSrc *camerasrc, GstEvent *event)
{
  GstCameraSrcClass *bclass;
  gboolean ret = FALSE;

  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);
  GST_LOG_OBJECT (camerasrc, "handling %s event", GST_EVENT_TYPE_NAME (event));

  if (bclass->event) {
    ret = bclass->event (camerasrc, event);
  }

  return ret;
}


/*
 * gst_camerasrc_send_event:
 * @element: GstElement object.
 * @event: GstEvent to be handled.
 *
 * Returns: TRUE if the event was handled.
 */
static gboolean
gst_camerasrc_send_event (GstElement * element, GstEvent * event)
{
  GstCameraSrc *camerasrc;
  gboolean ret = FALSE;

  camerasrc = GST_CAMERA_SRC_CAST (element);

  GST_LOG_OBJECT (camerasrc, "got %s event", GST_EVENT_TYPE_NAME (event));

  switch (GST_EVENT_TYPE (event)) {
      case GST_EVENT_CUSTOM_UPSTREAM:
        ret = gst_camerasrc_handle_event (camerasrc, event);
        break;
    default:
      break;
  }

  if (!ret) {
    ret = GST_ELEMENT_CLASS (parent_class)->send_event (element, event);
  }

  return ret;
}


/*
 * gst_camerasrc_event:
 * @src: #GstBaseSrc object.
 * @event: #GetEvent object.
 *
 * Returns: TRUE of the event was handled.
 */
static gboolean
gst_camerasrc_event (GstBaseSrc * src, GstEvent * event)
{
  GstCameraSrc *camerasrc;
  gboolean ret;

  camerasrc = GST_CAMERA_SRC_CAST (src);
  ret = gst_camerasrc_handle_event (camerasrc, event);

  if (!ret) {
    ret = GST_BASE_SRC_CLASS (parent_class)->event (src, event);
  }

  return ret;
}


/**
 * gst_camerasrc_update_max_zoom:
 * @camerasrc: #GstCameraSrc object.
 *
 * Check and update zoom property maximum value.
 */
static void
gst_camerasrc_update_max_zoom (GstCameraSrc *camerasrc)
{
  GstCameraSrcClass *bclass;
  GObjectClass *oclass;
  GParamSpec *pspec;
  GParamSpecFloat *pspec_f;
  gfloat maxzoom = 4.0;

  oclass = G_OBJECT_GET_CLASS (camerasrc);
  bclass = GST_CAMERA_SRC_GET_CLASS (camerasrc);
  pspec = g_object_class_find_property (oclass, "zoom");

  if (bclass->get_max_zoom) {
    if (!bclass->get_max_zoom (camerasrc, &maxzoom)) {
      maxzoom = 4.0;
    }
  }

  /* Update gobject property */
  if (pspec && (G_PARAM_SPEC_VALUE_TYPE (pspec) == G_TYPE_FLOAT)) {
    pspec_f = G_PARAM_SPEC_FLOAT (pspec);
    pspec_f->maximum = maxzoom;
    GST_DEBUG_OBJECT (camerasrc, "set maximum zoom as %f", pspec_f->maximum);
    /* Check if new maximum zoom is lower than current zoom level */
    if (pspec_f->maximum < camerasrc->photoconf.zoom) {
      GST_DEBUG_OBJECT (camerasrc, "current zoom level too high: %f",
          camerasrc->photoconf.zoom);
      g_object_set (G_OBJECT (camerasrc), "zoom", pspec_f->maximum, NULL);
    }
  } else {
    GST_WARNING_OBJECT (camerasrc, "updating maximum zoom failed");
  }
}
