/* GStreamer
 *
 * Copyright (C) 2001-2002 Ronald Bultje <rbultje@ronald.bitfreak.net>
 *               2006 Edgard Lima <edgard.lima@indt.org.br>
 *               2008-2010 Nokia Corporation <multimedia@maemo.org>
 *
 * gstcamerasrc.h: Abstract camera source base class
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef __GST_CAMSRC_H__
#define __GST_CAMSRC_H__

#include <gst/gst.h>
#include <gst/base/gstpushsrc.h>
#include <gst/interfaces/photography.h>
#include <gst/interfaces/colorbalance.h>

/**
 * GST_CAMERA_SRC_MAX_SIZE:
 *
 * Maximum frame width or height size.
 */
#define GST_CAMERA_SRC_MAX_SIZE (1<<15) /* 2^15 == 32768 */

G_BEGIN_DECLS

#define GST_TYPE_CAMERA_SRC            (gst_camerasrc_get_type())
#define GST_CAMERA_SRC(obj)            (G_TYPE_CHECK_INSTANCE_CAST((obj),GST_TYPE_CAMERA_SRC,GstCameraSrc))
#define GST_CAMERA_SRC_CLASS(klass)    (G_TYPE_CHECK_CLASS_CAST((klass),GST_TYPE_CAMERA_SRC,GstCameraSrcClass))
#define GST_CAMERA_SRC_GET_CLASS(obj)  (G_TYPE_INSTANCE_GET_CLASS((obj),GST_TYPE_CAMERA_SRC,GstCameraSrcClass))
#define GST_IS_CAMERA_SRC(obj)         (G_TYPE_CHECK_INSTANCE_TYPE((obj),GST_TYPE_CAMERA_SRC))
#define GST_IS_CAMERA_SRC_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE((klass),GST_TYPE_CAMERA_SRC))

#define GST_CAMERA_SRC_CAST(obj) ((GstCameraSrc *)(obj))
#define GST_CAMERA_SRC_CLASS_CAST(klass) ((GstCameraSrcClass *)(klass))

#define GST_TYPE_CAMERA_SRC_CAPTURE_MODE (gst_camerasrc_capture_mode_get_type())
#define GST_TYPE_CAMERA_SRC_VIEWFINDER_MODE (gst_camerasrc_viewfinder_mode_get_type())

/**
 * GST_CAMERA_SRC_CAPTURE_START:
 *
 * Message that is sent when capturing process is about to start.
 */
#define GST_CAMERA_SRC_CAPTURE_START "photo-capture-start"

/**
 * GST_CAMERA_SRC_CAPTURE_END:
 *
 * Message that is sent when capturing process has finished.
 */
#define GST_CAMERA_SRC_CAPTURE_END "photo-capture-end"

/**
 * GST_CAMERA_SRC_PREVIEW_IMAGE:
 *
 * This message is used to send the preview image to application. Message will
 * contain one field called "buffer" which will hold a GstBuffer containing
 * the preview image data.
 */
#define GST_CAMERA_SRC_PREVIEW_IMAGE "photo-capture-preview"

/**
 * GST_CAMERA_SRC_CAF_STATUS:
 *
 * Continuous autofocus algorithm status update message.
 */
#define GST_CAMERA_SRC_CAF_STATUS "caf-update"

/**
 * GST_CAMERA_SRC_FLASH_STATUS:
 *
 * Flash lamp status update message.
 */
#define GST_CAMERA_SRC_FLASH_STATUS "flash-status"

/**
 * GstCameraSrcCaptureMode:
 * @GST_CAMERA_SRC_CAPTURE_MODE_VIEWFINDER: Viewfinder is running.
 * @GST_CAMERA_SRC_CAPTURE_MODE_STILL: Still image capture mode.
 * @GST_CAMERA_SRC_CAPTURE_MODE_VIDEO: Video capture mode.
 *
 * Camera element capturing modes.
 */
typedef enum {
  GST_CAMERA_SRC_CAPTURE_MODE_VIEWFINDER,
  GST_CAMERA_SRC_CAPTURE_MODE_STILL,
  GST_CAMERA_SRC_CAPTURE_MODE_VIDEO
} GstCameraSrcCaptureMode;


/**
 * GstCameraSrcViewfinderMode:
 * @GST_CAMERA_SRC_VIEWFINDER_MODE_STILL: Still image capture VF mode.
 * @GST_CAMERA_SRC_VIEWFINDER_MODE_VIDEO: Video capture VF mode.
 *
 * Defines in which mode viewfinder should be run.
 */
typedef enum {
  GST_CAMERA_SRC_VIEWFINDER_MODE_STILL,
  GST_CAMERA_SRC_VIEWFINDER_MODE_VIDEO
} GstCameraSrcViewfinderMode;


/**
 * GstCameraSrcAFReqMode:
 * @AF_NONE_REQUESTED: No change to AF mode requested.
 * @AF_OFF_REQUESTED: AF should be turned off.
 * @AF_ON_REQUESTED: AF should be turned on.
 *
 * Requested autofocus status.
 */
typedef enum {
  AF_NONE_REQUESTED = 0,
  AF_OFF_REQUESTED,
  AF_ON_REQUESTED
} GstCameraSrcAFReqMode;


/**
 * GstCameraFlashStatus:
 * @GST_CAMERA_FLASH_STATUS_READY: Flash is ready for use.
 * @GST_CAMERA_FLASH_STATUS_COOLDOWN: Flash is cooling down.
 *
 * Flash lamp status.
 */
typedef enum {
  GST_CAMERA_FLASH_STATUS_READY = 0,
  GST_CAMERA_FLASH_STATUS_COOLDOWN
} GstCameraFlashStatus;


/**
 * GstCameraCapturePhase:
 * @GST_CAMERA_OFF: photo capturing functionality disabled.
 * @GST_CAMERA_VIEWFINDER: element is running in viewfinder mode.
 * @GST_CAMERA_AUTOFOCUS: element is running autofocus phase.
 * @GST_CAMERA_AUTOFOCUS_DONE: autofocus phase has ended.
 * @GST_CAMERA_CAPTURE_START: element is preparing to capture HQ image.
 * @GST_CAMERA_CAPTURE: element is capturing HQ image.
 * @GST_CAMERA_CAPTURE_DONE: finishing HQ capture operation.
 *
 * Photo capturing phases.
 */
typedef enum {
  GST_CAMERA_OFF = 0,
  GST_CAMERA_VIEWFINDER,
  GST_CAMERA_AUTOFOCUS,
  GST_CAMERA_AUTOFOCUS_DONE,
  GST_CAMERA_CAPTURE_START,
  GST_CAMERA_CAPTURE,
  GST_CAMERA_CAPTURE_DONE
} GstCameraCapturePhase;


typedef struct _GstCameraSrc GstCameraSrc;
typedef struct _GstCameraSrcClass GstCameraSrcClass;
typedef struct _GstCameraFocusStatus GstCameraFocusStatus;

/**
 * GstCameraSrc:
 * @element: the parent element.
 *
 * Opaque #GstCameraSrc object.
 */
struct _GstCameraSrc {
  GstPushSrc element;

  /*< private >*/
  GstCaps *vf_caps;
  GList *colors;

  /* TRUE if outgoing buffer should be always duplicated */
  gboolean always_copy;

  /* Current video device format */
  guint   current_w, current_h;   /* current capture frame size */
  guint   fps_d, fps_n;           /* framerate if device is open */
  guint32 current_fourcc;         /* current color format */
  GstClockTime duration;          /* duration of one frame */

  /* Photo interface -related */
  GstPhotoSettings        photoconf;    /* Caches the settings */
  GMutex                  *state_lock;
  GMutex                  *af_lock;
  GstPhotoCapturePrepared prep_func;
  gpointer                prep_udata;

  /* Image capture format */
  guint   capture_w, capture_h;
  guint   capture_fps_n, capture_fps_d;
  guint32 capture_fourcc;
  gboolean capture_resolution_set; /* Is capture resolution set already? */

  /* Preview image format */
  guint   preview_w, preview_h;
  guint32 preview_fourcc;
  gboolean preview_resolution_set; /* Is preview resolution set already? */

  GstCameraSrcCaptureMode capture_mode;
  GstCameraSrcViewfinderMode viewfinder_mode;

  /* Protected with state_lock */
  GstCameraCapturePhase photo_capture_phase;

  guint8 requested_af_mode;
};


/**
 * GstCameraSrcClass:
 * @parent_class: the parent class structure.
 * @open: Called when subclass should open the video device.
 * @is_open: Called to get information if video device is currently open.
 * @close: Called when subclass should close video device and free resources.
 * @get_attribute: Called to get V4L2 attribute from device.
 * @set_attribute: Called to set V4L2 attribute to device.
 * @set_capture: Called to set VF/capture/preview resolution & format.
 * @start: Capturing should start.
 * @is_active: Ask subclass whether capturing is ongoing.
 * @grab_frame: Ask the subclass to capture a frame.
 * @stop: Subclass should stop capturing frames from device.
 * @get_caps: Ask the subclass for supported output (viewfinder) caps.
 * @get_num_buffers: Ask the subclass how many video buffers it is using.
 * @unlock: Tell the subclass to stop any pending operation to video device.
 * @unlock_stop: Clear previous unlock request.
 * @fill_image_tags: Called after image capturing to retrieve image metadata.
 * @event: Used for passing custom events to subclass.
 * @get_capabilities: Ask what capabilities subclass has.
 * @set_vfinder_mode: Set viewfinder mode.
 * @set_capture_mode: Set capturing mode.
 * @set_autofocus: Turn on / off autofocus algorithm.
 * @set_autoexposure: Turn on / off auto exposure algorithm.
 * @write_settings: Write all GstPhotoSettings to subclass at once.
 * @read_settings: Read all device settings to given GstPhotoSettings structure.
 * @set_zoom: Set the zoom factor.
 * @get_max_zoom: Ask for maximum zoom factor that can be used.
 * @get_supported_caps: Ask subclass about supported caps for given mode.
 *
 * #GstCameraSrc class object.
 */
struct _GstCameraSrcClass
{
  GstPushSrcClass parent_class;

  /*< public >*/
 /* virtual methods for subclasses */
  gboolean (*open)              (GstCameraSrc *camsrc);

  gboolean (*is_open)           (GstCameraSrc *camsrc);

  gboolean (*close)             (GstCameraSrc *camsrc);

  /* attribute control */
  gboolean (*get_attribute)     (GstCameraSrc *camsrc, int attribute,
                                 int *value);

  gboolean (*set_attribute)     (GstCameraSrc *camsrc, int attribute,
                                 const int value);

  gboolean   (*set_capture)     (GstCameraSrc *camsrc,
                                 GstOperationMode mode,
                                 gboolean try_only,
                                 guint32 *pixelformat,
                                 guint *width, guint32 *height,
                                 guint *fps_n, guint *fps_d);

  gboolean   (*start)           (GstCameraSrc *camsrc, GstCaps *caps);

  gboolean   (*is_active)       (GstCameraSrc *camsrc);

  GstFlowReturn (*grab_frame)   (GstCameraSrc *camsrc, GstBuffer **buf,
                                 GstCameraCapturePhase phase);

  gboolean   (*stop)            (GstCameraSrc *camsrc);

  GstCaps*   (*get_caps)        (GstCameraSrc *camsrc);

  guint      (*get_num_buffers) (GstCameraSrc *camsrc);

  gboolean   (*unlock)          (GstCameraSrc *camsrc);

  gboolean   (*unlock_stop)     (GstCameraSrc *camsrc);

  gboolean   (*fill_image_tags) (GstCameraSrc *camsrc, GstTagList *tlist);

  gboolean   (*event)           (GstCameraSrc *camsrc, GstEvent *event);

  /* FORMER DRIVER-API */

  GstPhotoCaps
  (*get_capabilities)  (GstCameraSrc *camsrc);

  gboolean
  (*set_vfinder_mode)  (GstCameraSrc *camsrc,
                        GstCameraSrcViewfinderMode mode);

  gboolean
  (*set_capture_mode)  (GstCameraSrc *camsrc,
                        GstCameraSrcCaptureMode mode);

  gboolean
  (*set_autofocus)     (GstCameraSrc *camsrc, gboolean on_off);

  gboolean
  (*set_autoexposure)  (GstCameraSrc *camsrc, gboolean on_off);

  gboolean
  (*write_settings)    (GstCameraSrc *camsrc,
                        GstPhotoSettings *photoconf,
                        gboolean scene_mode_override);

  gboolean
  (*read_settings)     (GstCameraSrc *camsrc, GstPhotoSettings *photoconf);

  gboolean
  (*set_zoom)          (GstCameraSrc *camsrc, gfloat zoomfactor);

  gboolean
  (*get_max_zoom)     (GstCameraSrc *camsrc, gfloat *maxzoom);

  GstCaps *
  (*get_supported_caps)  (GstCameraSrc *camsrc, GstOperationMode mode);

  /*< private >*/
  gpointer       _gst_reserved[GST_PADDING_LARGE];
};


/**
 * GstCameraFocusStatus:
 * @status: GstFocusStatus indicating the status of focus operation.
 * @windows: Table of integers defining the focused windows.
 * @num_windows: Number of items in the @windows table.
 * @focus_rows: Number of rows in window matrix.
 * @focus_columns: Number of columns in window matrix.
 * @coverage: Percentage of viewfinder area being used for focusing.
 */
struct _GstCameraFocusStatus
{
  GstFocusStatus status;
  guint          *windows;
  guint          num_windows;
  guint8         focus_rows;
  guint8         focus_columns;
  guint          coverage;
};


GType gst_camerasrc_get_type (void);

void gst_camerasrc_add_color_channel (GstCameraSrc *camsrc,
                                      GstColorBalanceChannel *channel);

void gst_camerasrc_clear_color_channels (GstCameraSrc *camsrc);

void gst_camerasrc_send_capture_start_message (GstCameraSrc *camsrc);

void gst_camerasrc_send_capture_stop_message (GstCameraSrc *camsrc);

void gst_camerasrc_send_preview (GstCameraSrc *camsrc, GstBuffer *buf);

void gst_camerasrc_af_update (GstCameraSrc *camsrc, GstCameraFocusStatus *fs);

void gst_camerasrc_caf_update (GstCameraSrc *camsrc, GstCameraFocusStatus *fs);

void gst_camerasrc_shake_update (GstCameraSrc *camsrc, GstPhotoShakeRisk risk);

void gst_camerasrc_flash_status_update (GstCameraSrc *camsrc,
    GstCameraFlashStatus status);

GstCaps *
gst_camerasrc_get_caps_from_info (GstCameraSrc *camsrc, guint32 fourcc,
    guint width, guint height, guint fps_n, guint fps_d);

const gchar * gst_camerasrc_exposure_mode_from_exif_value (gint value);

const gchar * gst_camerasrc_scene_capture_type_from_exif_value (gint value);

const gchar * gst_camerasrc_gain_adjustment_from_exif_value (gint value);

const gchar * gst_camerasrc_contrast_from_exif_value (gint value);

const gchar * gst_camerasrc_saturation_from_exif_value (gint value);

const gchar * gst_camerasrc_flash_mode_from_exif_value (gint value);

const gchar * gst_camerasrc_sharpness_from_exif_value (gint value);

const gchar * gst_camerasrc_metering_mode_from_exif_value (gint value);

const gchar * gst_camerasrc_file_source_from_exif_value (gint value);

G_END_DECLS

#endif /* __GST_CAMSRC_H__ */
