/* GStreamer
 *
 * Copyright (C) 2001-2002 Ronald Bultje <rbultje@ronald.bitfreak.net>
 *               2006 Edgard Lima <edgard.lima@indt.org.br>
 *               2009 Texas Instruments, Inc - http://www.ti.com/
 *
 * bufferpool.c V4L2 buffer pool class
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <unistd.h>
#include <gst/gst.h>
#include <sys/ioctl.h>

#include "gstsubdevcamsrc.h"
#include "bufferpool.h"

GST_DEBUG_CATEGORY_EXTERN (gst_subdevsrc_debug);
#define GST_CAT_DEFAULT gst_subdevsrc_debug


/**
 * GstSubdevBuffer:
 */

static GstBufferClass *subdevbuffer_parent_class = NULL;

static void
gst_subdevsrc_buffer_finalize (GstSubdevBuffer * buffer)
{
  GstSubdevSrcBufferPool *pool;
  gboolean resuscitated = FALSE;
  gint index;

  pool = buffer->pool;
  index = buffer->vbuffer.index;

  GST_LOG ("finalizing buffer %p %d", buffer, index);

  g_mutex_lock (pool->lock);
  if (GST_BUFFER_SIZE (buffer) != 0)
    /* BUFFER_SIZE is only set if the frame was dequeued */
    pool->num_live_buffers--;

  if (pool->running) {
    if (ioctl (pool->video_fd, VIDIOC_QBUF, &buffer->vbuffer) < 0) {
      GST_WARNING ("could not requeue buffer %p %d", buffer, index);
    } else {
      /* FIXME: check that the caps didn't change */
      GST_LOG ("reviving buffer %p, %d", buffer, index);
      gst_buffer_ref (GST_BUFFER (buffer));
      GST_BUFFER_SIZE (buffer) = 0;
      pool->buffers[index] = buffer;
      pool->queued[index] = 1;
      g_cond_signal (pool->data_cond);
      resuscitated = TRUE;
    }
  } else {
    GST_LOG ("the pool is shutting down");
  }
  g_mutex_unlock (pool->lock);

  if (!resuscitated) {
    GST_LOG ("buffer %p not recovered, unmapping", buffer);
    gst_mini_object_unref (GST_MINI_OBJECT (pool));

    if (buffer->gbuffer) {
      /* It was allocated with gst_pad_alloc_buffer */
      gst_buffer_unref (buffer->gbuffer);
    } else {
      /* It was allocated with posix_memalign */
      free (GST_BUFFER_DATA (buffer));
    }
    GST_BUFFER_DATA (buffer) = NULL;

    GST_LOG ("free subdevbuffer");
    GST_MINI_OBJECT_CLASS (subdevbuffer_parent_class)->finalize (GST_MINI_OBJECT
        (buffer));
  }
}

static void
gst_subdevsrc_buffer_class_init (gpointer g_class, gpointer class_data)
{
  GstMiniObjectClass *mini_object_class = GST_MINI_OBJECT_CLASS (g_class);

  subdevbuffer_parent_class = g_type_class_peek_parent (g_class);

  mini_object_class->finalize =
      (GstMiniObjectFinalizeFunction) gst_subdevsrc_buffer_finalize;
}

GType
gst_subdevsrc_buffer_get_type (void)
{
  static GType _gst_subdevsrc_buffer_type;

  if (G_UNLIKELY (_gst_subdevsrc_buffer_type == 0)) {
    static const GTypeInfo subdevsrc_buffer_info = {
      sizeof (GstBufferClass),
      NULL,
      NULL,
      gst_subdevsrc_buffer_class_init,
      NULL,
      NULL,
      sizeof (GstSubdevBuffer),
      0,
      NULL,
      NULL
    };
    _gst_subdevsrc_buffer_type = g_type_register_static (GST_TYPE_BUFFER,
        "GstSubdevBuffer", &subdevsrc_buffer_info, 0);
  }
  return _gst_subdevsrc_buffer_type;
}

static GstSubdevBuffer *
gst_subdevsrc_buffer_new (GstSubdevSrcBufferPool * pool,
    GstSubdevSrc * subdevsrc, guint index, GstCaps * caps)
{
  GstSubdevBuffer *ret = NULL;
  GstFlowReturn flow_ret;

  ret = (GstSubdevBuffer *) gst_mini_object_new (GST_TYPE_SUBDEVSRC_BUFFER);
  GST_LOG ("creating buffer %u, %p in pool %p", index, ret, pool);
  ret->pool =
      (GstSubdevSrcBufferPool *) gst_mini_object_ref (GST_MINI_OBJECT (pool));

  ret->vbuffer.index = index;
  ret->vbuffer.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  ret->vbuffer.memory = V4L2_MEMORY_USERPTR;
  ret->vbuffer.length = subdevsrc->frame_byte_size;
  ret->gbuffer = NULL;

  if (ioctl (pool->video_fd, VIDIOC_QUERYBUF, &ret->vbuffer) < 0)
    goto querybuf_failed;

  if (gst_pad_is_linked (GST_BASE_SRC_PAD (subdevsrc))) {
    GST_LOG ("using pad_alloc, size = %d", subdevsrc->frame_byte_size);
    flow_ret = gst_pad_alloc_buffer_and_set_caps (GST_BASE_SRC_PAD (subdevsrc),
        0LL, subdevsrc->frame_byte_size, caps, &ret->gbuffer);
    if (flow_ret != GST_FLOW_OK)
      goto pad_alloc_failed;
    GST_BUFFER_DATA (ret) = ret->gbuffer->data;
  }
  else {
    void *data;

    GST_LOG ("using posix_memalign");
    if (posix_memalign (&data, getpagesize(), ret->vbuffer.length) != 0) {
      goto memalign_failed;
    }
   GST_BUFFER_DATA (ret) = (guint8 *) data;
  }

  GST_BUFFER_SIZE (ret) = ret->vbuffer.length;
  GST_BUFFER_FLAG_SET (ret, GST_BUFFER_FLAG_READONLY);
  gst_buffer_set_caps (GST_BUFFER (ret), caps);

#ifdef USE_MLOCK
  GST_DEBUG ("mlocking buffer data");
  if (mlock ((void *) GST_BUFFER_DATA (ret), subdevsrc->frame_byte_size) == -1)
    goto mlock_failed;
#endif

  /* mlocking succeeded, now we can set the pointer to vbuffer. The existence
   * of this pointer will be used later to determine if the munlock() is
   * needed */
  ret->vbuffer.m.userptr = (unsigned int) GST_BUFFER_DATA (ret);

  GST_LOG ("  index:     %u", ret->vbuffer.index);
  GST_LOG ("  type:      %d", ret->vbuffer.type);
  GST_LOG ("  bytesused: %u", ret->vbuffer.bytesused);
  GST_LOG ("  flags:     %08x", ret->vbuffer.flags);
  GST_LOG ("  field:     %d", ret->vbuffer.field);
  GST_LOG ("  memory:    %d", ret->vbuffer.memory);
  if (ret->vbuffer.memory == V4L2_MEMORY_MMAP)
    GST_LOG ("  MMAP offset:  %p", ret->vbuffer.m.offset);
  else if (ret->vbuffer.memory == V4L2_MEMORY_USERPTR)
    GST_LOG ("  user address: %p", ret->vbuffer.m.userptr);
  GST_LOG ("  length:    %u", ret->vbuffer.length);
  GST_LOG ("  input:     %u", ret->vbuffer.input);

  return ret;

  /* ERRORS */
#ifdef USE_MLOCK
mlock_failed:
  {
    GST_WARNING ("Failed to mlock memory: %s", g_strerror (errno));
    gst_buffer_unref (GST_BUFFER (ret));
    return NULL;
  }
#endif
querybuf_failed:
  {
    gint errnosave = errno;

    GST_WARNING ("Failed QUERYBUF: %s", g_strerror (errnosave));
    gst_buffer_unref (GST_BUFFER (ret));
    errno = errnosave;
    return NULL;
  }
memalign_failed:
  {
    GST_WARNING ("Failed to posix_memalign a buffer");
    return NULL;
  }
pad_alloc_failed:
  {
    GST_WARNING ("Failed to pad_alloc_buffer: %s",
        gst_flow_get_name (flow_ret));
    return NULL;
  }
}


#define GST_TYPE_SUBDEVSRC_BUFFER_POOL (gst_subdevsrc_buffer_pool_get_type())
#define GST_IS_SUBDEVSRC_BUFFER_POOL(obj) (G_TYPE_CHECK_INSTANCE_TYPE ((obj), GST_TYPE_SUBDEVSRC_BUFFER_POOL))
#define GST_SUBDEVSRC_BUFFER_POOL(obj) (G_TYPE_CHECK_INSTANCE_CAST ((obj), GST_TYPE_SUBDEVSRC_BUFFER_POOL, GstSubdevSrcBufferPool))

static GstMiniObjectClass *buffer_pool_parent_class = NULL;

static void
gst_subdevsrc_buffer_pool_finalize (GstSubdevSrcBufferPool * pool)
{
  g_mutex_free (pool->lock);
  pool->lock = NULL;

  if (pool->video_fd >= 0)
    close (pool->video_fd);

  if (pool->buffers) {
    g_free (pool->buffers);
    pool->buffers = NULL;
  }

  if (pool->queued) {
    g_free (pool->queued);
    pool->queued = NULL;
  }

  g_cond_free (pool->data_cond);
  pool->data_cond = NULL;

  GST_MINI_OBJECT_CLASS (buffer_pool_parent_class)->finalize (GST_MINI_OBJECT
      (pool));
}

/*
 */
static void
gst_subdevsrc_buffer_pool_init (GstSubdevSrcBufferPool * pool,
    gpointer g_class)
{
  pool->lock = g_mutex_new ();
  pool->running = FALSE;
  pool->num_live_buffers = 0;
  pool->data_cond = g_cond_new();
}

/*
 */
static void
gst_subdevsrc_buffer_pool_class_init (gpointer g_class, gpointer class_data)
{
  GstMiniObjectClass *mini_object_class = GST_MINI_OBJECT_CLASS (g_class);

  buffer_pool_parent_class = g_type_class_peek_parent (g_class);

  mini_object_class->finalize = (GstMiniObjectFinalizeFunction)
      gst_subdevsrc_buffer_pool_finalize;
}

/*
 */
static GType
gst_subdevsrc_buffer_pool_get_type (void)
{
  static GType _gst_subdevsrc_buffer_pool_type;

  if (G_UNLIKELY (_gst_subdevsrc_buffer_pool_type == 0)) {
    static const GTypeInfo subdevsrc_buffer_pool_info = {
      sizeof (GstBufferClass),
      NULL,
      NULL,
      gst_subdevsrc_buffer_pool_class_init,
      NULL,
      NULL,
      sizeof (GstSubdevSrcBufferPool),
      0,
      (GInstanceInitFunc) gst_subdevsrc_buffer_pool_init,
      NULL
    };
    _gst_subdevsrc_buffer_pool_type = g_type_register_static (GST_TYPE_MINI_OBJECT,
        "GstSubdevSrcBufferPool", &subdevsrc_buffer_pool_info, 0);
  }
  return _gst_subdevsrc_buffer_pool_type;
}


/**
 * gst_subdevsrc_buffer_pool_new:
 * @src: #GstSubdevSrc object.
 * @fd: file descriptor of the video device.
 * @caps: GstCaps of the buffers to be produced.
 *
 * Create a new bufferpool object.
 *
 * Returns: new #GstSubdevSrcBufferPool instance, or NULL on error.
 */
GstSubdevSrcBufferPool *
gst_subdevsrc_buffer_pool_new (GstElement *src, gint fd, GstCaps * caps)
{
  GstSubdevSrcBufferPool *pool;
  GstSubdevSrc *subdevsrc;
  gint n;

  subdevsrc = GST_SUBDEVSRC_CAST (src);

  pool = (GstSubdevSrcBufferPool *)
      gst_mini_object_new (GST_TYPE_SUBDEVSRC_BUFFER_POOL);

  pool->video_fd = dup (fd);
  if (pool->video_fd < 0)
    goto dup_failed;


  pool->buffer_count = subdevsrc->num_buffers;
  pool->buffers = g_new0 (GstSubdevBuffer *, pool->buffer_count);
  pool->queued = g_new0 (guint, pool->buffer_count);

  for (n = 0; n < pool->buffer_count; n++) {
    pool->buffers[n] = gst_subdevsrc_buffer_new (pool, subdevsrc, n, caps);
    if (!pool->buffers[n])
      goto buffer_new_failed;
  }

  pool->flushing = FALSE;

  return pool;

  /* ERRORS */
dup_failed:
  {
    gint errnosave = errno;

    gst_subdevsrc_buffer_pool_destroy (pool, src);

    errno = errnosave;

    return NULL;
  }
buffer_new_failed:
  {
    gint errnosave = errno;

    GST_LOG ("creating a new buffer failed");
    gst_mini_object_unref (GST_MINI_OBJECT (pool));

    errno = errnosave;

    return NULL;
  }
}


/**
 * gst_subdevsrc_buffer_pool_activate:
 * @pool: #GstSubdevSrcBufferPool object.
 * @src: #GstSubdevSrc element.
 *
 * Activates the pool. This queues all buffers to the device.
 *
 * Returns: TRUE on success.
 */
gboolean
gst_subdevsrc_buffer_pool_activate (GstSubdevSrcBufferPool * pool,
    GstElement * src)
{
  GstSubdevSrc *subdevsrc;
  gint n;

  subdevsrc = GST_SUBDEVSRC_CAST (src);

  g_mutex_lock (pool->lock);

  for (n = 0; n < pool->buffer_count; n++) {

    struct v4l2_buffer *buf;

    buf = &pool->buffers[n]->vbuffer;

    GST_LOG ("enqueue pool buffer %d, address %p", n, buf->m.userptr);

    if (ioctl (pool->video_fd, VIDIOC_QBUF, buf) < 0)
      goto queue_failed;

    pool->queued[n] = 1;
  }
  pool->running = TRUE;

  g_mutex_unlock (pool->lock);

  return TRUE;

  /* ERRORS */
queue_failed:
  {
    g_mutex_unlock (pool->lock);
    return FALSE;
  }
}


/**
 * gst_subdevsrc_buffer_pool_update:
 * @pool: #GstSubdevSrcBufferPool to be checked.
 *
 * Check that at least one buffer is queued to video node. If all buffers
 * are dequeued, this function will wait until one gets queued, or streaming
 * is stopping.
 *
 * Returns: TRUE if buffer was found in device queue, FALSE otherwise.
 */
gboolean
gst_subdevsrc_buffer_pool_update (GstSubdevSrcBufferPool * pool)
{
  gboolean ret = FALSE;
  gint n;

  if (!pool) return FALSE;

  g_mutex_lock (pool->lock);

  if (pool->flushing) {
    g_mutex_unlock (pool->lock);
    return FALSE;
  }

  for (n = 0; n < pool->buffer_count; n++) {
    if (pool->queued[n]) {
      /* This buffer is queued to video output node. We can leave now */
      g_mutex_unlock (pool->lock);
      return TRUE;
    }
  }

  /* We need a reference, so that another thread can't destroy the pool
   * while we are waiting for a buffer to appear */
  gst_mini_object_ref (GST_MINI_OBJECT (pool));

  /* Wait for buffer or shutdown */
  GST_LOG ("no free buffers available, waiting");
  g_cond_wait (pool->data_cond, pool->lock);
  ret = pool->running;

  g_mutex_unlock (pool->lock);

  gst_mini_object_unref (GST_MINI_OBJECT (pool));

  return ret;
}


/**
 * gst_subdevsrc_buffer_pool_destroy:
 * @pool: #GstSubdevSrcBufferPool to be destroyed.
 * @src: #GstSubdevSrc object.
 *
 * Destroys the given pool and frees the buffers.
 */
void
gst_subdevsrc_buffer_pool_destroy (GstSubdevSrcBufferPool * pool,
    GstElement *src)
{
  GstSubdevSrc *subdevsrc;
  gint n;

  subdevsrc = GST_SUBDEVSRC_CAST (src);

  g_mutex_lock (pool->lock);
  pool->running = FALSE;

  /* Someone might be waiting in _pool_update() */
  g_cond_signal (pool->data_cond);
  g_mutex_unlock (pool->lock);

  GST_DEBUG ("destroy pool");

  /* after this point, no more buffers will be queued or dequeued; no buffer
   * from pool->buffers that is NULL will be set to a buffer, and no buffer that
   * is not NULL will be pushed out. */

  /* miniobjects have no dispose, so they can't break ref-cycles, as buffers ref
   * the pool, we need to unref the buffer to properly finalize te pool */
  for (n = 0; n < pool->buffer_count; n++) {
    GstBuffer *buf;

    g_mutex_lock (pool->lock);
    buf = GST_BUFFER (pool->buffers[n]);
    g_mutex_unlock (pool->lock);

    if (buf) {
#ifdef USE_MLOCK
      if (pool->buffers[n].m.userptr) {
        GST_DEBUG ("munlocking buffer data");
        munlock ((void *)pool->buffers[n].m.userptr,
            subdevsrc->frame_byte_size);
      }
#endif
      /* we own the ref if the buffer is in pool->buffers; drop it. */
      gst_buffer_unref (buf);
    }
  }

  gst_mini_object_unref (GST_MINI_OBJECT (pool));
}
