/* GStreamer MediaController camera source
 * Copyright (C) 2010 Nokia Corporation <multimedia@maemo.org>
 *
 * Contact: Maemo Multimedia <multimedia@maemo.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <gst/gst.h>

#include "mediacontroller.h"
#include "configparser.h"

GST_DEBUG_CATEGORY_EXTERN (gst_subdevsrc_debug);
#define GST_CAT_DEFAULT gst_subdevsrc_debug

/*
 *
 */
static gboolean
_gst_subdevsrc_parser_pipeline (GstConfigParser *cp, gchar *pipeline,
    GstPipeConfiguration *conf)
{
  gboolean ret =  FALSE;

  if (g_strrstr (pipeline, "!")) {
    GstPipeConfigurationEntry *entry;
    gchar **vect;
    gchar *tmp, *token;
    gint i;

    vect = g_strsplit (pipeline, "!", 0);
    for (i = 0; (tmp = vect[i]) != NULL ; i++) {
      token = g_strstrip (tmp);
      entry = g_slice_new0 (GstPipeConfigurationEntry);

      if (!strncmp (token, "V4L2_MBUS_FMT", 13)) {
        GST_DEBUG ("    Item: %s (pixfmt)", token);
        entry->type = GST_PIPE_ENTRY_CHANGE_PIXFMT;
        entry->field.pcode = gst_subdevsrc_mc_map_pixfmt (token);
      }
      else if (!strncmp (token, "OUT_FORMAT", 10)) {
        GST_DEBUG ("    Item: %s (format cmd)", token);
        entry->type = GST_PIPE_ENTRY_SET_OUT_PIXFMT;
      }
      else if (!strncmp (token, "OUT_SIZE", 8)) {
        GST_DEBUG ("    Item: %s (size cmd)", token);
        entry->type = GST_PIPE_ENTRY_SET_OUT_RESOLUTION;
      }
      else {
        GST_DEBUG ("    Item: %s (entity)", token);
        entry->type = GST_PIPE_ENTRY_ENTITY;
        entry->field.entity_name = g_strdup (token);
      }
      conf->entries = g_slist_prepend (conf->entries, entry);
    }

    conf->entries = g_slist_reverse (conf->entries);
    g_strfreev (vect);
    ret = TRUE;
  }
  else {
    GST_WARNING ("Badly formatted pipeline");
  }

  return ret;
}


/*
 *
 */
static gboolean
_gst_subdevsrc_parser_primary (GstConfigParser *cp, gchar *str)
{
  GST_DEBUG ("Primary pipeline: %s", str);
  cp->primary = g_slice_new0 (GstPipeConfiguration);
  return _gst_subdevsrc_parser_pipeline (cp, str, cp->primary);
}


/*
 *
 */
static gboolean
_gst_subdevsrc_parser_secondary (GstConfigParser *cp, gchar *str)
{
  GST_DEBUG ("Secondary pipeline: %s", str);
  cp->secondary = g_slice_new0 (GstPipeConfiguration);
  return _gst_subdevsrc_parser_pipeline (cp, str, cp->secondary);
}


/*
 *
 */
static gboolean
_gst_subdevsrc_parser_assignment (GstConfigParser *cp, gchar *str)
{
  GstConfigAssignment *am = NULL;
  gboolean ret = FALSE;

  if (g_strrstr (str, "=")) {
    gchar **vect;

    am = g_slice_new0 (GstConfigAssignment);

    vect = g_strsplit (str, "=", 0);
    am->key = g_strdup (g_strstrip (vect[0]));
    am->value = g_strdup (g_strstrip (vect[1]));

    GST_LOG ("Assignment: %s = %s", am->key, am->value);

    cp->assignments = g_slist_append (cp->assignments, am);
    g_strfreev (vect);
    ret = TRUE;
  }

  if (!ret) {
    GST_WARNING ("Failed to parse assignment: %s", str);
  }

  return ret;
}


/*
 *
 */
static gboolean
_gst_subdevsrc_parser_line (GstConfigParser *cp, gchar *str)
{
  gboolean ret = FALSE;
  gchar *line;

  line = g_strstrip (str);

  if (strlen (line) == 0) {
    GST_LOG ("Empty line");
    ret = TRUE;
  }
  else if (line[0] == '#') {
    GST_LOG ("Comment line");
    ret = TRUE;
  }
  else if (!strncmp (line, "primary:", 8)) {
    ret = _gst_subdevsrc_parser_primary (cp, &line[8]);
  }
  else if (!strncmp (line, "secondary:", 10)) {
    ret = _gst_subdevsrc_parser_secondary (cp, &line[10]);
  }
  else {
    ret = _gst_subdevsrc_parser_assignment (cp, line);
  }

  return ret;
}


/*
 *
 */
static void
_gst_subdevsrc_parser_destroy_pconfiguration (GstPipeConfiguration *pipe)
{
  GSList *iter;

  if (pipe == NULL)
    return;

  for (iter = pipe->entries; iter != NULL; iter = iter->next) {
    GstPipeConfigurationEntry *e = (GstPipeConfigurationEntry *) iter->data;
    if (e->type == GST_PIPE_ENTRY_ENTITY) {
      g_free (e->field.entity_name);
    }
    g_slice_free (GstPipeConfigurationEntry, e);
  }
  g_slist_free (pipe->entries);
  g_slice_free (GstPipeConfiguration, pipe);
}


/*
 * gst_subdevsrc_parser_open_config:
 * @location: Path to configuration file.
 *
 * Returns: TRUE on success.
 */
static gboolean
_gst_subdevsrc_parser_open_config (GstConfigParser *cp, const gchar *location)
{
  GIOChannel *ch = NULL;
  gboolean ret = FALSE;
  GError *err = NULL;
  gsize len, termpos;
  GIOStatus status;
  gchar *str;
  int fd = -1;

  /* Use g_io_channel_new_file() */
  fd = open (location, O_RDONLY);
  if (fd < 0) {
    GST_WARNING ("Opening file %s failed", location);
    goto done;
  }

  ch = g_io_channel_unix_new (fd);
  g_io_channel_set_encoding (ch, "ISO-8859-1", &err);

  if (err) {
    GST_WARNING ("Failed to set encoding: %s", err->message);
    g_free (err);
    goto done;
  }

  status = g_io_channel_read_line (ch, &str, &len, &termpos, &err);

  while (status != G_IO_STATUS_EOF) {
    if (err) {
      GST_WARNING ("Error while reading the file: %s", err->message);
      g_free (err);
      goto done;
    }

    if (str) {
      str[termpos] = '\0';
      ret = _gst_subdevsrc_parser_line (cp, str);
      g_free (str);
      if (ret == FALSE) {
        break;
      }
    }

    status = g_io_channel_read_line (ch, &str, &len, &termpos, &err);
  }

done:

  if (ch) {
    g_io_channel_shutdown (ch, FALSE, NULL);
    g_io_channel_unref (ch);
  }

  if (fd != -1) {
    close (fd);
  }

  return ret;
}


/**
 * gst_subdevsrc_config_parser_new:
 * @location: Path to configuration file.
 *
 * Creates a new configuration parser object.
 */
GstConfigParser *
gst_subdevsrc_config_parser_new (const gchar *location)
{
  GstConfigParser *cp;

  cp = g_slice_new0 (GstConfigParser);

  if (!_gst_subdevsrc_parser_open_config (cp, location)) {
    g_slice_free (GstConfigParser, cp);
    return FALSE;
  }

  GST_DEBUG ("New config parser created: %p", cp);

  return cp;
}


/**
 * gst_subdevsrc_config_parser_destroy:
 * @cp: GstConfigParser object.
 *
 * Frees the given config parser and all associated resources.
 */
void
gst_subdevsrc_config_parser_destroy (GstConfigParser *cp)
{
  GSList *iter;

  if (cp == NULL)
    return;

  GST_DEBUG ("Destroying config parser: %p", cp);

  for (iter = cp->assignments; iter != NULL; iter = iter->next) {
    GstConfigAssignment *am = (GstConfigAssignment *) iter->data;
    g_free (am->key);
    g_free (am->value);
    g_slice_free (GstConfigAssignment, am);
  }
  g_slist_free (cp->assignments);

  _gst_subdevsrc_parser_destroy_pconfiguration (cp->primary);
  _gst_subdevsrc_parser_destroy_pconfiguration (cp->secondary);

  g_slice_free (GstConfigParser, cp);
}


/**
 * gst_subdevsrc_config_parser_get_mappings:
 * @cp: GstConfigParser object.
 * @key: Mapping key to be searched.
 *
 * Retrieve all values from configuration mappings which have the given key.
 * If no mapping was found, the key will be returned.
 *
 * Returns: A GSList of gchar strings. The strings must be freed after use.
 */
GSList *
gst_subdevsrc_config_parser_get_mappings (GstConfigParser *cp, const gchar *key)
{
  GSList *map = NULL;
  GSList *iter;

  for (iter = cp->assignments; iter != NULL; iter = iter->next) {
    GstConfigAssignment *am = (GstConfigAssignment *) iter->data;
    if (!strcmp (am->key, key)) {
      map = g_slist_prepend (map, g_strdup (am->value));
    }
  }

  if (!map) {
    /* No mappings found. Put the key into return list */
    map = g_slist_prepend (map, g_strdup (key));
  }
  map = g_slist_reverse (map);

  return map;
}