/* GStreamer MediaController camera source
 * Copyright (C) 2010 Nokia Corporation <multimedia@maemo.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
* SECTION:element-subdevsrc
* @short_description: Source element for subdev-based video devices
* <refsect2>
* subdevsrc can be used to capture video from Maemo devices. Compared to
* v4l2src element camerasrc is more  concentrated on high-quality image
* capturing than streaming video frames. Subdevsrc implements GstPhotography
* interface and thus provides wider choice of photography-specific parameters.
* It supports different resolutions for viewfinder and actual image capturing.
* </refsect2>
* <refsect2>
* <title>Example launch lines</title>
* <para>
* |[
* gst-launch subdevsrc ! xvimagesink
* ]|
* This pipeline shows the video captured from main camera device.
* </para>
* <para>
* |[
* gst-launch-0.10 subdevsrc camera-device=secondary capture-mode=video ! xvimagesink
* ]|
* This pipeline shows the video from main camera too, but it captures frames
* in video mode. In this case element keeps the privacy light lit during
* streaming.
* </para>
* </refsect2>
* <refsect2>
* <title>Messages</title>
* <para>
* During capturing process camerasrc sends a bunch of GstMessages to bus:
* <itemizedlist>
* <listitem>
* <para>
* GST_PHOTOGRAPHY_AUTOFOCUS_DONE is sent when application has started
* autofocus operation and it finishes. This message contains following fields:
* "status" tells whether the focusing succeeded or failed.
* "focus_window_rows" and "focus_window_columns" tell the focus matrix size.
* "focus_window_mask" is an integer containing a bitmask. It contains
* rows x columns bits, which mark the focus points in the focus matrix.
* Lowest bit (LSB) always represents the top-left corner of the focus matrix.
* This field is only valid when focusing status is SUCCESS.
* </para>
* </listitem>
* <listitem>
* <para>
* GST_PHOTOGRAPHY_SHAKE_RISK is sent during autofocusing process. It keeps
* application updated about the possible risk for shaken image.
* </para>
* </listitem>
* <listitem>
* <para>
* "photo-capture-start" is sent just before the actual high-quality capture
* operation is about to happen.
* </para>
* </listitem>
* <listitem>
* <para>
* "photo-capture-end" is sent immediately after high-quality image has been
* captured. This can be used for example for playing the shutter sound
* in sync with the capture operation.
* </para>
* </listitem>
* </itemizedlist>
* </para>
* </refsect2>
* <refsect2>
* <title>Supported formats</title>
* <para>
* Currently subdevsrc only supports UYVY and YUY2 color formats.
* </para>
* </refsect2>
*/

#include <string.h>
#include <unistd.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "gstsubdevcamsrc.h"
#include "subdev_calls.h"
#include "libomap3camd_calls.h"

GST_DEBUG_CATEGORY (gst_subdevsrc_debug);
#define GST_CAT_DEFAULT gst_subdevsrc_debug

static gboolean gst_subdevsrc_is_open (GstCameraSrc *camsrc);
static void gst_subdevsrc_finalize (GObject * object);
static void gst_subdevsrc_dispose (GObject * object);
static void gst_subdevsrc_set_property (GObject * object,
    guint prop_id, const GValue * value, GParamSpec * pspec);
static void gst_subdevsrc_get_property (GObject * object,
    guint prop_id, GValue * value, GParamSpec * pspec);
static gboolean gst_subdevsrc_unlock (GstCameraSrc * src);
static gboolean gst_subdevsrc_unlock_stop (GstCameraSrc * src);
static gboolean gst_subdevsrc_event (GstCameraSrc * src, GstEvent *event);
static gboolean gst_subdevsrc_iface_supported (GstImplementsInterface * iface,
    GType iface_type);
static GstCaps * gst_subdevsrc_get_supported (GstCameraSrc *camsrc,
    GstOperationMode mode);
static gboolean gst_subdevsrc_set_capture (GstCameraSrc *camsrc,
    GstOperationMode mode, gboolean try_only, guint32 *fourcc,
    guint *width, guint *height, guint *fps_n, guint *fps_d);
static gboolean gst_subdevsrc_set_image_tags (GstCameraSrc *camsrc,
    GstTagList *tlist);
static gboolean gst_subdevsrc_set_capture_mode (GstCameraSrc *camsrc,
    GstCameraSrcCaptureMode mode);
static gboolean gst_subdevsrc_set_viewfinder_mode (GstCameraSrc *camsrc,
    GstCameraSrcViewfinderMode mode);
static gboolean gst_subdevsrc_set_autofocus (GstCameraSrc *camsrc, gboolean on);
static gboolean gst_subdevsrc_set_autoexposure (GstCameraSrc *camsrc,
    gboolean on);
static gboolean gst_subdevsrc_read_settings (GstCameraSrc *camsrc,
    GstPhotoSettings *photoconf);
static gboolean gst_subdevsrc_write_settings (GstCameraSrc *camsrc,
    GstPhotoSettings *photoconf, gboolean scene_override);
static GstFlowReturn gst_subdevsrc_grab_frame (GstCameraSrc *camsrc,
    GstBuffer **buf, GstCameraCapturePhase phase);
static gboolean gst_subdevsrc_open (GstCameraSrc *camsrc);
static gboolean gst_subdevsrc_close (GstCameraSrc *camsrc);
static gboolean gst_subdevsrc_start (GstCameraSrc *camsrc, GstCaps *caps);
static gboolean gst_subdevsrc_stop (GstCameraSrc *camsrc);
static gboolean gst_subdevsrc_get_max_zoom (GstCameraSrc *camsrc, gfloat *zoom);
static gboolean gst_subdevsrc_set_zoom (GstCameraSrc *camsrc, gfloat zoom);
static gboolean gst_subdevsrc_set_attribute (GstCameraSrc *camsrc,
    int attribute, const int value);
static gboolean gst_subdevsrc_get_attribute (GstCameraSrc *camsrc,
    int attribute, int *value);

typedef enum {
  PROP_0,
  PROP_QUEUE_SIZE,
  PROP_CAMERA_DEVICE,
  PROP_NIGHT_MODE_CAPS,
  PROP_VIDEO_TORCH,
} GstSubdevSrcProperties;


static GstStaticPadTemplate gst_subdevsrc_src_template =
    GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("video/x-raw-yuv, "
        "format = (fourcc) UYVY, "
        "framerate = (fraction) [ 0, 200 ], "
        "width = (int) [ 1, MAX ], "
        "height = (int) [ 1, MAX ];"
        "video/x-raw-yuv, "
        "format = (fourcc) YUY2, "
        "framerate = (fraction) [ 0, 200 ], "
        "width = (int) [ 1, MAX ], "
        "height = (int) [ 1, MAX ]"));


#define DEFAULT_PROP_DEVICE_NAME  NULL
#define DEFAULT_PROP_DEVICE       "/dev/video0"
#define DEFAULT_PROP_DEVICE_FD    -1


/*
 *
 */
static void
gst_subdevsrc_interface_init (GstImplementsInterfaceClass * klass)
{
  /*
  * default virtual functions
  */
  klass->supported = gst_subdevsrc_iface_supported;
}


/*
 *
 */
void
gst_subdevsrc_init_interfaces (GType type)
{
  static const GInterfaceInfo iface_info = {
    (GInterfaceInitFunc) gst_subdevsrc_interface_init,
    NULL,
    NULL,
  };

  g_type_add_interface_static (type,
      GST_TYPE_IMPLEMENTS_INTERFACE, &iface_info);
}


GST_BOILERPLATE_FULL (GstSubdevSrc, gst_subdevsrc, GstCameraSrc,
    GST_TYPE_CAMERA_SRC, gst_subdevsrc_init_interfaces);


/*
 *
 */
static gboolean
gst_subdevsrc_iface_supported (GstImplementsInterface * iface, GType iface_type)
{
  GstCameraSrc *camsrc = GST_CAMERA_SRC (iface);

//   g_assert (iface_type == GST_TYPE_VIDEO_ORIENTATION);
//
//   if (gst_subdevsrc_is_open (camsrc)
//     return TRUE;

  if (GST_IS_IMPLEMENTS_INTERFACE (camsrc)) {
    GstImplementsInterfaceClass * pklass;

    pklass =
        g_type_interface_peek (parent_class, GST_TYPE_IMPLEMENTS_INTERFACE);
    return pklass->supported (iface, iface_type);
  }

  return FALSE;
}


/**
 * gst_subdevsrc_camera_device_get_type:
 *
 * Get camera device enumeration type.
 *
 * Returns: GType of camera device enumeration.
 */
GType
gst_subdevsrc_camera_device_get_type (void)
{
  static GType gst_subdevsrc_camera_device_type = 0;
  static GEnumValue gst_subdevsrc_camera_devices[] = {
    {GST_SUBDEVSRC_CAPTURE_DEVICE_PRIMARY, "Primary camera", "primary"},
    {GST_SUBDEVSRC_CAPTURE_DEVICE_SECONDARY, "Secondary camera", "secondary"},
    {0, NULL, NULL},
  };

  if (G_UNLIKELY (!gst_subdevsrc_camera_device_type)) {
    gst_subdevsrc_camera_device_type =
        g_enum_register_static ("GstSubdevSrcCameraDevice",
                                gst_subdevsrc_camera_devices);
  }
  return gst_subdevsrc_camera_device_type;
}


/*
 * gst_subdevsrc_reset_videomode:
 * @mode: #GstVideoMode structure.
 *
 * Resets given video #GstVideoMode contents to zero. It doesn't touch
 * structure's sub-objects, such as mutexes or buffer or buffer pool.
 */
static void
gst_subdevsrc_reset_videomode (GstVideoMode *mode)
{
  mode->width = 0;
  mode->height = 0;
  mode->color_code = 0;
  mode->fps_n = 0;
  mode->fps_d = 0;
}


/*
 * image_buffer_alloc_func:
 * @size: size of the buffer to be allocated.
 * @user_data: user data passed when registering the callback.
 *
 * RecyclePool calls this when it needs to allocate a memory block for new
 * buffer.
 *
 * Returns: Pointer to the allocated memory, or NULL.
 */
static void *
image_buffer_alloc_func (guint32 size, gpointer user_data)
{
  void *buf;

  GST_DEBUG ("Allocating buffer: %d bytes", size);
  if (posix_memalign (&buf, getpagesize(), size) != 0) {
    GST_WARNING ("Memory allocation failed");
    return NULL;
  }

  GST_DEBUG ("Memory allocated: %p", buf);

#ifdef USE_MLOCK
  GST_DEBUG ("Memory locking still image buffer");
  if (mlock (buf, size) == -1) {
    GST_WARNING ("Memory locking failed");
    free (nbuf);
    return NULL;
  }
#endif

  return buf;
}


/*
 * image_buffer_recycle_func:
 * @rbuf: #GstRecycleBuffer object to be recycled.
 * @user_data: user data passed when registering the callback.
 *
 * RecyclePool calls this when it receives an used image buffer and wants to
 * know whether we would like to recycle it.
 *
 * Returns: TRUE if the buffer was recycled.
 */
static gboolean
image_buffer_recycle_func (GstRecycleBuffer *rbuf, gpointer user_data)
{
  GstSubdevSrc *subdevsrc;
  gboolean ret = FALSE;

  subdevsrc = (GstSubdevSrc *) user_data;

  GST_DEBUG_OBJECT (subdevsrc, "Image buffer recycle callback");

  g_mutex_lock (subdevsrc->device_mutex);

  if (!subdevsrc->is_active) {
    GST_DEBUG_OBJECT (subdevsrc, "Shutting down, recycling disabled");
    goto out;
  }

#ifdef RECYCLE_BUFFERS
  if (g_mutex_trylock (subdevsrc->capture.buffer_lock)) {

    if (!subdevsrc->capture.buffer) {
      GST_DEBUG_OBJECT (subdevsrc, "Recycling raw image buffer %p", rbuf->data);

      /* Send the buffer back to library */
      if (gst_subdevsrc_libomap3camd_add_buffer (subdevsrc, rbuf,
          GST_PHOTOGRAPHY_OPERATION_MODE_IMAGE_CAPTURE))
      {
        GST_DEBUG_OBJECT (subdevsrc, "Buffer re-sent successfully");
        subdevsrc->capture.buffer = rbuf;
        ret = TRUE;
      }
      else {
        GST_WARNING_OBJECT (subdevsrc,
            "Error re-sending image buffer to libomap3camd");
      }
    }
    else {
      GST_DEBUG_OBJECT (subdevsrc, "Not recycling - buffer already allocated");
    }

    g_mutex_unlock (subdevsrc->capture.buffer_lock);
  }
#endif

out:
  g_mutex_unlock (subdevsrc->device_mutex);

  return ret;
}


/*
 * preview_buffer_recycle_func:
 * @rbuf: #GstRecycleBuffer object to be recycled.
 * @user_data: user data passed when registering the callback.
 *
 * RecyclePool calls this when it receives an used preview buffer and wants to
 * know whether we would like to recycle it.
 *
 * Returns: TRUE if the buffer was recycled.
 */
static gboolean
preview_buffer_recycle_func (GstRecycleBuffer *rbuf, gpointer user_data)
{
  GstSubdevSrc *subdevsrc;
  gboolean ret = FALSE;

  subdevsrc = (GstSubdevSrc *) user_data;

  GST_DEBUG_OBJECT (subdevsrc, "Preview buffer recycle callback");

  g_mutex_lock (subdevsrc->device_mutex);

  if (!subdevsrc->is_active) {
    GST_DEBUG_OBJECT (subdevsrc, "Shutting down, recycling disabled");
    goto out;
  }

#ifdef RECYCLE_BUFFERS
  if (g_mutex_trylock (subdevsrc->preview.buffer_lock)) {

    if (!subdevsrc->preview.buffer) {
      GST_DEBUG_OBJECT (subdevsrc, "Recycling preview buffer %p", rbuf->data);

      /* Send the buffer back to library */
      if (gst_subdevsrc_libomap3camd_add_buffer (subdevsrc, rbuf,
          GST_PHOTOGRAPHY_OPERATION_MODE_PREVIEW))
      {
        GST_DEBUG_OBJECT (subdevsrc, "Buffer re-sent successfully");
        subdevsrc->preview.buffer = rbuf;
        ret = TRUE;
      }
      else {
        GST_WARNING_OBJECT (subdevsrc,
            "Error re-sending preview buffer to libomap3camd");
      }
    }
    else {
      GST_DEBUG_OBJECT (subdevsrc, "Not recycling - buffer already allocated");
    }

    g_mutex_unlock (subdevsrc->preview.buffer_lock);
  }
#endif

out:
  g_mutex_unlock (subdevsrc->device_mutex);

  return ret;
}


/*
 * image_buffer_free_func:
 * @ptr: pointer to the memory to be freed.
 * @size: size of the memory chunk.
 * @user_data: user data passed when registering the callback.
 *
 * RecyclePool calls this when it is freeing a buffer data.
 */
static void
image_buffer_free_func (void *ptr, guint32 size, gpointer user_data)
{
  GST_DEBUG ("Freeing buffer %p (%d bytes)", ptr, size);

#ifdef USE_MLOCK
  munlock (ptr, size);
#endif

  free (ptr);
}


/*
 * gst_subdevsrc_is_open:
 * @camsrc: #GstCameraSrc object.
 *
 * Returns: TRUE if camera device is currently open.
 */
static gboolean
gst_subdevsrc_is_open (GstCameraSrc *camsrc)
{
  GstSubdevSrc *subdevsrc = GST_SUBDEVSRC_CAST (camsrc);

  return subdevsrc->is_open;
}

/*
 * gst_subdevsrc_is_active:
 * @camsrc: #GstCameraSrc object.
 *
 * Returns: TRUE if camera device is active (streaming frames).
 */
static gboolean
gst_subdevsrc_is_active (GstCameraSrc *camsrc)
{
  GstSubdevSrc *subdevsrc = GST_SUBDEVSRC_CAST (camsrc);

  return subdevsrc->is_active;
}


/*
 * gst_subdevsrc_get_caps:
 * @camsrc: #GstCameraSrc object.
 *
 * This function provides supported source pad caps to camerasrc baseclass.
 *
 * Returns: Supported source pad caps.
 */
static GstCaps *
gst_subdevsrc_get_caps (GstCameraSrc *camsrc)
{
  GstSubdevSrc *subdevsrc = GST_SUBDEVSRC_CAST (camsrc);
  GstCaps *ret;

  GST_DEBUG_OBJECT (subdevsrc, "Get caps");

  if (!gst_subdevsrc_is_open (camsrc)) {
    return
        gst_caps_copy (gst_pad_get_pad_template_caps (GST_BASE_SRC_PAD
            (subdevsrc)));
  }

  if (subdevsrc->probed_caps)
    return gst_caps_ref (subdevsrc->probed_caps);

  ret = gst_subdevsrc_probe_caps (subdevsrc, FALSE);
  subdevsrc->probed_caps = gst_caps_ref (ret);

  GST_INFO_OBJECT (subdevsrc, "probed caps: %" GST_PTR_FORMAT, ret);

  return ret;
}


/*
 * gst_subdevsrc_get_num_buffers:
 * @camsrc: #GstCameraSrc object.
 *
 * Provides the number of buffers that are used for retrieving frames from
 * camera device to camerasrc baseclass.
 *
 * Returns: Number of buffers allocated for streaming.
 */
static guint
gst_subdevsrc_get_num_buffers (GstCameraSrc *camsrc)
{
  GstSubdevSrc *subdevsrc = GST_SUBDEVSRC_CAST (camsrc);

  return subdevsrc->num_buffers;
}


/*
 *
 */
static GstPhotoCaps
gst_subdevsrc_get_capabilities (GstCameraSrc *camsrc)
{
  GstPhotoCaps pcaps;

  pcaps = GST_PHOTOGRAPHY_CAPS_ZOOM     | GST_PHOTOGRAPHY_CAPS_EV_COMP  |
          GST_PHOTOGRAPHY_CAPS_WB_MODE  | GST_PHOTOGRAPHY_CAPS_TONE     |
          GST_PHOTOGRAPHY_CAPS_SCENE    | GST_PHOTOGRAPHY_CAPS_FLASH    |
          GST_PHOTOGRAPHY_CAPS_FOCUS    | GST_PHOTOGRAPHY_CAPS_APERTURE |
          GST_PHOTOGRAPHY_CAPS_EXPOSURE | GST_PHOTOGRAPHY_CAPS_SHAKE    |
          GST_PHOTOGRAPHY_CAPS_ISO_SPEED;

  return pcaps;
}


/*
 * gst_subdevsrc_base_init:
 * @klass: #GstElementClass.
 *
 */
static void
gst_subdevsrc_base_init (gpointer klass)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (klass);

  gst_element_class_set_details_simple (element_class,
      "MC camera source",
      "Video/Src",
      "MediaController camera source element",
      "Maemo Multimedia <multimedia@maemo.org>");

  gst_element_class_add_pad_template (element_class,
       gst_static_pad_template_get (&gst_subdevsrc_src_template));
}

/*
 * gst_subdevsrc_class_init:
 * @klass: #GstSubdevSrcClass.
 *
 */
static void
gst_subdevsrc_class_init (GstSubdevSrcClass * klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  GstCameraSrcClass *camera_class = GST_CAMERA_SRC_CLASS (klass);

  gobject_class->set_property = gst_subdevsrc_set_property;
  gobject_class->get_property = gst_subdevsrc_get_property;
  gobject_class->dispose = gst_subdevsrc_dispose;
  gobject_class->finalize = gst_subdevsrc_finalize;

  /**
   * GstSubdevSrc:queue-size:
   *
   * Defines how many buffers should be allocated for streaming.
   */
  g_object_class_install_property (gobject_class, PROP_QUEUE_SIZE,
      g_param_spec_uint ("queue-size", "Queue size",
          "Number of buffers to be enqueud in the driver",
          GST_SUBDEVSRC_MIN_BUFFERS, GST_SUBDEVSRC_MAX_BUFFERS,
          GST_SUBDEVSRC_DEFAULT_BUFFERS,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

  /**
   * GstSubdevSrc:camera-device:
   *
   * Defines which camera device to use.
   */
  g_object_class_install_property (gobject_class, PROP_CAMERA_DEVICE,
      g_param_spec_enum ("camera-device", "Camera device",
          "Defines which camera device should be used",
          GST_TYPE_SUBDEVSRC_CAMERA_DEVICE,
          GST_SUBDEVSRC_CAPTURE_DEVICE_PRIMARY,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

 /**
   * GstSubdevSrc:night-mode-caps:
   *
   * Defines the viewfinder caps that must be used in night scene mode.
   */
  g_object_class_install_property (gobject_class, PROP_NIGHT_MODE_CAPS,
      g_param_spec_boxed ("night-mode-caps", "Night mode caps",
          "Defines viewfinder caps that must be used in night scene mode",
          GST_TYPE_CAPS, G_PARAM_READABLE | G_PARAM_STATIC_STRINGS));

 /**
   * GstSubdevSrc:video-recording-torch:
   *
   * Sets torch light on or off for video recording.
   */
  g_object_class_install_property (gobject_class, PROP_VIDEO_TORCH,
      g_param_spec_boolean("video-torch", "Video torch",
          "Sets torch light on or off for video recording",
          FALSE, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

  camera_class->is_open = GST_DEBUG_FUNCPTR (gst_subdevsrc_is_open);
  camera_class->open = GST_DEBUG_FUNCPTR (gst_subdevsrc_open);
  camera_class->close = GST_DEBUG_FUNCPTR (gst_subdevsrc_close);
  camera_class->get_attribute = GST_DEBUG_FUNCPTR (gst_subdevsrc_get_attribute);
  camera_class->set_attribute = GST_DEBUG_FUNCPTR (gst_subdevsrc_set_attribute);
  camera_class->set_capture = GST_DEBUG_FUNCPTR (gst_subdevsrc_set_capture);
  camera_class->start = GST_DEBUG_FUNCPTR (gst_subdevsrc_start);
  camera_class->is_active = GST_DEBUG_FUNCPTR (gst_subdevsrc_is_active);
  camera_class->grab_frame = GST_DEBUG_FUNCPTR (gst_subdevsrc_grab_frame);
  camera_class->stop = GST_DEBUG_FUNCPTR (gst_subdevsrc_stop);
  camera_class->get_caps = GST_DEBUG_FUNCPTR (gst_subdevsrc_get_caps);
  camera_class->get_num_buffers = GST_DEBUG_FUNCPTR (gst_subdevsrc_get_num_buffers);
  camera_class->get_max_zoom = GST_DEBUG_FUNCPTR (gst_subdevsrc_get_max_zoom);
  camera_class->unlock = GST_DEBUG_FUNCPTR (gst_subdevsrc_unlock);
  camera_class->unlock_stop = GST_DEBUG_FUNCPTR (gst_subdevsrc_unlock_stop);
  camera_class->event = GST_DEBUG_FUNCPTR (gst_subdevsrc_event);
  camera_class->set_zoom = GST_DEBUG_FUNCPTR (gst_subdevsrc_set_zoom);
  camera_class->get_supported_caps = GST_DEBUG_FUNCPTR (gst_subdevsrc_get_supported);
  camera_class->fill_image_tags = GST_DEBUG_FUNCPTR (gst_subdevsrc_set_image_tags);
  camera_class->set_capture_mode = GST_DEBUG_FUNCPTR (gst_subdevsrc_set_capture_mode);
  camera_class->set_vfinder_mode = GST_DEBUG_FUNCPTR (gst_subdevsrc_set_viewfinder_mode);
  camera_class->set_autofocus = GST_DEBUG_FUNCPTR (gst_subdevsrc_set_autofocus);
  camera_class->set_autoexposure = GST_DEBUG_FUNCPTR (gst_subdevsrc_set_autoexposure);
  camera_class->read_settings = GST_DEBUG_FUNCPTR (gst_subdevsrc_read_settings);
  camera_class->write_settings = GST_DEBUG_FUNCPTR (gst_subdevsrc_write_settings);
  camera_class->get_capabilities = GST_DEBUG_FUNCPTR (gst_subdevsrc_get_capabilities);
}


/*
 * gst_subdevsrc_init:
 * @subdevsrc: #GstSubdevSrc object.
 * @klass: #GstSubdevSrcClass.
 *
 * Initializes the given @subdevsrc object instance.
 */
static void
gst_subdevsrc_init (GstSubdevSrc * subdevsrc, GstSubdevSrcClass * klass)
{
  subdevsrc->num_buffers = GST_SUBDEVSRC_DEFAULT_BUFFERS;
  subdevsrc->poll = gst_poll_new (TRUE);
  subdevsrc->max_zoom_factor = MAX_RESIZER_FACTOR;
  subdevsrc->lib_max_zoom = MAX_RESIZER_FACTOR;
  subdevsrc->zoom_factor = 1.0;
  subdevsrc->is_open = FALSE;
  subdevsrc->is_active = FALSE;
  subdevsrc->sensor_fd = -1;
  subdevsrc->rsz_fd = -1;
  subdevsrc->video_fd = -1;
  subdevsrc->capture_mode = GST_CAMERA_SRC_CAPTURE_MODE_VIEWFINDER;
  subdevsrc->viewfinder_mode = GST_CAMERA_SRC_VIEWFINDER_MODE_STILL;
  subdevsrc->night_caps = NULL;
  subdevsrc->video_torch = FALSE;
  subdevsrc->flicker_counter = 0;
  subdevsrc->flash_updated = 0;

  subdevsrc->device_mutex = g_mutex_new ();

  memset (&subdevsrc->current, 0, sizeof (GstVideoMode));
  memset (&subdevsrc->sensor, 0, sizeof (GstVideoMode));
  memset (&subdevsrc->capture, 0, sizeof (GstVideoMode));
  memset (&subdevsrc->preview, 0, sizeof (GstVideoMode));
  memset (&subdevsrc->resizer, 0, sizeof (GstVideoMode));

  subdevsrc->capture.buffer_lock = g_mutex_new ();
  subdevsrc->preview.buffer_lock = g_mutex_new ();

  GST_DEBUG_OBJECT (subdevsrc, "initialized");
}


/*
 * gst_subdevsrc_dispose:
 * @object: #GstSubdevSrc object.
 *
 * Dispose the given object instance.
 */
static void
gst_subdevsrc_dispose (GObject * object)
{
  GstSubdevSrc *subdevsrc = GST_SUBDEVSRC_CAST (object);

  if (subdevsrc->probed_caps) {
    gst_caps_unref (subdevsrc->probed_caps);
    subdevsrc->probed_caps = NULL;
  }

  if (subdevsrc->night_caps) {
    gst_caps_unref (subdevsrc->night_caps);
    subdevsrc->night_caps = NULL;
  }

  if (subdevsrc->poll) {
    gst_poll_free (subdevsrc->poll);
  }

  g_mutex_free (subdevsrc->capture.buffer_lock);
  g_mutex_free (subdevsrc->preview.buffer_lock);
  g_mutex_free (subdevsrc->device_mutex);

  G_OBJECT_CLASS (parent_class)->dispose (object);
}


/*
 * gst_subdevsrc_finalize:
 * @object: #GstSubdevSrc object.
 *
 * Finalize the given object instance.
 */
static void
gst_subdevsrc_finalize (GObject * object)
{
  GstSubdevSrc *subdevsrc;

  subdevsrc = GST_SUBDEVSRC_CAST (object);

  G_OBJECT_CLASS (parent_class)->finalize (object);
}


/*
 * gst_subdevsrc_get_supported:
 * @camsrc: #GstCameraSrc object.
 * @mode: #GstOperationMode which caps to return.
 *
 * Provides information about supported output formats in each operation mode.
 * The result depends on current state of the element and from the selected
 * mode. In NULL mode CAPTURE caps are always the same as sourcepad template
 * caps. If camera device is opened, those caps are probed from the device,
 * and the result caps are more accurate. Also if capture resolution isn't set,
 * this function can't give accurate output caps for PREVIEW mode.
 *
 * Returns: Supported caps in given mode.
 */
static GstCaps *
gst_subdevsrc_get_supported (GstCameraSrc *camsrc, GstOperationMode mode)
{
  GstSubdevSrc *subdevsrc = GST_SUBDEVSRC_CAST (camsrc);
  GstCaps *retcaps = NULL;

  if (mode == GST_PHOTOGRAPHY_OPERATION_MODE_IMAGE_CAPTURE) {
    /* Capture caps are same as output caps */
    retcaps = gst_subdevsrc_get_caps (camsrc);
  }
  else if (mode == GST_PHOTOGRAPHY_OPERATION_MODE_PREVIEW) {
    if (subdevsrc->capture.width == 0 || subdevsrc->capture.height == 0) {
      /* Capture resolution not set yet. Return capture caps */
      GST_INFO_OBJECT (subdevsrc, "Preview caps without capture mode set");
      retcaps = gst_caps_new_empty ();
    }
    else {
      GstCaps *tmp;
      guint w, h, wmin, hmin;

      w = subdevsrc->capture.width;
      h = subdevsrc->capture.height;
      wmin = subdevsrc->preview_min_w;
      hmin = subdevsrc->preview_min_h;

      GST_INFO_OBJECT (subdevsrc, "Preview caps with capture: %dx%d", w, h);

      retcaps = gst_caps_new_simple ("video/x-raw-yuv",
        "format", GST_TYPE_FOURCC, GST_MAKE_FOURCC ('U', 'Y', 'V', 'Y'),
        "width", GST_TYPE_INT_RANGE, wmin, w,
        "height",  GST_TYPE_INT_RANGE, hmin, h,
        NULL);

      tmp = gst_caps_new_simple ("video/x-raw-yuv",
        "format", GST_TYPE_FOURCC, GST_MAKE_FOURCC ('Y', 'U', 'Y', '2'),
        "width", GST_TYPE_INT_RANGE, wmin, w,
        "height",  GST_TYPE_INT_RANGE, hmin, h,
        NULL);

      /* This will unref tmp caps */
      gst_caps_append (retcaps, tmp);
    }
  }
  else {
    GST_WARNING ("Unknown operation mode: %d", mode);
  }

  GST_DEBUG_OBJECT (subdevsrc, "Returning supported %s caps: %" GST_PTR_FORMAT,
      mode == GST_PHOTOGRAPHY_OPERATION_MODE_IMAGE_CAPTURE ?
      "CAPTURE" : "PREVIEW", retcaps);

  return retcaps;
}


/*
 * gst_subdevsrc_set_viewfinder_fmt:
 * @subdevsrc: #GstSubdevSrc object.
 * @fourcc: pointer to fourcc value.
 * @width: pointer to width value.
 * @height: pointer to height value.
 *
 * This must be called with device_mutex held
 */
static gboolean
gst_subdevsrc_set_viewfinder_fmt (GstSubdevSrc *subdevsrc,
    guint32 *fourcc, guint *width, guint *height)
{
  gboolean ret = TRUE;
  guint num_buffers;
  guint32 size;

  if (!gst_subdevsrc_libomap3camd_configure (subdevsrc, width, height, *fourcc,
      &size, &num_buffers, GST_PHOTOGRAPHY_OPERATION_MODE_VIEWFINDER))
  {
    goto res_failed;
  }

  GST_DEBUG_OBJECT (subdevsrc, "Setting viewfinder resolution SUCCESS");

out:

  return ret;

/* ERRORS*/
res_failed:
  {
    GST_WARNING_OBJECT (subdevsrc, "Setting viewfinder resolution FAILED");
    ret = FALSE;
    goto out;
  }
}


/*
 * gst_subdevsrc_set_capture_fmt:
 * @subdevsrc: #GstSubdevSrc object.
 * @fourcc: pointer to fourcc value.
 * @width: pointer to width value.
 * @height: pointer to height value.
 *
 * Configures libomap3camd to use given resolution and fourcc for image
 * capturing.
 *
 * Returns: TRUE on success.
 */
static gboolean
gst_subdevsrc_set_capture_fmt (GstSubdevSrc *subdevsrc,
    guint32 *fourcc, guint *width, guint *height)
{
  GstRecycleBuffer *buf;
  gboolean ret = TRUE;
  gboolean cret;
  guint rw, rh;

  /* Calculate rounded-up resolution */
  rw = (*width + CAPTURE_WIDTH_MODULO-1) & ~(CAPTURE_WIDTH_MODULO-1);
  rh = (*height + CAPTURE_HEIGHT_MODULO-1) & ~(CAPTURE_HEIGHT_MODULO-1);

  g_mutex_lock (subdevsrc->capture.buffer_lock);

  if (rw != subdevsrc->capture.width || rh != subdevsrc->capture.height ||
      *fourcc != subdevsrc->capture.color_code)
  {
    guint num_buffers = 0;
    guint32 size = 0;

    GST_DEBUG_OBJECT (subdevsrc, "Resolution changed: %dx%d -> %dx%d",
        subdevsrc->capture.width, subdevsrc->capture.height, rw, rh);
    GST_DEBUG_OBJECT (subdevsrc, "Color format: 0x%08x -> 0x%08x",
        subdevsrc->capture.color_code, *fourcc);

    /* Resolution changes so we need to free old HQ image buffers */
    if (subdevsrc->capture.pool) {
      if (subdevsrc->capture.buffer) {
        gst_recycle_pool_free_buffer (subdevsrc->capture.pool,
            subdevsrc->capture.buffer);
      }
      gst_recycle_pool_destroy (subdevsrc->capture.pool);
      subdevsrc->capture.buffer = NULL;
      subdevsrc->capture.pool = NULL;
    }

    /* Give new resolution to libomap3camd */
    g_mutex_lock (subdevsrc->device_mutex);
    ret = gst_subdevsrc_libomap3camd_configure (subdevsrc, &rw, &rh, *fourcc,
        &size, &num_buffers, GST_PHOTOGRAPHY_OPERATION_MODE_IMAGE_CAPTURE);
    g_mutex_unlock (subdevsrc->device_mutex);

    if (!ret) {
      goto res_failed;
    }

    subdevsrc->preview_min_w = 1;
    subdevsrc->preview_min_h = 1;

    /* Query minimum supported preview image size */
    gst_subdevsrc_libomap3camd_configure (subdevsrc, &subdevsrc->preview_min_w,
        &subdevsrc->preview_min_h, *fourcc, NULL, NULL,
        GST_PHOTOGRAPHY_OPERATION_MODE_PREVIEW);

    GST_DEBUG_OBJECT (subdevsrc, "Minimum preview size: %dx%d",
        subdevsrc->preview_min_w, subdevsrc->preview_min_h);

    subdevsrc->capture.pool = gst_recycle_pool_new (num_buffers * 2,
        num_buffers, size);

    if (!subdevsrc->capture.pool) {
      goto pool_alloc_failed;
    }

    gst_recycle_pool_set_alloc_func (subdevsrc->capture.pool,
        image_buffer_alloc_func, subdevsrc);

    gst_recycle_pool_set_recycle_func (subdevsrc->capture.pool,
        image_buffer_recycle_func, subdevsrc);

    gst_recycle_pool_set_free_func (subdevsrc->capture.pool,
        image_buffer_free_func, subdevsrc);

    subdevsrc->capture.color_code = *fourcc;
    *width = rw;
    *height = rh;
  }
  else {
    GST_DEBUG_OBJECT (subdevsrc, "Capture resolution not changed");
  }

  /* If active buffer is already selected, we can leave */
  if (subdevsrc->capture.buffer) {
    GST_DEBUG_OBJECT (subdevsrc,
        "No need to allocate new buffer, previous is recycled");
    goto out;
  }

  buf = gst_recycle_pool_get_buffer (subdevsrc->capture.pool);

  if (!buf) {
    goto buffer_alloc_failed;
  }

  GST_DEBUG_OBJECT (subdevsrc, "Got buffer %p", buf->data);

  g_mutex_lock (subdevsrc->device_mutex);
  cret = gst_subdevsrc_libomap3camd_add_buffer (subdevsrc, buf,
      GST_PHOTOGRAPHY_OPERATION_MODE_IMAGE_CAPTURE);
  g_mutex_unlock (subdevsrc->device_mutex);

  if (!cret) {
    goto set_buffer_failed;
  }

  subdevsrc->capture.buffer = buf;
  GST_DEBUG_OBJECT (subdevsrc, "Capture buffer set");

out:
  g_mutex_unlock (subdevsrc->capture.buffer_lock);

  GST_DEBUG_OBJECT (subdevsrc, "Allocating image capture buffer: %s",
                    ret ? "SUCCESS" : "FAILED");

  return ret;

/* ERRORS*/
res_failed:
  {
    GST_WARNING_OBJECT (subdevsrc, "Setting capture resolution FAILED");
    ret = FALSE;
    goto out;
  }

pool_alloc_failed:
  {
    GST_WARNING_OBJECT (subdevsrc, "Capture buffer pool allocation FAILED");
    ret = FALSE;
    goto out;
  }

buffer_alloc_failed:
  {
    GST_WARNING_OBJECT (subdevsrc, "Memory allocation FAILED");
    ret = FALSE;
    goto out;
  }

set_buffer_failed:
  {
    GST_WARNING_OBJECT (subdevsrc, "Setting capture buffer FAILED");
    gst_recycle_pool_free_buffer (subdevsrc->capture.pool, buf);
    ret = FALSE;
    goto out;
  }
}


/*
 * gst_subdevsrc_set_preview_fmt:
 * @subdevsrc: #GstSubdevSrc object.
 * @fourcc: pointer to fourcc value.
 * @width: pointer to width value.
 * @height: pointer to height value.
 *
 * Configures libomap3camd to use given resolution and fourcc for image
 * preview.
 *
 * Returns: TRUE on success.
 */
static gboolean
gst_subdevsrc_set_preview_fmt (GstSubdevSrc *subdevsrc,
    guint32 *fourcc, guint *width, guint *height)
{
  GstRecycleBuffer *buf;
  gboolean ret = TRUE;
  guint num_buffers;
  gboolean cret;
  guint32 size;
  guint rw, rh;

  /* Calculate rounded-up resolution */
  rw = (*width + PREVIEW_WIDTH_MODULO-1) & ~(PREVIEW_WIDTH_MODULO-1);
  rh = (*height + PREVIEW_HEIGHT_MODULO-1) & ~(PREVIEW_HEIGHT_MODULO-1);

  g_mutex_lock (subdevsrc->preview.buffer_lock);

  if (rw != subdevsrc->preview.width || rh != subdevsrc->preview.height ||
      *fourcc != subdevsrc->preview.color_code)
  {
    GST_DEBUG_OBJECT (subdevsrc, "Preview resolution changed: %dx%d -> %dx%d",
        subdevsrc->preview.width, subdevsrc->preview.height, rw, rh);

    /* Free preview image pool and possibly queued buffer */
    if (subdevsrc->preview.pool) {
      if (subdevsrc->preview.buffer) {
        gst_recycle_pool_free_buffer (subdevsrc->preview.pool,
          subdevsrc->preview.buffer);
      }
      gst_recycle_pool_destroy (subdevsrc->preview.pool);
      subdevsrc->preview.buffer = NULL;
      subdevsrc->preview.pool = NULL;
    }
  }
  else {
    GST_DEBUG_OBJECT (subdevsrc, "Preview resolution not changed");
  }

  /* This function can also be called with zero parameters. In that case
   * we need to invalidate the previous snapshot buffer, if any */
  if (*width != 0 && *height != 0 && *fourcc != 0) {

    /* If active preview buffer is already selected, we can leave */
    if (subdevsrc->preview.buffer) {
      GST_DEBUG_OBJECT (subdevsrc,
          "No need to allocate new buffer, previous is recycled");
      goto out;
    }

    g_mutex_lock (subdevsrc->device_mutex);
    ret = gst_subdevsrc_libomap3camd_configure (subdevsrc, &rw, &rh, *fourcc,
        &size, &num_buffers, GST_PHOTOGRAPHY_OPERATION_MODE_PREVIEW);
    g_mutex_unlock (subdevsrc->device_mutex);

    if (!ret) {
      goto res_failed;
    }

    subdevsrc->preview.color_code = *fourcc;
    *width = rw;
    *height = rh;

    if (!subdevsrc->preview.pool) {
      subdevsrc->preview.pool = gst_recycle_pool_new (num_buffers * 2,
          num_buffers, size);

      if (!subdevsrc->preview.pool) {
        goto pool_alloc_failed;
      }

      gst_recycle_pool_set_alloc_func (subdevsrc->preview.pool,
          image_buffer_alloc_func, subdevsrc);

      gst_recycle_pool_set_recycle_func (subdevsrc->preview.pool,
          preview_buffer_recycle_func, subdevsrc);

      gst_recycle_pool_set_free_func (subdevsrc->preview.pool,
          image_buffer_free_func, subdevsrc);
    }

    buf = gst_recycle_pool_get_buffer (subdevsrc->preview.pool);

    if (!buf) {
      goto buffer_alloc_failed;
    }

    GST_DEBUG_OBJECT (subdevsrc, "Got buffer %p", buf->data);

    g_mutex_lock (subdevsrc->device_mutex);
    cret = gst_subdevsrc_libomap3camd_add_buffer (subdevsrc, buf,
        GST_PHOTOGRAPHY_OPERATION_MODE_PREVIEW);
    g_mutex_unlock (subdevsrc->device_mutex);

    if (!cret) {
      goto set_buffer_failed;
    }

    subdevsrc->preview.buffer = buf;
    GST_DEBUG_OBJECT (subdevsrc, "Preview buffer set");
  }

out:
  g_mutex_unlock (subdevsrc->preview.buffer_lock);

  GST_DEBUG_OBJECT (subdevsrc, "Allocating preview buffer: %s",
      ret ? "SUCCESS" : "FAILED");

  return ret;

/* ERRORS*/
res_failed:
  {
    GST_WARNING_OBJECT (subdevsrc, "Setting preview resolution FAILED");
    ret = FALSE;
    goto out;
  }

pool_alloc_failed:
  {
    GST_WARNING_OBJECT (subdevsrc, "Preview buffer pool allocation FAILED");
    ret = FALSE;
    goto out;
  }

buffer_alloc_failed:
  {
    GST_WARNING_OBJECT (subdevsrc, "Memory allocation FAILED");
    ret = FALSE;
    goto out;
  }

set_buffer_failed:
  {
    GST_WARNING_OBJECT (subdevsrc, "Setting preview buffer FAILED");
    gst_recycle_pool_free_buffer (subdevsrc->preview.pool, buf);
    ret = FALSE;
    goto out;
  }
}


/*
 * gst_subdevsrc_set_capture:
 * @camsrc: #GstCameraSrc object.
 * @mode: #GstOperationMode describing which resolution to be set.
 * @try_only: If TRUE, the function will only check (and re-adjust) the params.
 * @fourcc: Color format to be used.
 * @width: Width to be used.
 * @height: Height to be used.
 * @fps_n: Framerate fraction numerator.
 * @fps_d: Framerate fraction denominator.
 *
 * Initialize the camera device to given parameters. If @try_only is set,
 * no hardware reconfiguration will be done. Instead, the incoming parameters
 * are just checked and changed to suitable values, if needed. If values are
 * totally unusable, return value FALSE will indicate this.
 *
 * Returns: TRUE on success. This means that given values were in acceptable
 * range and they may have been altered to the supported ones. FALSE means that
 * device can't be configured to the given parameters at all.
 */
static gboolean
gst_subdevsrc_set_capture (GstCameraSrc *camsrc,
    GstOperationMode mode, gboolean try_only, guint32 *fourcc,
    guint *width, guint *height, guint *fps_n, guint *fps_d)
{
  GstSubdevSrc *subdevsrc = GST_SUBDEVSRC_CAST (camsrc);
  gboolean ret = FALSE;

  GST_DEBUG_OBJECT (subdevsrc, "mode = %d, try_only = %d", mode, try_only);

  if (fourcc && width && height) {
    if (*fourcc) {
      switch (*fourcc) {
        case GST_MAKE_FOURCC ('Y', 'U', 'Y', '2'):
          *fourcc = GST_MAKE_FOURCC ('Y', 'U', 'Y', 'V');
          break;
        case GST_MAKE_FOURCC ('Y', 'U', 'Y', 'V'):
        case GST_MAKE_FOURCC ('U', 'Y', 'V', 'Y'):
          break;
        default:
          GST_WARNING_OBJECT (subdevsrc, "Unsupported fourcc: %" GST_FOURCC_FORMAT,
              GST_FOURCC_ARGS (*fourcc));
          return FALSE;
          break;
      }
    }

    if (mode == GST_PHOTOGRAPHY_OPERATION_MODE_IMAGE_CAPTURE) {
      if (try_only) {
        /* FIXME: We should REALLY try the parameters */
        ret = TRUE;
      }
      else {
        ret = gst_subdevsrc_set_capture_fmt (subdevsrc, fourcc, width, height);

        if (ret) {
          subdevsrc->capture.width = *width;
          subdevsrc->capture.height = *height;
          subdevsrc->capture.fps_n = 1;
          subdevsrc->capture.fps_d = 1;

          subdevsrc->zoom_factor = CLAMP (subdevsrc->zoom_factor, 1.0,
                                          subdevsrc->lib_max_zoom);

          g_mutex_lock (subdevsrc->device_mutex);
          gst_subdevsrc_libomap3camd_set_zoom (subdevsrc, subdevsrc->zoom_factor);
          g_mutex_unlock (subdevsrc->device_mutex);
        }
      }
    }
    else if (mode == GST_PHOTOGRAPHY_OPERATION_MODE_PREVIEW) {
      if (try_only) {
        /* FIXME: We should REALLY try the parameters */
        ret = TRUE;
      }
      else {
        ret = gst_subdevsrc_set_preview_fmt (subdevsrc, fourcc, width, height);

        if (ret) {
          subdevsrc->preview.width = *width;
          subdevsrc->preview.height = *height;
          subdevsrc->preview.color_code = *fourcc;
          subdevsrc->preview.fps_n = 1; /* Not used */
          subdevsrc->preview.fps_d = 1; /* Not used */
        }
      }
    }
    else {
      /* VIEWFINDER mode */
      /* FIXME: handle try_only mode */
      ret = gst_subdevsrc_configure_pipeline (subdevsrc, fourcc, width, height,
          fps_n, fps_d);

      if (ret) {
        g_mutex_lock (subdevsrc->device_mutex);
        gst_subdevsrc_set_viewfinder_fmt (subdevsrc, fourcc, width, height);
        g_mutex_unlock (subdevsrc->device_mutex);

        subdevsrc->current.width = *width;
        subdevsrc->current.height = *height;

        if (fps_n != NULL && fps_d != NULL) {
          subdevsrc->current.fps_n = *fps_n;
          subdevsrc->current.fps_d = *fps_d;
        }
      }
    }
  }

  return ret;
}


/*
 *
 */
static gboolean
gst_subdevsrc_set_viewfinder_mode (GstCameraSrc *camsrc,
    GstCameraSrcViewfinderMode mode)
{
  GstSubdevSrc *subdevsrc;
  gboolean ret = FALSE;

  subdevsrc = GST_SUBDEVSRC_CAST (camsrc);

  GST_DEBUG_OBJECT (subdevsrc, "Setting viewfinder mode");

  g_mutex_lock (subdevsrc->device_mutex);
  subdevsrc->viewfinder_mode = mode;

  ret = gst_subdevsrc_libomap3camd_set_vf_mode (subdevsrc, mode);

  if (ret && mode == GST_CAMERA_SRC_VIEWFINDER_MODE_VIDEO) {
    /* Turn on torch (assist light) in video mode, if the property is set */
    gst_subdevsrc_libomap3camd_set_assist_light (subdevsrc,
        subdevsrc->video_torch);
  }
  else {
    gst_subdevsrc_libomap3camd_set_assist_light (subdevsrc, FALSE);
  }

  g_mutex_unlock (subdevsrc->device_mutex);

  return ret;
}


/*
 *
 */
static gboolean
gst_subdevsrc_set_capture_mode (GstCameraSrc *camsrc,
    GstCameraSrcCaptureMode mode)
{
#ifndef GST_DISABLE_GST_DEBUG
  static const gchar *cmodes[] = { "VIEWFINDER", "STILL", "VIDEO" };
#endif
  GstSubdevSrc *subdevsrc;
  gboolean ret = FALSE;

  subdevsrc = GST_SUBDEVSRC_CAST (camsrc);

  GST_DEBUG_OBJECT (subdevsrc, "Setting capture mode");

  g_mutex_lock (subdevsrc->device_mutex);
  subdevsrc->capture_mode = mode;

  if (subdevsrc->capture_mode == GST_CAMERA_SRC_CAPTURE_MODE_VIDEO) {
    /* Turn on privacy light in video mode */
    gst_subdevsrc_libomap3camd_set_privacy_light (subdevsrc, TRUE);
  }
  else {
    gst_subdevsrc_libomap3camd_set_privacy_light (subdevsrc, FALSE);
  }

  g_mutex_unlock (subdevsrc->device_mutex);

  GST_DEBUG_OBJECT (subdevsrc, "Capture mode: %s", cmodes[mode]);

  return ret;
}


/*
 *
 */
static gboolean
gst_subdevsrc_set_autofocus (GstCameraSrc *camsrc, gboolean on)
{
  GstSubdevSrc *subdevsrc;
  gboolean ret = FALSE;
  gchar *mode;

  subdevsrc = GST_SUBDEVSRC_CAST (camsrc);
  mode = on ? "ON" : "OFF";

  GST_DEBUG_OBJECT (subdevsrc, "setting autofocus: %s", mode);

  g_mutex_lock (subdevsrc->device_mutex);
  ret = gst_subdevsrc_libomap3camd_set_autofocus (subdevsrc, on);
  g_mutex_unlock (subdevsrc->device_mutex);

  GST_DEBUG_OBJECT (subdevsrc, "Setting autofocus %s: %s", mode,
      ret ? "SUCCESS" : "FAILED");

  return ret;
}


/*
 *
 */
static gboolean
gst_subdevsrc_set_autoexposure (GstCameraSrc *camsrc, gboolean on)
{
  GstSubdevSrc *subdevsrc;
  gboolean ret = FALSE;
  gchar *mode;

  subdevsrc = GST_SUBDEVSRC_CAST (camsrc);
  mode = on ? "ON" : "OFF";

  GST_DEBUG_OBJECT (subdevsrc, "setting autoexposure: %s", mode);

  g_mutex_lock (subdevsrc->device_mutex);
  ret = gst_subdevsrc_libomap3camd_set_autoexposure (subdevsrc,on);
  g_mutex_unlock (subdevsrc->device_mutex);

  GST_DEBUG_OBJECT (subdevsrc, "Setting autoexposure %s: %s", mode,
                    ret ? "SUCCESS" : "FAILED");

  return ret;
}


/*
 *
 */
static gboolean
gst_subdevsrc_read_settings (GstCameraSrc *camsrc, GstPhotoSettings *photoconf)
{
  GstSubdevSrc *subdevsrc;
  gboolean ret;

  subdevsrc = GST_SUBDEVSRC_CAST (camsrc);

  g_mutex_lock (subdevsrc->device_mutex);
  ret = gst_subdevsrc_libomap3camd_read_settings (subdevsrc, photoconf);
  g_mutex_unlock (subdevsrc->device_mutex);

  return ret;
}


/*
 *
 */
static gboolean
gst_subdevsrc_write_settings (GstCameraSrc *camsrc,
    GstPhotoSettings *photoconf, gboolean scene_override)
{
  GstSubdevSrc *subdevsrc;
  gboolean ret;

  subdevsrc = GST_SUBDEVSRC_CAST (camsrc);

  g_mutex_lock (subdevsrc->device_mutex);
  ret = gst_subdevsrc_libomap3camd_write_settings (subdevsrc, photoconf,
      scene_override);
  g_mutex_unlock (subdevsrc->device_mutex);

  return ret;

}


/*
 * gst_subdevsrc_set_property:
 * @object: #GstSubdevSrc object.
 * @prop_id: Property ID to be set.
 * @value: GValue containing the new value for the property.
 * pspec: GParamSpec object.
 *
 */
static void
gst_subdevsrc_set_property (GObject * object,
    guint prop_id, const GValue * value, GParamSpec * pspec)
{
  GstSubdevSrc *subdevsrc = GST_SUBDEVSRC_CAST (object);

  switch (prop_id) {
    case PROP_QUEUE_SIZE:
      subdevsrc->num_buffers = g_value_get_uint (value);
      break;
    case PROP_CAMERA_DEVICE:
      subdevsrc->camera_device = g_value_get_enum (value);
      GST_DEBUG_OBJECT (subdevsrc, "Camera device set to %d",
          subdevsrc->camera_device);

      if (subdevsrc->probed_caps) {
        gst_caps_unref (subdevsrc->probed_caps);
        subdevsrc->probed_caps = NULL;
      }
      if (subdevsrc->night_caps) {
        gst_caps_unref (subdevsrc->night_caps);
        subdevsrc->night_caps = NULL;
      }
      break;
    case PROP_VIDEO_TORCH:
    {
      /* Turn on video torch (assist light) only in video mode,
         otherwise just cache the value */
      subdevsrc->video_torch = g_value_get_boolean (value);

      if (subdevsrc->viewfinder_mode == GST_CAMERA_SRC_VIEWFINDER_MODE_VIDEO) {
        g_mutex_lock (subdevsrc->device_mutex);
        gst_subdevsrc_libomap3camd_set_assist_light (subdevsrc,
            subdevsrc->video_torch);
        g_mutex_unlock (subdevsrc->device_mutex);
      }
      else {
        GST_DEBUG_OBJECT (subdevsrc,
            "Setting video torch: not in video mode - cached setting: %d",
            subdevsrc->video_torch);
      }
      break;
    }
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}


/*
 * gst_subdevsrc_get_property:
 * @object: #GstSubdevSrc object.
 * @prop_id: Property ID to be get.
 * @value: Pointer to GValue where the property value should be stored.
 * pspec: GParamSpec object.
 */
static void
gst_subdevsrc_get_property (GObject * object,
    guint prop_id, GValue * value, GParamSpec * pspec)
{
  GstSubdevSrc *subdevsrc = GST_SUBDEVSRC_CAST (object);

  switch (prop_id) {
    case PROP_QUEUE_SIZE:
      g_value_set_uint (value, subdevsrc->num_buffers);
      break;
    case PROP_CAMERA_DEVICE:
      g_value_set_enum (value, subdevsrc->camera_device);
      break;
    case PROP_NIGHT_MODE_CAPS:
    {
      if (gst_caps_is_empty (subdevsrc->night_caps))
        gst_caps_replace (&subdevsrc->night_caps, NULL);

      if (!subdevsrc->night_caps) {
        subdevsrc->night_caps = gst_subdevsrc_probe_caps (subdevsrc, TRUE);
      }

      GST_DEBUG_OBJECT (subdevsrc, "NIGHT MODE CAPS: %" GST_PTR_FORMAT,
        subdevsrc->night_caps);

      gst_value_set_caps (value, subdevsrc->night_caps);
      break;
    }
    case PROP_VIDEO_TORCH:
      g_value_set_boolean (value, subdevsrc->video_torch);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}


/*
 * gst_subdevsrc_unlock:
 * @src: #GstCameraSrc object.
 *
 * Returns: TRUE on success.
 */
static gboolean
gst_subdevsrc_unlock (GstCameraSrc * src)
{
  GstSubdevSrc *subdevsrc = GST_SUBDEVSRC_CAST (src);

  GST_LOG_OBJECT (subdevsrc, "Flushing");
  gst_poll_set_flushing (subdevsrc->poll, TRUE);

  /* If grab_frame is waiting for a buffer, this will wake it up */
  if (subdevsrc->pool) {
    g_mutex_lock (subdevsrc->pool->lock);
    subdevsrc->pool->flushing = TRUE;
    g_cond_signal (subdevsrc->pool->data_cond);
    g_mutex_unlock (subdevsrc->pool->lock);
  }

  return TRUE;
}


/*
 * gst_subdevsrc_unlock_stop:
 * @src: #GstCameraSrc object.
 *
 * Returns: TRUE on success.
 */
static gboolean
gst_subdevsrc_unlock_stop (GstCameraSrc * src)
{
  GstSubdevSrc *subdevsrc = GST_SUBDEVSRC_CAST (src);

  GST_LOG_OBJECT (subdevsrc, "No longer flushing");
  gst_poll_set_flushing (subdevsrc->poll, FALSE);

  if (subdevsrc->pool) {
    g_mutex_lock (subdevsrc->pool->lock);
    subdevsrc->pool->flushing = FALSE;
    g_mutex_unlock (subdevsrc->pool->lock);
  }

  return TRUE;
}


/*
 * gst_subdevsrc_event:
 *
 * Checks if the given event is targeted for us and handles it appropriately.
 */
static gboolean
gst_subdevsrc_event (GstCameraSrc * src, GstEvent *event)
{
  GstSubdevSrc *subdevsrc;
  const GstStructure *structure;
  const gchar *name;
  gboolean ret = FALSE;

  subdevsrc = GST_SUBDEVSRC_CAST (src);
  structure = gst_event_get_structure (event);

  if (!structure) return FALSE;

  name = gst_structure_get_name (structure);

  if (!g_strcmp0 (name, "regions-of-interest")) {
    g_mutex_lock (subdevsrc->device_mutex);
    gst_subdevsrc_libomap3camd_handle_roi_event (subdevsrc, structure);
    g_mutex_unlock (subdevsrc->device_mutex);
    ret = TRUE;
  }
  else if (!g_strcmp0 (name, "motion-estimate")) {
    gint64 timedelta;
    gfloat vector;

    GST_LOG_OBJECT (subdevsrc, "Motion estimate event");

    if (gst_structure_get ((GstStructure *) structure,
      "motion-vector", G_TYPE_FLOAT, &vector,
      "frame-time-delta", G_TYPE_INT64, &timedelta,
      NULL))
    {
      g_mutex_lock (subdevsrc->device_mutex);
      gst_subdevsrc_libomap3camd_handle_motion (subdevsrc, vector, timedelta);
      g_mutex_unlock (subdevsrc->device_mutex);
      ret = TRUE;
    }
    else {
      GST_WARNING_OBJECT (subdevsrc, "Can't read motion vector from event");
    }
  }

  return ret;
}


/*
 * gst_subdevsrc_set_image_tags:
 * @camsrc: @GstCameraSrc object.
 * @tlist: #GstTagList where the tags should be set.
 *
 * Add HQ image tags into taglist.
 *
 * Returns: TRUE on success.
 */
static gboolean
gst_subdevsrc_set_image_tags (GstCameraSrc *camsrc, GstTagList *tlist)
{
  GstSubdevSrc *subdevsrc;
  gboolean ret;

  subdevsrc = GST_SUBDEVSRC_CAST (camsrc);

  GST_DEBUG_OBJECT (subdevsrc, "Image parameters query");

  g_mutex_lock (subdevsrc->device_mutex);
  ret = gst_subdevsrc_libomap3camd_get_image_tags (subdevsrc, tlist);
  g_mutex_unlock (subdevsrc->device_mutex);

  return ret;
}


/*
 * gst_subdevsrc_send_makernote:
 * @subdevsrc: #GstSubdevSrc object
 *
 * Sends makernote event to downstream.
 *
 * Returns: TRUE on success.
 */
static gboolean
gst_subdevsrc_send_makernote (GstSubdevSrc *subdevsrc)
{
  guint8 *mndata;
  guint mlen;
  gboolean ret;

  g_mutex_lock (subdevsrc->device_mutex);
  ret = gst_subdevsrc_libomap3camd_get_makernote (subdevsrc, &mndata, &mlen);
  g_mutex_unlock (subdevsrc->device_mutex);

  GST_DEBUG_OBJECT (subdevsrc, "makernote query: %d",  ret);

  if (ret && mndata && mlen > 0) {
    GstTagList *maker_note;
    GstBuffer *mnbuf;
    GstEvent *event;

    mnbuf = gst_buffer_new ();
    GST_BUFFER_MALLOCDATA (mnbuf) = mndata;
    GST_BUFFER_DATA (mnbuf) = mndata;
    GST_BUFFER_SIZE (mnbuf) = mlen;

    GST_DEBUG_OBJECT (subdevsrc, "got makernote data: %d bytes", mlen);

    maker_note = gst_tag_list_new ();
    gst_tag_list_add (maker_note, GST_TAG_MERGE_REPLACE,
                      GST_TAG_APPLICATION_DATA, mnbuf, NULL);

    event = gst_event_new_tag (maker_note);
    gst_pad_push_event (GST_BASE_SRC_PAD (subdevsrc), event);
    /* We set the MALLOCDATA above, so freeing the buffer also frees the data */
    gst_buffer_unref (mnbuf);

    GST_DEBUG_OBJECT (subdevsrc, "makernote event sent");
  }

  return ret;
}


/*
 * gst_subdevsrc_send_postproc_params:
 * @subdevsrc: #GstSubdevSrc object
 *
 * Sends post-processing parameters to downstream as a custom event.
 *
 * Returns: TRUE on success.
 */
static gboolean
gst_subdevsrc_send_postproc_params (GstSubdevSrc *subdevsrc)
{
  GstStructure *pp_struct;
  gboolean ret;

  g_mutex_lock (subdevsrc->device_mutex);
  ret = gst_subdevsrc_libomap3camd_get_pp_params (subdevsrc, &pp_struct);
  g_mutex_unlock (subdevsrc->device_mutex);

  if (ret) {
    GstEvent *stats;

    stats = gst_event_new_custom (GST_EVENT_CUSTOM_DOWNSTREAM, pp_struct);

    gst_pad_push_event (GST_BASE_SRC_PAD (subdevsrc), stats);

    GST_DEBUG_OBJECT (subdevsrc, "postproc params event sent");
  }

  return ret;
}


/*
 * gst_subdevsrc_grab_frame:
 * @camsrc: #GstCameraSrc object.
 * @buf: Pointer to GstBuffer object where the grabbed frame should be placed.
 * @phase: #GstCameraCapturePhase describing the phase of capturing.
 *
 * Capture one single frame from camera device. This function captures both
 * viewfinder frames and high-quality still image frames.
 *
 * Returns: #GstFlowReturn describing the status of the capture operation.
 */
GstFlowReturn
gst_subdevsrc_grab_frame (GstCameraSrc *camsrc, GstBuffer **buf,
    GstCameraCapturePhase phase)
{
  GstSubdevSrc *subdevsrc = GST_SUBDEVSRC_CAST (camsrc);

  GST_LOG_OBJECT (subdevsrc, "gst_subdevsrc_grab_frame, phase: %d", phase);

  /* HQ capture mode */
  if (phase == GST_CAMERA_CAPTURE) {
    gboolean ret;

    /* FIXME: Make this function raise GST error already */
    g_mutex_lock (subdevsrc->device_mutex);
    ret = gst_subdevsrc_libomap3camd_grab_hq_frame (subdevsrc, buf);
    g_mutex_unlock (subdevsrc->device_mutex);

    if (!ret) {
      return GST_FLOW_ERROR;
    }

    GST_LOG_OBJECT (subdevsrc, "Retrieving makernote");
    gst_subdevsrc_send_makernote (subdevsrc);
    GST_LOG_OBJECT (subdevsrc, "Retrieving PP params");
    gst_subdevsrc_send_postproc_params (subdevsrc);
    GST_LOG_OBJECT (subdevsrc, "Done");

    /* Set preview resolution again, so that a new buffer gets preallocated */
    gst_subdevsrc_set_preview_fmt (subdevsrc, &subdevsrc->preview.color_code,
        &subdevsrc->preview.width, &subdevsrc->preview.height);

    return GST_FLOW_OK;
  }

  if (g_atomic_int_compare_and_exchange (&subdevsrc->caf_updated, 1, 0)) {
    GstCameraFocusStatus s;
    gboolean lret;

    g_mutex_lock (subdevsrc->device_mutex);
    lret = gst_subdevsrc_libomap3camd_get_focus_status (subdevsrc, &s, FALSE);
    g_mutex_unlock (subdevsrc->device_mutex);

    if (lret) {
      GST_LOG_OBJECT (subdevsrc, "CAF status: %d", s.status);
      gst_camerasrc_caf_update (GST_CAMERA_SRC (subdevsrc), &s);
    }
  }

  if (g_atomic_int_compare_and_exchange (&subdevsrc->af_updated, 1, 0) &&
      phase == GST_CAMERA_AUTOFOCUS)
  {
    GstCameraFocusStatus s = { 0, };
    gboolean lret;

    g_mutex_lock (subdevsrc->device_mutex);
    lret = gst_subdevsrc_libomap3camd_get_focus_status (subdevsrc, &s, TRUE);
    g_mutex_unlock (subdevsrc->device_mutex);

    if (lret) {
      GST_DEBUG_OBJECT (subdevsrc, "AF status: %d", s.status);
      gst_camerasrc_af_update (GST_CAMERA_SRC (subdevsrc), &s);
      if (s.windows) {
        g_free (s.windows);
      }
    }
  }

  if (g_atomic_int_compare_and_exchange (&subdevsrc->shake_updated, 1, 0)) {
    GstPhotoShakeRisk risk;
    gboolean lret;

    g_mutex_lock (subdevsrc->device_mutex);
    lret = gst_subdevsrc_libomap3camd_check_shaking (subdevsrc, &risk);
    g_mutex_unlock (subdevsrc->device_mutex);

    if (lret) {
      GST_LOG_OBJECT (subdevsrc, "Shake risk: %d", risk);
      gst_camerasrc_shake_update (GST_CAMERA_SRC (subdevsrc), risk);
    }
  }

  if (g_atomic_int_compare_and_exchange (&subdevsrc->flash_updated, 1, 0)) {
    GstCameraFlashStatus status;
    gboolean lret;

    g_mutex_lock (subdevsrc->device_mutex);
    lret = gst_subdevsrc_libomap3camd_flash_status (subdevsrc, &status);
    g_mutex_unlock (subdevsrc->device_mutex);

    if (lret) {
      GST_LOG_OBJECT (subdevsrc, "Flash status: %d", status);
      gst_camerasrc_flash_status_update (GST_CAMERA_SRC (subdevsrc), status);
    }
  }

  if (g_atomic_int_compare_and_exchange (&subdevsrc->flicker_detected, 1, 0)) {
    GstFlickerReductionMode fmode;
    gboolean lret;

    GST_DEBUG_OBJECT (subdevsrc, "Flicker detected");

    g_mutex_lock (subdevsrc->device_mutex);
    lret = gst_subdevsrc_libomap3camd_flicker_status (subdevsrc, &fmode);
    g_mutex_unlock (subdevsrc->device_mutex);

    if (lret) {
      camsrc->photoconf.flicker_mode = fmode;
    }
  }

  if (camsrc->photoconf.flicker_mode == GST_PHOTOGRAPHY_FLICKER_REDUCTION_AUTO)
  {
    subdevsrc->flicker_counter++;

    if (subdevsrc->flicker_counter > FLICKER_REACTIVATION_INTERVAL *
        subdevsrc->current.fps_n / subdevsrc->current.fps_d)
    {
      GST_DEBUG_OBJECT (subdevsrc, "Restarting flicker detection algorithm");

      gst_subdevsrc_libomap3camd_set_flicker_mode (subdevsrc,
          camsrc->photoconf.flicker_mode);
      subdevsrc->flicker_counter = 0;
    }
  }

  return gst_subdevsrc_grab_vf_frame (subdevsrc, buf);
}


/*
 * gst_subdevsrc_get_max_zoom:
 * @camsrc: @GstCameraSrc object.
 * @zoom: Pointer where to store the value.
 *
 * Retrieve max zoom factor. This requires the capture resolution to be set
 * first.
 *
 * Returns: TRUE on success.
 */
static gboolean
gst_subdevsrc_get_max_zoom (GstCameraSrc *camsrc, gfloat *zoom)
{
  GstSubdevSrc *subdevsrc = GST_SUBDEVSRC_CAST (camsrc);
  guint tmp_w, tmp_h;

  if (!subdevsrc->is_open) {
    return FALSE;
  }

  g_return_val_if_fail (subdevsrc->sensor_fmts != NULL, FALSE);

  GST_DEBUG_OBJECT (subdevsrc, "Getting max zoom from library");

  if (subdevsrc->capture.width && subdevsrc->capture.height) {

    GstVideoFrameInfo *finfo = NULL;
    GSList *iter;
    gfloat tmp;

    /* tmp_w x tmp_h tell the minimum crop rectangle size which still can
       be used to produce desired-sized image capture */
    tmp_w = subdevsrc->capture.width / MAX_RESIZER_FACTOR;
    tmp_h = subdevsrc->capture.height / MAX_RESIZER_FACTOR;

    /* Find the sensor mode that is used for image capture */
    for (iter = subdevsrc->sensor_fmts; iter; iter = iter->next) {
      finfo = (GstVideoFrameInfo *) iter->data;
      if (finfo->width > subdevsrc->capture.width &&
        finfo->height > subdevsrc->capture.height)
      {
        break;
      }
    }

    /* we know that we can safely crop sensor resolution down to tmp_w x tmp_h
     * and still produce the output size. Now we just need to calculate that
     * what is the size ratio of this crop rectangle and the full sensor size.
     * We calculate ratios separately for width and height and choose
     * the smaller one, because that is the one which actually restricts
     * the zooming. We also make sure that ratio is always bigger than 1 */

    tmp = MIN (finfo->width / tmp_w, finfo->height / tmp_h);
    *zoom = MAX (tmp, 1.0);

    /* If source is active, then we also know the limit from libomap3camd */
    if (subdevsrc->is_active) {
      *zoom = CLAMP (*zoom, 1.0, subdevsrc->lib_max_zoom);
    }
  }
  else {
    GST_WARNING_OBJECT (subdevsrc,
        "Getting max zoom without setting capture resolution first");
    /* We should be able to zoom at least as much as resizer supports for
     * all resolutions. */
    *zoom = (gfloat) MAX_RESIZER_FACTOR;
  }

  GST_DEBUG_OBJECT (subdevsrc, "MAX zoom factor for resolution %dx%d is %f",
      subdevsrc->capture.width, subdevsrc->capture.height, *zoom);

  return TRUE;
}


/*
 * gst_subdevsrc_set_zoom:
 * @camsrc: @GstCameraSrc object.
 * @zoom: zoom factor to be set.
 *
 * Set the zoom factor for outputted image.
 *
 * Returns: TRUE on success.
 */
static gboolean
gst_subdevsrc_set_zoom (GstCameraSrc *camsrc, gfloat zoom)
{
  GstSubdevSrc *subdevsrc;
  gboolean ret = TRUE;

  subdevsrc = GST_SUBDEVSRC_CAST (camsrc);

  GST_DEBUG_OBJECT (subdevsrc, "ZOOM: %f", zoom);

  if (subdevsrc->is_active) {
    /* Restrict zoom factor according to the limit from libomap3camd */
    subdevsrc->zoom_factor = CLAMP (zoom, 1.0, subdevsrc->lib_max_zoom);

    if (subdevsrc->capture_mode != GST_CAMERA_SRC_CAPTURE_MODE_STILL) {
      ret = gst_subdevsrc_update_cropping (subdevsrc, subdevsrc->current.width,
          subdevsrc->current.height, subdevsrc->zoom_factor);
    }

    if (ret) {
      g_mutex_lock (subdevsrc->device_mutex);
      ret = gst_subdevsrc_libomap3camd_set_zoom (subdevsrc,
          subdevsrc->zoom_factor);
      g_mutex_unlock (subdevsrc->device_mutex);
    }
  }
  else {
    subdevsrc->zoom_factor = zoom;
  }

  return ret;
}


/*
 * gst_subdevsrc_open:
 * @camsrc: #GstCameraSrc object.
 *
 * Open devices and initialize the element.
 *
 * Returns: TRUE on success.
 */
static gboolean
gst_subdevsrc_open (GstCameraSrc *camsrc)
{
  GstSubdevSrc *subdevsrc = GST_SUBDEVSRC_CAST (camsrc);
  gboolean ret;

  GST_DEBUG_OBJECT (subdevsrc, "Opening devices");

  if (!gst_subdevsrc_open_devices (subdevsrc)) {
    return FALSE;
  }

  g_mutex_lock (subdevsrc->device_mutex);
  ret = gst_subdevsrc_libomap3camd_init (subdevsrc);
  g_mutex_unlock (subdevsrc->device_mutex);

  if (!ret) {
    gst_subdevsrc_close_devices (subdevsrc);
  }

  return ret;
}


/*
 * gst_subdevsrc_close:
 * @camsrc: #GstCameraSrc object:
 *
 * Close all devices and deallocate resources.
 *
 * Returns: TRUE on success.
 */
static gboolean
gst_subdevsrc_close (GstCameraSrc *camsrc)
{
  GstSubdevSrc *subdevsrc = GST_SUBDEVSRC_CAST (camsrc);

  GST_DEBUG_OBJECT (subdevsrc, "gst_subdevsrc_close");

  /* Free allocated HQ image buffers */
  if (subdevsrc->capture.pool) {
    if (subdevsrc->capture.buffer) {
      gst_recycle_pool_free_buffer (subdevsrc->capture.pool,
          subdevsrc->capture.buffer);
      subdevsrc->capture.buffer = NULL;
    }
    gst_recycle_pool_destroy (subdevsrc->capture.pool);
    subdevsrc->capture.pool = NULL;
  }

  /* Free allocated preview buffers */
  if (subdevsrc->preview.pool) {
    if (subdevsrc->preview.buffer) {
      gst_recycle_pool_free_buffer (subdevsrc->preview.pool,
          subdevsrc->preview.buffer);
      subdevsrc->preview.buffer = NULL;
    }
    gst_recycle_pool_destroy (subdevsrc->preview.pool);
    subdevsrc->preview.pool = NULL;
  }

  g_mutex_lock (subdevsrc->device_mutex);
  gst_subdevsrc_libomap3camd_deinit (subdevsrc);
  g_mutex_unlock (subdevsrc->device_mutex);

  gst_subdevsrc_close_devices (subdevsrc);

  gst_subdevsrc_reset_videomode (&subdevsrc->current);
  gst_subdevsrc_reset_videomode (&subdevsrc->sensor);
  gst_subdevsrc_reset_videomode (&subdevsrc->capture);
  gst_subdevsrc_reset_videomode (&subdevsrc->preview);
  gst_subdevsrc_reset_videomode (&subdevsrc->resizer);

  return TRUE;
}


/*
 * gst_subdevsrc_start:
 * @camsrc: #GstCameraSrc object.
 * @caps: GstCaps to be assigned into GstBuffers.
 *
 * Prepare camera device for streaming. Allocates video buffers and queues
 * them to video device.
 *
 * Returns: TRUE on success.
 */
static gboolean
gst_subdevsrc_start (GstCameraSrc *camsrc, GstCaps *caps)
{
  GstSubdevSrc *subdevsrc = GST_SUBDEVSRC_CAST (camsrc);
  gboolean ret = TRUE;

  GST_DEBUG_OBJECT (subdevsrc, "starting up");

  subdevsrc->af_updated = 0;
  subdevsrc->caf_updated = 0;
  subdevsrc->shake_updated = 0;
  subdevsrc->flicker_detected = 0;

  if (subdevsrc->capture_mode == GST_CAMERA_SRC_CAPTURE_MODE_STILL) {
    GST_DEBUG_OBJECT (subdevsrc, "Initialization not needed in STILL mode");
    return TRUE;
  }
  else if (subdevsrc->is_active) {
    GST_DEBUG_OBJECT (subdevsrc, "Already active");
    return TRUE;
  }

  g_mutex_lock (subdevsrc->device_mutex);

  if (ret) {
    ret = gst_subdevsrc_libomap3camd_set_vf_mode (subdevsrc,
      subdevsrc->viewfinder_mode);
  }

  if (ret) {
    ret = gst_subdevsrc_libomap3camd_daemon_start (subdevsrc);
  }

  if (ret) {
    ret = gst_subdevsrc_libomap3camd_set_flicker_mode (subdevsrc,
        camsrc->photoconf.flicker_mode);
  }

  if (ret && subdevsrc->viewfinder_mode == GST_CAMERA_SRC_VIEWFINDER_MODE_VIDEO) {
    ret = gst_subdevsrc_libomap3camd_set_assist_light (subdevsrc,
        subdevsrc->video_torch);
  }

  g_mutex_unlock (subdevsrc->device_mutex);

  if (ret) {
    ret = gst_subdevsrc_start_streaming (subdevsrc, caps);
  }

  if (ret) {
    subdevsrc->is_active = TRUE;
  }
  else {
    GST_ELEMENT_ERROR (subdevsrc, RESOURCE, SETTINGS,
      ("Error initializing libomap3camd subsystem"), GST_ERROR_SYSTEM);
  }

  return ret;
}


/*
 * gst_subdevsrc_stop:
 * @camsrc: #GstCameraSrc object.
 *
 * Stop video streaming.
 *
 * Returns: TRUE on success.
 */
static gboolean
gst_subdevsrc_stop (GstCameraSrc *camsrc)
{
  GstSubdevSrc *subdevsrc = GST_SUBDEVSRC_CAST (camsrc);
  gboolean ret;

  if (!subdevsrc->is_active) {
    GST_DEBUG_OBJECT (subdevsrc, "Streaming not active");
    return TRUE;
  }

  GST_DEBUG_OBJECT (subdevsrc, "Stop streaming");

  g_mutex_lock (subdevsrc->device_mutex);
  gst_subdevsrc_libomap3camd_set_assist_light (subdevsrc, FALSE);
  gst_subdevsrc_libomap3camd_daemon_stop (subdevsrc);
  subdevsrc->is_active = FALSE;
  g_mutex_unlock (subdevsrc->device_mutex);

  ret = gst_subdevsrc_stop_streaming (subdevsrc);

  return ret;
}


/*
 * gst_subdevsrc_get_attribute:
 * @camsrc: #GstCameraSrc object.
 * @attribute: Attribute ID to be get.
 * @value: Pointer where the value should be set.
 *
 * Retrieve a value of certain V4L2 attribute from the device.
 *
 * Returns: TRUE on success.
 */
static gboolean
gst_subdevsrc_get_attribute (GstCameraSrc *camsrc, int attribute, int *value)
{
#ifndef GST_DISABLE_GST_DEBUG
  GstSubdevSrc *subdevsrc = GST_SUBDEVSRC_CAST (camsrc);
#endif

  GST_DEBUG_OBJECT (subdevsrc, "Get attribute %d", attribute);

  return FALSE;
}


/*
 * gst_subdevsrc_set_attribute:
 * @camsrc: #GstCameraSrc object.
 * @attribute: Attribute ID to be set.
 * @value: The value to be set.
 *
 * Set certain V4L2 attribute value to the device.
 *
 * Returns: TRUE on success.
 */
static gboolean
gst_subdevsrc_set_attribute (GstCameraSrc *camsrc, int attribute,
    const int value)
{
#ifndef GST_DISABLE_GST_DEBUG
  GstSubdevSrc *subdevsrc = GST_SUBDEVSRC_CAST (camsrc);
#endif

  GST_DEBUG_OBJECT (subdevsrc, "Set attribute: %d = %d", attribute, value);

  return FALSE;
}


/*
 * plugin_init:
 * @plugin: GstPlugin
 *
 * Returns: TRUE on success.
 */
static gboolean
plugin_init (GstPlugin * plugin)
{
  GST_DEBUG_CATEGORY_INIT (gst_subdevsrc_debug, "subdevsrc", 0, \
      "MediaController camera source");

  return gst_element_register (plugin, "subdevsrc",
      GST_RANK_NONE, GST_TYPE_SUBDEVSRC);
}

GST_PLUGIN_DEFINE (GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "subdevcam",
    "MediaController image capturing element",
    plugin_init,
    VERSION,
    "LGPL",
    "Nokia", "www.nokia.com")
