/* GStreamer MediaController camera source
 * Copyright (C) 2010 Nokia Corporation <multimedia@maemo.org>
 *
 * Contact: Maemo Multimedia <multimedia@maemo.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef __GST_SUBDEVSRC_H__
#define __GST_SUBDEVSRC_H__

#include <linux/videodev2.h>

#include <gst/gst.h>
#include <gst/camera/gstcamerasrc.h>
#include <omap3cam.h>

#include "mediacontroller.h"
#include "configparser.h"
#include "bufferpool.h"
#include "recyclepool.h"

G_BEGIN_DECLS

#define GST_TYPE_SUBDEVSRC              (gst_subdevsrc_get_type())
#define GST_SUBDEVSRC(obj)              (G_TYPE_CHECK_INSTANCE_CAST((obj), GST_TYPE_SUBDEVSRC, GstSubdevSrc))
#define GST_SUBDEVSRC_CLASS(klass)      (G_TYPE_CHECK_CLASS_CAST((klass), GST_TYPE_SUBDEVSRC, GstSubdevSrcClass))
#define GST_IS_SUBDEVSRC(obj)           (G_TYPE_CHECK_INSTANCE_TYPE((obj), GST_TYPE_SUBDEVSRC))
#define GST_IS_SUBDEVSRC_CLASS(klass)   (G_TYPE_CHECK_CLASS_TYPE((klass), GST_TYPE_SUBDEVSRC))
#define GST_SUBDEVSRC_CAST(obj)         ((GstSubdevSrc *)(obj))
#define GST_SUBDEVSRC_CLASS_CAST(klass) ((GstSubdevSrcClass *)(klass))


/**
 * MAX_RESIZER_FACTOR:
 *
 * Maximum scale factor for resizer.
 */
#define MAX_RESIZER_FACTOR (4)

#define CAPTURE_WIDTH_MODULO 16
#define CAPTURE_HEIGHT_MODULO 1
#define PREVIEW_WIDTH_MODULO 16
#define PREVIEW_HEIGHT_MODULO 1

#define RECYCLE_BUFFERS

/* Re-activate flicker detection algorithm every 10 seconds in AUTO mode */
#define FLICKER_REACTIVATION_INTERVAL 10

typedef struct _GstSubdevSrc GstSubdevSrc;
typedef struct _GstSubdevSrcClass GstSubdevSrcClass;
typedef struct _GstVideoFrameInfo GstVideoFrameInfo;
typedef struct _GstVideoMode GstVideoMode;

#define GST_TYPE_SUBDEVSRC_CAMERA_DEVICE (gst_subdevsrc_camera_device_get_type())

/**
 * GstSubdevSrcCameraDevice:
 * @GST_SUBDEVSRC_CAPTURE_DEVICE_PRIMARY: Primary camera device
 * @GST_SUBDEVSRC_CAPTURE_DEVICE_SECONDARY: Secondary camera device
 *
 * Enumeration for camera devices.
 */
typedef enum {
  GST_SUBDEVSRC_CAPTURE_DEVICE_PRIMARY,
  GST_SUBDEVSRC_CAPTURE_DEVICE_SECONDARY,
} GstSubdevSrcCameraDevice;


/**
 * GstVideoFrameInfo:
 * @width: Frame width.
 * @height: Frame height.
 * @color_code: Frame color code in subdev format (v4l2_mbus_pixelcode).
 * @framerates: GValue list of fractions.
 * @ar_category: Aspect ratio category.
 *
 * Structure containing information about frame format.
 */
struct _GstVideoFrameInfo
{
  guint width, height;
  guint32 color_code;
  GValue *framerates;
  guint ar_category;
};


/**
 * GstVideoMode:
 * @width: Frame width.
 * @height: Frame height.
 * @fps_n: Frame rate numerator.
 * @fps_d: Frame rate denominator.
 * @color_code: Frame color code in subdev format (v4l2_mbus_pixelcode).
 * @pool: Image pool related to this mode (if any).
 * @buffer: Currently active buffer related to this mode (if any).
 * @buffer_lock: Mutex protecting the buffer (if any).
 *
 * Structure containing information about video mode and related buffer pool,
 * active buffer and lock protecting it.
 */
struct _GstVideoMode
{
  guint width, height;
  guint fps_n, fps_d;
  guint32 color_code;
  GstRecyclePool *pool;
  GstRecycleBuffer *buffer;
  GMutex *buffer_lock;
};


/**
 * GstSubdevSrc:
 * @element: the parent element.
 * @sensor_fd: sensor subdev file descriptor.
 * @rsz_fd: resizer subdev file descriptor.
 * @video_fd: video output device file descriptor.
 * @is_open: TRUE, when all subdevs are opened.
 * @is_active: TRUE, when streaming buffers.
 * @current: currently configured output resolution.
 * @sensor: currently configured sensor resolution.
 * @capture: capture resolution.
 * @preview: preview resolution.
 * @resizer: Incoming resizer resolution.
 * @camera_device: Camera device enumeration.
 * @pool: Video buffer pool object.
 * @mc: Media controller struct.
 * @cp: Pipeline configuration parser.
 * @num_buffers: Number of video buffers used.
 * @poll: GstPoll object for monitoring video output device.
 * @probed_caps: GstCaps describing possible output formats.
 * @frame_byte_size: The size of currently configured format.
 * @sensor_fmts: sensor resolutions. A list of GstVideoFrameInfo structures.
 * @capture_formats: available capture formats (v4l2_subdev_pad_mbus_code_enum).
 * @max_zoom_factor: Maximum zoom factor.
 * @zoom_factor: Current zoom factor.
 * @device_mutex: Mutex used for accessing the devices.
 * @offset: Offset
 * @libhnd: Libomap3camd instance handle.
 * @scene_map: Scene mode mapping table. This is different in VIDEO/STILL modes.
 * @capture_mode: Currently used capturing mode.
 * @viewfinder_mode: Currently used viewfinder mode.
 * @initialized: TRUE when libomap3camd has been initialized.
 * @af_updated: TRUE when libomap3camd has notified about AF status change.
 * @caf_updated: TRUE when libomap3camd has notified about CAF status change.
 * @shake_updated: TRUE when libomap3camd has notified about shake risk change.
 *
 * The opaque GstSubdevSrc data structure.
 */
struct _GstSubdevSrc {
  GstCameraSrc element;

  gint sensor_fd;                /* sensor subdev file descriptor */
  gint rsz_fd;                   /* resizer subdev file descriptor */
  gint video_fd;                 /* video output device file descriptor */
  gboolean is_open;              /* TRUE, when all subdevs are opened */
  gboolean is_active;            /* TRUE, when streaming buffers */

  GstVideoMode current;          /* currently configured output resolution */
  GstVideoMode sensor;           /* currently configured sensor resolution */
  GstVideoMode capture;          /* image capture resolution */
  GstVideoMode preview;          /* image preview resolution */
  GstVideoMode resizer;          /* resizer input resolution */

  guint preview_min_w;
  guint preview_min_h;

  GstSubdevSrcCameraDevice camera_device;
  GstSubdevSrcBufferPool *pool;
  GstMediaController *mc;        /* Media controller struct */
  GstConfigParser *cp;           /* Pipeline configuration parser */
  guint32 num_buffers;
  GstPoll *poll;
  GstCaps *probed_caps;
  guint32 frame_byte_size;       /* The size of currently configured format */
  GSList *sensor_fmts;           /* sensor resolutions */
  GSList *capture_formats;       /* available capture formats */
  GstCaps *night_caps;           /* Night mode caps for selected VF resolution */
  gboolean video_torch;          /* Video recording torch light on/off */

  gfloat max_zoom_factor;        /* Maximum zoom factor */
  gdouble lib_max_zoom;          /* Max zoom supported by libomap3camd */
  gfloat zoom_factor;            /* Currently used zoom factor */

  GMutex *device_mutex;
  guint64 offset;
  guint flicker_counter;

  void *libhnd;
  const gint (*scene_map)[2];

  GstCameraSrcCaptureMode capture_mode;
  GstCameraSrcViewfinderMode viewfinder_mode;

  gboolean initialized;
  gint af_updated;
  gint caf_updated;
  gint shake_updated;
  gint flash_updated;
  gint flicker_detected;
};

/**
* GstSubdevSrcClass:
* @parent_class: Element parent class.
*
* The opaque GstSubdevSrcClass data structure.
*/
struct _GstSubdevSrcClass {
  GstCameraSrcClass parent_class;
};

GType gst_subdevsrc_get_type (void);

G_END_DECLS

#endif /* __GST_SUBDEVSRC_H__ */
