/* GStreamer
 *
 * Copyright (C) 2010 Nokia Corporation. <multimedia@maemo.org>
 *
 * libomap3camd_calls.c: camera library calls
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <omap3cam/omap3cam.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>
#include <unistd.h>
#include <math.h>

#include <gst/tag/tag.h>

#include "gstsubdevcamsrc.h"
#include "libomap3camd_calls.h"

/* Define this to enable usage of VIDEO_NIGHT and VIDEO_AUTO modes */
#define USE_SEPARATE_VIDEO_SCENE_MODES

GST_DEBUG_CATEGORY_EXTERN (gst_subdevsrc_debug);
#define GST_CAT_DEFAULT gst_subdevsrc_debug

/* edge enchancement setting names */
#define EENF_EE_STRENGTH               "edge-enhancement-strength"
#define EENF_EE_WE_THRESHOLD           "weak-edge-threshold"
#define EENF_EE_SE_THRESHOLD           "strong-edge-threshold"
#define EENF_LNF_STRENGTH_LOW          "luma-noise-filter-strength-low"
#define EENF_LNF_STRENGTH_MID          "luma-noise-filter-strength-mid"
#define EENF_LNF_STRENGTH_HI           "luma-noise-filter-strength-high"
#define EENF_CBF_STRENGTH_LOW          "cb-noise-filter-strength-low"
#define EENF_CBF_STRENGTH_MID          "cb-noise-filter-strength-mid"
#define EENF_CBF_STRENGTH_HI           "cb-noise-filter-strength-high"
#define EENF_CRF_STRENGTH_LOW          "cr-noise-filter-strength-low"
#define EENF_CRF_STRENGTH_MID          "cr-noise-filter-strength-mid"
#define EENF_CRF_STRENGTH_HI           "cr-noise-filter-strength-high"
#define EENF_SHADING_VERT_PARAM1       "shading-vert-param1"
#define EENF_SHADING_VERT_PARAM2       "shading-vert-param2"
#define EENF_SHADING_HORZ_PARAM1       "shading-horz-param1"
#define EENF_SHADING_HORZ_PARAM2       "shading-horz-param2"
#define EENF_SHADING_GAIN_SCALE        "shading-gain-scale"
#define EENF_SHADING_GAIN_OFFSET       "shading-gain-offset"
#define EENF_SHADING_GAIN_MAXVALUE     "shading-gain-maxvalue"
#define EENF_DOWNSAMPLE_CB_CR          "ratio-downsample-cb-cr"


/* These map tables help to adapt the general photography interface to
 * different device implementations.
 *
 * Table index 0 = Enumeration value from GstPhotography
 * Table index 1 = Device/Implementation -specific setting value
 */

static const gint gst_subdevsrc_wb_map[][2] =
{
  { GST_PHOTOGRAPHY_WB_MODE_AUTO,        CAM_AWB_MODE_AUTO },
  { GST_PHOTOGRAPHY_WB_MODE_DAYLIGHT,    CAM_AWB_MODE_DAYLIGHT },
  { GST_PHOTOGRAPHY_WB_MODE_SUNSET,      CAM_AWB_MODE_SUNSET },
  { GST_PHOTOGRAPHY_WB_MODE_CLOUDY,      CAM_AWB_MODE_CLOUDY },
  { GST_PHOTOGRAPHY_WB_MODE_TUNGSTEN,    CAM_AWB_MODE_TUNGSTEN },
  { GST_PHOTOGRAPHY_WB_MODE_FLUORESCENT, CAM_AWB_MODE_FLUORESCENT },
  { -1, -1 }
};

static const gint gst_subdevsrc_effect_map[][2] =
{
  { GST_PHOTOGRAPHY_COLOUR_TONE_MODE_NORMAL,       CAM_GENERAL_EFFECT_TYPE_NORMAL },
  { GST_PHOTOGRAPHY_COLOUR_TONE_MODE_SEPIA,        CAM_GENERAL_EFFECT_TYPE_SEPIA },
  { GST_PHOTOGRAPHY_COLOUR_TONE_MODE_NEGATIVE,     CAM_GENERAL_EFFECT_TYPE_NEGATIVE },
  { GST_PHOTOGRAPHY_COLOUR_TONE_MODE_GRAYSCALE,    CAM_GENERAL_EFFECT_TYPE_GRAYSCALE },
  { GST_PHOTOGRAPHY_COLOUR_TONE_MODE_NATURAL,      CAM_GENERAL_EFFECT_TYPE_NATURAL },
  { GST_PHOTOGRAPHY_COLOUR_TONE_MODE_VIVID,        CAM_GENERAL_EFFECT_TYPE_VIVID },
  { GST_PHOTOGRAPHY_COLOUR_TONE_MODE_COLORSWAP,    CAM_GENERAL_EFFECT_TYPE_COLORSWAP },
  { GST_PHOTOGRAPHY_COLOUR_TONE_MODE_SOLARIZE,     CAM_GENERAL_EFFECT_TYPE_SOLARIZE },
  { GST_PHOTOGRAPHY_COLOUR_TONE_MODE_OUT_OF_FOCUS, CAM_GENERAL_EFFECT_TYPE_OUT_OF_FOCUS },
  { GST_PHOTOGRAPHY_COLOUR_TONE_MODE_SKY_BLUE,     CAM_GENERAL_EFFECT_TYPE_NORMAL },
  { GST_PHOTOGRAPHY_COLOUR_TONE_MODE_GRASS_GREEN,  CAM_GENERAL_EFFECT_TYPE_NORMAL },
  { GST_PHOTOGRAPHY_COLOUR_TONE_MODE_SKIN_WHITEN,  CAM_GENERAL_EFFECT_TYPE_NORMAL },
  { -1, -1 }
};

static const gint gst_subdevsrc_scene_map_image[][2] =
{
  { GST_PHOTOGRAPHY_SCENE_MODE_AUTO,      CAM_GENERAL_SCENE_MODE_AUTO },
  { GST_PHOTOGRAPHY_SCENE_MODE_CLOSEUP,   CAM_GENERAL_SCENE_MODE_CLOSEUP },
  { GST_PHOTOGRAPHY_SCENE_MODE_PORTRAIT , CAM_GENERAL_SCENE_MODE_PORTRAIT },
  { GST_PHOTOGRAPHY_SCENE_MODE_LANDSCAPE, CAM_GENERAL_SCENE_MODE_LANDSCAPE },
  { GST_PHOTOGRAPHY_SCENE_MODE_SPORT,     CAM_GENERAL_SCENE_MODE_SPORT },
  { GST_PHOTOGRAPHY_SCENE_MODE_NIGHT,     CAM_GENERAL_SCENE_MODE_NIGHT_PORTRAIT },
  { GST_PHOTOGRAPHY_SCENE_MODE_MANUAL,    CAM_GENERAL_SCENE_MODE_AUTO },
  { -1, -1 }
};

static const gint gst_subdevsrc_scene_map_video[][2] =
{
  { GST_PHOTOGRAPHY_SCENE_MODE_AUTO,      CAM_GENERAL_SCENE_MODE_VIDEO_AUTO },
  { GST_PHOTOGRAPHY_SCENE_MODE_CLOSEUP,   CAM_GENERAL_SCENE_MODE_CLOSEUP },
  { GST_PHOTOGRAPHY_SCENE_MODE_PORTRAIT,  CAM_GENERAL_SCENE_MODE_PORTRAIT },
  { GST_PHOTOGRAPHY_SCENE_MODE_LANDSCAPE, CAM_GENERAL_SCENE_MODE_LANDSCAPE },
  { GST_PHOTOGRAPHY_SCENE_MODE_SPORT,     CAM_GENERAL_SCENE_MODE_SPORT },
  { GST_PHOTOGRAPHY_SCENE_MODE_NIGHT,     CAM_GENERAL_SCENE_MODE_VIDEO_NIGHT },
  { GST_PHOTOGRAPHY_SCENE_MODE_MANUAL,    CAM_GENERAL_SCENE_MODE_VIDEO_AUTO },
  { -1, -1 }
};

static const gint gst_subdevsrc_flash_map[][2] =
{
  { GST_PHOTOGRAPHY_FLASH_MODE_AUTO,    CAM_LIGHT_FLASH_MODE_AUTO },
  { GST_PHOTOGRAPHY_FLASH_MODE_OFF,     CAM_LIGHT_FLASH_MODE_OFF },
  { GST_PHOTOGRAPHY_FLASH_MODE_ON,      CAM_LIGHT_FLASH_MODE_ON },
  { GST_PHOTOGRAPHY_FLASH_MODE_FILL_IN, CAM_LIGHT_FLASH_MODE_FILL_IN },
  { GST_PHOTOGRAPHY_FLASH_MODE_RED_EYE, CAM_LIGHT_FLASH_MODE_RED_EYE },
  { -1, -1 }
};

static const gint gst_subdevsrc_focus_map[][2] =
{
  { GST_PHOTOGRAPHY_FOCUS_MODE_AUTO,                CAM_FOCUS_MODE_AUTO },
  { GST_PHOTOGRAPHY_FOCUS_MODE_MACRO,               CAM_FOCUS_MODE_MACRO },
  { GST_PHOTOGRAPHY_FOCUS_MODE_PORTRAIT,            CAM_FOCUS_MODE_PORTRAIT },
  { GST_PHOTOGRAPHY_FOCUS_MODE_INFINITY,            CAM_FOCUS_MODE_INFINITY },
  { GST_PHOTOGRAPHY_FOCUS_MODE_HYPERFOCAL,          CAM_FOCUS_MODE_HYPERFOCAL },
  { GST_PHOTOGRAPHY_FOCUS_MODE_EXTENDED,            CAM_FOCUS_MODE_EXTENDED },
  { GST_PHOTOGRAPHY_FOCUS_MODE_CONTINUOUS_NORMAL,   CAM_FOCUS_MODE_AUTO },
  { GST_PHOTOGRAPHY_FOCUS_MODE_CONTINUOUS_EXTENDED, CAM_FOCUS_MODE_AUTO },
  { -1, -1 }
};

static const gint gst_subdevsrc_flicker_map[][2] =
{
  { GST_PHOTOGRAPHY_FLICKER_REDUCTION_OFF,  CAM_GENERAL_FLICKER_REDUCTION_MODE_OFF },
  { GST_PHOTOGRAPHY_FLICKER_REDUCTION_50HZ, CAM_GENERAL_FLICKER_REDUCTION_MODE_50HZ },
  { GST_PHOTOGRAPHY_FLICKER_REDUCTION_60HZ, CAM_GENERAL_FLICKER_REDUCTION_MODE_60HZ },
  { GST_PHOTOGRAPHY_FLICKER_REDUCTION_AUTO, CAM_GENERAL_FLICKER_REDUCTION_MODE_AUTO },
  { -1, -1 }
};

static const gint gst_subdevsrc_iso_map[][2] =
{
  { CAM_EXPOSURE_ISO_VALUE_800,  800 },
  { CAM_EXPOSURE_ISO_VALUE_400,  400 },
  { CAM_EXPOSURE_ISO_VALUE_200,  200 },
  { CAM_EXPOSURE_ISO_VALUE_100,  100 },
  { CAM_EXPOSURE_ISO_VALUE_AUTO, 0 }
};

static const gint gst_subdevsrc_shakerisk_map[][2] =
{
  { GST_PHOTOGRAPHY_SHAKE_RISK_LOW,    CAM_EXPOSURE_SHAKE_RISK_LOW },
  { GST_PHOTOGRAPHY_SHAKE_RISK_MEDIUM, CAM_EXPOSURE_SHAKE_RISK_MEDIUM },
  { GST_PHOTOGRAPHY_SHAKE_RISK_HIGH,   CAM_EXPOSURE_SHAKE_RISK_HIGH },
  { -1, -1 }
};

static const gint gst_subdevsrc_focus_status_map[][2] =
{
  { GST_PHOTOGRAPHY_FOCUS_STATUS_NONE,    CAM_FOCUS_STATUS_IDLE },
  { GST_PHOTOGRAPHY_FOCUS_STATUS_RUNNING, CAM_FOCUS_STATUS_RUNNING },
  { GST_PHOTOGRAPHY_FOCUS_STATUS_SUCCESS, CAM_FOCUS_STATUS_SUCCESS },
  { GST_PHOTOGRAPHY_FOCUS_STATUS_FAIL,    CAM_FOCUS_STATUS_FAIL },
  { -1, -1 }
};

static const cam_engine_t gst_op_mode_to_engine_type_map[] =
{
  CAM_ENGINE_VIEWFINDER,    /* GST_PHOTOGRAPHY_OPERATION_MODE_VIEWFINDER */
  CAM_ENGINE_STILL_CAPTURE, /* GST_PHOTOGRAPHY_OPERATION_MODE_IMAGE_CAPTURE */
  CAM_ENGINE_SNAP_SHOT      /* GST_PHOTOGRAPHY_OPERATION_MODE_PREVIEW */
};


#ifndef GST_DISABLE_GST_DEBUG
/*
 * Libomap3camd error string mapping table.
 */
static const gchar *cameralib_error_map[] =
{
  "CAM_ERR_NONE",
  "CAM_ERR_PARAM",
  "CAM_ERR_UNSUPP",
  "CAM_ERR_HW",
  "CAM_ERR_SYS",
  "CAM_ERR_LEXIT",
  "CAM_ERR_DEPRECATED",
  "CAM_ERR_INVALID_STATE",
  "CAM_ERR_INTERNAL"
};
#endif


/*
 *
 */
static cam_exposure_iso_value_t
gst_subdevsrc_libomap3camd_map_iso (guint iso)
{
  guint i = 0;

  while (gst_subdevsrc_iso_map[i][1] != 0) {
    if (iso >= gst_subdevsrc_iso_map[i][1]) {
      break;
    }
    i++;
  }

  return gst_subdevsrc_iso_map[i][0];
}


/*
 *
 */
static guint
gst_subdevsrc_libomap3camd_reverse_map_iso (cam_exposure_iso_value_t iso)
{
  guint i = 0;

  while (gst_subdevsrc_iso_map[i][1] != 0) {
    if (iso == gst_subdevsrc_iso_map[i][0]) {
      break;
    }
    i++;
  }

  return gst_subdevsrc_iso_map[i][1];
}


/*
 * Return mapping for item from given table. If @reverse is FALSE, @item is
 * considered as a GstPhotography enumeration value and the function will
 * return matching libomap3camd value for it. If @reverse is TRUE, @item
 * is considered as a libomap3camd value and this function returns
 * corresponding GstPhotography value. If mapping is not found, mapping
 * for table index 0 is returned.
 */
static gint
find_item (const gint table[][2], const gint item, gboolean reverse)
{
  guint i = 0;
  guint8 index = reverse ? 1 : 0;

  /* Many times this works as a shortcut, since mappings are almost 1-to-1 */
  if  (table[item][index] == item) {
    return table[item][index ^ 1];
  }

  /* Shortcut didn't work, go through the whole table to find the item */
  while (table[i][0] != -1) {
    if (table[i][index] == item) {
      return table[i][index ^ 1];
    }
    i++;
  }

  return table[0][index ^ 1];
}


/*
 * gst_subdevsrc_libomap3camd_fourcc_to_camformat:
 * @pixelformat: V4L2 color format in fourcc.
 * @pcode: pointer where to place corresponding subdev color code.
 *
 * Convert V4l2 pixelformat into libomap3camd pixel format.
 *
 * Returns: TRUE if pixel format was supported.
 */
static gboolean
gst_subdevsrc_libomap3camd_fourcc_to_camformat (guint32 fourcc,
    cam_pixel_format_t *pformat)
{
  gboolean ret = TRUE;

  switch (fourcc) {
    case GST_MAKE_FOURCC ('Y', 'U', 'Y', 'V'):
    case GST_MAKE_FOURCC ('Y', 'U', 'Y', '2'):
      *pformat = CAM_FRAME_FORMAT_YUYV;
      break;
    case GST_MAKE_FOURCC ('U', 'Y', 'V', 'Y'):
      *pformat = CAM_PIXEL_FORMAT_UYVY;
      break;
    case GST_MAKE_FOURCC ('B', 'A', '1', '0'):
      *pformat = CAM_PIXEL_FORMAT_BAYER10;
      break;
    default:
      GST_WARNING ("Unsupported fourcc requested");
      ret = FALSE;
  }

  return ret;
}


/*
 * Capture start callback
 */
static void
gst_subdevsrc_libomap3camd_capture_start_cb (void *user_data)
{
  GstSubdevSrc *subdevsrc;

  subdevsrc = GST_SUBDEVSRC_CAST (user_data);
  GST_DEBUG_OBJECT (subdevsrc, "Capture start callback");

  /* Notify application that capture phase starts now */
  gst_camerasrc_send_capture_start_message (GST_CAMERA_SRC (subdevsrc));
}


/*
 * Capture stop callback
 */
static void
gst_subdevsrc_libomap3camd_capture_stop_cb (void *user_data)
{
  GstSubdevSrc *subdevsrc;

  subdevsrc = GST_SUBDEVSRC_CAST (user_data);
  GST_DEBUG_OBJECT (subdevsrc, "Capture stop callback");

  /* Notify application that capture phase ends now */
  gst_camerasrc_send_capture_stop_message (GST_CAMERA_SRC (subdevsrc));
}


/*
 * Autofocus update callback
 */
static void
gst_subdevsrc_libomap3camd_af_update_cb (void *user_data)
{
  GstSubdevSrc *subdevsrc;

  subdevsrc = GST_SUBDEVSRC_CAST (user_data);
  g_atomic_int_set (&subdevsrc->af_updated, 1);
}


/*
 * Continuous autofocus update callback
 */
static void
gst_subdevsrc_libomap3camd_caf_update_cb (void *user_data)
{
  GstSubdevSrc *subdevsrc;

  subdevsrc = GST_SUBDEVSRC_CAST (user_data);
  g_atomic_int_set (&subdevsrc->caf_updated, 1);
}


/*
 * Shake risk update callback
 */
static void
gst_subdevsrc_libomap3camd_shake_update_cb (void *user_data)
{
  GstSubdevSrc *subdevsrc;

  subdevsrc = GST_SUBDEVSRC_CAST (user_data);
  g_atomic_int_set (&subdevsrc->shake_updated, 1);
}


/*
 * Flash update callback
 */
static void
gst_subdevsrc_libomap3camd_flash_update_cb (void *user_data)
{
  GstSubdevSrc *subdevsrc;

  subdevsrc = GST_SUBDEVSRC_CAST (user_data);
  g_atomic_int_set (&subdevsrc->flash_updated, 1);
}


/*
 * Flicker detected callback
 */
static void
gst_subdevsrc_libomap3camd_flicker_cb (void *user_data)
{
  GstSubdevSrc *subdevsrc;

  subdevsrc = GST_SUBDEVSRC_CAST (user_data);
  g_atomic_int_set (&subdevsrc->flicker_detected, 1);
}


/*
 * gst_subdevsrc_libomap3camd_register_algo_callback:
 * @subdevsrc: #GstSubdevSrc object.
 * @func: callback function to be registered.
 * @algo: algorithm ID to which the callback should be registered.
 *
 * Register a function to algorithm status change callback.
 *
 * Returns: TRUE on success.
 */
static gboolean
gst_subdevsrc_libomap3camd_register_algo_callback (GstSubdevSrc *subdevsrc,
    AlgoNotifyCallback func, cam_algo_id_t algo)
{
  cam_status_changed_t notify;
  gboolean ret = TRUE;
  cam_err_t err;

  notify.callback = func;
  notify.argument = subdevsrc;

  err = cam_feature_set (subdevsrc->libhnd, CAM_STATUS_CHANGED, algo, &notify);
  if (err != CAM_ERR_NONE) {
    GST_WARNING ("Error registering algo %d callback: err = %s", algo,
        cameralib_error_map[err]);
    ret = FALSE;
  }

  return ret;
}


/*
 * Select correct scene map according to currently used viewfinder mode
 */
static void
gst_subdevsrc_libomap3camd_select_scene_map (GstSubdevSrc *subdevsrc,
    GstCameraSrcViewfinderMode vf_mode)
{
  if (vf_mode == GST_CAMERA_SRC_VIEWFINDER_MODE_VIDEO) {
    subdevsrc->scene_map = gst_subdevsrc_scene_map_video;
  } else {
    subdevsrc->scene_map = gst_subdevsrc_scene_map_image;
  }
}


/*
 * Retrieve focus information data from libomap3camd
 */
static gboolean
gst_subdevsrc_libomap3camd_get_focus_info (GstSubdevSrc *subdevsrc,
    cam_focus_info_t *info)
{
  gboolean ret = TRUE;
  cam_err_t err;

  err = cam_feature_get (subdevsrc->libhnd, CAM_FOCUS_INFO, info);
  if (err != CAM_ERR_NONE) {
    GST_WARNING_OBJECT (subdevsrc, "FOCUS_INFO failed: %s",
        cameralib_error_map[err]);
    ret = FALSE;
  }
  else {
    GST_DEBUG_OBJECT (subdevsrc, "Coverage:      %d", info->coverage);
    GST_DEBUG_OBJECT (subdevsrc, "Zones (vert):  %d", info->zones_vertical);
    GST_DEBUG_OBJECT (subdevsrc, "Zones (horz):  %d", info->zones_horizontal);
    GST_DEBUG_OBJECT (subdevsrc, "Focus zone:    %d", info->zone);
    GST_DEBUG_OBJECT (subdevsrc, "Lens position: %d", info->lens_position);
  }

  return ret;
}

/******************************************************************************/
/*                                PUBLIC API                                  */
/******************************************************************************/


/**
 * gst_subdevsrc_libomap3camd_add_buffer:
 * @subdevsrc: #GstSubdevSrc object.
 * @buf: #GstRecycleBuffer object to be sent to libomap3camd.
 * @mode: #GstOperationMode which tells the purpose of the buffer.
 *
 * Queue a buffer to libomap3camd. This can be used for both image capture
 * and preview buffers.
 *
 * Returns: TRUE on success.
 */
gboolean
gst_subdevsrc_libomap3camd_add_buffer (GstSubdevSrc *subdevsrc,
    GstRecycleBuffer *buf, GstOperationMode mode)
{
  cam_frame_buffer_t buffer;
  cam_engine_t type;
  gboolean ret = TRUE;
  cam_err_t err;

  type = gst_op_mode_to_engine_type_map[mode];

  buffer.b_address = buf->data;
  buffer.b_size = buf->pool->buffer_size;
  buffer.index = buf->index;
  buffer.status = CAM_ERR_NONE;
  buffer.notify.callback = NULL;
  buffer.notify.argument = NULL;

  GST_DEBUG_OBJECT (subdevsrc, "Queueing buffer %p", buffer.b_address);
  GST_LOG_OBJECT (subdevsrc, "    Size: %d", buffer.b_size);

  err = cam_feature_set (subdevsrc->libhnd, CAM_BUFFER, type, &buffer);

  if (err != CAM_ERR_NONE) {
    GST_WARNING ("Queueing buffer %p FAILED: %s", buffer.b_address,
        cameralib_error_map[err]);
    ret = FALSE;
  }

  return ret;
}


/**
 * gst_subdevsrc_libomap3camd_configure:
 * @subdevsrc: #GstSubdevSrc object.
 * @width: width of the frame.
 * @height: height of the frame.
 * @fourcc: fourcc code of the format to be configured.
 * @size: pointer where the size of one buffer should be stored.
 * @num_buffers: pointer where the amount of buffers should be stored.
 * @mode: #GstOperationMode into which libomap3camd should be configured.
 *
 * Configure libomap3camd for certain resolution, format and operation mode.
 * When mode is PREVIEW or IMAGE_CAPTURE, then @size and @num_buffers parameters
 * will be used to store corresponding into those pointers.
 *
 * Returns: TRUE on success.
 */
gboolean
gst_subdevsrc_libomap3camd_configure (GstSubdevSrc *subdevsrc,
    guint *width, guint *height, guint32 fourcc, guint32 *size,
    guint *num_buffers, GstOperationMode mode)
{
  cam_pixel_format_t pformat;
  cam_window_rect_t rect;
  gboolean ret = FALSE;
  cam_engine_t type;
  cam_err_t err;
  guint nbuf;

  rect.top = 0;
  rect.left = 0;
  rect.width = *width;
  rect.height = *height;

  type = gst_op_mode_to_engine_type_map[mode];

  GST_DEBUG_OBJECT (subdevsrc, "Setting the resolution %dx%d to library",
      *width, *height);

  if (subdevsrc->initialized) {

    if (!gst_subdevsrc_libomap3camd_fourcc_to_camformat (fourcc, &pformat)) {
      GST_WARNING_OBJECT (subdevsrc, "Format not supported");
      goto done;
    }

    err = cam_feature_set (subdevsrc->libhnd, CAM_PIXEL_FORMAT, type, pformat);

    if (err != CAM_ERR_NONE) {
      GST_WARNING_OBJECT (subdevsrc, "Setting the color format failed: %s",
          cameralib_error_map[err]);
      goto done;
    }

    err = cam_feature_set (subdevsrc->libhnd, CAM_RESOLUTION, type, &rect,
        &nbuf);

    if (err == CAM_ERR_NONE) {
      /* This might not be needed */
      err = cam_feature_get (subdevsrc->libhnd, CAM_RESOLUTION, type, &rect,
          &nbuf);

      /* In still capture case query the max zoom factor */
      if (type == CAM_ENGINE_STILL_CAPTURE) {
        cam_fractional_t oldzoom, test;
        double tmpzoom;

        /* Read the current zoom factor */
        err = cam_feature_get (subdevsrc->libhnd, CAM_DIGITAL_ZOOM, &oldzoom);
        if (err != CAM_ERR_NONE) {
          GST_WARNING_OBJECT (subdevsrc, "Fetching zoom failed: %s",
              cameralib_error_map[err]);
          goto done;
        }

        /* Try to set a big zoom factor (2^10.0) */
        test.numenator = 10 * 10;
        test.denumenator = 1 * 10;
        cam_feature_set (subdevsrc->libhnd, CAM_DIGITAL_ZOOM, &test);

        GST_DEBUG_OBJECT (subdevsrc, "Max lib zoom: %u/%u", test.numenator,
            test.denumenator);

        /* Put back the old zoom value, if set */
        if (oldzoom.numenator && oldzoom.denumenator) {
          cam_feature_set (subdevsrc->libhnd, CAM_DIGITAL_ZOOM, &oldzoom);
        }

        gst_util_fraction_to_double (test.numenator,
            test.denumenator, &tmpzoom);

        subdevsrc->lib_max_zoom = pow (2.0, tmpzoom);
        GST_DEBUG_OBJECT (subdevsrc, "Library max zoom: %f",
            subdevsrc->lib_max_zoom);
      }
    }
    else {
      GST_WARNING_OBJECT (subdevsrc, "Setting the resolution failed: %s",
          cameralib_error_map[err]);
    }

    if (err == CAM_ERR_NONE) {
      *width = rect.width;
      *height = rect.height;
      if (size) {
        *size = rect.size;
      }
      if (num_buffers) {
        *num_buffers = nbuf;
      }
      GST_DEBUG_OBJECT (subdevsrc, "Resolution set: %dx%d", rect.width,
          rect.height);

      if (type != CAM_ENGINE_VIEWFINDER) {
        GST_DEBUG_OBJECT (subdevsrc, "Buffer size = %d, buffers = %d",
            rect.size, nbuf);
      }
      ret = TRUE;
    }
  }
  else {
    GST_DEBUG_OBJECT (subdevsrc, "Device not initialized");
  }

done:

  return ret;
}


/**
 * gst_subdevsrc_libomap3camd_set_vf_mode:
 * @subdevsrc: #GstSubdevSrc object.
 * @mode: #GstCameraSrcViewfinderMode to be used for viewfinder.
 *
 * Configure libomap3camd to use given viewfinder mode.
 *
 * Returns: TRUE on success.
 */
gboolean
gst_subdevsrc_libomap3camd_set_vf_mode (GstSubdevSrc *subdevsrc,
    GstCameraSrcViewfinderMode mode)
{
#ifndef GST_DISABLE_GST_DEBUG
  static const gchar *vfmodes[] = { "STILL", "VIDEO" };
#endif
  cam_err_t err = CAM_ERR_NONE;
  gboolean ret = FALSE;

  gst_subdevsrc_libomap3camd_select_scene_map (subdevsrc, mode);

  GST_DEBUG_OBJECT (subdevsrc, "Viewfinder mode: %s", vfmodes[mode]);

  if (subdevsrc->initialized) {
    err = cam_feature_set (subdevsrc->libhnd, CAM_VIEWFINDER_MODE,
        mode == GST_CAMERA_SRC_VIEWFINDER_MODE_VIDEO ?
            CAM_VIEWFINDER_MODE_VIDEO_RECORD : CAM_VIEWFINDER_MODE_STILL_CAPTURE);
    ret = (err == CAM_ERR_NONE);

    GST_DEBUG_OBJECT (subdevsrc, "Setting VF mode: %s", ret ? "OK" : "FAIL");
  }
  else {
    GST_WARNING_OBJECT (subdevsrc, "Libomap3camd not initialized");
  }

  return ret;
}


/**
 * gst_subdevsrc_libomap3camd_read_settings:
 * @subdevsrc: #GstSubdevSrc object.
 * @photoconf: pointer to GstPhotoSettings where to read the settings.
 *
 * Read libomap3camd settings into given configuration structure.
 *
 * Returns: TRUE on success.
 */
gboolean
gst_subdevsrc_libomap3camd_read_settings (GstSubdevSrc *subdevsrc,
    GstPhotoSettings *photoconf)
{
  cam_awb_mode_t wbmode;
  cam_general_effect_type_t effect;
  cam_general_scene_mode_t scene;
  cam_light_flash_mode_t flash;
  cam_focus_mode_t focus;
  gint32 ev, iso, aperture, exposure;
  cam_err_t err;

  if (!subdevsrc->initialized) {
    GST_WARNING_OBJECT (subdevsrc, "Device not initialized");
    return FALSE;
  }

  GST_DEBUG_OBJECT (subdevsrc, "Retrieving settings from daemon");

  err = cam_feature_get (subdevsrc->libhnd, CAM_AWB_MODE, &wbmode);
  if (err != CAM_ERR_NONE) {
    GST_WARNING_OBJECT (subdevsrc, "CAM_AWB_MODE query failed: %s",
        cameralib_error_map[err]);
    goto error;
  }

  err = cam_feature_get (subdevsrc->libhnd, CAM_GENERAL_EFFECT_TYPE, &effect);
  if (err != CAM_ERR_NONE) {
    GST_WARNING_OBJECT (subdevsrc, "CAM_GENERAL_EFFECT_TYPE query failed: %s",
        cameralib_error_map[err]);
    goto error;
  }

  err = cam_feature_get (subdevsrc->libhnd, CAM_GENERAL_SCENE_MODE, &scene);
  if (err != CAM_ERR_NONE) {
    GST_WARNING_OBJECT (subdevsrc, "CAM_GENERAL_SCENE_MODE query failed: %s",
        cameralib_error_map[err]);
    goto error;
  }

  err = cam_feature_get (subdevsrc->libhnd, CAM_LIGHT_FLASH_MODE, &flash);
  if (err != CAM_ERR_NONE) {
    GST_WARNING_OBJECT (subdevsrc, "CAM_LIGHT_FLASH_MODE query failed: %s",
        cameralib_error_map[err]);
    goto error;
  }

  err = cam_feature_get (subdevsrc->libhnd, CAM_EXPOSURE_COMPENSATION,  &ev);
  if (err != CAM_ERR_NONE) {
    GST_WARNING_OBJECT (subdevsrc, "CAM_EXPOSURE_COMPENSATION query failed: %s",
        cameralib_error_map[err]);
    goto error;
  }

  err = cam_feature_get (subdevsrc->libhnd, CAM_EXPOSURE_ISO_VALUE, &iso);
  if (err != CAM_ERR_NONE) {
    GST_WARNING_OBJECT (subdevsrc, "CAM_EXPOSURE_ISO_VALUE query failed: %s",
        cameralib_error_map[err]);
    goto error;
  }

  err = cam_feature_get (subdevsrc->libhnd, CAM_EXPOSURE_MANUAL_APERTURE,
      &aperture);
  if (err != CAM_ERR_NONE) {
    GST_WARNING_OBJECT (subdevsrc,
        "CAM_EXPOSURE_MANUAL_APERTURE query failed: %s",
        cameralib_error_map[err]);
    goto error;
  }

  err = cam_feature_get (subdevsrc->libhnd, CAM_EXPOSURE_MANUAL_TIME, &exposure);
  if (err != CAM_ERR_NONE) {
    GST_WARNING_OBJECT (subdevsrc, "CAM_EXPOSURE_MANUAL_TIME query failed: %s",
        cameralib_error_map[err]);
    goto error;
  }

  err = cam_feature_get (subdevsrc->libhnd, CAM_FOCUS_MODE, &focus);
  if (err != CAM_ERR_NONE) {
    GST_WARNING_OBJECT (subdevsrc, "CAM_FOCUS_MODE query failed: %s",
        cameralib_error_map[err]);
    goto error;
  }

  photoconf->wb_mode = find_item (gst_subdevsrc_wb_map, wbmode, TRUE);
  photoconf->tone_mode = find_item (gst_subdevsrc_effect_map, effect, TRUE);
  photoconf->scene_mode = find_item (subdevsrc->scene_map, scene, TRUE);
  photoconf->flash_mode = find_item (gst_subdevsrc_flash_map, flash, TRUE);
  photoconf->focus_mode = find_item (gst_subdevsrc_focus_map, focus, TRUE);
  photoconf->ev_compensation = (gfloat) ev / 10;
  photoconf->iso_speed = gst_subdevsrc_libomap3camd_reverse_map_iso (iso);
  photoconf->aperture = aperture;
  photoconf->exposure = exposure;

  GST_DEBUG_OBJECT (subdevsrc, "Scene mode ext: %d, lib: %d",
      photoconf->scene_mode, scene);
  GST_DEBUG_OBJECT (subdevsrc, "Focus mode ext: %d, lib: %d",
      photoconf->focus_mode, focus);
  GST_DEBUG_OBJECT (subdevsrc, "Flash mode ext: %d, lib: %d",
      photoconf->flash_mode, flash);
  GST_DEBUG_OBJECT (subdevsrc, "ISO: %d, EV_comp: %f (%d)", iso,
      photoconf->ev_compensation, ev);

  return TRUE;

/* ERRORS */
error:
  {
    return FALSE;
  }
}


/**
 * gst_subdevsrc_libomap3camd_write_settings:
 * @subdevsrc: #GstSubdevSrc object.
 * @photoconf: GstPhotoSettings to be used as new settings.
 * @scene_override: TRUE, if scene mode should override other settings.
 *
 * Configure libomap3camd to use given new settings.
 *
 * Returns: TRUE on success.
 */
gboolean
gst_subdevsrc_libomap3camd_write_settings (GstSubdevSrc *subdevsrc,
GstPhotoSettings *photoconf, gboolean scene_override)
{
  GST_DEBUG_OBJECT (subdevsrc, "Writing settings to libomap3camd");

  if (!subdevsrc->initialized) {
    GST_WARNING_OBJECT (subdevsrc, "Device not initialized");
    return FALSE;
  }

  cam_feature_set (subdevsrc->libhnd, CAM_GENERAL_SCENE_MODE,
      find_item (subdevsrc->scene_map, photoconf->scene_mode, FALSE));

  if (scene_override &&
      photoconf->scene_mode != GST_PHOTOGRAPHY_SCENE_MODE_MANUAL)
  {
    /* If scene override flag is set, we just leave the new settings in use */
    GST_DEBUG_OBJECT (subdevsrc, "SCENE OVERRIDE, ext=%d, lib=%d",
        photoconf->scene_mode, find_item (subdevsrc->scene_map,
            photoconf->scene_mode, FALSE));
  }
  else {
    GST_DEBUG_OBJECT (subdevsrc, "Normal settings");

    cam_feature_set (subdevsrc->libhnd, CAM_AWB_MODE,
        find_item (gst_subdevsrc_wb_map, photoconf->wb_mode, FALSE));

    cam_feature_set (subdevsrc->libhnd, CAM_GENERAL_EFFECT_TYPE,
        find_item (gst_subdevsrc_effect_map, photoconf->tone_mode, FALSE));

    cam_feature_set (subdevsrc->libhnd, CAM_LIGHT_FLASH_MODE,
        find_item (gst_subdevsrc_flash_map, photoconf->flash_mode, FALSE));

    /* These will set exposure mode to MANUAL, is value is != 0 */
    cam_feature_set (subdevsrc->libhnd, CAM_EXPOSURE_MANUAL_TIME,
        photoconf->exposure);
    cam_feature_set (subdevsrc->libhnd, CAM_EXPOSURE_MANUAL_APERTURE,
        photoconf->aperture);

    cam_feature_set (subdevsrc->libhnd, CAM_EXPOSURE_COMPENSATION,
        (gint) (photoconf->ev_compensation * 10));

    cam_feature_set (subdevsrc->libhnd, CAM_EXPOSURE_ISO_VALUE,
        gst_subdevsrc_libomap3camd_map_iso (photoconf->iso_speed));

    cam_feature_set (subdevsrc->libhnd, CAM_FOCUS_MODE,
        find_item (gst_subdevsrc_focus_map, photoconf->focus_mode, FALSE));
  }

  /* Flicker mode AUTO is automatically re-activated within certain invervals */
  /* in gst_subdevsrc_grab_frame() function. Therefore we don't handle */
  /* that mode in here */
  if (photoconf->flicker_mode != GST_PHOTOGRAPHY_FLICKER_REDUCTION_AUTO) {
    gst_subdevsrc_libomap3camd_set_flicker_mode (subdevsrc,
        photoconf->flicker_mode);
  }

  return TRUE;
}


/**
 * gst_subdevsrc_libomap3camd_set_autofocus:
 * @subdevsrc: #GstSubdevSrc object.
 * @on: TRUE when autofocus algorithm should be started.
 *
 * Start/stop autofocus algorithm.
 *
 * Returns: TRUE on success.
 */
gboolean
gst_subdevsrc_libomap3camd_set_autofocus (GstSubdevSrc *subdevsrc, gboolean on)
{
  gboolean ret = FALSE;

  if (subdevsrc->is_active) {
    cam_err_t err;

    err = cam_feature_set (subdevsrc->libhnd, CAM_ALGO_CONTROL,
        CAM_ALGO_FOCUS_SINGLE, on ? CAM_CONTROL_RUN : CAM_CONTROL_STOPED);

    ret = (err == CAM_ERR_NONE);
  }
  else {
    GST_WARNING_OBJECT (subdevsrc, "Streaming not active");
  }

  return ret;
}


/**
 * gst_subdevsrc_libomap3camd_set_autoexposure:
 * @subdevsrc: #GstSubdevSrc object.
 * @on: TRUE when autoexposure algorithm should be started.
 *
 * Start/stop autoexposure algorithm.
 *
 * Returns: TRUE on success.
 */
gboolean
gst_subdevsrc_libomap3camd_set_autoexposure (GstSubdevSrc *subdevsrc,
    gboolean on)
{
  gboolean ret = FALSE;

  if (subdevsrc->is_active) {
    cam_err_t err;

    err = cam_feature_set (subdevsrc->libhnd, CAM_ALGO_CONTROL,
        CAM_ALGO_EXPOSURE, on ? CAM_CONTROL_RUN : CAM_CONTROL_STOPED);

    ret = (err == CAM_ERR_NONE);
  }
  else {
    GST_WARNING_OBJECT (subdevsrc, "Streaming not active");
  }

  return ret;
}


/**
 * gst_subdevsrc_libomap3camd_set_flicker_mode:
 * @subdevsrc: #GstSubdevSrc object.
 * @mode: GstFlickerReductionMode to be set.
 *
 * Set flicker reduction mode to given setting.
 *
 * Returns: TRUE on success.
 */
gboolean
gst_subdevsrc_libomap3camd_set_flicker_mode (GstSubdevSrc *subdevsrc,
    GstFlickerReductionMode mode)
{
  gboolean ret = FALSE;

  if (subdevsrc->initialized) {
    cam_err_t err;

    err = cam_feature_set (subdevsrc->libhnd, CAM_GENERAL_FLICKER_REDUCTION_MODE,
        find_item (gst_subdevsrc_flicker_map, mode, FALSE));

    GST_DEBUG_OBJECT (subdevsrc, "Flicker mode ext: %d, lib: %d",
      mode, find_item (gst_subdevsrc_flicker_map, mode, FALSE));

    ret = (err == CAM_ERR_NONE);
  }
  else {
    GST_WARNING_OBJECT (subdevsrc, "Streaming not active");
  }

  return ret;
}


/**
 * gst_subdevsrc_libomap3camd_get_image_tags:
 * @subdevsrc: #GstSubdevSrc object.
 * @tlist: #GstTagList where the tags should be set.
 *
 * Add HQ image tags into taglist.
 *
 * Returns: TRUE on success.
 */
gboolean
gst_subdevsrc_libomap3camd_get_image_tags (GstSubdevSrc *subdevsrc,
    GstTagList *tlist)
{
  gboolean ret = FALSE, tret;
  frame_info_t iinfo;
  cam_err_t err;
  const gchar *tag_string;

  GST_DEBUG_OBJECT (subdevsrc, "Image parameters query");

  err = cam_feature_get (subdevsrc->libhnd, CAM_IMAGE_INFO, &iinfo);

  if (err == CAM_ERR_NONE) {

    guint aperture, shutter, exposure, fnumber, flen;
    gint iso;

    aperture = shutter = exposure = fnumber = flen = iso = 0;

    if (iinfo.aperture_value.denumenator) {
      aperture = 100  * (gfloat) iinfo.aperture_value.numenator /
          iinfo.aperture_value.denumenator;
    }
    if (iinfo.shutter_speed_value.denumenator) {
      shutter = 100 * iinfo.shutter_speed_value.numenator /
          iinfo.shutter_speed_value.denumenator;
    }
    if (iinfo.exposure_time.denumenator) {
      exposure = 1000 * (guint64) iinfo.exposure_time.numenator /
          iinfo.exposure_time.denumenator;
    }
    if (iinfo.fnumber.denumenator) {
      fnumber = 100 * (gfloat) iinfo.fnumber.numenator /
          iinfo.fnumber.denumenator;
    }
    if (iinfo.focal_length.denumenator) {
      flen = 100 * (gfloat) iinfo.focal_length.numenator /
          iinfo.focal_length.denumenator;
    }

    tret = gst_tag_list_get_int (tlist, GST_TAG_CAPTURING_ISO_SPEED, &iso);

    /* Check if parent class has already set the ISO value. ISO speed 0 means
     * "automatic" and in this case we take the value from  daemon's image
     * information structure */
    if (!tret || (tret && iso == 0)) {
      iso = iinfo.iso_speed_ratings;
    }

    /* FIXME: Remove that stupid fraction -> integer -> fraction thing */
    gst_tag_list_add (tlist, GST_TAG_MERGE_REPLACE,
        GST_TAG_CAPTURING_ISO_SPEED, iso,             /* G_TYPE_INT */
        GST_TAG_CAPTURING_SHUTTER_SPEED, exposure, 1000,
        GST_TAG_CAPTURING_FLASH_FIRED, (iinfo.flash & 0x1) != 0,
        GST_TAG_CAPTURING_FOCAL_RATIO, fnumber / 100.0,
        GST_TAG_CAPTURING_FOCAL_LENGTH, flen / 100.0, NULL);

    tag_string = gst_camerasrc_contrast_from_exif_value ((gint) iinfo.contrast);
    if (tag_string)
      gst_tag_list_add (tlist, GST_TAG_MERGE_REPLACE,
          GST_TAG_CAPTURING_CONTRAST, tag_string, NULL);

    tag_string = gst_camerasrc_saturation_from_exif_value (
        (gint) iinfo.saturation);
    if (tag_string)
      gst_tag_list_add (tlist, GST_TAG_MERGE_REPLACE,
          GST_TAG_CAPTURING_SATURATION, tag_string, NULL);

    tag_string = gst_camerasrc_exposure_mode_from_exif_value (
        (gint) iinfo.exposure_mode);
    if (tag_string)
      gst_tag_list_add (tlist, GST_TAG_MERGE_REPLACE,
          GST_TAG_CAPTURING_EXPOSURE_MODE, tag_string, NULL);

    tag_string = gst_camerasrc_flash_mode_from_exif_value ((gint) iinfo.flash);
    if (tag_string)
      gst_tag_list_add (tlist, GST_TAG_MERGE_REPLACE,
          GST_TAG_CAPTURING_FLASH_MODE, tag_string, NULL);

    tag_string = gst_camerasrc_gain_adjustment_from_exif_value (
        (gint) iinfo.gain);
    if (tag_string)
      gst_tag_list_add (tlist, GST_TAG_MERGE_REPLACE,
          GST_TAG_CAPTURING_GAIN_ADJUSTMENT, tag_string, NULL);

    tag_string = gst_camerasrc_scene_capture_type_from_exif_value (
        (gint) iinfo.scene_capture_type);
    if (tag_string)
      gst_tag_list_add (tlist, GST_TAG_MERGE_REPLACE,
          GST_TAG_CAPTURING_SCENE_CAPTURE_TYPE, tag_string, NULL);

    tag_string = gst_camerasrc_sharpness_from_exif_value (
        (gint) iinfo.sharpness);
    if (tag_string)
      gst_tag_list_add (tlist, GST_TAG_MERGE_REPLACE,
          GST_TAG_CAPTURING_SHARPNESS, tag_string, NULL);

    tag_string = gst_camerasrc_metering_mode_from_exif_value (
        (gint) iinfo.metering_mode);
    if (tag_string)
      gst_tag_list_add (tlist, GST_TAG_MERGE_REPLACE,
          GST_TAG_CAPTURING_METERING_MODE, tag_string, NULL);

    tag_string = gst_camerasrc_file_source_from_exif_value (3); /* 3 = DSC */
    if (tag_string)
      gst_tag_list_add (tlist, GST_TAG_MERGE_REPLACE,
          GST_TAG_CAPTURING_SOURCE, tag_string, NULL);

    /*
     * There's no mapping for: "capture-light-source", iinfo.light_source
     */

    ret = TRUE;
  }

  return ret;
}


/**
 * gst_subdevsrc_libomap3camd_init:
 * @subdevsrc: #GstSubdevSrc object.
 *
 * Initializes libomap3camd instance and configures it to default settings.
 *
 * Returns: TRUE on success
 */
gboolean
gst_subdevsrc_libomap3camd_init (GstSubdevSrc *subdevsrc)
{
  cam_still_capture_mode_t cmode;
  cam_err_t err = CAM_ERR_NONE;

  subdevsrc->scene_map = gst_subdevsrc_scene_map_image;

  subdevsrc->libhnd = cam_library_create ();

  GST_DEBUG_OBJECT (subdevsrc, "Library instance: %p", subdevsrc->libhnd);

  if (subdevsrc->libhnd) {
    /* Pass the device FD to lib */
    GST_DEBUG_OBJECT (subdevsrc, "SETTING DEVICE: %d", subdevsrc->sensor_fd);
    err = cam_feature_set (subdevsrc->libhnd, CAM_CAMERA_DEVICE,
                           subdevsrc->sensor_fd);
    GST_DEBUG_OBJECT (subdevsrc, "SETTING DEVICE %s",
                      err == CAM_ERR_NONE ? "OK" : "FAILED");
    if (err != CAM_ERR_NONE) {
      GST_WARNING ("Error while setting sensor device to library");
      goto error;
    }
  }
  else {
    GST_WARNING ("libomap3camd subsystem initialization failed");
    goto error;
  }

  if (subdevsrc->camera_device == GST_SUBDEVSRC_CAPTURE_DEVICE_PRIMARY) {
    cmode = CAM_STILL_CAPTURE_MODE_HIGH_QUALITY_NO_POSTPROCESS;
  }
  else {
    cmode = CAM_STILL_CAPTURE_MODE_PERFORMANCE;
  }

  /* Set AE mode: AUTO */
  err = cam_feature_set (subdevsrc->libhnd, CAM_EXPOSURE_MODE,
      CAM_EXPOSURE_MODE_AUTO);
  if (err != CAM_ERR_NONE) {
    GST_WARNING_OBJECT (subdevsrc, "CAM_EXPOSURE_MODE error: %s",
        cameralib_error_map[err]);
    goto error;
  }

  /* Set AWB mode: AUTO */
  err = cam_feature_set (subdevsrc->libhnd, CAM_AWB_MODE,
      CAM_AWB_MODE_AUTO);
  if (err != CAM_ERR_NONE) {
    GST_WARNING_OBJECT (subdevsrc, "CAM_AWB_MODE error: %s",
        cameralib_error_map[err]);
    goto error;
  }

  /* Set AF mode: AUTO */
  err = cam_feature_set (subdevsrc->libhnd, CAM_FOCUS_MODE, CAM_FOCUS_MODE_AUTO);
  if (err != CAM_ERR_NONE) {
    GST_WARNING_OBJECT (subdevsrc, "CAM_FOCUS_MODE error: %s",
        cameralib_error_map[err]);
    goto error;
  }

  /* Set privacy light mode: AUTO */
  err = cam_feature_set (subdevsrc->libhnd, CAM_LIGHT_PRIVACY_MODE,
      CAM_LIGHT_PRIVACY_MODE_AUTO);
  if (err != CAM_ERR_NONE) {
    GST_WARNING_OBJECT (subdevsrc, "CAM_LIGHT_PRIVACY_MODE error: %s",
        cameralib_error_map[err]);
    goto error;
  }

  /* Set focus assist light mode: AUTO */
  err = cam_feature_set (subdevsrc->libhnd, CAM_LIGHT_ASSIST_MODE,
      CAM_LIGHT_ASSIST_MODE_AUTO);
  if (err != CAM_ERR_NONE) {
    GST_WARNING_OBJECT (subdevsrc, "CAM_LIGHT_ASSIST_MODE error: %s",
        cameralib_error_map[err]);
    goto error;
  }

  /* Set flicker reduction mode: AUTO */
  err = cam_feature_set (subdevsrc->libhnd, CAM_GENERAL_FLICKER_REDUCTION_MODE,
      CAM_GENERAL_FLICKER_REDUCTION_MODE_AUTO);
  if (err != CAM_ERR_NONE) {
    GST_WARNING_OBJECT (subdevsrc, "CAM_GENERAL_FLICKER_REDUCTION_MODE error: %s",
        cameralib_error_map[err]);
    goto error;
  }

  /* Set scene mode */
  err = cam_feature_set (subdevsrc->libhnd, CAM_GENERAL_SCENE_MODE,
      CAM_GENERAL_SCENE_MODE_AUTO);
  if (err != CAM_ERR_NONE) {
    GST_WARNING_OBJECT ("CAM_GENERAL_SCENE_MODE error: %s",
        cameralib_error_map[err]);
    goto error;
  }

  /* Set capture quality */
  err = cam_feature_set (subdevsrc->libhnd, CAM_STILL_CAPTURE_MODE, cmode);
  if (err != CAM_ERR_NONE) {
    GST_WARNING_OBJECT (subdevsrc, "CAM_STILL_CAPTURE_MODE error: %s",
        cameralib_error_map[err]);
    goto error;
  }

  /* Set capture format */
  err = cam_feature_set (subdevsrc->libhnd, CAM_FRAME_FORMAT,
      CAM_ENGINE_STILL_CAPTURE, CAM_FRAME_FORMAT_UYVY);
  if (err != CAM_ERR_NONE) {
    GST_WARNING_OBJECT (subdevsrc, "CAM_FRAME_FORMAT error: %s",
        cameralib_error_map[err]);
    goto error;
  }

  /* Start AE algorithm */
  err = cam_feature_set (subdevsrc->libhnd, CAM_ALGO_CONTROL, CAM_ALGO_EXPOSURE,
      CAM_CONTROL_RUN);
  if (err != CAM_ERR_NONE) {
    GST_WARNING_OBJECT (subdevsrc, "CAM_ALGO_CONTROL:AE error: %s",
        cameralib_error_map[err]);
    goto error;
  }

  /* Start AWB algorithm */
  err = cam_feature_set (subdevsrc->libhnd, CAM_ALGO_CONTROL,
      CAM_ALGO_WHITE_BALANCE, CAM_CONTROL_RUN);
  if (err != CAM_ERR_NONE) {
    GST_WARNING_OBJECT (subdevsrc, "CAM_ALGO_CONTROL:AWB error: %s",
        cameralib_error_map[err]);
    goto error;
  }

  subdevsrc->initialized = TRUE;
  GST_DEBUG_OBJECT (subdevsrc, "Library initialized");

  return TRUE;

/* ERRORS */
error:
 {
   GST_WARNING_OBJECT (subdevsrc, "Library initialization failed");
   return FALSE;
 }
}


/**
 * gst_subdevsrc_libomap3camd_deinit:
 * @subdevsrc: #GstSubdevSrc object.
 *
 * Deinitializes libomap3camd instance.
 *
 * Returns: TRUE on success
 */
void
gst_subdevsrc_libomap3camd_deinit (GstSubdevSrc *subdevsrc)
{
  if (subdevsrc->libhnd) {
    /* Disable remaining camera daemon algorithms */
    cam_feature_set (subdevsrc->libhnd, CAM_ALGO_CONTROL, CAM_ALGO_EXPOSURE,
        CAM_CONTROL_STOPED);

    GST_DEBUG_OBJECT (subdevsrc, "AE stopped");

    cam_feature_set (subdevsrc->libhnd, CAM_ALGO_CONTROL,
        CAM_ALGO_WHITE_BALANCE, CAM_CONTROL_STOPED);

    GST_DEBUG_OBJECT (subdevsrc, "AWB stopped");

    cam_feature_set (subdevsrc->libhnd, CAM_ALGO_CONTROL, CAM_ALGO_FOCUS_SINGLE,
        CAM_CONTROL_STOPED);

    GST_DEBUG_OBJECT (subdevsrc, "AF stopped");

    cam_feature_set (subdevsrc->libhnd, CAM_LIGHT_PRIVACY_MODE,
        CAM_LIGHT_PRIVACY_MODE_OFF);

    GST_DEBUG_OBJECT (subdevsrc, "Privacy led off");

    cam_feature_set (subdevsrc->libhnd, CAM_LIGHT_ASSIST_MODE,
        CAM_LIGHT_ASSIST_MODE_OFF);

    GST_DEBUG_OBJECT (subdevsrc, "Assist light off");

    cam_library_delete (subdevsrc->libhnd);
    subdevsrc->libhnd = NULL;
    subdevsrc->initialized = FALSE;
    GST_DEBUG_OBJECT (subdevsrc, "Camera library deinitialized");
  }
}


/**
 * gst_subdevsrc_libomap3camd_set_privacy_light:
 * @subdevsrc: GstSubdevSrc object.
 * @on: TRUE if the privacy light should be lit.
 *
 * Turns privacy light on or off.
 *
 * Returns: TRUE on success.
 */
gboolean
gst_subdevsrc_libomap3camd_set_privacy_light (GstSubdevSrc *subdevsrc,
    gboolean on)
{
  cam_light_privacy_mode_t mode;
  gboolean ret = FALSE;

  if (subdevsrc->initialized) {
    cam_err_t err;

    GST_DEBUG_OBJECT (subdevsrc, "Setting privacy light");

    mode = on ? CAM_LIGHT_PRIVACY_MODE_ON : CAM_LIGHT_PRIVACY_MODE_AUTO;
    err = cam_feature_set (subdevsrc->libhnd, CAM_LIGHT_PRIVACY_MODE, mode);
    ret = (err == CAM_ERR_NONE);

    if (ret)
      GST_DEBUG_OBJECT (subdevsrc, "Privacy light set: %d (%s)",
          mode, on ? "ON" : "OFF");
    else
      GST_DEBUG_OBJECT (subdevsrc, "PRIVACY_MODE failed: %s",
          cameralib_error_map[err]);
  }
  else {
    GST_DEBUG_OBJECT (subdevsrc, "Streaming not active");
  }

  return ret;
}


/**
 * gst_subdevsrc_libomap3camd_set_assist_light:
 * @subdevsrc: GstSubdevSrc object.
 * @on: TRUE if the assist light should be lit.
 *
 * Turns assist light on or off.
 *
 * Returns: TRUE on success.
 */
gboolean
gst_subdevsrc_libomap3camd_set_assist_light (GstSubdevSrc *subdevsrc,
    gboolean on)
{
  cam_light_assist_mode_t mode;
  gboolean ret = FALSE;

  if (subdevsrc->initialized) {
    cam_err_t err;

    GST_DEBUG_OBJECT (subdevsrc, "Setting assist light");

    mode = on ? CAM_LIGHT_ASSIST_MODE_ON : CAM_LIGHT_ASSIST_MODE_AUTO;
    err = cam_feature_set (subdevsrc->libhnd, CAM_LIGHT_ASSIST_MODE, mode);
    ret = (err == CAM_ERR_NONE);

    if (ret)
      GST_DEBUG_OBJECT (subdevsrc, "Assist light set: %d (%s)",
          mode, on ? "ON" : "OFF");
    else
      GST_DEBUG_OBJECT (subdevsrc, "ASSIST_MODE failed: %s",
          cameralib_error_map[err]);
  }
  else {
    GST_DEBUG_OBJECT (subdevsrc, "Library not initialized");
  }

  return ret;
}


/**
 * gst_subdevsrc_libomap3camd_handle_roi_event:
 * @subdevsrc: #GstSubdevSrc object
 * @structure: Event structure to be handled
 *
 * Handles regions-of-interest event.
 *
 * Returns: TRUE if the event was handled successfully.
 */
gboolean
gst_subdevsrc_libomap3camd_handle_roi_event (GstSubdevSrc *subdevsrc,
    const GstStructure *structure)
{
  guint width, height, objcount, i, objnum;
  gboolean ret = TRUE;
  cam_err_t err;

  GST_DEBUG_OBJECT (subdevsrc, "OBJECT TRACKING EVENT");

  if (!subdevsrc->initialized) {
    GST_WARNING_OBJECT (subdevsrc, "Library not initialized");
    return FALSE;
  }

  if (gst_structure_get_uint (structure, "frame-width", &width) &&
      gst_structure_get_uint (structure, "frame-height", &height))
  {
    cam_object_tracking_t tracking;
    const GValue *regions, *region;
    const GstStructure *rstructure;

    memset (&tracking, 0, sizeof (cam_object_tracking_t));

    /* This is a GValue list of GstStructure objects */
    regions = gst_structure_get_value (structure, "regions");
    if (!regions) {
      GST_WARNING ("RoI event missing regions data");
      return TRUE;
    }

    objcount = gst_value_list_get_size (regions);

    /* FIXME: in this case should we select highest-priority objects? */
    if (objcount > OBJECT_TRACKING_MAX_COUNT) {
      objcount = OBJECT_TRACKING_MAX_COUNT;
      GST_DEBUG_OBJECT (subdevsrc, "Too many regions, limiting to %d", objcount);
    }

    tracking.frame_size.top = 0;
    tracking.frame_size.left = 0;
    tracking.frame_size.width = width;
    tracking.frame_size.height = height;
    tracking.rotation = CAM_OBJECT_ORIENTATION_UNKNOWN;
    objnum = 0;

    for (i = 0; i<objcount; i++) {
      guint left, top, w, h, prio;

      region = gst_value_list_get_value (regions, i);
      rstructure = gst_value_get_structure (region);

      if (gst_structure_get_uint (rstructure, "region-x", &left) &&
          gst_structure_get_uint (rstructure, "region-y", &top) &&
          gst_structure_get_uint (rstructure, "region-w", &w) &&
          gst_structure_get_uint (rstructure, "region-h", &h))
      {
        tracking.objects[objnum].left = left;
        tracking.objects[objnum].top = top;
        tracking.objects[objnum].width = w;
        tracking.objects[objnum].height = h;

        /* Priority field is optional; if not set, use priority 1 */
        if (!gst_structure_get_uint (rstructure, "region-priority", &prio))
          prio = 1;

        tracking.priority[objnum] = prio;

        GST_DEBUG_OBJECT (subdevsrc, "RoI %d : x=%d, y=%d, w=%d, h=%d, pri=%d",
            objnum, tracking.objects[objnum].left, tracking.objects[objnum].top,
            tracking.objects[objnum].width, tracking.objects[objnum].height,
            tracking.priority[objnum]);
        objnum++;
      }
      else {
        GST_WARNING_OBJECT (subdevsrc, "Missing information in RoI structure");
        ret = FALSE;
      }
    }

    if (objnum > 0) {
      err = cam_feature_set (subdevsrc->libhnd, CAM_OBJECT_TRACING, &tracking);
      GST_DEBUG_OBJECT (subdevsrc, "OBJECT_TRACING returns: %s",
          cameralib_error_map[err]);
      ret = (err == CAM_ERR_NONE);
    }
  }
  else {
    GST_WARNING_OBJECT (subdevsrc, "Can't read width / height from event");
    ret = FALSE;
  }

  return ret;
}


/**
 * gst_subdevsrc_libomap3camd_handle_motion:
 * @subdevsrc: #GstSubdevSrc object
 * @motionvector: Amount of the movement.
 * @timedelta: Amount of time between frames used for calculating this vector.
 *
 * Passes motion estimation vector to libomap3camd.
 *
 * Returns: TRUE if the data was handled successfully.
 */
gboolean
gst_subdevsrc_libomap3camd_handle_motion (GstSubdevSrc *subdevsrc,
    gfloat motionvector, gint64 timedelta)
{
  cam_motion_vector_t vector;
  cam_err_t err;

  /* It doesn't matter which vector to use */
  vector.x_vector = motionvector;
  vector.y_vector = 0;
  vector.d_time = timedelta;

  err = cam_feature_set (subdevsrc->libhnd, CAM_MOTION_DETECTION, &vector);
  GST_LOG_OBJECT (subdevsrc, "Library returns: %d (%s)", err,
      cameralib_error_map[err]);

  return (err == CAM_ERR_NONE);
}


/**
 * gst_subdevsrc_libomap3camd_daemon_start:
 * @subdevsrc: #GstSubdevSrc object.
 *
 * Starts running H3A algorithms in libomap3camd.
 *
 * Returns: TRUE on success.
 */
gboolean
gst_subdevsrc_libomap3camd_daemon_start (GstSubdevSrc *subdevsrc)
{
  cam_err_t err;

  GST_DEBUG_OBJECT (subdevsrc, "Starting up...");

  /* Start running viewfinder */
  err = cam_feature_set (subdevsrc->libhnd, CAM_ENGINE_CONTROL,
      CAM_ENGINE_VIEWFINDER, CAM_CONTROL_RUN);
  if (err != CAM_ERR_NONE) {
    GST_WARNING_OBJECT (subdevsrc, "CAM_ENGINE_CONTROL error: %s",
        cameralib_error_map[err]);
    goto set_config_err;
  }

  if (!gst_subdevsrc_libomap3camd_register_algo_callback (subdevsrc,
      gst_subdevsrc_libomap3camd_capture_start_cb, CAM_STILL_CAPTURE_START))
  {
    GST_WARNING_OBJECT (subdevsrc, "Error registering CAPTURE_START callback");
    goto set_config_err;
  }

  if (!gst_subdevsrc_libomap3camd_register_algo_callback (subdevsrc,
    gst_subdevsrc_libomap3camd_capture_stop_cb, CAM_STILL_CAPTURE_DONE))
  {
    GST_DEBUG_OBJECT (subdevsrc, "Error registering CAPTURE_DONE callback");
    goto set_config_err;
  }

  if (!gst_subdevsrc_libomap3camd_register_algo_callback (subdevsrc,
    gst_subdevsrc_libomap3camd_af_update_cb, CAM_ALGO_FOCUS_SINGLE))
  {
    GST_DEBUG_OBJECT (subdevsrc, "Error registering FOCUS_SINGLE callback");
    goto set_config_err;
  }

  if (!gst_subdevsrc_libomap3camd_register_algo_callback (subdevsrc,
    gst_subdevsrc_libomap3camd_caf_update_cb, CAM_ALGO_FOCUS_CONTINUOUS))
  {
    GST_DEBUG_OBJECT (subdevsrc, "Error registering FOCUS_CONTINUOUS callback");
    goto set_config_err;
  }

  if (!gst_subdevsrc_libomap3camd_register_algo_callback (subdevsrc,
    gst_subdevsrc_libomap3camd_shake_update_cb, CAM_ALGO_SHAKE_RISK))
  {
    GST_DEBUG_OBJECT (subdevsrc, "Error registering SHAKE_RISK callback");
    goto set_config_err;
  }

  if (!gst_subdevsrc_libomap3camd_register_algo_callback (subdevsrc,
    gst_subdevsrc_libomap3camd_flash_update_cb, CAM_ALGO_FLASHLIGHT_CONTROL))
  {
    GST_DEBUG_OBJECT (subdevsrc, "Error registering FLASHLIGHT_CONTROL callback");
    goto set_config_err;
  }

  if (!gst_subdevsrc_libomap3camd_register_algo_callback (subdevsrc,
    gst_subdevsrc_libomap3camd_flicker_cb, CAM_ALGO_LINE_FREQ_DETECTED))
  {
    GST_DEBUG_OBJECT (subdevsrc, "Error registering LINE_FREQ_DETECTED callback");
    goto set_config_err;
  }

  GST_DEBUG_OBJECT (subdevsrc, "Daemon started successfully");

  return TRUE;

set_config_err:
{
  GST_WARNING_OBJECT (subdevsrc, "Camera Daemon configuration failed");
  goto error;
}

error:
  return FALSE;
}


/**
 * gst_subdevsrc_libomap3camd_daemon_stop:
 * @subdevsrc: #GstSubdevSrc object.
 *
 * Stops running H3A algorithms in libomap3camd.
 *
 * Returns: TRUE on success.
 */
gboolean
gst_subdevsrc_libomap3camd_daemon_stop (GstSubdevSrc *subdevsrc)
{
  GST_DEBUG_OBJECT (subdevsrc, "Stopping...");

  /* Stop viewfinder */
  cam_feature_set (subdevsrc->libhnd, CAM_ENGINE_CONTROL,
      CAM_ENGINE_VIEWFINDER, CAM_CONTROL_STOPED);

  GST_DEBUG_OBJECT (subdevsrc, "Viewfinder stopped");

  return TRUE;
}


/**
 * gst_subdevsrc_libomap3camd_grab_hq_frame:
 * @subdevsrc: #GstSubdevSrc object
 * @buf: Pointer to GstBuffer where to place the image frame
 *
 * Grab a High-Quality image by using libomap3camd.
 *
 * Returns: TRUE if capture succeeds, FALSE otherwise
 */
gboolean
gst_subdevsrc_libomap3camd_grab_hq_frame (GstSubdevSrc *subdevsrc,
    GstBuffer **buf)
{
  GstRecycleBuffer *imgbuf;
  cam_frame_buffer_t buffer = { 0, };
  gboolean ret = FALSE;
  cam_err_t err;

  g_mutex_lock (subdevsrc->capture.buffer_lock);

  if (subdevsrc->capture.buffer == NULL) {
    GST_WARNING_OBJECT (subdevsrc, "Image buffer not allocated");
    goto grab_leave;
  }

  imgbuf = subdevsrc->capture.buffer;

  GST_DEBUG_OBJECT (subdevsrc, "Starting capture thread");

  /* Start capturing engine */
  err = cam_feature_set (subdevsrc->libhnd, CAM_ENGINE_CONTROL,
      CAM_ENGINE_STILL_CAPTURE, CAM_CONTROL_RUN);

  if (err != CAM_ERR_NONE) {
    GST_WARNING_OBJECT (subdevsrc, "Error starting capture thread: %s",
        cameralib_error_map[err]);
    goto grab_leave;
  }

  g_mutex_lock (subdevsrc->preview.buffer_lock);

  if (subdevsrc->preview.buffer) {
    GST_DEBUG_OBJECT (subdevsrc, "Active preview");
    cam_frame_buffer_t prv;

    /* Start preview engine */
    err = cam_feature_set (subdevsrc->libhnd, CAM_ENGINE_CONTROL,
        CAM_ENGINE_SNAP_SHOT, CAM_CONTROL_RUN);

    if (err != CAM_ERR_NONE) {
      GST_WARNING_OBJECT (subdevsrc, "Error starting preview engine: %s",
          cameralib_error_map[err]);
      goto skip_preview;
    }

    GST_LOG_OBJECT (subdevsrc, "Snapshot engine running");

    prv.b_address = subdevsrc->preview.buffer->data;
    prv.b_size = subdevsrc->preview.buffer->pool->buffer_size;

    err = cam_feature_get (subdevsrc->libhnd, CAM_BUFFER,
        CAM_ENGINE_SNAP_SHOT, &prv);

    if (err != CAM_ERR_NONE) {
      GST_WARNING_OBJECT (subdevsrc,
          "Error generating preview image: err = %s, prv.status = %s",
          cameralib_error_map[err], cameralib_error_map[prv.status]);
    }
    else {
      GstBuffer *buf;

      GST_LOG_OBJECT (subdevsrc, "Snapshot image retrieved");
      buf = gst_recycle_pool_convert_buffer (subdevsrc->preview.buffer);
      if (buf != NULL) {
        subdevsrc->preview.buffer = NULL;
        gst_camerasrc_send_preview (GST_CAMERA_SRC (subdevsrc), buf);
        gst_buffer_unref (buf);
      }
      else {
        GST_WARNING_OBJECT (subdevsrc, "Can't allocate buffer for preview image");
      }
    }

    /* Stop preview engine */
    cam_feature_set (subdevsrc->libhnd, CAM_ENGINE_CONTROL,
        CAM_ENGINE_SNAP_SHOT, CAM_CONTROL_STOPED);
  }

skip_preview:

  g_mutex_unlock (subdevsrc->preview.buffer_lock);

  buffer.b_address = imgbuf->data;
  buffer.b_size = imgbuf->pool->buffer_size;

  GST_DEBUG_OBJECT (subdevsrc, "Starting capture");

  err = cam_feature_get (subdevsrc->libhnd, CAM_BUFFER,
      CAM_ENGINE_STILL_CAPTURE, &buffer);

  GST_DEBUG_OBJECT (subdevsrc, "Capture done");

  /* Stop capture engine */
  cam_feature_set (subdevsrc->libhnd, CAM_ENGINE_CONTROL,
      CAM_ENGINE_STILL_CAPTURE, CAM_CONTROL_STOPED);

  if (err != CAM_ERR_NONE || buffer.status != CAM_ERR_NONE) {
    GST_WARNING_OBJECT (subdevsrc,
        "Error capturing image, err = %s, buffer.status = %s",
        cameralib_error_map[err], cameralib_error_map[buffer.status]);
    goto grab_leave;
  }

  *buf = gst_recycle_pool_convert_buffer (imgbuf);

  /* Always UYVY, so we can calculate the image size */
  /* This is needed because icapture gives us a buffer that is 8kb-aligned */
  /* which is bigger than actual frame size */
  GST_BUFFER_SIZE (*buf) =
      subdevsrc->capture.width * subdevsrc->capture.height * 2;
  ret = TRUE;

grab_leave:

  if (ret) {
    /* We let the buffer go, it will be handled now by GstRecyclePool */
    subdevsrc->capture.buffer = NULL;
  }
  g_mutex_unlock (subdevsrc->capture.buffer_lock);

  return ret;
}


/**
 * gst_subdevsrc_libomap3camd_get_makernote:
 * @subdevsrc: #GstSubdevSrc object.
 * @maker_note: Pointer where the makernote data should be placed.
 * @maker_note_size: Pointer where to place makernote size.
 *
 * Create a makernote data by asking it directly from libomap3camd.
 *
 * Returns: TRUE on success.
 */
gboolean
gst_subdevsrc_libomap3camd_get_makernote (GstSubdevSrc *subdevsrc,
    guint8 **maker_note, guint *maker_note_size)
{
  gboolean ret = FALSE;
  cam_err_t err;

  if (subdevsrc->initialized) {
    *maker_note_size = sizeof (maker_note_t);
    *maker_note = (guint8 *) g_malloc (*maker_note_size);

    err = cam_feature_get (subdevsrc->libhnd, CAM_MAKER_NOTE, *maker_note);

    if (err != CAM_ERR_NONE) {
      GST_WARNING_OBJECT (subdevsrc, "Failed getting makernote data: %s",
          cameralib_error_map[err]);
      g_free (*maker_note);
      *maker_note_size = 0;
    }
    else {
      ret = TRUE;
    }
  }
  else {
    GST_WARNING_OBJECT (subdevsrc, "Device not initialized");
  }

  return ret;
}


/**
 * gst_subdevsrc_libomap3camd_get_pp_params:
 * @subdevsrc: #GstSubdevSrc object.
 * @pp_struct: pointer where to store GstStructure containing PP parameters.
 *
 * Retrieve image post-processing parameters.
 *
 * Returns: TRUE on success.
 */
gboolean
gst_subdevsrc_libomap3camd_get_pp_params (GstSubdevSrc *subdevsrc,
    GstStructure **pp_struct)
{
  cam_ipp_parameters_t params = { 0, };
  cam_err_t err = CAM_ERR_NONE;
  gboolean ret = FALSE;

  GST_DEBUG_OBJECT (subdevsrc, "Post-processing parameters query");

  params.version = CAM_IPP_VERSION_1_9;

  if (subdevsrc->initialized) {
    err = cam_feature_get (subdevsrc->libhnd, CAM_IPP_PARAMETERS, &params);

    if (err == CAM_ERR_NONE) {

      GST_DEBUG ("EE strength:              %u", params.info.ipp_1_9.edge_enhancement_strength);
      GST_DEBUG ("EE weak edge threshold:   %u", params.info.ipp_1_9.weak_edge_threshold);
      GST_DEBUG ("EE strong edge threshold: %u", params.info.ipp_1_9.strong_edge_threshold);
      GST_DEBUG ("LUMA noise low:           %u", params.info.ipp_1_9.freq_luma_noise.low);
      GST_DEBUG ("LUMA noise mid:           %u", params.info.ipp_1_9.freq_luma_noise.mid);
      GST_DEBUG ("LUMA noise hi:            %u", params.info.ipp_1_9.freq_luma_noise.high);
      GST_DEBUG ("CB noise low:             %u", params.info.ipp_1_9.freq_cb_noise.low);
      GST_DEBUG ("CB noise mid:             %u", params.info.ipp_1_9.freq_cb_noise.mid);
      GST_DEBUG ("CB noise hi:              %u", params.info.ipp_1_9.freq_cb_noise.high);
      GST_DEBUG ("CR noise low:             %u", params.info.ipp_1_9.freq_cr_noise.low);
      GST_DEBUG ("CR noise mid:             %u", params.info.ipp_1_9.freq_cr_noise.mid);
      GST_DEBUG ("CR noise hi:              %u", params.info.ipp_1_9.freq_cr_noise.high);
      GST_DEBUG ("Shading vert param1:      %u", params.info.ipp_1_9.shading_vert_param1);
      GST_DEBUG ("Shading vert param2:      %u", params.info.ipp_1_9.shading_vert_param2);
      GST_DEBUG ("Shading horz param1:      %u", params.info.ipp_1_9.shading_horz_param1);
      GST_DEBUG ("Shading horz param2:      %u", params.info.ipp_1_9.shading_horz_param2);
      GST_DEBUG ("Shading gain scale:       %u", params.info.ipp_1_9.shading_gain_scale);
      GST_DEBUG ("Shading gain offset:      %u", params.info.ipp_1_9.shading_gain_offset);
      GST_DEBUG ("Shading gain maxvalue:    %u", params.info.ipp_1_9.shading_gain_maxvalue);
      GST_DEBUG ("Ratio downsample Cb Cr:   %u", params.info.ipp_1_9.ratio_downsample_cb_cr);

      *pp_struct = gst_structure_new ("application/x-gst-ipp",
          EENF_EE_STRENGTH, G_TYPE_UINT, params.info.ipp_1_9.edge_enhancement_strength,
          EENF_EE_WE_THRESHOLD, G_TYPE_UINT, params.info.ipp_1_9.weak_edge_threshold,
          EENF_EE_SE_THRESHOLD, G_TYPE_UINT, params.info.ipp_1_9.strong_edge_threshold,

          EENF_LNF_STRENGTH_LOW, G_TYPE_UINT, params.info.ipp_1_9.freq_luma_noise.low,
          EENF_LNF_STRENGTH_MID, G_TYPE_UINT, params.info.ipp_1_9.freq_luma_noise.mid,
          EENF_LNF_STRENGTH_HI,  G_TYPE_UINT, params.info.ipp_1_9.freq_luma_noise.high,

          EENF_CBF_STRENGTH_LOW, G_TYPE_UINT, params.info.ipp_1_9.freq_cb_noise.low,
          EENF_CBF_STRENGTH_MID, G_TYPE_UINT, params.info.ipp_1_9.freq_cb_noise.mid,
          EENF_CBF_STRENGTH_HI,  G_TYPE_UINT, params.info.ipp_1_9.freq_cb_noise.high,

          EENF_CRF_STRENGTH_LOW, G_TYPE_UINT, params.info.ipp_1_9.freq_cr_noise.low,
          EENF_CRF_STRENGTH_MID, G_TYPE_UINT, params.info.ipp_1_9.freq_cr_noise.mid,
          EENF_CRF_STRENGTH_HI,  G_TYPE_UINT, params.info.ipp_1_9.freq_cr_noise.high,

          EENF_SHADING_VERT_PARAM1, G_TYPE_UINT, params.info.ipp_1_9.shading_vert_param1,
          EENF_SHADING_VERT_PARAM2, G_TYPE_UINT, params.info.ipp_1_9.shading_vert_param2,
          EENF_SHADING_HORZ_PARAM1, G_TYPE_UINT, params.info.ipp_1_9.shading_horz_param1,
          EENF_SHADING_HORZ_PARAM2, G_TYPE_UINT, params.info.ipp_1_9.shading_horz_param2,

          EENF_SHADING_GAIN_SCALE, G_TYPE_UINT, params.info.ipp_1_9.shading_gain_scale,
          EENF_SHADING_GAIN_OFFSET, G_TYPE_UINT, params.info.ipp_1_9.shading_gain_offset,
          EENF_SHADING_GAIN_MAXVALUE, G_TYPE_UINT, params.info.ipp_1_9.shading_gain_maxvalue,

          EENF_DOWNSAMPLE_CB_CR, G_TYPE_UINT, params.info.ipp_1_9.ratio_downsample_cb_cr,
          NULL);
      ret = TRUE;
    }
    else {
      GST_WARNING_OBJECT (subdevsrc,
          "Getting post-processing parameters failed: %s",
          cameralib_error_map[err]);
    }
  }
  else {
    GST_WARNING_OBJECT (subdevsrc, "Streaming not active");
  }

  return ret;
}


/**
 * gst_subdevsrc_libomap3camd_check_shaking:
 * @subdevsrc: #GstSubdevSrc object
 * @risk: pointer where to put the risk value
 *
 * Retrieve image shake risk level from library.
 *
 * Returns: TRUE on success.
 */
gboolean
gst_subdevsrc_libomap3camd_check_shaking (GstSubdevSrc *subdevsrc,
    GstPhotoShakeRisk *risk)
{
  cam_exposure_shake_risk_t riskparm;
  gboolean ret = FALSE;
  cam_err_t err;

  err = cam_feature_get (subdevsrc->libhnd, CAM_EXPOSURE_SHAKE_RISK, &riskparm);

  if (err == CAM_ERR_NONE) {
    *risk = find_item (gst_subdevsrc_shakerisk_map, riskparm, TRUE);
    ret = TRUE;
  }
  else {
    GST_WARNING_OBJECT (subdevsrc, "SHAKE_RISK failed: %s",
        cameralib_error_map[err]);
  }

  return ret;
}


/**
 * gst_subdevsrc_libomap3camd_flash_status:
 * @subdevsrc: #GstSubdevSrc object
 * @status: pointer where to put the flash status value
 *
 * Retrieve flash status value from library.
 *
 * Returns: TRUE on success.
 */
gboolean
gst_subdevsrc_libomap3camd_flash_status (GstSubdevSrc *subdevsrc,
    GstCameraFlashStatus *status)
{
  cam_light_flash_status_t fstatus;
  gboolean ret = FALSE;
  cam_err_t err;

  err = cam_feature_get (subdevsrc->libhnd, CAM_LIGHT_FLASH_STATUS, &fstatus);

  if (err == CAM_ERR_NONE) {
    /* the mapping is 1:1 */
    *status = fstatus;
    ret = TRUE;
  }
  else {
    GST_WARNING_OBJECT (subdevsrc, "FLASH_STATUS failed: %s",
        cameralib_error_map[err]);
  }

  return ret;
}


/**
 * gst_subdevsrc_libomap3camd_flicker_status:
 * @subdevsrc: #GstSubdevSrc object
 * @status: pointer where to put the flicker reduction mode value
 *
 * Retrieve flicker reduction mode from library.
 *
 * Returns: TRUE on success.
 */
gboolean
gst_subdevsrc_libomap3camd_flicker_status (GstSubdevSrc *subdevsrc,
    GstFlickerReductionMode *mode)
{
  cam_general_flicker_reduction_mode_t fmode;
  gboolean ret = FALSE;
  cam_err_t err;

  err = cam_feature_get (subdevsrc->libhnd, CAM_GENERAL_FLICKER_REDUCTION_MODE,
      &fmode);

  if (err == CAM_ERR_NONE) {
    *mode = find_item (gst_subdevsrc_flicker_map, fmode, TRUE);
    ret = TRUE;
  }
  else {
    GST_WARNING_OBJECT (subdevsrc, "FLICKER_REDUCTION_MODE failed: %s",
        cameralib_error_map[err]);
  }

  return ret;
}


/**
 * gst_subdevsrc_libomap3camd_set_zoom:
 * @subdevsrc: #GstSubdevSrc object.
 * @zoom: Desired zoom factor.
 *
 * Set the zoom factor for captured image.
 *
 * Returns: TRUE on success.
 */
gboolean
gst_subdevsrc_libomap3camd_set_zoom (GstSubdevSrc *subdevsrc, gfloat zoom)
{
  cam_err_t err = CAM_ERR_NONE;
  cam_fractional_t zoomval;
  gboolean ret;
  gfloat tmp;

  GST_DEBUG_OBJECT (subdevsrc, "ZOOM: %f", zoom);

  if (zoom > subdevsrc->lib_max_zoom) {
    zoom = subdevsrc->lib_max_zoom;
  }

  tmp = logf (zoom) / logf (2);
  zoomval.numenator = tmp * 100;
  zoomval.denumenator = 100;

  GST_DEBUG_OBJECT (subdevsrc, "Setting zoom fraction: %d/%d (%f)",
      zoomval.numenator, zoomval.denumenator,
      (float) zoomval.numenator / zoomval.denumenator);

  err = cam_feature_set (subdevsrc->libhnd, CAM_DIGITAL_ZOOM, &zoomval);
  ret = err == CAM_ERR_NONE;

  GST_DEBUG_OBJECT (subdevsrc, "Setting zoom: %s", ret ? "SUCCESS" : "FAIL");

  return ret;
}


/**
 * gst_subdevsrc_libomap3camd_get_focus_status:
 * @subdevsrc: #GstSubdevSrc object.
 * @fs: pointer to GstCameraFocusStatus where to store status information.
 * @detailed: set to TRUE if successful focus status should return extended
 *    focusing information (fill the whole status structure).
 *
 * Retrieve focus status. If status is not GST_PHOTOGRAPHY_FOCUS_STATUS_SUCCESS,
 * only the "status" field of the given @fs structure will be updated.
 * Otherwise the @detailed parameter defines whether the all fields in @status
 * structure should be filled. If set to TRUE, they will be set and also
 * memory for "windows" structure will be allocated. This memory must be freed
 * by the caller.
 *
 * Returns: TRUE on success.
 */
gboolean
gst_subdevsrc_libomap3camd_get_focus_status (GstSubdevSrc *subdevsrc,
    GstCameraFocusStatus *fs, gboolean detailed)
{
  cam_focus_status_t stat;
  gboolean ret = FALSE;
  cam_err_t err;

  GST_DEBUG_OBJECT (subdevsrc, "Retrieving focus status");

  err = cam_feature_get (subdevsrc->libhnd, CAM_FOCUS_STATUS, &stat);

  if (err == CAM_ERR_NONE) {
    fs->status = find_item (gst_subdevsrc_focus_status_map, stat, TRUE);
    ret = TRUE;

    GST_DEBUG_OBJECT (subdevsrc, "Focus status: %d", fs->status);

    if (detailed && stat == CAM_FOCUS_STATUS_SUCCESS) {
      cam_focus_info_t info;

      GST_DEBUG_OBJECT (subdevsrc, "Retrieve focus window information");

      ret = gst_subdevsrc_libomap3camd_get_focus_info (subdevsrc, &info);

      if (ret) {
        fs->focus_rows = info.zones_vertical;
        fs->focus_columns = info.zones_horizontal;
        fs->windows = g_malloc (sizeof (guint));
        fs->windows[0] = info.zone;
        fs->num_windows = 1;
        fs->coverage = info.coverage;
      }
    }
  }
  else {
    GST_WARNING_OBJECT (subdevsrc, "FOCUS_STATUS failed: %s",
        cameralib_error_map[err]);
  }

  return ret;
}
