/* GStreamer MediaController camera source
 * Copyright (C) 2010 Nokia Corporation <multimedia@maemo.org>
 *
 * Contact: Maemo Multimedia <multimedia@maemo.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <glib/gprintf.h>
#include <gst/gst.h>
#include <sys/ioctl.h>

#include "mediacontroller.h"

#define MEDIA_DEVICE "/dev/media0"

GST_DEBUG_CATEGORY_EXTERN (gst_subdevsrc_debug);
#define GST_CAT_DEFAULT gst_subdevsrc_debug

#define MAKE_PIXFMT(a, b) { a, b, #a }

#define MEDIA_ENTITY_TYPE(x) ((x) & MEDIA_ENTITY_TYPE_MASK)

struct pixfmt_data_table {
  enum v4l2_mbus_pixelcode pcode;
  guint32 v4l2_pixfmt;
  gchar *str;
};

/* Pixfmt string table */

const struct pixfmt_data_table pixfmt_data[] =
{
  MAKE_PIXFMT (V4L2_MBUS_FMT_FIXED, 0),
  MAKE_PIXFMT (V4L2_MBUS_FMT_RGB444_2X8_PADHI_BE, V4L2_PIX_FMT_RGB444),
  MAKE_PIXFMT (V4L2_MBUS_FMT_RGB444_2X8_PADHI_LE, V4L2_PIX_FMT_RGB444),
  MAKE_PIXFMT (V4L2_MBUS_FMT_RGB555_2X8_PADHI_BE, V4L2_PIX_FMT_RGB555X),
  MAKE_PIXFMT (V4L2_MBUS_FMT_RGB555_2X8_PADHI_LE, V4L2_PIX_FMT_RGB555),
  MAKE_PIXFMT (V4L2_MBUS_FMT_BGR565_2X8_BE, 0),
  MAKE_PIXFMT (V4L2_MBUS_FMT_BGR565_2X8_LE, 0),
  MAKE_PIXFMT (V4L2_MBUS_FMT_RGB565_2X8_BE, V4L2_PIX_FMT_RGB565X),
  MAKE_PIXFMT (V4L2_MBUS_FMT_RGB565_2X8_LE, V4L2_PIX_FMT_RGB565),
  MAKE_PIXFMT (V4L2_MBUS_FMT_Y8_1X8, V4L2_PIX_FMT_GREY),
  MAKE_PIXFMT (V4L2_MBUS_FMT_UYVY8_1_5X8, 0),
  MAKE_PIXFMT (V4L2_MBUS_FMT_VYUY8_1_5X8, 0),
  MAKE_PIXFMT (V4L2_MBUS_FMT_YUYV8_1_5X8, 0),
  MAKE_PIXFMT (V4L2_MBUS_FMT_YVYU8_1_5X8, 0),
  MAKE_PIXFMT (V4L2_MBUS_FMT_UYVY8_2X8, 0),
  MAKE_PIXFMT (V4L2_MBUS_FMT_VYUY8_2X8, 0),
  MAKE_PIXFMT (V4L2_MBUS_FMT_YUYV8_2X8, 0),
  MAKE_PIXFMT (V4L2_MBUS_FMT_YVYU8_2X8, 0),
  MAKE_PIXFMT (V4L2_MBUS_FMT_Y10_1X10, 0),
  MAKE_PIXFMT (V4L2_MBUS_FMT_YUYV10_2X10, 0),
  MAKE_PIXFMT (V4L2_MBUS_FMT_YVYU10_2X10, 0),
  MAKE_PIXFMT (V4L2_MBUS_FMT_UYVY8_1X16, V4L2_PIX_FMT_UYVY),
  MAKE_PIXFMT (V4L2_MBUS_FMT_VYUY8_1X16, V4L2_PIX_FMT_VYUY),
  MAKE_PIXFMT (V4L2_MBUS_FMT_YUYV8_1X16, V4L2_PIX_FMT_YUYV),
  MAKE_PIXFMT (V4L2_MBUS_FMT_YVYU8_1X16, V4L2_PIX_FMT_YVYU),
  MAKE_PIXFMT (V4L2_MBUS_FMT_YUYV10_1X20, 0),
  MAKE_PIXFMT (V4L2_MBUS_FMT_YVYU10_1X20, 0),
  MAKE_PIXFMT (V4L2_MBUS_FMT_SBGGR8_1X8, V4L2_PIX_FMT_SBGGR8),
  MAKE_PIXFMT (V4L2_MBUS_FMT_SGRBG8_1X8, V4L2_PIX_FMT_SGRBG8),
  MAKE_PIXFMT (V4L2_MBUS_FMT_SBGGR10_DPCM8_1X8, 0),
  MAKE_PIXFMT (V4L2_MBUS_FMT_SGBRG10_DPCM8_1X8, 0),
  MAKE_PIXFMT (V4L2_MBUS_FMT_SGRBG10_DPCM8_1X8, V4L2_PIX_FMT_SGRBG10DPCM8),
  MAKE_PIXFMT (V4L2_MBUS_FMT_SRGGB10_DPCM8_1X8, 0),
  MAKE_PIXFMT (V4L2_MBUS_FMT_SBGGR10_2X8_PADHI_BE, V4L2_PIX_FMT_SBGGR16),
  MAKE_PIXFMT (V4L2_MBUS_FMT_SBGGR10_2X8_PADHI_LE, V4L2_PIX_FMT_SBGGR16),
  MAKE_PIXFMT (V4L2_MBUS_FMT_SBGGR10_2X8_PADLO_BE, V4L2_PIX_FMT_SBGGR16),
  MAKE_PIXFMT (V4L2_MBUS_FMT_SBGGR10_2X8_PADLO_LE, V4L2_PIX_FMT_SBGGR16),
  MAKE_PIXFMT (V4L2_MBUS_FMT_SBGGR10_1X10, 0),
  MAKE_PIXFMT (V4L2_MBUS_FMT_SGBRG10_1X10, 0),
  MAKE_PIXFMT (V4L2_MBUS_FMT_SGRBG10_1X10, V4L2_PIX_FMT_SGRBG10),
  MAKE_PIXFMT (V4L2_MBUS_FMT_SRGGB10_1X10, 0),
  MAKE_PIXFMT (V4L2_MBUS_FMT_SBGGR12_1X12, 0)
};


/*
 * gst_subdevsrc_mc_query_entities:
 * @mc: GstMediaController object
 *
 * Returns: TRUE on success
 */
static gboolean
gst_subdevsrc_mc_query_entities (GstMediaController *mc)
{
  GstMediaEntity *e;
  struct stat devstat;
  gchar devname[32];
  gchar sysname[32];
  gchar target[256];
  gchar *nodename;
  int id, ret, endpos;

  GST_DEBUG ("Querying media entities");

  e = g_slice_new (GstMediaEntity);

  for (id = 1; ; id++) {

    /* Make valgrind happy about not using unintialized memory in ioctl call */
    memset (&e->info, 0, sizeof (struct media_entity_desc));

    e->fd = -1;
    e->info.id = id;

    if (ioctl (mc->fd, MEDIA_IOC_ENUM_ENTITIES, &e->info) < 0)
      break;

    if (e->info.type != MEDIA_ENTITY_TYPE_NODE_V4L &&
        MEDIA_ENTITY_TYPE (e->info.type) != MEDIA_ENTITY_TYPE_SUBDEV)
    {
      continue;
    }

    /* Find the corresponding device name based on major/minor numbers */
    g_sprintf (sysname, "/sys/dev/char/%u:%u", e->info.v4l.major,
        e->info.v4l.minor);

    ret = readlink (sysname, target, sizeof(target));
    if (ret < 0) {
      continue;
    }

    /* make sure we don't address outside the buffer */
    endpos = ret < sizeof(target) ? ret : sizeof(target)-1;

    target[endpos] = '\0';
    nodename = strrchr (target, '/');
    if (nodename == NULL) {
      continue;
    }

    g_sprintf (devname, "/dev/%s", nodename + 1);
    ret = stat (devname, &devstat);
    if (ret < 0)
      continue;

    e->devname = g_strdup (devname);
    e->pads = g_new0 (struct media_pad_desc, e->info.pads);
    e->links = g_new0 (struct media_link_desc, e->info.links);

    mc->entities = g_list_append (mc->entities, e);
    mc->entities_count++;

    GST_DEBUG ("Found entity '%s' (id=%d, device=%s) which is %s has %d pads and %d links",
        e->info.name, e->info.id, e->devname,
        MEDIA_ENTITY_TYPE (e->info.type) == MEDIA_ENTITY_TYPE_NODE ? "node" : "subdev",
        e->info.pads, e->info.links);

    /* New entity for next round */
    e = g_slice_new (GstMediaEntity);
  }

  /* The last one which wasn't used */
  g_slice_free (GstMediaEntity, e);

  GST_DEBUG ("Entities count = %d", mc->entities_count);

  return TRUE;
}


/*
 * gst_subdevsrc_mc_query_links:
 * @mc: GstMediaController object
 *
 * Returns: TRUE on success
 */
static gboolean
gst_subdevsrc_mc_query_links (GstMediaController *mc)
{
  struct media_links_enum links;
  GstMediaEntity *e;
  GList *iter;
  guint i;

  GST_DEBUG ("Querying links");

  for (iter = mc->entities; iter != NULL; iter = iter->next) {
    e = (GstMediaEntity *) iter->data;

    links.entity = e->info.id;
    links.pads = e->pads;
    links.links = e->links;

    if (ioctl (mc->fd, MEDIA_IOC_ENUM_LINKS, &links) < 0)
      break;

    if (e->info.links == 0)
      continue;

    GST_DEBUG ("Entity '%s':", e->info.name);

    for (i=0; i < e->info.links; i++) {

      if (e->links[i].sink.entity-1 >= mc->entities_count) {
        GST_WARNING ("    Incorrect data in entity link table");
        return FALSE;
      }

      GST_DEBUG ("    pad %d -> %s:%d %s %s",
                 e->links[i].source.index,
                 ((GstMediaEntity *)g_list_nth_data (mc->entities, e->links[i].sink.entity-1))->info.name,
                 e->links[i].sink.index,
                 e->links[i].flags & MEDIA_LINK_FLAG_ACTIVE ? "ACTIVE" : "",
                 e->links[i].flags & MEDIA_LINK_FLAG_IMMUTABLE ? "IMMUTABLE" : "");
    }
  }

  return TRUE;
}


/*
 * gst_subdevsrc_mc_setup_link:
 * @mc: GstMediaController object
 * @src: Source GstMediaEntity object
 * @sink: Sink GstMediaEntity object
 * @format: Media format to be configured between entities
 * @active: TRUE if the link should be activated and FALSE if deactivated
 *
 * Returns: TRUE on success
 */
static gboolean
gst_subdevsrc_mc_setup_link (GstMediaController *mc, const GstMediaEntity *src,
    const GstMediaEntity *sink, struct v4l2_mbus_framefmt *format,
    gboolean active)
{
  struct media_link_desc link;
  struct v4l2_subdev_format pformat;
  int i, ret, srcpad = -1, sinkpad = -1;
  guint flags = 0;

  /* Find correct src pad */
  for (i=0; i < src->info.links; i++) {
    if (sink->info.id == src->links[i].sink.entity) {
      srcpad = src->links[i].source.index;
      sinkpad = src->links[i].sink.index;
      flags = src->links[i].flags;
    }
  }

  if (srcpad == -1 || sinkpad == -1) {
    GST_DEBUG ("Invalid link");
    return FALSE;
  }

  if (format) {
    pformat.which = V4L2_SUBDEV_FORMAT_ACTIVE;
    pformat.pad = srcpad;
    pformat.format = *format;

    GST_LOG ("S_FMT (%s:%d) : res = %dx%d, code = %d",
        src->info.name, srcpad, pformat.format.width, pformat.format.height,
        pformat.format.code);

    /* Set the format on src pad first */
    if (ioctl (src->fd, VIDIOC_SUBDEV_S_FMT, &pformat) < 0) {
      /* FIXME: This must be an error */
      GST_DEBUG ("Setting src pad format for entity '%s' failed", src->info.name);
      GST_DEBUG ("ERRNO: %d (%s)", errno, g_strerror (errno));
      return FALSE;
    }
    GST_DEBUG ("Setting format on entity [%s:%d] SUCCESS",
               src->info.name, srcpad);

    GST_LOG ("G_FMT (%s:%d) : res = %dx%d, code = %d",
      src->info.name, srcpad, pformat.format.width, pformat.format.height,
      pformat.format.code);
  }

  /* source pad */
  link.source.entity = src->info.id;
  link.source.index = srcpad;
  link.source.flags = MEDIA_PAD_FLAG_OUTPUT;

  /* sink pad */
  link.sink.entity = sink->info.id;
  link.sink.index = sinkpad;
  link.sink.flags = MEDIA_PAD_FLAG_INPUT;

  /* Set / reset the ACTIVE flag */
  if (active) {
    link.flags = flags | MEDIA_LINK_FLAG_ACTIVE;
  } else {
    link.flags = flags & ~MEDIA_LINK_FLAG_ACTIVE;
  }

  GST_DEBUG ("Setting link [%s:%d] -> [%s:%d] %s (flags: %d)",
      src->info.name, srcpad, sink->info.name, sinkpad,
      active ? "ACTIVE" : "INACTIVE", link.flags);

  ret = ioctl (mc->fd, MEDIA_IOC_SETUP_LINK, &link);
  if (ret < 0) {
    GST_DEBUG ("Setting up a link failed");
    return FALSE;
  }

  if (format) {
    /* Set format to sink pad, if the entity isn't an output node */
    if (MEDIA_ENTITY_TYPE (sink->info.type) != MEDIA_ENTITY_TYPE_NODE) {

      pformat.pad = sinkpad;

      GST_LOG ("S_FMT (%s:%d) : res = %dx%d, code = %d",
          sink->info.name, sinkpad, pformat.format.width, pformat.format.height,
          pformat.format.code);

      /* Now set the format on sink entity's pad */
      if (ioctl (sink->fd, VIDIOC_SUBDEV_S_FMT, &pformat) < 0) {
        /* FIXME: This must be an error */
        GST_DEBUG ("Setting sink format for entity '%s' failed", sink->info.name);
        return FALSE;
      }

      GST_DEBUG ("Setting format on entity [%s:%d] SUCCESS",
                 sink->info.name, sinkpad);

      GST_LOG ("G_FMT (%s:%d) : res = %dx%d, code = %d",
          sink->info.name, sinkpad, pformat.format.width, pformat.format.height,
          pformat.format.code);

      /* Copy the configured format into original format structure */
      *format = pformat.format;
    }

    /* It was output node. Configure it separately */
    else {
      struct v4l2_format format;

      GST_DEBUG ("Setting format for output node '%s'", sink->info.name);

      memset (&format, 0x00, sizeof (struct v4l2_format));
      format.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
      if (ioctl (sink->fd, VIDIOC_G_FMT, &format) < 0) {
        /* FIXME: This must be an error */
        GST_DEBUG ("G_FMT for output node '%s' failed", sink->info.name);
        return FALSE;
      }

      format.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
      format.fmt.pix.width = pformat.format.width;
      format.fmt.pix.height = pformat.format.height;
      format.fmt.pix.pixelformat =
          gst_subdevsrc_mc_pcode_to_v4l2_format (pformat.format.code);
      format.fmt.pix.field = V4L2_FIELD_NONE;

      if (ioctl (sink->fd, VIDIOC_S_FMT, &format) < 0) {
        /* FIXME: This must be an error */
        GST_DEBUG ("S_FMT for output node '%s' failed", sink->info.name);
        return FALSE;
      }

      mc->output_size = format.fmt.pix.sizeimage;
    }
  }

  return TRUE;
}


/*
 * gst_subdevsrc_mc_get_entity:
 * @mc: GstMediaController object
 * @ename: Name of the entity.
 *
 * Searches for the given name from entity list. This performs a partial
 * match, so even the beginning of the name will be enough.
 *
 * Returns: Found GstMediaEntity, or NULL
 */
static GstMediaEntity *
gst_subdevsrc_mc_get_entity (GstMediaController *mc, const gchar *ename)
{
  guint ename_len;
  GList *iter;

  if (!ename) {
    return NULL;
  }

  GST_DEBUG ("Finding entity '%s'", ename);

  ename_len = strlen (ename);

  for (iter = mc->entities; iter != NULL; iter = iter->next) {
    GstMediaEntity *entity = (GstMediaEntity *) iter->data;
    if (!strncmp (ename, entity->info.name, ename_len)) {
      GST_LOG ("Matching entity '%s' found", entity->info.name);
      if (entity->fd == -1) {
        entity->fd = open (entity->devname, O_RDWR);
        GST_LOG ("Opening device '%s' %s", entity->devname,
                 entity->fd != -1 ? "SUCCESS" : "FAILED");
      }
      return entity;
    }
  }

  return NULL;
}



/* ======== PUBLIC API ======== */



/**
 * gst_subdevsrc_mc_init:
 *
 * Creates a new GstMediaController object.
 *
 * Returns: New GstMediaController object
 */
GstMediaController *
gst_subdevsrc_mc_init (void)
{
  GstMediaController *mc;

  mc = g_slice_new0 (GstMediaController);

  GST_DEBUG ("Opening media controller device");
  mc->fd = open (MEDIA_DEVICE, O_RDWR);
  if (mc->fd < 0) {
    GST_DEBUG ("Opening media controller device failed");
    g_slice_free (GstMediaController, mc);
    return NULL;
  }
  GST_DEBUG ("Media controller device opened");

  if (!gst_subdevsrc_mc_query_entities (mc)) {
    GST_WARNING ("Querying entities failed");
    return NULL;
  }

  if (!gst_subdevsrc_mc_query_links (mc)) {
    GST_WARNING ("Querying links failed");
    return NULL;
  }

  gst_subdevsrc_mc_reset_links (mc);

  return mc;
}

/**
 * gst_subdevsrc_mc_deinit:
 * @mc: GstMediaController object
 *
 * Destroys previously allocated GstMediaController instance.
 */
void
gst_subdevsrc_mc_deinit (GstMediaController *mc)
{
  GList *iter;

  if (mc == NULL) {
    return;
  }

  GST_DEBUG ("Closing media controller device");

  for (iter = mc->entities; iter != NULL; iter = iter->next) {
    GstMediaEntity *entity = (GstMediaEntity *) iter->data;

    if (entity->fd != -1) {
      GST_DEBUG ("  Closing device '%s'", entity->devname);
      close (entity->fd);
    }

    g_free (entity->pads);
    g_free (entity->links);
    g_free (entity->devname);
    g_slice_free (GstMediaEntity, entity);
  }

  g_list_free (mc->entities);

  close (mc->fd);
  g_slice_free (GstMediaController, mc);
}


/**
 * gst_subdevsrc_mc_get_device_fd:
 * @mc: #GstMediaController object.
 * @ename: device entity name to be looked for.
 *
 * Retrieve subdevice file descriptor. The descriptor is still owned by
 * GstMediaController object and must NOT be closed.
 *
 * Returns: opened device file descriptor, or -1 if not found.
 */
int
gst_subdevsrc_mc_get_device_fd (GstMediaController *mc, const gchar *ename)
{
  GstMediaEntity *entity;

  entity = gst_subdevsrc_mc_get_entity (mc, ename);
  if (entity) {
    GST_DEBUG ("Device mapped: %s -> %s", ename, entity->devname);
    return entity->fd;
  }

  GST_WARNING ("Couldn't find mapping for entity '%s'", ename);
  return -1;
}


/**
 * gst_subdevsrc_mc_reset_links:
 * @mc: GstMediaController object.
 *
 * Reset (disconnect) all links between entities.
 *
 * Returns: TRUE on success
 */
gboolean
gst_subdevsrc_mc_reset_links (GstMediaController *mc)
{
  GstMediaEntity *src;
  GstMediaEntity *sink;
  GList *iter;
  int i;

  GST_DEBUG ("Resetting all links between entities");

  for (iter = mc->entities; iter != NULL; iter = iter->next) {
    src = (GstMediaEntity *) iter->data;

    for (i=0; i < src->info.links; i++) {
      if (src->links[i].flags & MEDIA_LINK_FLAG_IMMUTABLE)
        continue;

      sink = (GstMediaEntity *) g_list_nth_data (mc->entities, src->links[i].sink.entity-1);
      gst_subdevsrc_mc_setup_link (mc, src, sink, NULL, FALSE);
    }
  }

  return TRUE;
}


/**
 * gst_subdevsrc_mc_setup_pipeline:
 * @mc: GstMediaController object
 * @pipeline: string representation of the pipeline to be built
 * @format: v4l2_mbus_framefmt structure defining the pipeline frame format
 *
 * Activate the given subdev path in ISP.
 *
 * Returns: TRUE on success
 */
gboolean
gst_subdevsrc_mc_setup_pipeline (GstMediaController *mc, const gchar *pipeline,
    struct v4l2_mbus_framefmt *format)
{
  gchar **tokens;
  gchar *sinkname;
  const GstMediaEntity *src, *sink;
  gint i;

  tokens = g_strsplit (pipeline, "!", 0);
  src = gst_subdevsrc_mc_get_entity (mc, g_strstrip (tokens[0]));

  if (!src) {
    GST_DEBUG ("Entity '%s' not found", tokens[0]);
    return FALSE;
  }

  for (i=1; (sinkname = tokens[i]) != NULL; i++) {
    g_strstrip (sinkname);
    sink = gst_subdevsrc_mc_get_entity (mc, sinkname);
    if (!sink) {
      GST_DEBUG ("Entity '%s' not found", sinkname);
      goto error;
    }
    GST_DEBUG ("Activating link '%s' -> '%s'", src->info.name, sink->info.name);
    if (!gst_subdevsrc_mc_setup_link (mc, src, sink, format, TRUE)) {
      GST_DEBUG ("Activating link failed");
      goto error;
    }
    src = sink;
  }

  g_strfreev (tokens);

  return TRUE;

error:
  gst_subdevsrc_mc_reset_links (mc);
  g_strfreev (tokens);

  return FALSE;
}


/**
 * gst_subdevsrc_mc_link_entities:
 * @mc: GstMediaController object
 * @srcname: source entity name
 * @sinkname: sink entity name
 * @format: v4l2_mbus_framefmt structure defining the pipeline frame format
 *
 * Activate the given subdev path in ISP.
 *
 * Returns: TRUE on success
 */
gboolean
gst_subdevsrc_mc_link_entities (GstMediaController *mc, const gchar *srcname,
    const gchar *sinkname, struct v4l2_mbus_framefmt *format)
{
  const GstMediaEntity *src, *sink;

  src = gst_subdevsrc_mc_get_entity (mc, srcname);

  if (!src) {
    GST_DEBUG ("Entity '%s' not found", srcname);
    return FALSE;
  }

  sink = gst_subdevsrc_mc_get_entity (mc, sinkname);

  if (!sink) {
    GST_DEBUG ("Entity '%s' not found", sinkname);
    return FALSE;
  }

  GST_DEBUG ("Activating link '%s' -> '%s'", src->info.name, sink->info.name);
  if (!gst_subdevsrc_mc_setup_link (mc, src, sink, format, TRUE)) {
    GST_DEBUG ("Activating link failed");
    goto error;
  }

  return TRUE;

error:

  return FALSE;
}


/**
 * gst_subdevsrc_mc_get_device_identifier:
 * @mc: GstMediaController object
 * @fd: File descriptor of previously opened device
 *
 * Return media device identifier string of the given file descriptor.
 *
 * Returns: Device identifier string, or NULL.
 */
const gchar *
gst_subdevsrc_mc_get_device_identifier (GstMediaController *mc, int fd)
{
  const gchar *ret = NULL;
  GstMediaEntity *entity;
  GList *iter;

  g_return_val_if_fail (mc != NULL, NULL);

  for (iter = mc->entities; iter != NULL; iter = iter->next) {
    entity = (GstMediaEntity *) iter->data;
    if (entity->fd == fd) {
      ret = entity->info.name;
      break;
    }
  }

  GST_DEBUG ("Returning device identifier: %s", ret ? ret : "NULL");

  return ret;
}


/**
 * gst_subdevsrc_mc_map_pixfmt:
 * @pixfmt: Pixel format string
 *
 * Convert given string to mediabus pixel code.
 *
 * Returns: corresponding pixel code, or 0 if not found.
 */
guint32
gst_subdevsrc_mc_map_pixfmt (const gchar *pixfmt)
{
  guint32 pcode = 0;
  guint i;

  for (i = 0; i < G_N_ELEMENTS (pixfmt_data); i++) {
    if (!strcmp (pixfmt_data[i].str, pixfmt)) {
      pcode = pixfmt_data[i].pcode;
      GST_DEBUG ("Pixfmt found: %s = %d", pixfmt, pcode);
      break;
    }
  }

  return pcode;
}


/**
 * gst_subdevsrc_mc_pcode_to_v4l2_format:
 * @pcode: Mediabus pixel code
 *
 * Map the given mediabus pixelcode to V4L2 pixel format.
 *
 * Returns: corresponding pixel format, or 0 if not found.
 */
guint32
gst_subdevsrc_mc_pcode_to_v4l2_format (enum v4l2_mbus_pixelcode code)
{
  guint32 pcode = 0;
  guint i;

  for (i = 0; i < G_N_ELEMENTS (pixfmt_data); i++) {
    if (pixfmt_data[i].pcode == code) {
      pcode = pixfmt_data[i].v4l2_pixfmt;
      GST_DEBUG ("V4L2 pixelformat found: %s = 0x%04x (%" GST_FOURCC_FORMAT ")",
          pixfmt_data[i].str, pcode, GST_FOURCC_ARGS (pcode));
      break;
    }
  }

  return pcode;
}
