/* GStreamer
 *
 * Copyright (C) 2010 Nokia Corporation <multimedia@maemo.org>
 *
 * recyclepool.c Pool which recycles memory pointers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:recyclepool
 * @short_description: pool for recycling buffers over and over again
 *
 * Recyclepool is used to keep track of certain amount of buffers and reusing
 * (recycling) them back to use when they are released.
 *
 * A new recycle pool is created with gst_recycle_pool_new(). It can be
 * destroyed with gst_recycle_pool_destroy().
 *
 * To allow application to have control over the buffer
 * allocation/recycling/freeing operations, recycle pool provides appropriate
 * gst_recycle_pool_set_alloc_func(), gst_recycle_pool_set_recycle_func() and
 * gst_recycle_pool_set_free_func() functions for registering custom handlers
 * for those operations. None of them are obligatory, though.
 *
 * Theory of operation: when application wants to get a new unused buffer from
 * pool, it calls gst_recycle_pool_get_buffer(). The buffer type is
 * #GstRecycleBuffer, which can be converted to GstBuffer with
 * gst_recycle_pool_convert_buffer() call. After this, given GstBuffer can
 * be used just normally. However, when the buffer is being unreffed by some
 * element, it will not be freed. Instead, recycle pool's callback is invoked
 * and the buffer is put back into the pool. If application has registered
 * a recycle_func, it will be used to ask whether the application would like
 * to take it into use immediately, or should it be left in the buffer pool.
 *
 * Recycle pool can be "cleaned up" with gst_recycle_pool_cleanup() call.
 * It will go through the pool and free all buffers, which are not being used
 * at that moment.
 *
 * If recycle pool is full, a.k.a. maximum amount of buffers are in use,
 * gst_recycle_pool_get_buffer() call will block until one of the used buffers
 * gets released.
 */

#include <unistd.h>

#include "gstsubdevcamsrc.h"
#include "recyclepool.h"


GST_DEBUG_CATEGORY_STATIC (gst_recycle_pool_debug);
#define GST_CAT_DEFAULT gst_recycle_pool_debug


/*****************************************************************************/
/******************************** PRIVATE API ********************************/
/*****************************************************************************/


#define GST_TYPE_RECYCLE_POOL (gst_recycle_pool_get_type())
#define GST_IS_RECYCLE_POOL(obj) (G_TYPE_CHECK_INSTANCE_TYPE ((obj), GST_TYPE_RECYCLE_POOL))
#define GST_RECYCLE_POOL(obj) (G_TYPE_CHECK_INSTANCE_CAST ((obj), GST_TYPE_RECYCLE_POOL, GstRecyclePool))

#if 0
#define DONT_STORE_FREED_BUFFERS
#endif

static GstMiniObjectClass *recycle_pool_parent_class = NULL;


/*
 *
 */
static void
_gst_recycle_pool_assign_index_unlocked (GstRecyclePool *pool,
    GstRecycleBuffer *buf)
{
  GstRecycleBuffer *tmp;
  gboolean index_used;
  gboolean found = FALSE;
  GSList *iter;
  guint i;

  GST_DEBUG ("Finding an empty index");

  for (i = 0; i < pool->optimal_size; i++) {
    index_used = FALSE;

    for (iter = pool->bufs; iter != NULL; iter = iter->next) {
      tmp = (GstRecycleBuffer *) iter->data;

      if (tmp->index == i) {
        index_used = TRUE;
        break;
      }
    }

    if (!index_used) {
      GST_DEBUG ("Unused index %d found", i);
      buf->index = i;
      found = TRUE;
      break;
    }
  }

  if (!found) {
    /* Pool has reached its optimal buffer count. Replace some index
     * and reassign it to this new buffer. This should always succeed, because
     * pool maximum size is restricted to (optimal size * 2) buffers */
    for (iter = pool->bufs; iter != NULL; iter = iter->next) {
      tmp = (GstRecycleBuffer *) iter->data;

      if (!tmp->replaced) {
        GST_DEBUG ("Replacing index %d (buffer %p)", tmp->index, tmp->data);
        tmp->replaced = TRUE;
        buf->index = tmp->index;
        found = TRUE;
        break;
      }
    }

    if (!found) {
      /* This should never happen */
      GST_WARNING ("Failed to properly replace a buffer. Using index 0");
      buf->index = 0;
      buf->replaced = TRUE; /* This causes the buffer to get freed after use */
    }
  }
}


/*
 *
 */
static GstRecycleBuffer *
_gst_recycle_pool_alloc_buffer_unlocked (GstRecyclePool *pool, guint32 size)
{
  GstRecycleBuffer *buf;

  GST_DEBUG ("Pool allocating buffer");

  buf = g_new0 (GstRecycleBuffer, 1);

  if (pool->alloc_func) {
    buf->data = pool->alloc_func (size, pool->alloc_user_data);
  }
  else {
    buf->data = g_malloc (size);
  }

  if (!buf->data) {
    GST_WARNING ("Buffer not allocated");
    g_free (buf);
    buf = NULL;
    goto out;
  }
  else {
    gst_mini_object_ref (GST_MINI_OBJECT (pool));
    buf->pool = pool;
  }

  _gst_recycle_pool_assign_index_unlocked (pool, buf);
  pool->bufs = g_slist_append (pool->bufs, buf);

  GST_DEBUG ("New buffer %p (index %d) allocated, count = %d", buf->data,
      buf->index, g_slist_length (pool->bufs));

out:

  return buf;
}


/*
 *
 */
static void
_gst_recycle_pool_free_buffer (gpointer data, gpointer user_data)
{
  GstRecycleBuffer *buf;

  buf = (GstRecycleBuffer *) data;

  GST_DEBUG ("Pool freeing buffer: %p (index %d)", buf->data, buf->index);

  if (buf->pool->free_func) {
    buf->pool->free_func (buf->data, buf->pool->buffer_size,
        buf->pool->free_user_data);
  }
  else {
    g_free (buf->data);
  }
  gst_mini_object_unref (GST_MINI_OBJECT (buf->pool));

  g_free (buf);
}


/*
 *
 */
static void
_gst_recycle_pool_release_buffer_unlocked (GstRecyclePool *pool,
    GstRecycleBuffer *buffer)
{
  GSList *item;

  item = g_slist_find (pool->bufs, buffer);
  if (item) {
    /* Always free a replaced buffer, don't recycle it */
    if (buffer->replaced) {
      _gst_recycle_pool_free_buffer (buffer, NULL);
      pool->bufs = g_slist_remove (pool->bufs, buffer);
    }
    else {
#ifdef DONT_STORE_FREED_BUFFERS
      _gst_recycle_pool_free_buffer (buffer, NULL);
      pool->bufs = g_slist_remove (pool->bufs, buffer);
#else
      GST_DEBUG ("Pool storing buffer %p (index %d) to pool", buffer->data,
          buffer->index);
      buffer->in_use = FALSE;
#endif
    }
  }
}


/*
 *
 */
static void
_gst_recycle_pool_cleanup_unlocked (GstRecyclePool *pool)
{
  GstRecycleBuffer *buf;
  GSList *removed = NULL;
  GSList *iter;

  GST_DEBUG ("Cleaning up pool");

  /* First collect all buffers which are not in use and free them */
  for (iter = pool->bufs; iter != NULL; iter = iter->next) {
    buf = (GstRecycleBuffer *) iter->data;

    if (!buf->in_use) {
      GST_DEBUG ("  Freeing buffer %p", buf->data);
      removed = g_slist_prepend (removed, buf);
      _gst_recycle_pool_free_buffer (buf, NULL);
    }
  }

  /* Remove the deleted buffers from pool's buffer list */
  for (iter = removed; iter != NULL; iter = iter->next) {
    pool->bufs = g_slist_remove (pool->bufs, iter->data);
  }
  g_slist_free (removed);
}


/*
 *
 */
static void
_gst_recycle_pool_buffer_recycle_func (gpointer mem)
{
  GstRecycleBuffer *buf;
  GstRecyclePool *pool;
  gboolean recycled = FALSE;

  buf = (GstRecycleBuffer *) mem;
  pool = buf->pool;

  /* If subdevsrc has already unreffed the pool and this is the last buffer
   * holding a reference to it, we must temporarily ref the pool to prevent it
   * from being finalized during _gst_recycle_pool_cleanup_unlocked() call */
  gst_mini_object_ref (GST_MINI_OBJECT (pool));

  g_mutex_lock (pool->lock);

  if (buf->replaced) {
    /* If the buffer index is replaced, don't recycle it */
    GST_DEBUG ("Release replaced buffer %p, index = %d", buf->data, buf->index);
    _gst_recycle_pool_release_buffer_unlocked (pool, buf);
    goto done;
  }

  /* Recycle the buffer immediately, but only when nobody is waiting for it */
  if (pool->running && !pool->waiting && pool->recycle_func) {
    recycled = pool->recycle_func (buf, pool->recycle_user_data);
  }

  if (recycled) {
    GST_DEBUG ("Pool recycling buffer %p (index %d)", buf->data, buf->index);
    buf->in_use = TRUE;
  }
  else {
    buf->in_use = FALSE;
#ifndef DONT_STORE_FREED_BUFFERS
    if (!pool->running)
#endif
    {
      GST_DEBUG ("Pool freeing buffer %p (index %d)", buf->data, buf->index);
      _gst_recycle_pool_release_buffer_unlocked (pool, buf);
    }
    else {
      GST_DEBUG ("Pool storing buffer %p (index %d)", buf->data, buf->index);
    }

    g_cond_signal (pool->cond);
  }

done:

  g_mutex_unlock (pool->lock);

  gst_mini_object_unref (GST_MINI_OBJECT (pool));
}


/*
 *
 */
static void
gst_recycle_pool_finalize (GstRecyclePool * pool)
{
  GST_DEBUG ("Pool finalize");

  g_mutex_free (pool->lock);
  pool->lock = NULL;

  g_cond_free (pool->cond);
  pool->cond = NULL;

  if (pool->bufs) {
    /* buffer list should be empty at this point */
    g_slist_foreach (pool->bufs, _gst_recycle_pool_free_buffer, NULL);
    g_slist_free (pool->bufs);
    pool->bufs = NULL;
  }

  GST_MINI_OBJECT_CLASS (recycle_pool_parent_class)->finalize (GST_MINI_OBJECT
      (pool));
}


/*
 *
 */
static void
gst_recycle_pool_init (GstRecyclePool * pool, GstRecyclePoolClass * g_class)
{
  pool->lock = g_mutex_new ();
  pool->cond = g_cond_new();
}


/*
 *
 */
static void
gst_recycle_pool_base_init (gpointer g_class)
{
  /* nop */
}


/*
 *
 */
static void
gst_recycle_pool_class_init (GstRecyclePoolClass * klass)
{
  GstMiniObjectClass *mini_object_class = GST_MINI_OBJECT_CLASS (klass);

  recycle_pool_parent_class = g_type_class_peek_parent (klass);

  mini_object_class->finalize = (GstMiniObjectFinalizeFunction)
      gst_recycle_pool_finalize;
}


#define _do_init(thing) \
  GST_DEBUG_CATEGORY_INIT (gst_recycle_pool_debug, "recyclepool", 0, "object to recycle buffer data chunks")

GST_BOILERPLATE_FULL (GstRecyclePool, gst_recycle_pool, GstMiniObject,
    GST_TYPE_MINI_OBJECT, _do_init);


/*****************************************************************************/
/********************************* PUBLIC API ********************************/
/*****************************************************************************/



/**
 * gst_recycle_pool_new:
 * @max_pool_size: Maximum number of buffers this pool should use.
 * @optimal_pool_size: Number of buffers this pool should try to maintain.
 * @buffer_size: Size (in bytes) of the buffers to be allocated.
 *
 * Creates a new #GstRecyclePool instance.
 *
 * Returns: a new #GstRecyclePool object.
 */
GstRecyclePool *
gst_recycle_pool_new (guint max_pool_size, guint optimal_pool_size,
    guint32 buffer_size)
{
  GstRecyclePool *pool;

  g_return_val_if_fail (max_pool_size != 0, NULL);
  g_return_val_if_fail (optimal_pool_size != 0, NULL);
  g_return_val_if_fail (optimal_pool_size <= max_pool_size, NULL);
  g_return_val_if_fail (buffer_size != 0, NULL);

  pool = (GstRecyclePool *) gst_mini_object_new (GST_TYPE_RECYCLE_POOL);

  GST_DEBUG ("Creating pool, optimal size %d, max size %d", optimal_pool_size,
      max_pool_size);

  /* Pool can't get bigger than 2 x optimal size, because all buffers can
   * be replaced only once */
  if (max_pool_size > optimal_pool_size * 2) {
   GST_WARNING ("Restricting MAX pool size to %d", optimal_pool_size * 2);
   max_pool_size = optimal_pool_size * 2;
  }

  pool->max_size = max_pool_size;
  pool->optimal_size = optimal_pool_size;
  pool->buffer_size = buffer_size;
  pool->running = TRUE;

  GST_DEBUG ("New recycle pool created: %p", pool);

  return pool;
}


/**
 * gst_recycle_pool_destroy:
 * @pool: #GstRecyclePool object.
 *
 * Destroys the pool and frees the unused buffers hold by it. If some buffers
 * are still in use, they will keep a reference to the pool, preventing
 * it to get destroyed. Therefore all buffers which are taken out from pool
 * should be freed first, or made sure that they will get freed at some point.
 */
void
gst_recycle_pool_destroy (GstRecyclePool *pool)
{
  g_return_if_fail (pool != NULL);

  GST_DEBUG ("Pool destroy");

  g_mutex_lock (pool->lock);
  pool->running = FALSE;

  /* Free all unused buffers. The rest will be freed automatically, when
   * they are released */
  _gst_recycle_pool_cleanup_unlocked (pool);

  g_mutex_unlock (pool->lock);

  gst_mini_object_unref (GST_MINI_OBJECT (pool));
}


/**
 * gst_recycle_pool_set_alloc_func:
 * @pool: #GstRecyclePool object.
 * @func: Pointer to buffer allocation function.
 * @user_data: User data to be passed to @func.
 *
 * Sets a callback which is called when #GstRecyclePool wants to allocate
 * a new buffer. This enables writing a custom allocator for the buffers.
 */
void
gst_recycle_pool_set_alloc_func (GstRecyclePool *pool, AllocFunc func,
    gpointer user_data)
{
  g_return_if_fail (pool != NULL);

  pool->alloc_func = func;
  pool->alloc_user_data = user_data;
}


/**
 * gst_recycle_pool_set_recycle_func:
 * @pool: #GstRecyclePool object.
 * @func: Pointer to recycling function.
 * @user_data: User data to be passed to @func.
 *
 * Sets a callback which is called when #GstRecyclePool receives a buffer
 * which was used and got freed. This callback makes it possible to directly
 * recycle the same buffer again.
 */
void
gst_recycle_pool_set_recycle_func (GstRecyclePool *pool, RecycleFunc func,
    gpointer user_data)
{
  g_return_if_fail (pool != NULL);

  pool->recycle_func = func;
  pool->recycle_user_data = user_data;
}


/**
 * gst_recycle_pool_set_free_func:
 * @pool: #GstRecyclePool object.
 * @func: Pointer to freeing function.
 * @user_data: User data to be passed to @func.
 *
 * Sets a callback which is called when a buffer is freed. This should be used
 * together with @gst_recycle_pool_set_alloc_func.
 */
void
gst_recycle_pool_set_free_func (GstRecyclePool *pool, FreeFunc func,
    gpointer user_data)
{
  g_return_if_fail (pool != NULL);

  pool->free_func = func;
  pool->free_user_data = user_data;
}


/**
 * gst_recycle_pool_get_buffer:
 * @pool: #GstRecyclePool object.
 *
 * Retrieve an unused buffer from the pool. If maximum pool size is reached
 * and no buffers are available, this function waits for one to be freed.  If
 * maximum size isn't reached yet, a new buffer will be allocated.
 *
 * Returns: New #GstRecycleBuffer object or NULL on error.
 */
GstRecycleBuffer *
gst_recycle_pool_get_buffer (GstRecyclePool *pool)
{
  GstRecycleBuffer *buf = NULL;
  gboolean found = FALSE;
  GSList *iter;

  g_return_val_if_fail (pool != NULL, NULL);

  g_mutex_lock (pool->lock);

recheck:

  /* First try to find unused recycled buffer from pool */
  for (iter = pool->bufs; iter ; iter = iter->next) {
    buf = iter->data;
    if (buf->in_use == FALSE) {
      GST_LOG ("Found available buffer %p (index %d) from pool", buf->data,
          buf->index);
      found = TRUE;
      break;
    }
  }

  if (!found) {
    if (pool->max_size == 0 || g_slist_length (pool->bufs) < pool->max_size) {
      /* No free buffers, but we can expand the pool, size max size isn't hit */
      buf = _gst_recycle_pool_alloc_buffer_unlocked (pool, pool->buffer_size);
      buf->in_use = TRUE;
    }
    else {
      /* All buffers in use -> we need to wait for one to get freed */
      GST_LOG ("Pool is full. Waiting for a buffer to become available");
      pool->waiting++;
      g_cond_wait (pool->cond, pool->lock);
      pool->waiting--;
      GST_LOG ("Waiting ended. Re-checking pool for buffers");
      goto recheck;
    }
  }

  if (!buf) {
    g_mutex_unlock (pool->lock);
    goto buffer_alloc_failed;
  }

  buf->in_use = TRUE;
  g_mutex_unlock (pool->lock);

  GST_DEBUG ("Returning buffer %p", buf->data);

out:

  return buf;

/* ERRORS */
buffer_alloc_failed:
  {
    GST_WARNING ("Buffer allocation failed");
    /* outbuf is NULL */
    goto out;
  }
}


/**
 * gst_recycle_pool_convert_buffer:
 * @rbuf: buffer to be converted
 *
 * Converts a #GstRecycleBuffer into a #GstBuffer so that it can be pushed
 * into pipeline. This function sets up its own FREE_FUNC for the buffer,
 * so that the memory can be recycled, when buffer is going to be freed.
 *
 * Returns: converted buffer.
 */
GstBuffer *
gst_recycle_pool_convert_buffer (GstRecycleBuffer *rbuf)
{
  GstBuffer *outbuf;

  g_return_val_if_fail (rbuf != NULL, NULL);

  outbuf = gst_buffer_new ();

  GST_BUFFER_DATA (outbuf) = (guint8 *) rbuf->data;
  GST_BUFFER_SIZE (outbuf) = rbuf->pool->buffer_size;
  GST_BUFFER_MALLOCDATA (outbuf) = (guint8 *) rbuf;
  GST_BUFFER_FREE_FUNC (outbuf) = _gst_recycle_pool_buffer_recycle_func;

  return outbuf;
}


/**
 * gst_recycle_pool_free_buffer:
 * @pool: #GstRecyclePool object.
 * @buffer: #GstRecycleBuffer to be freed.
 *
 * Frees a previously allocated @buffer from pool.
 */
void
gst_recycle_pool_free_buffer (GstRecyclePool *pool, GstRecycleBuffer *buffer)
{
  g_return_if_fail (pool != NULL);
  g_return_if_fail (buffer != NULL);

  g_mutex_lock (pool->lock);

  _gst_recycle_pool_release_buffer_unlocked (pool, buffer);

  g_mutex_unlock (pool->lock);
}


/**
 * gst_recycle_pool_cleanup:
 * @pool: #GstRecyclePool object.
 *
 * Clean up the pool by freeing all buffers which are not in use.
 */
void
gst_recycle_pool_cleanup (GstRecyclePool *pool)
{
  g_return_if_fail (pool != NULL);

  g_mutex_lock (pool->lock);

  _gst_recycle_pool_cleanup_unlocked (pool);

  g_mutex_unlock (pool->lock);
}
