/* GStreamer
 *
 * Copyright (C) 2010 Nokia Corporation <multimedia@maemo.org>
 *
 * recyclepool.h Pool which recycles memory pointers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef __GSTRECYCLEPOOL_H__
#define __GSTRECYCLEPOOL_H__

#include <gst/gst.h>

G_BEGIN_DECLS

typedef struct _GstRecyclePool GstRecyclePool;
typedef struct _GstRecyclePoolClass GstRecyclePoolClass;
typedef struct _GstRecycleBuffer GstRecycleBuffer;

/**
 * AllocFunc:
 * @size: number of bytes to allocate
 * @user_data: caller-specifiec data pointer
 *
 * Specifies the type of function passed to @gst_recycle_pool_set_alloc_func.
 */
typedef void * (*AllocFunc) (guint32 size, gpointer user_data);

/**
 * RecycleFunc:
 * @buf: #GstRecycleBuffer to be recycled
 * @user_data: called-specifiec data pointer
 *
 * Specifies the type of function passed to gst_recycle_pool_set_recycle_func.
 */
typedef gboolean (*RecycleFunc) (GstRecycleBuffer *buf, gpointer user_data);

/**
 * FreeFunc:
 * @mem: pointer to memory to be freed
 * @size: size of the memory chunk
 * @user_data: called-specifiec data pointer
 *
 * Specifies the type of function passed to gst_recycle_pool_set_free_func.
 */
typedef void (*FreeFunc) (void *mem, guint32 size, gpointer user_data);


/**
 * GstRecycleBuffer:
 * @data: Buffer data pointer.
 * @in_use: Whether this buffer is in use.
 * @index: Index of the buffer.
 * @replaced: TRUE if this some other buffer is using the same index as this.
 * @pool: #GstRecyclePool where this buffer belongs to.
 *
 * Structure containing information about HQ image buffer.
 */
struct _GstRecycleBuffer {
  void           *data;
  gboolean       in_use;
  guint          index;
  gboolean       replaced;
  GstRecyclePool *pool;
};

/**
 * GstRecyclePool:
 * @parent: parent #GstMiniObject
 * @bufs: List of #GstRecycleBuffer objects
 * @cond: GCond used to signal when a buffer is available in the pool
 * @lock: Mutex for protecting buffer list from simultaneous access
 * @max_size: Max amount of buffers in the pool
 * @optimal_size: Optimal amount of buffers the pool should try to keep
 * @buffer_size: size of the buffers (in bytes) to be allocated
 * @waiting: TRUE if someone is waiting for a buffer to be available
 * @running: TRUE if the pool is running
 * @alloc_func: Custom buffer allocation function
 * @alloc_user_data: data for @alloc_func
 * @recycle_func: Custom buffer recycling function
 * @recycle_user_data: data for @recycle_func
 * @free_func: Custom buffer freeing function
 * @free_user_data: data for @free_func
 *
 * Structure containing information about recycled buffers.
 */
struct _GstRecyclePool {
  GstMiniObject parent;

  GSList *bufs;
  GMutex *lock;
  GCond *cond;
  guint max_size;             /* 0 = unlimited */
  guint optimal_size;
  guint32 buffer_size;
  guint waiting;              /* >0 if someone is waiting in _get_buffer() */
  gboolean running;

  AllocFunc alloc_func;
  gpointer alloc_user_data;

  RecycleFunc recycle_func;
  gpointer recycle_user_data;

  FreeFunc free_func;
  gpointer free_user_data;
};

/**
 * GstRecyclePoolClass:
 * @parent_class: parent class object
 *
 * #GstRecyclePoolClass class object.
 */
struct _GstRecyclePoolClass {
  GstMiniObjectClass  parent_class;

  /*< private >*/
  gpointer _gst_reserved[GST_PADDING];
};

GstRecyclePool * gst_recycle_pool_new (guint max_pool_size,
    guint optimal_pool_size, guint32 buffer_size);

void gst_recycle_pool_destroy (GstRecyclePool *pool);

GstRecycleBuffer * gst_recycle_pool_get_buffer (GstRecyclePool *pool);

GstBuffer * gst_recycle_pool_convert_buffer (GstRecycleBuffer *rbuf);

void gst_recycle_pool_free_buffer (GstRecyclePool *pool,
    GstRecycleBuffer *buffer);

void gst_recycle_pool_cleanup (GstRecyclePool *pool);

void gst_recycle_pool_set_alloc_func (GstRecyclePool *pool, AllocFunc func,
    gpointer user_data);

void gst_recycle_pool_set_recycle_func (GstRecyclePool *pool, RecycleFunc func,
    gpointer user_data);

void gst_recycle_pool_set_free_func (GstRecyclePool *pool, FreeFunc func,
    gpointer user_data);

G_END_DECLS

#endif  /* __GSTRECYCLEPOOL_H__ */
