/* GStreamer
 *
 * Copyright (C) 2010 Nokia Corporation <multimedia@maemo.org>
 *
 * subdev_calls.c - subdev system calls
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <linux/v4l2-subdev.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "gstsubdevcamsrc.h"
#include "subdev_calls.h"
#include "mediacontroller.h"
#include "bufferpool.h"
#include "configparser.h"

GST_DEBUG_CATEGORY_EXTERN (gst_subdevsrc_debug);
#define GST_CAT_DEFAULT gst_subdevsrc_debug

/* complete made up ranking, the values themselves are meaningless */
#define YUV_BASE_RANK     1000
#define BAYER_BASE_RANK   15

#define RESIZER_MAX_DOWNSCALE_FACTOR 4
#define SUBDEVSRC_POLL_TIMEOUT (3 * GST_SECOND)

static gfloat aspect_category_thresholds[] =
{
  1.76,   /* infinity -> 16:9 */
  1.46,   /* 16:9 -> 3:2 */
  1.34,   /* 3:2 -> 4:3 */
  0.00    /* 4:3 -> 0 */
};

/* ======== INTERNAL FUNCTIONS ======== */


/*
 * gst_subdevsrc_format_get_rank:
 * @fourcc: Color format in fourcc.
 *
 * Returns: Numeric rank value for given color format.
 */
static gint
gst_subdevsrc_format_get_rank (guint32 fourcc)
{
  switch (fourcc) {

    case V4L2_MBUS_FMT_UYVY8_1X16:   /* UYVY, 16 bits per pixel */
      return YUV_BASE_RANK + 10;
    case V4L2_MBUS_FMT_YUYV8_1X16:   /* YUY2, 16 bits per pixel */
      return YUV_BASE_RANK + 9;
    case V4L2_MBUS_FMT_SGRBG10_1X10:
      return BAYER_BASE_RANK;

    default:
      break;
  }

  return 0;
}


/*
 * gst_subdevsrc_get_ar_category:
 * @width: Resolution width.
 * @height: Resolution height.
 *
 * Returns: Aspect ratio category number.
 */
static guint
gst_subdevsrc_get_ar_category (guint width, guint height)
{
  gfloat ar = (gfloat) width / height;
  guint i = 0;

  while (aspect_category_thresholds[i] != 0.0) {
    if (ar > aspect_category_thresholds[i]) break;
    i++;
  }

  return i;
}


/*
 * gst_subdevsrc_fourcc_to_pcode:
 * @fourcc: Color format in fourcc.
 * @pcode: pointer where to place corresponding subdev color code.
 *
 * Convert fourcc color format into subdev color code.
 *
 * Returns: TRUE if the format was supported.
 */
static gboolean
gst_subdevsrc_fourcc_to_pcode (guint32 fourcc, enum v4l2_mbus_pixelcode *pcode)
{
  gboolean ret = TRUE;

  switch (fourcc) {
    case GST_MAKE_FOURCC ('Y', 'U', 'Y', 'V'):
    case GST_MAKE_FOURCC ('Y', 'U', 'Y', '2'):
      *pcode = V4L2_MBUS_FMT_YUYV8_1X16;
      break;
    case GST_MAKE_FOURCC ('U', 'Y', 'V', 'Y'):
      *pcode = V4L2_MBUS_FMT_UYVY8_1X16;
      break;
    default:
      GST_WARNING ("Unsupported fourcc requested");
      ret = FALSE;
  }

  return ret;
}


/*
 * gst_subdevsrc_format_cmp_func:
 * @a: pointer to v4l2_subdev_mbus_code_enum structure.
 * @b: pointer to v4l2_subdev_mbus_code_enum structure.
 *
 * Compare two color formats. This method is a comparator function for
 * g_slist_insert_sorted().
 *
 * Returns: Value > 0 if format @b is greater than format @a.
 */
static gint
gst_subdevsrc_format_cmp_func (gconstpointer a, gconstpointer b)
{
  guint32 pf1 = ((struct v4l2_subdev_mbus_code_enum *) a)->code;
  guint32 pf2 = ((struct v4l2_subdev_mbus_code_enum *) b)->code;

  if (pf1 == pf2)
    return 0;

  return gst_subdevsrc_format_get_rank (pf2) -
        gst_subdevsrc_format_get_rank (pf1);
}


/*
 * gst_subdevsrc_size_cmp_func:
 * @a: #GstVideoFrameInfo object.
 * @b: #GstVideoFrameInfo object.
 *
 * Compare frame sizes. this method is a comparator function for g_slist_sort().
 * The comparison is performed by calculating the acreage for both frames and
 * comparing those values.
 *
 * Returns: Value greater than 0 if frame @b is bigger than frame @a.
 */
static gint
gst_subdevsrc_size_cmp_func (gconstpointer a, gconstpointer b)
{
  GstVideoFrameInfo *s1 = (GstVideoFrameInfo *) a;
  GstVideoFrameInfo *s2 = (GstVideoFrameInfo *) b;
  guint32 p1 = s1->width * s1->height;
  guint32 p2 = s2->width * s2->height;

  return p1 - p2;
}


/*
 * gst_frameinfo_free:
 * @finfo: #GstVideoFrameInfo object.
 *
 * Deallocates the given #GstVideoFrameInfo object and its contents.
 */
static void
gst_frameinfo_free (GstVideoFrameInfo *finfo)
{
  if (finfo->framerates) {
    g_value_unset (finfo->framerates);
    g_free (finfo->framerates);
  }
  g_slice_free (GstVideoFrameInfo, finfo);
}


/*
 * gst_fmt_free:
 * @fmt: v4l2_subdev_mbus_code_enum structure.
 *
 * Deallocates the given v4l2_subdev_mbus_code_enum structure.
 */
static void
gst_fmt_free (struct v4l2_subdev_mbus_code_enum *fmt)
{
  g_slice_free (struct v4l2_subdev_mbus_code_enum, fmt);
}


/*
 * gst_subdevsrc_sensor_probe_framerates:
 * @subdevsrc: #GstSubdevSrc object.
 * @fd: Sensor device file descriptor.
 * @width: width of the resolution
 * @height: height of the resolution
 * @pixelformat: fourcc of the desired format
 *
 * Retrieve a list of supported framerates for given resolution and color
 * format.
 *
 * Returns: Pointer to GValue containing list of fraction values. Return value
 * must be freed with g_free().
 */
static GValue *
gst_subdevsrc_sensor_probe_framerates (GstSubdevSrc *subdevsrc, guint fd,
    guint width, guint height, guint32 pixelformat)
{
  struct v4l2_subdev_frame_interval_enum ival;
  guint32 num, denom;
  GValue rate = { 0, };
  GValue *rates = NULL;

  memset (&ival, 0, sizeof (struct v4l2_frmivalenum));
  ival.index = 0;
  ival.pad = 0;     /* Sensors have only one pad */
  ival.code = pixelformat;
  ival.width = width;
  ival.height = height;

  GST_LOG_OBJECT (subdevsrc, "Get frame interval for %ux%u", width, height);

  /* keep in mind that v4l2 gives us frame intervals (durations); we invert the
   * fraction to get framerate */
  if (ioctl (fd, VIDIOC_SUBDEV_ENUM_FRAME_INTERVAL, &ival) < 0)
    goto enum_frameintervals_failed;

  rates = g_new0 (GValue, 1);
  g_value_init (rates, GST_TYPE_LIST);
  g_value_init (&rate, GST_TYPE_FRACTION);

  do {
    num = ival.interval.numerator;
    denom = ival.interval.denominator;

    GST_LOG_OBJECT (subdevsrc, "  Adding framerate: %d/%d", denom, num);

    /* swap to get the framerate */
    gst_value_set_fraction (&rate, denom, num);
    gst_value_list_append_value (rates, &rate);

    ival.index++;
  } while (ioctl (fd, VIDIOC_SUBDEV_ENUM_FRAME_INTERVAL, &ival) >= 0);

return_data:

  return rates;

    /* ERRORS */
enum_frameintervals_failed:
  {
    GST_WARNING_OBJECT (subdevsrc,
        "Unable to enumerate intervals for %" GST_FOURCC_FORMAT "@%ux%u",
        GST_FOURCC_ARGS (pixelformat), width, height);
    goto return_data;
  }
}


/*
 * gst_subdevsrc_probe_sensor_resolutions:
 * @subdevsrc: #GstSubdevSrc object.
 * @fd: Sensor device file descriptor.
 *
 * Query the sensor for supported output resolutions and store them into
 * subdevsrc's internal "sensor_fmts" list.
 *
 * Returns: TRUE on success.
 */
static gboolean
gst_subdevsrc_probe_sensor_resolutions (GstSubdevSrc *subdevsrc, guint fd)
{
  struct v4l2_subdev_mbus_code_enum format;
  struct v4l2_subdev_frame_size_enum size;
  GstVideoFrameInfo *finfo;
  gboolean ret = FALSE;
  guint n;

  GST_DEBUG_OBJECT (subdevsrc, "Probing sensor resolutions");

  /* format enumeration */
  for (n = 0;; n++) {

    /* It has only one pad */
    format.pad = 0;
    format.index = n;

    if (ioctl (fd, VIDIOC_SUBDEV_ENUM_MBUS_CODE, &format) < 0) {
      if (errno == EINVAL) {
        break;                  /* end of enumeration */
      } else {
        goto enum_codes_failed;
      }
    }

    GST_LOG_OBJECT (subdevsrc, "got code: %d", format.code);

    size.index = 0;
    size.pad = 0;        /* Sensor has always only one pad */
    size.code = format.code;

    GST_DEBUG_OBJECT (subdevsrc,
        "querying frame size, index = %d, pad = %d, code = %d",
        size.index, size.pad, size.code);

    if (ioctl (fd, VIDIOC_SUBDEV_ENUM_FRAME_SIZE, &size) < 0) {
      goto enum_framesizes_failed;
    }

    do {
      finfo = g_slice_new (GstVideoFrameInfo);
      finfo->width = size.max_width & ~(CAPTURE_WIDTH_MODULO-1);
      finfo->height = size.max_height & ~(CAPTURE_HEIGHT_MODULO-1);
      finfo->color_code = size.code;
      finfo->ar_category = gst_subdevsrc_get_ar_category (finfo->width,
          finfo->height);

      GST_DEBUG_OBJECT (subdevsrc, "Got resolution: %ux%u - %ux%u",
          size.min_width, size.min_height, size.max_width, size.max_height);

      finfo->framerates = gst_subdevsrc_sensor_probe_framerates (subdevsrc, fd,
          size.max_width, size.max_height, size.code);

      if (!finfo->framerates) {
        gst_frameinfo_free (finfo);
        goto enum_framerates_failed;
      }
      subdevsrc->sensor_fmts = g_slist_append (subdevsrc->sensor_fmts, finfo);

      size.index++;
    } while (ioctl (fd, VIDIOC_SUBDEV_ENUM_FRAME_SIZE, &size) >= 0);
  }

  ret = TRUE;

  /* Sort the resolutions from smallest to largest */
  subdevsrc->sensor_fmts = g_slist_sort (subdevsrc->sensor_fmts,
    gst_subdevsrc_size_cmp_func);

out:

  return ret;

    /* ERRORS */
enum_codes_failed:
  {
    GST_WARNING_OBJECT (subdevsrc, "Unable to enumerate sensor pixel formats");
    goto out;
  }
enum_framesizes_failed:
  {
    GST_WARNING_OBJECT (subdevsrc, "Unable to enumerate sensor frame sizes");
    goto out;
  }
enum_framerates_failed:
  {
    GST_WARNING_OBJECT (subdevsrc, "Unable to enumerate sensor frame rates");
    goto out;
  }
}


/*
 * gst_subdevsrc_fill_format_list:
 * @subdevsrc: #GstSubdevSrc object.
 * @fd: Subdev device node where to probe the color formats.
 *
 * Query previewer entity which output color formats it supports.
 *
 * Returns: TRUE on success.
 */
static gboolean
gst_subdevsrc_fill_format_list (GstSubdevSrc *subdevsrc, gint fd)
{
  struct v4l2_subdev_mbus_code_enum *format;
  guint n;

  GST_DEBUG_OBJECT (subdevsrc, "Getting format enumerations from previewer");

  /* format enumeration */
  for (n = 0;; n++) {
    format = g_slice_new (struct v4l2_subdev_mbus_code_enum);

    /* It has only one pad */
    format->pad = 1;
    format->index = n;

    if (ioctl (fd, VIDIOC_SUBDEV_ENUM_MBUS_CODE, format) < 0) {
      if (errno == EINVAL) {
        g_slice_free (struct v4l2_subdev_mbus_code_enum, format);
        break;                  /* end of enumeration */
      } else {
        goto failed;
      }
    }

    GST_LOG_OBJECT (subdevsrc, "index:  %u", format->index);
    GST_LOG_OBJECT (subdevsrc, "  code: %d", format->code);
    GST_LOG_OBJECT (subdevsrc, "  pad:  %d", format->pad);

    /* sort formats according to our preference;  we do this, because caps
     * are probed in the order the formats are in the list, and the order of
     * formats in the final probed caps matters for things like fixation */
    subdevsrc->capture_formats =
        g_slist_insert_sorted (subdevsrc->capture_formats, format,
        (GCompareFunc) gst_subdevsrc_format_cmp_func);
  }

  GST_DEBUG_OBJECT (subdevsrc, "Got %d format(s)", n);

  return TRUE;

  /* ERRORS */
failed:
  {
    GST_ELEMENT_ERROR (subdevsrc, RESOURCE, SETTINGS,
        ("Failed to enumerate possible video formats"),
        ("Failed to get number %d in pixelformat enumeration. (%d - %s)",
        n, errno, g_strerror (errno)));
    gst_fmt_free (format);
    return FALSE;
  }
}


/*
 * Go through the pipeline configuration and open all devices in it. If
 * pipeline uses assignments which have multiple definitions (e.g. sensor)
 * the devices will be opened in the same order as defined in configuration
 * file. First successfully opened device will be used for the rest of the
 * pipeline operations.
 *
 * This also replaces possible mapping keys in pipeline definition with
 * actual entity identifiers.
 */
static gboolean
gst_subdevsrc_open_pipeline_devices (GstSubdevSrc *subdevsrc,
    GstPipeConfiguration *pipe)
{
  GstPipeConfigurationEntry *e;
  GSList *map, *iter, *iter2;
  gint fd = -1;
  gchar *ename;

  for (iter = pipe->entries; iter != NULL; iter = iter->next) {
    e = (GstPipeConfigurationEntry *) iter->data;

    /* Handle all entity elements in the pipeline */
    if (e->type == GST_PIPE_ENTRY_ENTITY) {

      /* Find all mappings for the given entity */
      map = gst_subdevsrc_config_parser_get_mappings (subdevsrc->cp,
          e->field.entity_name);

      for (iter2 = map; iter2 != NULL; iter2 = iter2->next) {
        ename = (gchar *) iter2->data;
        fd = gst_subdevsrc_mc_get_device_fd (subdevsrc->mc, ename);
        if (fd != -1) {
          /* Device opened. Replace the name with entity identifier */
          g_free (e->field.entity_name);
          e->field.entity_name = g_strdup (ename);
          break;
        }
      }

      g_slist_foreach (map, (GFunc) g_free, NULL);
      g_slist_free (map);

      if (fd < 0) {
        goto open_failed;
      }

      if (iter == pipe->entries) {
        /* It was the first element, so it is a sensor. Store fd */
        subdevsrc->sensor_fd = fd;
        if (!gst_subdevsrc_probe_sensor_resolutions (subdevsrc, fd)) {
          goto resolutions_failed;
        }
      }
    }
    else if (e->type == GST_PIPE_ENTRY_SET_OUT_PIXFMT) {
      /* This is the point where output color format will be set. It means
       * that previous node is the one where we need to query supported
       * output formats. Currently "fd" contains last entity file handle */
      gst_subdevsrc_fill_format_list (subdevsrc, fd);
    }
    else if (e->type == GST_PIPE_ENTRY_SET_OUT_RESOLUTION) {
      /* This is the point where output resolution will be set. It means
       * that previous node is the one where we need to set the output
       * size. Currently "fd" contains last entity file handle */
      subdevsrc->rsz_fd = fd;
    }
  }

  /* The last entity is always the video output node */
  if (fd != -1) {
    subdevsrc->video_fd  = fd;
  }

  return TRUE;

open_failed:
  {
    GST_ELEMENT_ERROR (subdevsrc, RESOURCE, OPEN_READ_WRITE,
        ("Opening a device failed"), GST_ERROR_SYSTEM);
    goto error;
  }
resolutions_failed:
  {
    GST_ELEMENT_ERROR (subdevsrc, RESOURCE, OPEN_READ_WRITE,
        ("Failed enumerating sensor resolutions"), GST_ERROR_SYSTEM);
    goto error;
  }
error:
  {
    return FALSE;
  }
}


/*
 * gst_subdevsrc_resizer_set_crop:
 * @subdevsrc: GstSubdevSrc object:
 * @x: X coordinate of the top-left corner of cropping area.
 * @y: Y coordinate of the top-left corner of cropping area.
 * @w: Cropping area width.
 * @h: Cropping area height.
 *
 * Configure the cropping to given rectangle.
 *
 * Returns: TRUE on success.
 */
static gboolean
gst_subdevsrc_resizer_set_crop (GstSubdevSrc *subdevsrc, guint x, guint y,
    guint w, guint h)
{
  gboolean ret = FALSE;
  struct v4l2_subdev_crop scrop;
  struct v4l2_subdev_crop gcrop;
  int fd;

  fd = subdevsrc->rsz_fd;

  memset (&scrop, 0, sizeof (scrop));
  memset (&gcrop, 0, sizeof (scrop));

  scrop.pad = 0;    /* It has only one sink pad */
  scrop.which = V4L2_SUBDEV_FORMAT_ACTIVE;
  scrop.rect.left = x;
  scrop.rect.top = y;
  scrop.rect.width = w;
  scrop.rect.height = h;

  GST_DEBUG_OBJECT (subdevsrc,
      "Crop request: left = %d, top = %d, w = %d, h = %d",
      scrop.rect.left, scrop.rect.top, scrop.rect.width, scrop.rect.height);

  if (-1 == ioctl (fd, VIDIOC_SUBDEV_S_CROP, &scrop)) {
    goto s_crop_failed;
  }

  GST_DEBUG_OBJECT (subdevsrc, "S_CROP success");

  gcrop.pad = 0;    /* It has only one sink pad */
  gcrop.which = V4L2_SUBDEV_FORMAT_ACTIVE;

  if (-1 == ioctl (fd, VIDIOC_SUBDEV_G_CROP, &gcrop)) {
    goto g_crop_failed;
  }

  GST_DEBUG_OBJECT (subdevsrc,
      "Crop selected: left = %d, top = %d, w = %d, h = %d",
      gcrop.rect.left, gcrop.rect.top, gcrop.rect.width, gcrop.rect.height);

#if 0
  if (gcrop.c.left != scrop.c.left || gcrop.c.top != scrop.c.top ||
      gcrop.c.width != scrop.c.width || gcrop.c.height != scrop.c.height)
  {
    goto crop_not_supported;
  }
#endif

  ret = TRUE;

done:

  return ret;


/* ERRORS */
s_crop_failed:
  GST_WARNING_OBJECT (subdevsrc, "VIDIOC_SUBDEV_S_CROP not supported: %s",
      g_strerror (errno));
  goto done;

g_crop_failed:
  GST_WARNING_OBJECT (subdevsrc, "VIDIOC_SUBDEV_G_CROP not supported: %s",
      g_strerror (errno));
  goto done;

#if 0
crop_not_supported:
  GST_WARNING_OBJECT (subdevsrc, "Given crop value not accepted");
  goto done;
#endif
}


/*
 * gst_subdevsrc_update_cropping_unlocked:
 * @subdevsrc: #GstSubdevSrc object.
 * @width: Output width.
 * @height: Output height.
 * @zoom: Desired zoom factor.
 *
 * Set up cropping according to given output resolution and zoom factor.
 *
 * Returns: TRUE on success.
 */
static gboolean
gst_subdevsrc_update_cropping_unlocked (GstSubdevSrc *subdevsrc, guint width,
    guint height, gfloat zoom)
{
  gfloat input_ar, reso_ar;
  guint crop_x, crop_y, crop_w, crop_h;

  g_return_val_if_fail (width != 0, FALSE);
  g_return_val_if_fail (height != 0, FALSE);
  g_return_val_if_fail (zoom != 0, FALSE);

  if (zoom < 1.0 || zoom > subdevsrc->max_zoom_factor) {
    GST_WARNING_OBJECT (subdevsrc, "invalid zoom = %.2f", zoom);
    return FALSE;
  }

  input_ar = (gfloat) subdevsrc->resizer.width / subdevsrc->resizer.height;
  reso_ar = (gfloat) width / height;

  if (input_ar > reso_ar) {
    crop_w = (width * subdevsrc->resizer.height / height) / zoom;
    crop_h = MAX (subdevsrc->resizer.height, height) / zoom;
  }
  else {
    crop_w = MAX (subdevsrc->resizer.width, width) / zoom;
    crop_h = (height * subdevsrc->resizer.width / width) / zoom;
  }

  crop_x = ABS ((gint)(subdevsrc->resizer.width - crop_w)) / 2;
  crop_y = ABS ((gint)(subdevsrc->resizer.height - crop_h)) / 2;

  GST_LOG_OBJECT (subdevsrc, "set cropping: x: %d, y: %d, w: %d, h: %d",
      crop_x, crop_y, crop_w, crop_h);

  return gst_subdevsrc_resizer_set_crop (subdevsrc, crop_x, crop_y,
      crop_w, crop_h);
}


/*
 * Go through the pipeline configuration and link entities according
 * to it. Also set up output resolution and color formats.
 */
static gboolean
gst_subdevsrc_link_pipeline (GstSubdevSrc *subdevsrc,
    GstPipeConfiguration *pipe, enum v4l2_mbus_pixelcode out_pcode,
    guint *width, guint *height)
{
  struct v4l2_mbus_framefmt format;
  const gchar *src = NULL;
  const gchar *sink = NULL;
  GstPipeConfigurationEntry *e;
  GSList *iter;

  format.width = subdevsrc->sensor.width;
  format.height = subdevsrc->sensor.height;
  format.field = V4L2_FIELD_ANY;
  format.colorspace = V4L2_COLORSPACE_SRGB;
  format.code = subdevsrc->sensor.color_code;

  for (iter = pipe->entries; iter != NULL; iter = iter->next) {
    e = (GstPipeConfigurationEntry *) iter->data;

    if (e->type == GST_PIPE_ENTRY_ENTITY) {
      if (src == NULL) {
        src = e->field.entity_name;
      }
      else {
        sink = e->field.entity_name;
        if (!gst_subdevsrc_mc_link_entities (subdevsrc->mc, src, sink, &format))
        {
          GST_WARNING_OBJECT (subdevsrc, "Activating a link failed");
          return FALSE;
        }
        /* Jump to next element */
        src = sink;
      }
    }
    else if (e->type == GST_PIPE_ENTRY_CHANGE_PIXFMT) {
      GST_DEBUG_OBJECT (subdevsrc, "=== CHANGING FORMAT PIXCODE ===");
      format.code = e->field.pcode;
    }
    else if (e->type == GST_PIPE_ENTRY_SET_OUT_PIXFMT) {
      GST_DEBUG_OBJECT (subdevsrc, "=== SETTING OUTPUT PIXCODE ===");
      format.code = out_pcode;
    }
    else if (e->type == GST_PIPE_ENTRY_SET_OUT_RESOLUTION) {
      subdevsrc->resizer.width = format.width;
      subdevsrc->resizer.height = format.height;

      /* Set the desired output size */
      format.width = *width;
      format.height = *height;

      GST_DEBUG_OBJECT (subdevsrc, "Resizer input resolution: %dx%d",
          subdevsrc->resizer.width, subdevsrc->resizer.height);

      gst_subdevsrc_update_cropping_unlocked (subdevsrc, *width, *height,
          subdevsrc->zoom_factor);
    }
  }

  *width = format.width;
  *height = format.height;

  return TRUE;
}


/*
 * gst_subdevsrc_select_sensor_mode:
 * @subdevsrc: #GstSubdevSrc object.
 * @width: desired width to be used.
 * @height: desired height to be used.
 * @fps_n: Framerate fraction numerator.
 * @fps_d: Framerate fraction denominator.
 *
 * Find best sensor resolution and framerate for given output format.
 *
 * Returns: TRUE on success.
 */
static gboolean
gst_subdevsrc_select_sensor_mode (GstSubdevSrc *subdevsrc, guint32 width,
   guint32 height, gint fps_n, gint fps_d)
{
  GstVideoFrameInfo *finfo = NULL;
  gboolean test_ar_categories = TRUE;
  guint output_ar_category;
  guint32 mwidth, mheight;
  GSList *iter;

  g_return_val_if_fail (subdevsrc->sensor_fmts != NULL, FALSE);

  output_ar_category = gst_subdevsrc_get_ar_category (width, height);
  mwidth = width * MAX_RESIZER_FACTOR;
  mheight = height * MAX_RESIZER_FACTOR;

try_again:

  for (iter = subdevsrc->sensor_fmts; iter; iter = iter->next) {
    finfo = (GstVideoFrameInfo *) iter->data;

    /* 2 rules:                                                           */
    /*   - Never select smaller sensor resolution than output size        */
    /*   - Select only resolution which can be scaled down to output size */

    if (finfo->width >= width && finfo->height >= height &&
        finfo->width <= mwidth && finfo->height <= mheight)
    {
      guint i, size;

      size = gst_value_list_get_size (finfo->framerates);
      GST_LOG_OBJECT (subdevsrc, "Found candidate resolution: %dx%d",
                      finfo->width, finfo->height);

      if (test_ar_categories && output_ar_category != finfo->ar_category) {
        GST_LOG_OBJECT (subdevsrc, "  Aspect ratios do not match!");
        continue;
      }

      for (i=0; i<size; i++) {
        const GValue *fract = gst_value_list_get_value (finfo->framerates, i);
        gint fps_num = gst_value_get_fraction_numerator (fract);
        gint fps_denom = gst_value_get_fraction_denominator (fract);

        GST_LOG_OBJECT (subdevsrc, "  Comparing framerates %d/%d == %d/%d",
            fps_n, fps_d, fps_num, fps_denom);

        if (fps_n == fps_num && fps_denom == fps_d) {
          GST_LOG_OBJECT (subdevsrc, "  This resolution & FPS is usable");
          goto resolution_found;
        }
      }
    }
  }

  if (test_ar_categories) {
    GST_DEBUG_OBJECT (subdevsrc, "Searching resolution again without AR check");
    test_ar_categories = FALSE;
    goto try_again;
  }

  GST_WARNING_OBJECT (subdevsrc, "Unable to find suitable sensor resolution");
  return FALSE;

resolution_found:

  GST_DEBUG_OBJECT (subdevsrc, "Best sensor resolution is %ux%u",
      finfo->width, finfo->height);

  subdevsrc->sensor.width = finfo->width;
  subdevsrc->sensor.height = finfo->height;
  subdevsrc->sensor.fps_n = fps_n;
  subdevsrc->sensor.fps_d = fps_d;
  subdevsrc->sensor.color_code = finfo->color_code;

  return TRUE;
}


/*
 * gst_subdevsrc_set_sensor_framerate:
 * @subdevsrc: #GstSubdevSrc object.
 * @fd: Sensor device file descriptor.
 *
 * Configure sensor device to given framerate.
 *
 * Returns: TRUE on success.
 */
static gboolean
gst_subdevsrc_set_sensor_framerate (GstSubdevSrc *subdevsrc, guint fd)
{
  struct v4l2_subdev_frame_interval ival;

  GST_DEBUG_OBJECT (subdevsrc, "Setting framerate %d/%d to sensor",
      subdevsrc->sensor.fps_n, subdevsrc->sensor.fps_d);

  /* Convert from FPS to interval --> change numerator and denominator */
  ival.interval.numerator = subdevsrc->sensor.fps_d;
  ival.interval.denominator = subdevsrc->sensor.fps_n;

  if (ioctl (fd, VIDIOC_SUBDEV_S_FRAME_INTERVAL, &ival) < 0) {
    GST_WARNING_OBJECT (subdevsrc, "VIDIOC_SUBDEV_S_FRAME_INTERVAL failed");
    return FALSE;
  }

  GST_DEBUG_OBJECT (subdevsrc, "Framerate successfully set");

  return TRUE;
}


/* gst_subdevsrc_add_night_mode_fps:
 * @structure: GstStructure where to add the framerate
 * @framerates: A GValueList containing all possible framerates
 *
 * Go through the @framerates list and choose one which is best for night
 * capture viewfinder. Add the selected framerate into given @structure.
 *
 * Returns: TRUE if proper framerate was found and added.
 */
static gboolean
gst_subdevsrc_add_night_mode_fps (GstStructure *structure, GValue *framerates)
{
  const GValue *fract;
  gint fps_num, fps_denom;
  gint nearest_idx = -1;
  gfloat tmp, diff = 9999.9;
  gboolean ret = FALSE;
  guint i, size;

  size = gst_value_list_get_size (framerates);

  for (i=0; i<size; i++) {
    gdouble fps;

    fract = gst_value_list_get_value (framerates, i);
    fps_num = gst_value_get_fraction_numerator (fract);
    fps_denom = gst_value_get_fraction_denominator (fract);

    GST_DEBUG ("Checking framerate %d/%d", fps_num, fps_denom);

    gst_util_fraction_to_double (fps_num, fps_denom, &fps);

    tmp = ABS (fps - 15.0);
    if (tmp < diff && tmp < 2.0) {
      nearest_idx = i;
      diff = tmp;
      GST_DEBUG ("New nearest_idx %d (diff %f)", i, diff);
    }
  }

  if (nearest_idx != -1) {
    fract = gst_value_list_get_value (framerates, nearest_idx);
    gst_structure_set_value (structure, "framerate", fract);
    ret = TRUE;
  }

  return ret;
}


/******************************************************************************/
/*                                PUBLIC API                                  */
/******************************************************************************/


/**
 * gst_subdevsrc_open_devices:
 * @subdevsrc: #GstSubdevSrc object.
 *
 * Open all necessary (sub)devices that are needed for video streaming.
 *
 * Returns: TRUE on success.
 */
gboolean
gst_subdevsrc_open_devices (GstSubdevSrc *subdevsrc)
{
  GstPollFD pollfd = GST_POLL_FD_INIT;
  gboolean pret;

  g_mutex_lock (subdevsrc->device_mutex);

  subdevsrc->mc = gst_subdevsrc_mc_init ();
  if (!subdevsrc->mc) {
    goto mc_fail;
  }

  subdevsrc->cp = gst_subdevsrc_config_parser_new ("/etc/camera/pipelines.conf");
  if (!subdevsrc->cp) {
    goto config_fail;
  }

  if (subdevsrc->camera_device == GST_SUBDEVSRC_CAPTURE_DEVICE_PRIMARY) {
    pret = gst_subdevsrc_open_pipeline_devices (subdevsrc, subdevsrc->cp->primary);
  }
  else {
    pret = gst_subdevsrc_open_pipeline_devices (subdevsrc, subdevsrc->cp->secondary);
  }

  if (!pret) {
    goto sensor_not_open;
  }

  GST_INFO_OBJECT (subdevsrc, "Subdevs opened successfully");

  subdevsrc->is_open = TRUE;

  pollfd.fd = subdevsrc->video_fd;
  gst_poll_add_fd (subdevsrc->poll, &pollfd);
  gst_poll_fd_ctl_read (subdevsrc->poll, &pollfd, TRUE);

  g_mutex_unlock (subdevsrc->device_mutex);

  return TRUE;


/* ERRORS */
sensor_not_open:
  {
    GST_ELEMENT_ERROR (subdevsrc, RESOURCE, OPEN_READ_WRITE,
        ("Could not open sensor device for reading and writing."),
        GST_ERROR_SYSTEM);
    goto error;
  }
mc_fail:
  {
    GST_ELEMENT_ERROR (subdevsrc, RESOURCE, OPEN_READ_WRITE,
        ("Initializing the media controller failed"), GST_ERROR_SYSTEM);
    goto error;
  }
config_fail:
  {
    GST_ELEMENT_ERROR (subdevsrc, RESOURCE, OPEN_READ_WRITE,
        ("Reading pipeline configurations file failed"), GST_ERROR_SYSTEM);
    goto error;
  }
error:
  {
    g_mutex_unlock (subdevsrc->device_mutex);
    gst_subdevsrc_close_devices (subdevsrc);
    return FALSE;
  }
}


/**
 * gst_subdevsrc_close_devices:
 * @subdevsrc: #GstSubdevSrc object:
 *
 * Closes all currently opened devices.
 *
 * Returns: TRUE on success.
 */
gboolean
gst_subdevsrc_close_devices (GstSubdevSrc *subdevsrc)
{
  GstPollFD pollfd = GST_POLL_FD_INIT;

  GST_DEBUG_OBJECT (subdevsrc, "Trying to close devices");

  g_mutex_lock (subdevsrc->device_mutex);

  g_slist_foreach (subdevsrc->capture_formats, (GFunc) gst_fmt_free, NULL);
  g_slist_free (subdevsrc->capture_formats);
  subdevsrc->capture_formats = NULL;

  g_slist_foreach (subdevsrc->sensor_fmts, (GFunc) gst_frameinfo_free, NULL);
  g_slist_free (subdevsrc->sensor_fmts);
  subdevsrc->sensor_fmts = NULL;

  if (subdevsrc->video_fd != -1) {
    pollfd.fd = subdevsrc->video_fd;
    gst_poll_remove_fd (subdevsrc->poll, &pollfd);
  }

  gst_subdevsrc_mc_deinit (subdevsrc->mc);
  gst_subdevsrc_config_parser_destroy (subdevsrc->cp);
  subdevsrc->is_open = FALSE;
  subdevsrc->sensor_fd = -1;
  subdevsrc->rsz_fd = -1;
  subdevsrc->video_fd = -1;

  g_mutex_unlock (subdevsrc->device_mutex);

  return TRUE;
}


/**
 * gst_subdevsrc_configure_pipeline:
 * @subdevsrc: #GstSubdevSrc object.
 * @fourcc: Color format in V4L2 format.
 * @width: width of the frame.
 * @height: height of the frame.
 * @fps_n: framerate fraction numerator.
 * @fps_d: framerate fraction denominator.
 *
 * Configure subdev pipeline to given resolution & format & framerate. The
 * parameters given may be altered during the call and final settings being
 * used may differ a bit from given ones.
 *
 * Returns: TRUE on success.
 */
gboolean
gst_subdevsrc_configure_pipeline (GstSubdevSrc *subdevsrc,
    guint32 *fourcc, guint32 *width, guint32 *height, guint32 *fps_n,
    guint32 *fps_d)
{
  enum v4l2_mbus_pixelcode pcode;
  GstPipeConfiguration *pipe;
  gboolean ret = FALSE;

  if (subdevsrc->capture_mode == GST_CAMERA_SRC_CAPTURE_MODE_STILL) {
    GST_DEBUG_OBJECT (subdevsrc, "Initialization not needed in STILL mode");
    return TRUE;
  }

  GST_DEBUG_OBJECT (subdevsrc, "Capture res: %dx%d @ %f fps",
      *width, *height, (float) *fps_n / *fps_d);

  if (!gst_subdevsrc_fourcc_to_pcode (*fourcc, &pcode)) {
    GST_WARNING_OBJECT (subdevsrc, "Color format not supported");
    return FALSE;
  }

  g_mutex_lock (subdevsrc->device_mutex);

  gst_subdevsrc_mc_reset_links (subdevsrc->mc);

  /* Step 1: select suitable sensor resolution */
  if (!gst_subdevsrc_select_sensor_mode (subdevsrc, *width, *height, *fps_n,
    *fps_d))
  {
    goto sensor_error;
  }

  if (subdevsrc->camera_device == GST_SUBDEVSRC_CAPTURE_DEVICE_PRIMARY) {
    pipe = subdevsrc->cp->primary;
  }
  else {
    pipe = subdevsrc->cp->secondary;
  }

  /* Step 2: build the pipeline */
  if (!gst_subdevsrc_link_pipeline (subdevsrc, pipe, pcode, width, height)) {
    goto pipeline_error;
  }

  if (!gst_subdevsrc_set_sensor_framerate (subdevsrc, subdevsrc->sensor_fd)) {
    goto framerate_error;
  }

  subdevsrc->frame_byte_size = subdevsrc->mc->output_size;
  ret = TRUE;

done:
  g_mutex_unlock (subdevsrc->device_mutex);

  return ret;

/* ERRORS */
sensor_error:
  {
    GST_ELEMENT_ERROR (subdevsrc, RESOURCE, SETTINGS,
        ("Failed to configure sensor"),
        ("Setting sensor resolution failed"));
    ret = FALSE;
    goto done;
  }

pipeline_error:
  {
    GST_ELEMENT_ERROR (subdevsrc, RESOURCE, SETTINGS,
        ("Failed to link entities"),
        ("Linking of pipeline failed"));
    ret = FALSE;
    goto done;
  }

framerate_error:
  {
    GST_ELEMENT_ERROR (subdevsrc, RESOURCE, SETTINGS,
        ("Failed configure sensor framerate"),
        ("Configuring framerate %d/%d failed",
            subdevsrc->sensor.fps_n, subdevsrc->sensor.fps_d));
    ret = FALSE;
    goto done;
  }
}


/**
 * gst_subdevsrc_start_streaming:
 * @subdevsrc: #GstSubdevSrc object.
 * @caps: GstCaps to be assigned into GstBuffers.
 *
 * Prepare camera device for streaming. Allocates video buffers and queues
 * them to video device.
 *
 * Returns: TRUE on success.
 */
gboolean
gst_subdevsrc_start_streaming (GstSubdevSrc *subdevsrc, GstCaps *caps)
{
  gint type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  struct v4l2_requestbuffers breq;
  gint fd = subdevsrc->video_fd;
  gboolean ret = FALSE;

  GST_DEBUG_OBJECT (subdevsrc, "starting up");

  if (subdevsrc->capture_mode == GST_CAMERA_SRC_CAPTURE_MODE_STILL) {
    GST_DEBUG_OBJECT (subdevsrc, "Start not needed in STILL mode");
    return TRUE;
  }

  memset (&breq, 0, sizeof (struct v4l2_requestbuffers));

  GST_DEBUG_OBJECT (subdevsrc, "STREAMING, requesting %d CAPTURE buffers",
                    subdevsrc->num_buffers);

  breq.count = subdevsrc->num_buffers;
  breq.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  breq.memory = V4L2_MEMORY_USERPTR;

  g_mutex_lock (subdevsrc->device_mutex);

  if (ioctl (fd, VIDIOC_REQBUFS, &breq) < 0)
    goto reqbufs_failed;

  GST_LOG_OBJECT (subdevsrc, " count:  %u", breq.count);
  GST_LOG_OBJECT (subdevsrc, " type:   %d", breq.type);
  GST_LOG_OBJECT (subdevsrc, " memory: %d", breq.memory);

  if (breq.count < GST_SUBDEVSRC_MIN_BUFFERS)
    goto no_buffers;

  if (subdevsrc->num_buffers != breq.count) {
    GST_WARNING_OBJECT (subdevsrc, "using %u buffers instead", breq.count);
    subdevsrc->num_buffers = breq.count;
    g_object_notify (G_OBJECT (subdevsrc), "queue-size");
  }

  /* Map the buffers */
  GST_LOG_OBJECT (subdevsrc, "initiating buffer pool");

  if (!(subdevsrc->pool =
        gst_subdevsrc_buffer_pool_new (GST_ELEMENT_CAST (subdevsrc), fd, caps)))
    goto buffer_pool_new_failed;

  GST_INFO_OBJECT (subdevsrc, "capturing buffers");

  subdevsrc->offset = 0;

  if (!gst_subdevsrc_buffer_pool_activate (subdevsrc->pool,
      GST_ELEMENT_CAST (subdevsrc)))
  {
    goto pool_activate_failed;
  }

  if (ioctl (fd, VIDIOC_STREAMON, &type) < 0)
    goto streamon_failed;

  GST_DEBUG_OBJECT (subdevsrc, "STREAMON called");

  ret = TRUE;

done:
  g_mutex_unlock (subdevsrc->device_mutex);

  return ret;

  /* ERRORS */
reqbufs_failed:
  {
    GST_ELEMENT_ERROR (subdevsrc, RESOURCE, READ,
        ("Could not get buffers from output device"),
        ("error requesting %d buffers: %s",
         subdevsrc->num_buffers, g_strerror (errno)));
    ret = FALSE;
    goto done;
  }
no_buffers:
  {
    GST_ELEMENT_ERROR (subdevsrc, RESOURCE, READ,
        ("Could not get enough buffers from video output device."),
        ("we received %d buffers, we want at least %d",
            breq.count, GST_SUBDEVSRC_MIN_BUFFERS));
    ret = FALSE;
    goto done;
  }
buffer_pool_new_failed:
  {
    GST_ELEMENT_ERROR (subdevsrc, RESOURCE, READ,
        ("Could not map buffers from device '%d'", fd),
        ("Failed to create buffer pool: %s", g_strerror (errno)));
    ret = FALSE;
    goto done;
  }
pool_activate_failed:
  {
    GST_ELEMENT_ERROR (subdevsrc, RESOURCE, READ,
        ("Could not enqueue buffers"),
        ("Activating buffer pool failed: %s", g_strerror (errno)));
    ret = FALSE;
    goto done;
  }
streamon_failed:
  {
    GST_ELEMENT_ERROR (subdevsrc, RESOURCE, OPEN_READ,
        ("Error starting streaming capture from output device: %s",
         g_strerror (errno)), GST_ERROR_SYSTEM);
    ret = FALSE;
    goto done;
  }
}


/**
 * gst_subdevsrc_stop_streaming:
 * @subdevsrc: #GstSubdevSrc object.
 *
 * Stop video streaming.
 *
 * Returns: TRUE on success.
 */
gboolean
gst_subdevsrc_stop_streaming (GstSubdevSrc *subdevsrc)
{
  gint type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  gboolean ret = FALSE;

  GST_DEBUG_OBJECT (subdevsrc, "Stop streaming");

  g_mutex_lock (subdevsrc->device_mutex);

  /* we actually need to sync on all queued buffers but not
   * on the non-queued ones */
  if (ioctl (subdevsrc->video_fd, VIDIOC_STREAMOFF, &type) < 0)
    goto streamoff_failed;

  GST_DEBUG_OBJECT (subdevsrc, "STREAMOFF called");

  if (subdevsrc->pool) {
    gst_subdevsrc_buffer_pool_destroy (subdevsrc->pool,
        GST_ELEMENT_CAST (subdevsrc));
    subdevsrc->pool = NULL;
  }
  ret = TRUE;

done:
  g_mutex_unlock (subdevsrc->device_mutex);

  return ret;

  /* ERRORS */
streamoff_failed:
  {
    GST_ELEMENT_ERROR (subdevsrc, RESOURCE, CLOSE,
        ("Error stopping streaming capture from output device"),
        GST_ERROR_SYSTEM);
    goto done;
  }
}


/**
 * gst_subdevsrc_probe_caps:
 * @subdevsrc: #GstSubdevSrc object.
 * @night_mode: TRUE if the caps should be generated for night mode only.
 *
 * Create a GstCaps object which describes supported output formats.
 *
 * Returns: #GstCaps containing supported output caps.
 */
GstCaps *
gst_subdevsrc_probe_caps (GstSubdevSrc *subdevsrc, gboolean night_mode)
{
  GstStructure *structure = NULL;
  GstCaps *ret = gst_caps_new_empty ();
  guint32 fcc = 0;
  gboolean add;
  GSList *walk;
  GSList *iter;

  GST_DEBUG_OBJECT (subdevsrc, "Probing supported caps");

  if (!subdevsrc->capture_formats) {
    GST_WARNING_OBJECT (subdevsrc, "Formats list is empty");
    goto done;
  }

  for (walk = subdevsrc->capture_formats; walk; walk = walk->next) {
    struct v4l2_subdev_mbus_code_enum *format;

    format = (struct v4l2_subdev_mbus_code_enum *) walk->data;

    GST_DEBUG_OBJECT (subdevsrc, "handling 0x%08x", format->code);

    switch (format->code) {
      case V4L2_MBUS_FMT_YUYV8_1X16:
        fcc = GST_MAKE_FOURCC ('Y', 'U', 'Y', '2');
        break;
      case V4L2_MBUS_FMT_UYVY8_1X16:
        fcc = GST_MAKE_FOURCC ('U', 'Y', 'V', 'Y');
        break;
      case V4L2_MBUS_FMT_SGRBG10_1X10:
      case V4L2_MBUS_FMT_SGRBG10_DPCM8_1X8:
        GST_DEBUG_OBJECT (subdevsrc, "Ignoring bayer formats");
        break;
      default:
        GST_DEBUG_OBJECT (subdevsrc, "Unknown fourcc 0x%08x", format->code);
        break;
    }

    if (fcc != 0) {
      guint min_w, min_h, max_w, max_h;

      min_w = min_h = G_MAXUINT;
      max_w = max_h = 0;
      GST_DEBUG_OBJECT (subdevsrc, "FCC found: %" GST_FOURCC_FORMAT,
          GST_FOURCC_ARGS (fcc));

      for (iter = subdevsrc->sensor_fmts; iter; iter = iter->next) {
        GstVideoFrameInfo *finfo;

        finfo = (GstVideoFrameInfo *) iter->data;

        min_w = finfo->width / RESIZER_MAX_DOWNSCALE_FACTOR;
        min_h = finfo->height / RESIZER_MAX_DOWNSCALE_FACTOR;
        max_w = finfo->width;
        max_h = finfo->height;

        GST_DEBUG_OBJECT (subdevsrc, "min: %ux%u, max: %ux%u", min_w, min_h,
            max_w, max_h);

        add = FALSE;

        structure = gst_structure_new ("video/x-raw-yuv",
            "format", GST_TYPE_FOURCC, fcc,
            "width", GST_TYPE_INT_RANGE, min_w, max_w,
            "height", GST_TYPE_INT_RANGE, min_h,max_h,
            NULL);

        if (G_IS_VALUE (finfo->framerates)) {
          if (!night_mode) {
            gst_structure_set_value (structure, "framerate", finfo->framerates);
            add = TRUE;
          }
          else {
            add = gst_subdevsrc_add_night_mode_fps (structure, finfo->framerates);
          }
        }

        if (add) {
          gst_caps_append_structure (ret, structure);
        }
        else {
          gst_structure_free (structure);
        }
      }
    }
    else {
      GST_DEBUG_OBJECT (subdevsrc, "FCC not found");
    }
  }

done:

  return ret;
}


/**
 * gst_subdevsrc_grab_vf_frame:
 * @subdevsrc: #GstSubdevSrc object.
 * @buf: Pointer to GstBuffer object where the grabbed frame should be placed.
 *
 * Capture one single viewfinder frame from camera device.
 *
 * Returns: #GstFlowReturn describing the status of the capture operation.
 */
GstFlowReturn
gst_subdevsrc_grab_vf_frame (GstSubdevSrc *subdevsrc, GstBuffer **buf)
{
  #define NUM_TRIALS 50

  struct v4l2_buffer buffer;
  gint32 trials = NUM_TRIALS;
  GstBuffer *pool_buffer;
  gboolean need_copy = FALSE;
  gint index, ret, count = 0;

  GST_LOG_OBJECT (subdevsrc, "Grabbing frame");

again:

  /* queue buffers */
  if (!gst_subdevsrc_buffer_pool_update (subdevsrc->pool)) {
    goto stopped;
  }

  memset (&buffer, 0x00, sizeof (buffer));
  buffer.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  buffer.memory = V4L2_MEMORY_USERPTR;

  for (;;) {
    ret = gst_poll_wait (subdevsrc->poll, SUBDEVSRC_POLL_TIMEOUT);
    if (G_UNLIKELY (ret < 0)) {
      if (errno == EBUSY)
        goto stopped;
      if (errno != EAGAIN && errno != EINTR)
        goto select_error;
    }

    if (G_UNLIKELY (ret == 0)) {
      goto timeout;
    }
    g_mutex_lock (subdevsrc->device_mutex);
    ret = ioctl (subdevsrc->video_fd, VIDIOC_DQBUF, &buffer);
    g_mutex_unlock (subdevsrc->device_mutex);

    if (ret >= 0)
      break;

    GST_WARNING_OBJECT (subdevsrc,
        "problem grabbing frame %d (ix=%d), trials=%d, pool-ct=%d, buf.flags=%d",
        buffer.sequence, buffer.index, trials,
        GST_MINI_OBJECT_REFCOUNT (subdevsrc->pool), buffer.flags);

    /* if the sync() got interrupted, we can retry */
    switch (errno) {
      case EAGAIN:
        GST_WARNING_OBJECT (subdevsrc,
            "Non-blocking I/O has been selected using O_NONBLOCK and"
            " no buffer was in the outgoing queue.");
        break;
      case EINVAL:
        goto einval;
      case ENOMEM:
        goto enomem;
      case EIO:
        GST_INFO_OBJECT (subdevsrc,
            "VIDIOC_DQBUF failed due to an internal error."
            " Can also indicate temporary problems like signal loss."
            " Note the driver might dequeue an (empty) buffer despite"
            " returning an error, or even stop capturing.");
        /* have we de-queued a buffer ? */
        if (!(buffer.flags & (V4L2_BUF_FLAG_QUEUED | V4L2_BUF_FLAG_DONE))) {
          /* this fails
             if ((buffer.index >= 0) && (buffer.index < v4l2camsrc->breq.count)) {
             GST_DEBUG_OBJECT (v4l2camsrc, "reenqueing buffer (ix=%ld)", buffer.index);
             gst_v4l2camsrc_queue_frame (v4l2camsrc, buffer.index);
             }
             else {
           */
          GST_DEBUG_OBJECT (subdevsrc, "reenqueing buffer");
        }
        break;
      case EINTR:
        GST_WARNING_OBJECT (subdevsrc, "could not sync on a buffer");
        break;
      default:
        GST_WARNING_OBJECT (subdevsrc,
            "Grabbing frame got interrupted. No expected reason.");
        break;
    }

    /* check nr. of attempts to capture */
    if (--trials == -1) {
      goto too_many_trials;
    } else {
      memset (&buffer, 0x00, sizeof (buffer));
      buffer.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
      buffer.memory = V4L2_MEMORY_USERPTR;
    }
  }

  if (subdevsrc->frame_byte_size > 0) {
    /* if size does not match what we expected, try again */
    if (buffer.bytesused != subdevsrc->frame_byte_size) {
      GST_ELEMENT_WARNING (subdevsrc, RESOURCE, READ,
          ("Got unexpected frame size of %u instead of %u.",
           buffer.bytesused, subdevsrc->frame_byte_size), (NULL));
      if (count++ > 50)
        goto size_error;

      goto again;
    }
  }

  g_mutex_lock (subdevsrc->pool->lock);

  index = buffer.index;

  /* get our GstBuffer with that index from the pool, if the buffer is
   * in use (NULL) we have a serious problem (should only be possible in mmap
   * case). */
  pool_buffer = GST_BUFFER (subdevsrc->pool->buffers[index]);

  if (G_UNLIKELY (pool_buffer == NULL))
    goto no_buffer;

  GST_LOG_OBJECT (subdevsrc, "grabbed buffer %p at index %d (refct = %d)",
    pool_buffer, index, GST_MINI_OBJECT_REFCOUNT (pool_buffer));

  /* ref the buffer and requeue, when if becomes writable again */
  subdevsrc->pool->buffers[index] = NULL;
  subdevsrc->pool->num_live_buffers++;
  subdevsrc->pool->queued[index] = 0;

  g_mutex_unlock (subdevsrc->pool->lock);

  /* this can change at every frame, esp. with jpeg */
  GST_BUFFER_SIZE (pool_buffer) = buffer.bytesused;
  GST_BUFFER_OFFSET (pool_buffer) = subdevsrc->offset++;
  GST_BUFFER_OFFSET_END (pool_buffer) = subdevsrc->offset;
  /* This causes GstCameraSrc to apply a timestamp to the buffer */
  GST_BUFFER_TIMESTAMP (pool_buffer) =  GST_CLOCK_TIME_NONE;

  if (G_UNLIKELY (need_copy)) {
    GST_DEBUG_OBJECT (subdevsrc, "copying buffer");
    *buf = gst_buffer_copy (pool_buffer);
    GST_BUFFER_FLAG_UNSET (*buf, GST_BUFFER_FLAG_READONLY);
    /* this will requeue */
    gst_buffer_unref (pool_buffer);
  } else {
    *buf = pool_buffer;
  }

  GST_LOG_OBJECT (subdevsrc, "grabbed frame %d (ix=%d), flags %08x, pool-ct=%d",
      buffer.sequence, buffer.index, buffer.flags,
                  subdevsrc->pool->num_live_buffers);

  return GST_FLOW_OK;

  /* ERRORS */
size_error:
  {
    GST_ELEMENT_ERROR (subdevsrc, RESOURCE, READ,
        ("Error grabbing frame from device, size mismatch"), (NULL));
    return GST_FLOW_ERROR;
  }
select_error:
  {
    GST_ELEMENT_ERROR (subdevsrc, RESOURCE, READ, (NULL),
        ("select error %d: %s (%d)", ret, g_strerror (errno), errno));
    return GST_FLOW_ERROR;
  }
stopped:
  {
    GST_DEBUG_OBJECT (subdevsrc, "stop called");
    return GST_FLOW_WRONG_STATE;
  }
einval:
  {
    GST_ELEMENT_ERROR (subdevsrc, RESOURCE, FAILED,
        ("Failed trying to get video frames"),
        ("The buffer type is not supported, or the index is out of bounds,"
                " or no buffers have been allocated yet, or the userptr"
                " or length are invalid."));
    return GST_FLOW_ERROR;
  }
enomem:
  {
    GST_ELEMENT_ERROR (subdevsrc, RESOURCE, FAILED,
        ("Failed trying to get video frames. Not enough memory."),
             ("insufficient memory to enqueue a user "
             "pointer buffer."));
    return GST_FLOW_ERROR;
  }
too_many_trials:
  {
    GST_ELEMENT_ERROR (subdevsrc, RESOURCE, FAILED,
        ("Failed trying to get video frames"),
        ("Failed after %d tries. system error: %s",
            NUM_TRIALS, g_strerror (errno)));
    return GST_FLOW_ERROR;
  }
no_buffer:
  {
    GST_ELEMENT_ERROR (subdevsrc, RESOURCE, FAILED,
        ("Failed trying to get video frames"),
        ("No free buffers found in the pool at index %d.", index));
    g_mutex_unlock (subdevsrc->pool->lock);
    return GST_FLOW_ERROR;  }
timeout:
  {
    GST_ELEMENT_ERROR (subdevsrc, RESOURCE, FAILED,
        ("Timeout when trying to get video frames from device"), NULL);
    return GST_FLOW_ERROR;
  }
}


/**
 * gst_subdevsrc_update_cropping:
 * @subdevsrc: #GstSubdevSrc object.
 * @width: Output width.
 * @height: Output height.
 * @zoom: Desired zoom factor.
 *
 * Set up cropping according to given output resolution and zoom factor.
 *
 * Returns: TRUE on success.
 */
gboolean
gst_subdevsrc_update_cropping (GstSubdevSrc *subdevsrc, guint width,
    guint height, gfloat zoom)
{
  gboolean ret;

  g_mutex_lock (subdevsrc->device_mutex);
  ret = gst_subdevsrc_update_cropping_unlocked (subdevsrc, width, height, zoom);
  g_mutex_unlock (subdevsrc->device_mutex);

  return ret;
}
