/* GStreamer V4L2 camera source
 * Copyright (C) 2010 Nokia Corporation <multimedia@maemo.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-v4l2newcamsrc
 * @short_description: V4L2 camera source
 * @see_also: #GstCameraSrc
 *
 * <refsect2>
 * <para>
 * Bla bla...
 * </para>
 * <para>
 * Foo bar
 * </para>
 * <title>Example launch line</title>
 * <para>
 * <programlisting>
 * gst-launch v4l2camsrc ! xvimagesink
 * </programlisting>
 * </para>
 * </refsect2>
 */

#include <string.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "gstv4l2camsrc.h"
#include "v4l2camsrc_calls.h"
#include "gstv4l2camvidorient.h"


GST_DEBUG_CATEGORY (gst_v4l2camsrc_debug);
#define GST_CAT_DEFAULT gst_v4l2camsrc_debug

static gboolean gst_v4l2camsrc_is_open (GstCameraSrc *camsrc);
static void gst_v4l2camsrc_finalize (GObject * object);
static void gst_v4l2camsrc_dispose (GObject * object);
static void gst_v4l2camsrc_set_property (GObject * object,
    guint prop_id, const GValue * value, GParamSpec * pspec);
static void gst_v4l2camsrc_get_property (GObject * object,
    guint prop_id, GValue * value, GParamSpec * pspec);
static gboolean gst_v4l2camsrc_unlock (GstCameraSrc * src);
static gboolean gst_v4l2camsrc_unlock_stop (GstCameraSrc * src);
static gboolean gst_v4l2camsrc_iface_supported (GstImplementsInterface * iface,
    GType iface_type);

typedef enum {
  PROP_0,
  PROP_DEVICE,
  PROP_DEVICE_FD,
  PROP_DEVICE_NAME,
  PROP_QUEUE_SIZE,
} GstV4L2CamSrcProperties;


static GstStaticPadTemplate gst_v4l2camsrc_src_template =
    GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("video/x-raw-yuv, "
        "format = (fourcc) UYVY, "
        "framerate = (fraction) [ 0, 200 ], "
        "width = (int) [ 1, MAX ], "
        "height = (int) [ 1, MAX ];"
        "video/x-raw-yuv, "
        "format = (fourcc) I420, "
        "framerate = (fraction) [ 0, 200 ], "
        "width = (int) [ 1, MAX ], "
        "height = (int) [ 1, MAX ];"
        "video/x-raw-yuv, "
        "format = (fourcc) YUY2, "
        "framerate = (fraction) [ 0, 200 ], "
        "width = (int) [ 1, MAX ], "
        "height = (int) [ 1, MAX ]"));


#define DEFAULT_PROP_DEVICE_NAME  NULL
#define DEFAULT_PROP_DEVICE       "/dev/video0"
#define DEFAULT_PROP_DEVICE_FD    -1

GST_IMPLEMENT_V4L2CAMSRC_VIDORIENT_METHODS (GstV4l2NewCamSrc, gst_v4l2camsrc);


static void
gst_v4l2camsrc_interface_init (GstImplementsInterfaceClass * klass)
{
  /*
  * default virtual functions
  */
  klass->supported = gst_v4l2camsrc_iface_supported;
}

void
gst_v4l2camsrc_init_interfaces (GType type)
{
  static const GInterfaceInfo v4l2camsrc_iface_info = {
    (GInterfaceInitFunc) gst_v4l2camsrc_interface_init,
    NULL,
    NULL,
  };
  static const GInterfaceInfo v4l2camsrc_videoorientation_info = {
    (GInterfaceInitFunc) gst_v4l2camsrc_video_orientation_interface_init,
    NULL,
    NULL,
  };

  g_type_add_interface_static (type,
      GST_TYPE_IMPLEMENTS_INTERFACE, &v4l2camsrc_iface_info);
  g_type_add_interface_static (type,
      GST_TYPE_VIDEO_ORIENTATION, &v4l2camsrc_videoorientation_info);
}


GST_BOILERPLATE_FULL (GstV4l2NewCamSrc, gst_v4l2camsrc, GstCameraSrc,
    GST_TYPE_CAMERA_SRC, gst_v4l2camsrc_init_interfaces);



static gboolean
gst_v4l2camsrc_iface_supported (GstImplementsInterface * iface, GType iface_type)
{
  GstCameraSrc *camsrc = GST_CAMERA_SRC (iface);

  if (gst_v4l2camsrc_is_open (camsrc) &&
      iface_type == GST_TYPE_VIDEO_ORIENTATION)
  {
    return TRUE;
  }

  else if (GST_IS_IMPLEMENTS_INTERFACE (camsrc)) {
    GstImplementsInterfaceClass * parent_klass;

    parent_klass =
        g_type_interface_peek (parent_class, GST_TYPE_IMPLEMENTS_INTERFACE);
    return parent_klass->supported (iface, iface_type);
  }

  return FALSE;
}


/*
 * gst_v4l2camsrc_is_open:
 *
 */
static gboolean
gst_v4l2camsrc_is_open (GstCameraSrc *camsrc)
{
  GstV4l2NewCamSrc *v4l2camsrc = GST_V4L2CAMSRC (camsrc);

  return GST_V4L2CAMSRC_IS_OPEN (v4l2camsrc);
}

/*
 * gst_v4l2camsrc_is_active:
 *
 */
static gboolean
gst_v4l2camsrc_is_active (GstCameraSrc *camsrc)
{
  GstV4l2NewCamSrc *v4l2camsrc = GST_V4L2CAMSRC (camsrc);

  return GST_V4L2CAMSRC_IS_ACTIVE (v4l2camsrc);
}

/*
 * gst_v4l2camsrc_v4l2fourcc_to_structure:
 *
 */
static GstStructure *
gst_v4l2camsrc_v4l2fourcc_to_structure (guint32 fourcc)
{
  GstStructure *structure = NULL;
  guint32 fcc = 0;

  switch (fourcc) {
    case V4L2_PIX_FMT_YUYV:
      fcc = GST_MAKE_FOURCC ('Y', 'U', 'Y', '2');
      break;
    case V4L2_PIX_FMT_UYVY:
      fcc = GST_MAKE_FOURCC ('U', 'Y', 'V', 'Y');
      break;
#ifdef V4L2_PIX_FMT_SGRBG10
    case V4L2_PIX_FMT_SGRBG10:
      GST_DEBUG ("Ignoring bayer formats");
      break;
#endif
    case V4L2_PIX_FMT_YUV420:
      fcc = GST_MAKE_FOURCC ('I', '4', '2', '0');
      break;
    default:
      GST_DEBUG ("Unknown fourcc 0x%08x %" GST_FOURCC_FORMAT,
                 fourcc, GST_FOURCC_ARGS (fourcc));
      break;
  }

  if (fcc != 0)
    structure = gst_structure_new ("video/x-raw-yuv",
        "format", GST_TYPE_FOURCC, fcc, NULL);

  return structure;
}

/*
 * gst_v4l2camsrc_get_caps:
 *
 */
static GstCaps *
gst_v4l2camsrc_get_caps (GstCameraSrc *camsrc)
{
  GstV4l2NewCamSrc *v4l2camsrc = GST_V4L2CAMSRC (camsrc);
  GstStructure *template;
  GstCaps *ret;
  GSList *walk;

  if (!gst_v4l2camsrc_is_open (camsrc)) {
    /* FIXME: should it probe the device? */
    return NULL;
  }

  if (!v4l2camsrc->formats)
    gst_v4l2camsrc_fill_format_list (v4l2camsrc);

  ret = gst_caps_new_empty ();

  for (walk = v4l2camsrc->formats; walk; walk = walk->next) {
    struct v4l2_fmtdesc *format;

    /* FIXME: Introduce own format structure */
    format = (struct v4l2_fmtdesc *) walk->data;

    template = gst_v4l2camsrc_v4l2fourcc_to_structure (format->pixelformat);

    if (template) {
      GstCaps *tmp;

      tmp = gst_v4l2camsrc_probe_caps_for_format (v4l2camsrc,
                format->pixelformat, template);
      if (tmp)
        gst_caps_append (ret, tmp);

      gst_structure_free (template);
    } else {
      GST_DEBUG_OBJECT (v4l2camsrc, "unknown format %u", format->pixelformat);
    }
  }

  v4l2camsrc->probed_caps = gst_caps_ref (ret);

  GST_INFO_OBJECT (v4l2camsrc, "probed caps: %" GST_PTR_FORMAT, ret);

  return ret;
}

/*
 * gst_v4l2camsrc_get_num_buffers:
 *
 */
static guint
gst_v4l2camsrc_get_num_buffers (GstCameraSrc *camsrc)
{
  GstV4l2NewCamSrc *v4l2camsrc = GST_V4L2CAMSRC (camsrc);

  return v4l2camsrc->num_buffers;
}


/*
 * gst_v4l2camsrc_base_init:
 * @klass: #GstElementClass.
 *
 */
static void
gst_v4l2camsrc_base_init (gpointer klass)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (klass);

  gst_element_class_set_details_simple (element_class,
      "V4L2 camera source",
      "Video/Src",
      "Video4Linux2 camera source element",
      "Maemo Multimedia <multimedia@maemo.org>");

  gst_element_class_add_pad_template (element_class,
       gst_static_pad_template_get (&gst_v4l2camsrc_src_template));
}

/*
 * gst_v4l2camsrc_class_init:
 * @klass: #GstV4l2NewCamSrcClass.
 *
 */
static void
gst_v4l2camsrc_class_init (GstV4l2NewCamSrcClass * klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  GstCameraSrcClass *camera_class = GST_CAMERA_SRC_CLASS (klass);

  gobject_class->set_property = GST_DEBUG_FUNCPTR (gst_v4l2camsrc_set_property);
  gobject_class->get_property = GST_DEBUG_FUNCPTR (gst_v4l2camsrc_get_property);
  gobject_class->dispose = GST_DEBUG_FUNCPTR (gst_v4l2camsrc_dispose);
  gobject_class->finalize = GST_DEBUG_FUNCPTR (gst_v4l2camsrc_finalize);

  g_object_class_install_property (gobject_class, PROP_DEVICE,
      g_param_spec_string ("device", "Device", "Device location",
          DEFAULT_PROP_DEVICE,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

  g_object_class_install_property (gobject_class, PROP_DEVICE_NAME,
      g_param_spec_string ("device-name", "Device name",
          "Name of the device", DEFAULT_PROP_DEVICE_NAME,
          G_PARAM_READABLE | G_PARAM_STATIC_STRINGS));

  g_object_class_install_property (gobject_class, PROP_DEVICE_FD,
      g_param_spec_int ("device-fd", "File descriptor",
          "File descriptor of the device", -1, G_MAXINT, DEFAULT_PROP_DEVICE_FD,
          G_PARAM_READABLE | G_PARAM_STATIC_STRINGS));

  g_object_class_install_property (gobject_class, PROP_QUEUE_SIZE,
      g_param_spec_uint ("queue-size", "Queue size",
          "Number of buffers to be enqueud in the driver",
          GST_V4L2CAMSRC_MIN_BUFFERS, GST_V4L2CAMSRC_MAX_BUFFERS,
          GST_V4L2CAMSRC_DEFAULT_BUFFERS,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

  camera_class->is_open = GST_DEBUG_FUNCPTR (gst_v4l2camsrc_is_open);
  camera_class->open = GST_DEBUG_FUNCPTR (gst_v4l2camsrc_open);
  camera_class->close = GST_DEBUG_FUNCPTR (gst_v4l2camsrc_close);
  camera_class->get_attribute = GST_DEBUG_FUNCPTR (gst_v4l2camsrc_get_attribute);
  camera_class->set_attribute = GST_DEBUG_FUNCPTR (gst_v4l2camsrc_set_attribute);
  camera_class->set_capture = GST_DEBUG_FUNCPTR (gst_v4l2camsrc_set_capture);
  camera_class->start = GST_DEBUG_FUNCPTR (gst_v4l2camsrc_capture_start);
  camera_class->is_active = GST_DEBUG_FUNCPTR (gst_v4l2camsrc_is_active);
  camera_class->grab_frame = GST_DEBUG_FUNCPTR (gst_v4l2camsrc_grab_frame);
  camera_class->stop = GST_DEBUG_FUNCPTR (gst_v4l2camsrc_capture_stop);
  camera_class->get_caps = GST_DEBUG_FUNCPTR (gst_v4l2camsrc_get_caps);
  camera_class->get_num_buffers = GST_DEBUG_FUNCPTR (gst_v4l2camsrc_get_num_buffers);
  camera_class->unlock = GST_DEBUG_FUNCPTR (gst_v4l2camsrc_unlock);
  camera_class->unlock_stop = GST_DEBUG_FUNCPTR (gst_v4l2camsrc_unlock_stop);
}


/*
 * gst_v4l2camsrc_init:
 * @v4l2camsrc: #GstV4l2NewCamSrc.
 * @klass: #GstV4l2NewCamSrcClass.
 *
 */
static void
gst_v4l2camsrc_init (GstV4l2NewCamSrc * v4l2camsrc, GstV4l2NewCamSrcClass * klass)
{
  v4l2camsrc->num_buffers = GST_V4L2CAMSRC_DEFAULT_BUFFERS;
  v4l2camsrc->tmp_num_buffers = v4l2camsrc->num_buffers;
  v4l2camsrc->videodev = g_strdup (DEFAULT_PROP_DEVICE);
  v4l2camsrc->video_fd = DEFAULT_PROP_DEVICE_FD;
  v4l2camsrc->poll = gst_poll_new (TRUE);
  v4l2camsrc->buffer = NULL;
  v4l2camsrc->crop_supported = FALSE;
  v4l2camsrc->max_zoom_factor = 1.0;

  v4l2camsrc->device_mutex = g_mutex_new ();

  GST_DEBUG ("initialized");
}


static void
gst_v4l2camsrc_dispose (GObject * object)
{
  GstV4l2NewCamSrc *v4l2camsrc = GST_V4L2CAMSRC (object);

  if (v4l2camsrc->formats) {
    gst_v4l2camsrc_clear_format_list (v4l2camsrc);
  }

  if (v4l2camsrc->probed_caps) {
    gst_caps_unref (v4l2camsrc->probed_caps);
    v4l2camsrc->probed_caps = NULL;
  }

  /* FIXME: driver cleanup function */
  if (v4l2camsrc->videodev) {
    g_free (v4l2camsrc->videodev);
    v4l2camsrc->videodev = NULL;
  }

  if (v4l2camsrc->poll) {
    gst_poll_free (v4l2camsrc->poll);
  }

  g_mutex_free (v4l2camsrc->device_mutex);

  G_OBJECT_CLASS (parent_class)->dispose (object);
}


/*
 * gst_v4l2camsrc_finalize:
 * @object:
 *
 */
static void
gst_v4l2camsrc_finalize (GObject * object)
{
  G_OBJECT_CLASS (parent_class)->finalize (object);
}




















/*
 */
static void
gst_v4l2camsrc_set_property (GObject * object,
    guint prop_id, const GValue * value, GParamSpec * pspec)
{
  GstV4l2NewCamSrc *v4l2camsrc = GST_V4L2CAMSRC (object);

  switch (prop_id) {
    case PROP_DEVICE:
      g_free (v4l2camsrc->videodev);
      v4l2camsrc->videodev = g_value_dup_string (value);
      break;
    case PROP_QUEUE_SIZE:
      v4l2camsrc->num_buffers = g_value_get_uint (value);
      v4l2camsrc->tmp_num_buffers = v4l2camsrc->num_buffers;
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

/*
 */
static void
gst_v4l2camsrc_get_property (GObject * object,
    guint prop_id, GValue * value, GParamSpec * pspec)
{
  GstV4l2NewCamSrc *v4l2camsrc = GST_V4L2CAMSRC (object);

  switch (prop_id) {
    case PROP_DEVICE:
      g_value_set_string (value, v4l2camsrc->videodev);
      break;
    case PROP_DEVICE_NAME:
    {
      const guchar *new = NULL;

      if (gst_v4l2camsrc_is_open (GST_CAMERA_SRC (v4l2camsrc))) {
        new = v4l2camsrc->vcap.card;
      } else if (gst_v4l2camsrc_open (GST_CAMERA_SRC (v4l2camsrc))) {
        new = v4l2camsrc->vcap.card;
        gst_v4l2camsrc_close (GST_CAMERA_SRC (v4l2camsrc));
        gst_camerasrc_clear_color_channels (GST_CAMERA_SRC (v4l2camsrc));
      }
      g_value_set_string (value, (gchar *) new);
      break;
    }
    case PROP_DEVICE_FD:
    {
      if (gst_v4l2camsrc_is_open (GST_CAMERA_SRC (v4l2camsrc)))
        g_value_set_int (value, v4l2camsrc->video_fd);
      else
        g_value_set_int (value, DEFAULT_PROP_DEVICE_FD);
      break;
    }
    case PROP_QUEUE_SIZE:
      g_value_set_uint (value, v4l2camsrc->num_buffers);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}



static gboolean
gst_v4l2camsrc_unlock (GstCameraSrc * src)
{
  GstV4l2NewCamSrc *v4l2camsrc = GST_V4L2CAMSRC (src);

  GST_LOG_OBJECT (v4l2camsrc, "Flushing");
  gst_poll_set_flushing (v4l2camsrc->poll, TRUE);

  return TRUE;
}

static gboolean
gst_v4l2camsrc_unlock_stop (GstCameraSrc * src)
{
  GstV4l2NewCamSrc *v4l2camsrc = GST_V4L2CAMSRC (src);

  GST_LOG_OBJECT (v4l2camsrc, "No longer flushing");
  gst_poll_set_flushing (v4l2camsrc->poll, FALSE);

  return TRUE;
}









/*
 * plugin_init:
 * @plugin: GstPlugin
 *
 * Returns: TRUE on success.
 */
static gboolean
plugin_init (GstPlugin * plugin)
{
  GST_DEBUG_CATEGORY_INIT (gst_v4l2camsrc_debug, "v4l2newcamsrc", 0, \
      "V4L2 camera source");

  return gst_element_register (plugin, "v4l2newcamsrc",
      GST_RANK_NONE, GST_TYPE_V4L2CAMSRC);
}

GST_PLUGIN_DEFINE (GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "v4l2newcam",
    "V4L2 camera image capturing element",
    plugin_init,
    VERSION,
    "LGPL",
    "Nokia", "www.nokia.com")
