/* GStreamer V4L2 camera source
 * Copyright (C) 2010 Nokia Corporation <multimedia@maemo.org>
 *
 * Contact: Maemo Multimedia <multimedia@maemo.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef __GST_V4L2CAMSRC_H__
#define __GST_V4L2CAMSRC_H__

#include <linux/videodev2.h>

#include <gst/gst.h>
#include <gst/camera/gstcamerasrc.h>

G_BEGIN_DECLS

#define GST_TYPE_V4L2CAMSRC \
(gst_v4l2camsrc_get_type())
#define GST_V4L2CAMSRC(obj) \
(G_TYPE_CHECK_INSTANCE_CAST((obj), GST_TYPE_V4L2CAMSRC, GstV4l2NewCamSrc))
#define GST_V4L2CAMSRC_CLASS(klass) \
(G_TYPE_CHECK_CLASS_CAST((klass), GST_TYPE_V4L2CAMSRC, GstV4l2NewCamSrcClass))
#define GST_IS_V4L2CAMSRC(obj) \
(G_TYPE_CHECK_INSTANCE_TYPE((obj), GST_TYPE_V4L2CAMSRC))
#define GST_IS_V4L2CAMSRC_CLASS(klass) \
(G_TYPE_CHECK_CLASS_TYPE((klass), GST_TYPE_V4L2CAMSRC))


typedef struct _GstV4l2NewCamSrc GstV4l2NewCamSrc;
typedef struct _GstV4l2NewCamSrcClass GstV4l2NewCamSrcClass;

typedef struct _GstV4l2Buffer GstV4l2Buffer;
typedef struct _GstV4l2NewCamSrcBufferPool GstV4l2NewCamSrcBufferPool;

/**
 * GstV4l2Buffer:
 * @buffer: parent GstBuffer
 * @pool: reference to the owning #GstV4l2NewCamSrcBufferPool
 * @gbuffer: GstBuffer allocated for this buffer for holding the data
 * @vbuffer: v4l2_buffer structure
 *
 * Opaque object.
*/

struct _GstV4l2Buffer {
  GstBuffer   buffer;
  GstV4l2NewCamSrcBufferPool *pool;
  GstBuffer *gbuffer;
  void *vbuffer;
};


/**
 * GstV4l2NewCamSrcBufferPool:
 * @parent: the parent object
 * @lock: pool lock
 * @running: TRUE if the pool is being used
 * @num_live_buffers: number of buffers being pushed outside element
 * @buffer_count: total number of buffers used
 * @buffers: buffer table
 * @video_fd: video device which produces the buffers
 * @queued: lookup table for buffers whether they're queued to driver
 * @data_cond: condition being used for waiting a buffer to become available
 *
 * Opaque #GstV4l2NewCamSrcBufferPool object.
 */
struct _GstV4l2NewCamSrcBufferPool
{
  GstMiniObject parent;

  GMutex *lock;
  gboolean running;          /* with lock */
  gint num_live_buffers;     /* number of buffers not with driver */
  guint buffer_count;
  GstV4l2Buffer **buffers;

  gint video_fd;           /* a dup(2) of the v4l2object's video_fd */
  guint *queued;
  GCond* data_cond;
};

/**
* GstV4l2NewCamSrc:
* @element: the parent element.
*
* The opaque #GstV4l2NewCamSrc data structure.
*/
struct _GstV4l2NewCamSrc {
  GstCameraSrc element;

  /*< private >*/
  char *videodev;               /* video device file name */
  gint video_fd;                /* the video-device's file descriptor */
  GstV4l2NewCamSrcBufferPool *pool;
  guint32 num_buffers;
  guint32 tmp_num_buffers;
  guint64 offset;
  GstPoll * poll;
  guint8 **buffer;              /* the video buffer (mmap()'ed) */
  GSList *formats;              /* list of available capture formats */
  GstCaps *probed_caps;
  guint32 frame_byte_size;
  GMutex *device_mutex;

  guint vf_w, vf_h;
  guint vf_fps_n, vf_fps_d;
  guint32 vf_fourcc;

  guint capture_w, capture_h;
  guint capture_fps_n, capture_fps_d;
  guint32 capture_fourcc;

  struct v4l2_capability vcap;  /* the video device's capabilities */
  struct v4l2_cropcap vcrop;    /* cropping & scaling capabilities */
  gboolean crop_supported;
  gfloat max_zoom_factor;
};


/**
* GstV4l2NewCamSrcClass:
* @parent_class: Element parent class.
*
* The opaque GstV4l2NewCamSrcClass data structure.
*/
struct _GstV4l2NewCamSrcClass {
  GstCameraSrcClass parent_class;
};

GType gst_v4l2camsrc_get_type (void);

G_END_DECLS

#endif /* __GST_V4L2CAMSRC_H__ */