/* GStreamer
 *
 * Copyright (C) 2010 Nokia Corporation <multimedia@maemo.org>
 *
 * subdevsrc.c: Photo interface API test for subdevsrc
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <gst/gst.h>
#include <gst/interfaces/photography.h>
#include <gst/camera/gstcamerasrc.h>
#include <gst/check/gstcheck.h>

/*
* Test all getters and setters over Photography API
* Test only one (valid) value in setting values oevr the api.
*/


GstElement * initialize_photo_if ( )
{
  GstElement *pipeline;

  GstStateChangeReturn state_ret;
  GstElement *source, *cf, *sink;
  GstCaps *caps;
  GstPad *sourcepad;

  gint width = 640;
  gint height = 480;

  gboolean ret;

  gst_init (NULL, NULL);

  pipeline = gst_pipeline_new ("pipe");

  source = gst_element_factory_make ("subdevsrc", "src");
  if (source == NULL)
  {
    g_print ("source fail\n");
    return NULL;
  }

  cf = gst_element_factory_make ("capsfilter", "filter");
  if (cf == NULL)
  {
    g_print ("cf fail\n");
    return NULL;
  }

  sink = gst_element_factory_make ("fakesink", "sink");
  if (sink == NULL)
  {
    g_print ("sink fail\n");
    return NULL;
  }

  sourcepad = gst_element_get_static_pad (source, "src");
  if (sourcepad == NULL)
  {
    g_print ("sourcepad fail\n");
    return NULL;
  }

  /*  Not using real driver implementation, but dummy driver. Using real
  *   is not relevant when considering camsrc element testing.

  g_object_set (G_OBJECT (source), "camera-device", "primary", NULL);
  */

  caps = gst_caps_new_simple ("video/x-raw-yuv",
        "format", GST_TYPE_FOURCC, GST_MAKE_FOURCC ('U', 'Y', 'V', 'Y'),
        "width", G_TYPE_INT, width,
        "height", G_TYPE_INT, height,
        NULL);

  g_object_set (G_OBJECT (cf), "caps", caps, NULL);
  GST_INFO ("Using caps: %" GST_PTR_FORMAT, caps);

  gst_bin_add_many (GST_BIN (pipeline), source, cf, sink, NULL);

  ret =  gst_element_link_many (source, cf, sink, NULL);
  if (!ret)
  {
    g_print ("link fail\n");
    return NULL;
  }

  state_ret = gst_element_set_state (pipeline, GST_STATE_PLAYING);
  g_print ("Pipeline set to PLAYING: %d\n", state_ret);

  return pipeline;
}

GST_START_TEST (test_device_photography_interface)
{

  GstElement *pipeline = NULL;
  GstElement *source = NULL;

  gboolean ret = FALSE;
  GstStateChangeReturn state_ret;

  GstPhotography *photo_if = NULL;

  pipeline = initialize_photo_if ();

  source = gst_bin_get_by_name (GST_BIN(pipeline), "src");
  photo_if = GST_PHOTOGRAPHY (source);

  /* ev_compensation range is from -2.5 to 2.5 */
  /* test both cases */
  gfloat evcomp = 2.5;
  ret = gst_photography_set_ev_compensation (photo_if, evcomp);
  g_print ("Set EV compensation : \t%s\n", ret ? "Success" : "Fail");
  ret = gst_photography_get_ev_compensation (photo_if, &evcomp);
  g_print ("Get EV compensation : \t%s\n", ret ? "Success" : "Fail");
  fail_unless_equals_float(evcomp, 2.5);

  /* value out of range, call should return false */
  evcomp = 5.0;
  ret = gst_photography_set_ev_compensation (photo_if, evcomp);
  g_print ("Set invalid EV compensation : \t%s\n", ret ? "Success" : "Fail");
  fail_unless (ret == FALSE);

  /* There are only 100, 200, 400 and 800 modes. 600 should be rounded to 400 */
  guint iso_speed = 600;
  ret = gst_photography_set_iso_speed (photo_if, iso_speed);
  ret = gst_photography_set_iso_speed (photo_if, iso_speed);
  g_print ("Set ISO speed : \t%s\n", ret ? "Success" : "Fail");
  ret = gst_photography_get_iso_speed (photo_if, &iso_speed);
  g_print ("Get ISO speed : \t%s\n", ret ? "Success" : "Fail");
  fail_unless_equals_int (iso_speed, 400);

  guint aperture = 100;
  ret = gst_photography_set_aperture (photo_if, aperture);
  g_print ("Set aperture : \t\t%s\n", ret ? "Success" : "Fail");
  ret = gst_photography_get_aperture (photo_if, &aperture);
  g_print ("Get aperture : \t\t%s\n", ret ? "Success" : "Fail");
  fail_unless_equals_int (aperture, 100);

  guint exposure = 10;
  ret = gst_photography_set_exposure (photo_if, exposure);
  g_print ("Set exposure : \t\t%s\n", ret ? "Success" : "Fail");
  ret = gst_photography_get_exposure (photo_if, &exposure);
  g_print ("Get exposure : \t\t%s\n", ret ? "Success" : "Fail");
  fail_unless_equals_int (exposure, 10);

  GstWhiteBalanceMode wb_mode = GST_PHOTOGRAPHY_WB_MODE_CLOUDY;
  ret = gst_photography_set_white_balance_mode (photo_if, wb_mode);
  g_print ("Set WB mode : \t\t%s\n", ret ? "Success" : "Fail");
  ret = gst_photography_get_white_balance_mode (photo_if, &wb_mode);
  g_print ("Get WB mode : \t\t%s\n", ret ? "Success" : "Fail");
  fail_unless ( wb_mode == GST_PHOTOGRAPHY_WB_MODE_CLOUDY);

  GstColourToneMode c_tone = GST_PHOTOGRAPHY_COLOUR_TONE_MODE_GRAYSCALE;
  ret = gst_photography_set_colour_tone_mode (photo_if, c_tone);
  g_print ("Set colour tone mode : \t%s\n", ret ? "Success" : "Fail");
  ret = gst_photography_get_colour_tone_mode (photo_if, &c_tone);
  g_print ("Get colour tone mode : \t%s\n", ret ? "Success" : "Fail");
  fail_unless ( c_tone == GST_PHOTOGRAPHY_COLOUR_TONE_MODE_GRAYSCALE);

  GstSceneMode sc_mode = GST_PHOTOGRAPHY_SCENE_MODE_LANDSCAPE;
  ret = gst_photography_set_scene_mode (photo_if, sc_mode);
  g_print ("Set scene mode : \t%s\n", ret ? "Success" : "Fail");
  ret = gst_photography_get_scene_mode (photo_if, &sc_mode);
  g_print ("Get scene mode : \t%s\n", ret ? "Success" : "Fail");
  fail_unless ( sc_mode == GST_PHOTOGRAPHY_SCENE_MODE_LANDSCAPE);

  GstFlashMode fmode = GST_PHOTOGRAPHY_FLASH_MODE_RED_EYE;
  ret = gst_photography_set_flash_mode (photo_if, fmode);
  g_print ("Set flash mode : \t%s\n", ret ? "Success" : "Fail");
  ret = gst_photography_get_flash_mode (photo_if, &fmode);
  g_print ("Get flash mode : \t%s\n", ret ? "Success" : "Fail");
  fail_unless ( fmode == GST_PHOTOGRAPHY_FLASH_MODE_RED_EYE);

  gfloat zoom = 2.0;
  ret = gst_photography_set_zoom (photo_if, zoom);
  g_print ("Set zoom to %f: \t%s\n", zoom, ret ? "Success" : "Fail");

  ret = gst_photography_get_zoom (photo_if, &zoom);
  g_print ("Get zoom : \t\t%s\n", ret ? "Success" : "Fail");
  fail_unless_equals_float (zoom, 2.0);

  GstFlickerReductionMode flicker_mode = GST_PHOTOGRAPHY_FLICKER_REDUCTION_50HZ;
  ret = gst_photography_set_flicker_mode (photo_if, flicker_mode);
  g_print ("Set flicker mode: \t%s\n", ret ? "Success" : "Fail");
  ret = gst_photography_get_flicker_mode (photo_if, &flicker_mode);
  g_print ("Get flicker mode: \t%s\n", ret ? "Success" : "Fail");
  fail_unless (flicker_mode == GST_PHOTOGRAPHY_FLICKER_REDUCTION_50HZ);

  GstFocusMode focus_mode = GST_PHOTOGRAPHY_FOCUS_MODE_INFINITY;
  ret = gst_photography_set_focus_mode (photo_if, focus_mode);
  g_print ("Setting focus mode: \t%s\n", ret ? "Success" : "Fail");
  ret = gst_photography_get_focus_mode (photo_if, &focus_mode);
  g_print ("Get focus mode: \t%s\n", ret ? "Success" : "Fail");
  fail_unless (focus_mode == GST_PHOTOGRAPHY_FOCUS_MODE_INFINITY);

  /* set_format & get_format */
  GstCaps *mode_caps;
  mode_caps = gst_caps_new_simple ("video/x-raw-yuv",
        "format", GST_TYPE_FOURCC, GST_MAKE_FOURCC ('U', 'Y', 'V', 'Y'),
        "width", G_TYPE_INT, 640,
        "height", G_TYPE_INT, 480,
        "framerate", GST_TYPE_FRACTION, 2406, 100,
        NULL);

  GstOperationMode op_mode = GST_PHOTOGRAPHY_OPERATION_MODE_IMAGE_CAPTURE;
  ret = gst_photography_set_format (photo_if, op_mode, mode_caps);
  g_print ("Set format : \t\t%s\n", ret ? "Success" : "Fail");

  GstCaps * omode_caps;
  GstStructure *structure;
  gint w;
  gint h;
  gint fps_n = 0;
  gint fps_d = 0;
  guint32 fourcc;
  gchar *capsstr;
  gboolean afocus = TRUE;

  omode_caps = gst_photography_get_format (photo_if, op_mode);
  capsstr = gst_caps_to_string (omode_caps);
  g_print ("omode_caps : %s\n", capsstr);
  g_free (capsstr);
  fail_unless (omode_caps != NULL);

  structure = gst_caps_get_structure (omode_caps, 0);
  gst_structure_get_int (structure, "width", &w);
  gst_structure_get_int (structure, "height", &h);
  gst_structure_get_fraction (structure, "framerate", &fps_n, &fps_d);
  gst_structure_get_fourcc (structure, "format", &fourcc);
  g_print ("Get format: %dx%d @ %d/%d, format = %" GST_FOURCC_FORMAT,
              w, h, fps_n, fps_d, GST_FOURCC_ARGS(fourcc));
  fail_unless_equals_int (w, 640);
  fail_unless_equals_int (h, 480);

  gst_photography_set_autofocus (photo_if, afocus);

  /* set_config & get_config */
  GstPhotoSettings photoconf;

  photoconf.wb_mode = GST_PHOTOGRAPHY_WB_MODE_AUTO;
  photoconf.tone_mode = GST_PHOTOGRAPHY_COLOUR_TONE_MODE_NORMAL;
  photoconf.exposure = 0;
  photoconf.aperture = 0;
  photoconf.iso_speed = 0;
  photoconf.ev_compensation = 0.0;
  photoconf.scene_mode = GST_PHOTOGRAPHY_SCENE_MODE_MANUAL;
  photoconf.flash_mode = GST_PHOTOGRAPHY_FLASH_MODE_AUTO;
  photoconf.zoom = 1.0;
  photoconf.flicker_mode = GST_PHOTOGRAPHY_FLICKER_REDUCTION_OFF;
  photoconf.focus_mode = GST_PHOTOGRAPHY_FOCUS_MODE_AUTO;

  ret = gst_photography_set_config (photo_if, &photoconf);
  g_print ("Set config : \t%s\n", ret ? "Success" : "Fail");
  fail_unless(ret == TRUE);

  ret = gst_photography_get_config (photo_if, &photoconf);
  g_print ("Get config : \t%s\n", ret ? "Success" : "Fail");
  fail_unless (photoconf.wb_mode == GST_PHOTOGRAPHY_WB_MODE_AUTO);
  fail_unless (photoconf.flash_mode == GST_PHOTOGRAPHY_FLASH_MODE_AUTO);

  /* cleanup */
  state_ret = gst_element_set_state (pipeline, GST_STATE_NULL);
  g_print ("Pipeline set to NULL: %d\n", state_ret);
  gst_object_unref (GST_OBJECT (pipeline));

}
GST_END_TEST;

GST_START_TEST (test_device_photography_properties)
{
  GstElement *pipeline = NULL;
  GstElement *source = NULL;
  GstPhotography *photo_if = NULL;
  GstStateChangeReturn state_ret;
  GstSceneMode scene_mode;
  GstFlickerReductionMode flicker_mode;
  GstColourToneMode colour_tone_mode;
  GstFlashMode flash_mode;
  GstWhiteBalanceMode wb_mode;
  GstFocusMode focus_mode;
  GstCameraSrcCaptureMode capture_mode;
  GstCameraSrcViewfinderMode viewfinder_mode;
  GstPhotoCaps photocaps = GST_PHOTOGRAPHY_CAPS_ZOOM;
  guint exposure,aperture, iso_speed;
  gfloat ev_compensation, zoom;

  pipeline = initialize_photo_if ();
  source = gst_bin_get_by_name (GST_BIN(pipeline), "src");
  photo_if = GST_PHOTOGRAPHY (source);

  g_object_set (G_OBJECT (source), "scene-mode", 5, NULL);
  g_object_get (G_OBJECT (source), "scene-mode", &scene_mode, NULL);
  fail_unless_equals_int (scene_mode, 5);

  g_object_set (G_OBJECT (source), "flicker-mode", 2, NULL);
  g_object_get (G_OBJECT (source), "flicker-mode", &flicker_mode, NULL);
  fail_unless_equals_int (flicker_mode, 2);

  g_object_set (G_OBJECT (source), "exposure", 5, NULL);
  g_object_get (G_OBJECT (source), "exposure", &exposure, NULL);
  fail_unless_equals_int (exposure, 5);

  g_object_set (G_OBJECT (source), "aperture", 100, NULL);
  g_object_get (G_OBJECT (source), "aperture", &aperture, NULL);
  fail_unless_equals_int (aperture, 100);

  g_object_set (G_OBJECT (source), "flash-mode", 4, NULL);
  g_object_get (G_OBJECT (source), "flash-mode", &flash_mode, NULL);
  fail_unless_equals_int (flash_mode, 4);

  g_object_set (G_OBJECT (source), "iso-speed", 800, NULL);
  g_object_get (G_OBJECT (source), "iso-speed", &iso_speed, NULL);
  fail_unless_equals_int (iso_speed, 800);

  g_object_set (G_OBJECT (source), "colour-tone-mode", 3, NULL);
  g_object_get (G_OBJECT (source), "colour-tone-mode", &colour_tone_mode, NULL);
  fail_unless_equals_int (colour_tone_mode, 3);

  g_object_set (G_OBJECT (source), "ev-compensation", 2.1, NULL);
  g_object_get (G_OBJECT (source), "ev-compensation", &ev_compensation, NULL);
  fail_unless_equals_float (ev_compensation, 2.1);

  g_object_set (G_OBJECT (source), "white-balance-mode", 1, NULL);
  g_object_get (G_OBJECT (source), "white-balance-mode", &wb_mode, NULL);
  fail_unless_equals_int (wb_mode, 1);

  g_object_set (G_OBJECT (source), "focus-mode", 4, NULL);
  g_object_get (G_OBJECT (source), "focus-mode", &focus_mode, NULL);
  fail_unless_equals_int (focus_mode, 4);

  g_object_set (G_OBJECT (source), "capture-mode", 1, NULL);
  g_object_get (G_OBJECT (source), "capture-mode", &capture_mode, NULL);
  fail_unless_equals_int (capture_mode, 1);

  g_object_set (G_OBJECT (source), "viewfinder-mode", 0, NULL);
  g_object_get (G_OBJECT (source), "viewfinder-mode", &viewfinder_mode, NULL);
  fail_unless_equals_int (viewfinder_mode, 0);

  photocaps = gst_photography_get_capabilities (photo_if);
  fail_unless_equals_int (GST_PHOTOGRAPHY_CAPS_ZOOM,
      photocaps & GST_PHOTOGRAPHY_CAPS_ZOOM);

  g_object_set (G_OBJECT (source), "zoom", 3.5, NULL);
  g_object_get (G_OBJECT (source), "zoom", &zoom, NULL);
  fail_unless_equals_float (zoom, 3.5);

  /* cleanup */
  state_ret = gst_element_set_state (pipeline, GST_STATE_NULL);
  g_print ("Pipeline set to NULL: %d\n", state_ret);
  gst_object_unref (GST_OBJECT (pipeline));

}
GST_END_TEST;

static Suite *
camsrc_suite (void)
{
  Suite *s = suite_create ("camsrc");
  TCase *tc_chain = tcase_create ("general");
  tcase_set_timeout(tc_chain, 20);
  suite_add_tcase (s, tc_chain);

  tcase_add_test (tc_chain, test_device_photography_interface);
  tcase_add_test (tc_chain, test_device_photography_properties);

  return s;
}
GST_CHECK_MAIN (camsrc)
